unit FormulaParser;

//Copyright 2013-2017 Ren Dahms (rene.dahms at math.ethz.ch)
//
//This file is part of LSRMTools.
//
//    LSRMTools is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    LSRMTools is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with LSRMTools.  If not, see <http://www.gnu.org/licenses/>.
//
//LSRMTools use some parts of ALGLIB which is a free software distributed
//under a GPL license - version 2 or (at your option) any later version.

interface

uses System.Generics.Collections,System.Generics.Defaults,SysUtils,windows,classes;

type
{$REGION 'Token separators'}
  TTokenChars = packed record
    WhiteChars: TSysCharSet;
    DecimalChar, PlusChar, MinusChar, MulChar, DivChar, PowerChar, OpeningBraceChar,
    ClosingBraceChar, BeginVecChar, EndVecChar, FormulaSepChar, ParaSepChar: Char;
    FuncCumString, FuncIncString, FuncMaxString, FuncMinString, FuncAbsString,
    FuncSqrtString, ClaimPropString, FuncColumnString: String;
  end;
{$ENDREGION}

{$REGION 'basic tokens'}
  TToken = class;
  TTokenClass = class of TToken;
  TToken_Operator = class;
  TToken = class(TObject)
    private {$REGION}
      FParas: TObjectList<TToken>;
      FNumberOfParas: integer;
      FStart: integer;
      FLevel: integer;
      FName: string;
      FLinear: boolean;
    {$ENDREGION}
    protected {$REGION}
      function GetEndPos: integer; virtual;
      function GetLen: integer; virtual;
      function GetLevel: integer; virtual;
      function GetName: string; virtual;
      function GetNumberOfParas: integer; virtual;
      function GetLinear: boolean; virtual;
      function GetParas: TObjectList<TToken>; virtual;
      function GetStart: integer; virtual;
      procedure SetLevel(const Value: integer); virtual;
      procedure SetName(const Value:string); virtual;
      procedure SetNumberOfParas(const Value: integer); virtual;
      procedure SetLinear(const Value: boolean); virtual;
      procedure SetStart(const Value: integer); virtual;
    {$ENDREGION}
    public {$REGION}
      constructor Create(AName:string); virtual;
      destructor Destroy; override;
      procedure CheckNumberOfParas(i: integer); virtual;
      procedure Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer); virtual; abstract;
      procedure ProcessOPs(CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; Level, Priority: integer; Paras: Pointer); virtual;
      procedure Calc(Result, Paras: pointer); virtual;
      property EndPos: integer read GetEndPos;
      property Len: integer read GetLen;
      property Level: integer read GetLevel write SetLevel;
      property Name: string read GetName write SetName;
      property NumberOfParas: integer read GetNumberOfParas write SetNumberOfParas;
      property Linear: boolean read GetLinear write SetLinear ;
      property Paras: TObjectList<TToken> read GetParas;
      property Start: integer read GetStart write SetStart;
    {$ENDREGION}
  end;

  TToken_Evaluable = class(TToken);
  TToken_NotEvaluable = class(TToken)
    public
      constructor Create(AName: string); override;
  end;

  TToken_Number = class(TToken_Evaluable)
    private {$REGION}
      FNumber: Double;
    {$ENDREGION}
    protected {$REGION}
      function GetValue: Double; virtual;
      procedure SetName(const Value: string); override;
    {$ENDREGION}
    public {$REGION}
      constructor Create(AName: string); override;
      procedure Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer); override;
      procedure Calc(Result, Paras: pointer); override;
      property Value: Double read GetValue;
    {$ENDREGION}
  end;

  TToken_Unknown = class(TToken_NotEvaluable)
    private {$REGION}
      FWhiteChars:TSysCharSet;
    {$ENDREGION}
    protected {$REGION}
      function GetWhiteChars: TSysCharSet; virtual;
      procedure SetWhiteChars(const Value: TSysCharSet); virtual;
    {$ENDREGION}
    public {$REGION}
      constructor Create(AName: string); override;
      procedure Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer); override;
      property WhiteChars: TSysCharSet read GetWhiteChars write SetWhiteChars;
    {$ENDREGION}
  end;

{$ENDREGION}

{$REGION 'operator token'}

  TToken_Operator = class(TToken_Evaluable)
    public
      constructor Create(AName: string); override;
      procedure CheckNumberOfParas(i: integer); override;
      procedure Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer); override;
      function Priority: integer; virtual;
  end;

  TToken_OP_FixedNumberOfParas = class(TToken_Operator);
  TToken_OP_VariableNumberOfParas = class(TToken_Operator)
    public
      procedure Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer); override;
  end;

  TToken_Plus = class(TToken_OP_VariableNumberOfParas)
    public
      constructor Create(AName: string); override;
      function Priority: integer; override;
      procedure Calc(Result, Paras: pointer); override;
  end;

  TToken_Minus = class(TToken_OP_VariableNumberOfParas)
    public
      constructor Create(AName: string); override;
      function Priority: integer; override;
      procedure Calc(Result, Paras: pointer); override;
  end;

  TToken_Mul = class(TToken_OP_FixedNumberOfParas)
    public
      constructor Create(AName: string); override;
      function Priority: integer; override;
      procedure Calc(Result, Paras: pointer); override;
  end;

  TToken_Div = class(TToken_OP_FixedNumberOfParas)
    public
      constructor Create(AName: string); override;
      function Priority: integer; override;
      procedure Calc(Result, Paras: pointer); override;
  end;

  TToken_Power = class(TToken_OP_FixedNumberOfParas)
    public
      function Priority: integer; override;
      procedure Calc(Result, Paras: pointer); override;
  end;

{$ENDREGION}

{$REGION 'brace token'}

  TToken_OpeningBrace = class(TToken_NotEvaluable)
    public
      procedure Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer); override;
  end;

  TToken_ClosingBrace = class(TToken_Evaluable)
    public
      procedure Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer); override;
  end;

  TToken_BeginVec = class(TToken_NotEvaluable)
    public
      procedure Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer); override;
  end;

  TToken_EndVec = class(TToken_Evaluable)
    public
      procedure Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer); override;
  end;

{$ENDREGION}

{$REGION 'vector token'}

  TToken_Vector = class(TToken_Evaluable)
    private
      FValues: array of Double;
    public
      constructor Create(AName: string); override;
      procedure Calc(Result, Paras: pointer); override;
      procedure Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer); override;
  end;

{$ENDREGION}

{$REGION 'separator token'}

  TToken_Separator = class(TToken_NotEvaluable);

  TToken_FormulaSep = class(TToken_Separator)
    public
      procedure Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer); override;
  end;

  TToken_ParaSep = class(TToken_Separator)
    public
      procedure Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer); override;
  end;

{$ENDREGION}

{$REGION 'function token'}

  TToken_Func = class(TToken_Evaluable)
    protected {$REGION}
      function GetPara(i: integer): TToken;
      procedure SetPara(i: integer; const Value: TToken);
    {$ENDREGION}
    public {$REGION}
      procedure Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer); override;
      property Para[i: integer]: TToken read GetPara write SetPara;
    {$ENDREGION}
  end;

  TToken_Abs = class(TToken_Func)
    public
      constructor Create(AName: string); override;
      procedure Calc(Result, Paras: pointer); override;
  end;

  TToken_ClaimProperty = class(TToken_Func)
    public
      constructor Create(AName: string); override;
      procedure Calc(Result, Paras: pointer); override;
      procedure Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer); override;
  end;

  TToken_Column = class(TToken_Func)
    public
      constructor Create(AName: string); override;
      procedure Calc(Result, Paras: pointer); override;
      procedure Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer); override;
  end;

  TToken_Cum = class(TToken_Func)
    public
      constructor Create(AName: string); override;
      procedure Calc(Result, Paras: pointer); override;
  end;

  TToken_ID = class(TToken_Func)
    public
      constructor Create(AName: string); override;
      procedure Calc(Result, Paras: pointer); override;
      procedure Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer); override;
  end;

  TToken_Inc = class(TToken_Func)
    public
      constructor Create(AName: string); override;
      procedure Calc(Result, Paras: pointer); override;
  end;

  TToken_Max = class(TToken_Func)
    public
      constructor Create(AName: string); override;
      procedure Calc(Result, Paras: pointer); override;
  end;

  TToken_Min = class(TToken_Func)
    public
      constructor Create(AName: string); override;
      procedure Calc(Result, Paras: pointer); override;
  end;

  TToken_Sqrt = class(TToken_Func)
    public
      constructor Create(AName: string); override;
      procedure Calc(Result, Paras: pointer); override;
  end;

{$ENDREGION}

{$REGION 'formula parser'}

  TTokenIdentifier = record
    Name: string;
    TokenClass: TTokenClass;
  end;

  TTokenCompare = class(TComparer<TTokenIdentifier>)
    function Compare(const Left, Right: TTokenIdentifier): integer; override;
  end;

  TFormulaParser= class(TObject)
    private {$REGION}
      FFormula: string;
      FTokenList: TList<TTokenIdentifier>;
      FFormulas: TObjectList<TToken>;
      FCaseSensitive: Boolean;
      FDecimalSep: char;
      FWhiteChars: TSysCharSet;
      function GetNumber(Formula: string; i: integer): string; virtual;
      procedure InitTokens(TokenChars: TTokenChars);
    {$ENDREGION}
    protected {$REGION}
      function GetCaseSensitive: Boolean; virtual;
      function GetCount: integer; virtual;
      function GetFormula(i: integer): TToken; virtual;
      function GetFormulaString: string; virtual;
      procedure SetCaseSensitive(const Value: Boolean); virtual;
    {$ENDREGION}
    public {$REGION}
      constructor Create(TokenChars: TTokenChars);
      destructor Destroy; override;
      function LastErrorMessage(AException: Exception): string; virtual;
      procedure Parse(Formula: string; Paras: pointer); virtual;
      procedure RegisterToken(Name: string; TokenClass: TTokenClass); virtual;
      property CaseSensitive: Boolean read GetCaseSensitive write SetCaseSensitive default false;
      property Count: integer read GetCount;
      property Formula[i: integer]: TToken read GetFormula;
      property FormulaString: string read GetFormulaString;
    {$ENDREGION}
  end;

{$ENDREGION}

{$REGION 'Exceptions'}

  EParseException = class(Exception)
    public
      Start: integer;
      constructor Create(i: integer); overload;
  end;

  EParseException_InvalidCharacter = class(EParseException);
  EParseException_SyntaxError = class(EParseException);
  EParseException_MissingFormula = class(EParseException);
  EParseException_MissingParameter = class(EParseException);
  EParseException_MissingOpeningBrace = class(EParseException);
  EParseException_MissingClosingBrace = class(EParseException);
  EParseException_MissingParameterSeparator = class(EParseException);
  EParseException_MissingFormulaSeparator = class(EParseException);
  EParseException_MissingBeginVector= class(EParseException);
  EParseException_MissingEndVector= class(EParseException);
  EParseException_NumberExpected= class(EParseException);
  EParseException_ClaimPropertyOutOfRange= class(EParseException);
  EParseException_OnlyLinearAllowed= class(EParseException);

  EInitParserException = class(Exception);
  EInitParserException_InvalidDelimiter= class(EInitParserException);
  EInitParserException_DoubleDelimiter= class(EInitParserException);
  EInitParserException_InvalidFunction= class(EInitParserException);
  EInitParserException_DoubleFunction= class(EInitParserException);

{$ENDREGION}

implementation

uses StrUtils, Math, LSRM;

const {$REGION 'digits and priorities'}
      C_SetOfdigits: TSysCharSet = ['0','1','2','3','4','5','6','7','8','9'];
      C_Priority_Sign: integer  = 60;
      C_Priority_Power: integer = 50;
      C_Priority_Div: integer   = 40;
      C_Priority_Mul: integer   = 30;
      C_Priority_Minus: integer = 20;
      C_Priority_Plus: integer  = 10;
{$ENDREGION}

{$REGION 'TFormulaParser'}

{$REGION 'TFormulaParser private'}

function TFormulaParser.GetNumber(Formula: string; i: integer): string;
  var j: integer;
begin
  // get digits
  j:=i;
  while (i<=length(Formula)) and CharInSet(Formula[i],C_SetOfdigits) do inc(i);
  // check if it is a decimal separator
  if (i<length(Formula)) and (Formula[i]=FDecimalSep)
    and CharInSet(FOrmula[i+1],C_SetOfdigits) then begin
    inc(i,2);
    while (i<=length(Formula)) and CharInSet(Formula[i],C_SetOfdigits) do inc(i);
  end;
  // check if the next char equals 'E' or 'e' followed by a number
  if ((i<length(Formula)) and (UpperCase(Formula[i])='E')
        and CharInSet(Formula[i+1],C_SetOfdigits))
     or ((i<length(Formula)-1) and (UpperCase(Formula[i])='E')
        and CharInSet(Formula[i+1],['+','-'])
        and CharInSet(Formula[i+2],C_SetOfdigits)) then begin
    if CharInSet(Formula[i+1],['+','-']) then inc(i,3) else inc(i,2);
    // get digits
    while (i<=length(Formula)) and CharInSet(Formula[i],C_SetOfdigits) do inc(i);
    // check if it is a decimal separator
    if (i<length(Formula)) and (Formula[i]=FDecimalSep)
      and CharInSet(FOrmula[i+1],C_SetOfdigits) then begin
      inc(i,2);
      while (i<=length(Formula)) and CharInSet(Formula[i],C_SetOfdigits) do inc(i);
    end;
  end;
  result:=MidStr(Formula,j,i-j);
end;

{$ENDREGION}

{$REGION 'TFormulaParser protected'}

function TFormulaParser.GetCaseSensitive: Boolean;
begin
  result:=FCaseSensitive;
end;

function TFormulaParser.GetCount: integer;
begin
  result:=FFormulas.Count;
end;

function TFormulaParser.GetFormula(i: integer): TToken;
begin
  result:=FFormulas[i];
end;

function TFormulaParser.GetFormulaString: string;
begin
  result:=FFormula;
end;

procedure TFormulaParser.SetCaseSensitive(const Value: Boolean);
begin
  FCaseSensitive:=Value;
end;

{$ENDREGION}

{$REGION 'TFormulaParser public'}

procedure TFormulaParser.InitTokens(TokenChars: TTokenChars);
  function IsLetter(letter: Char): boolean;
  begin
    result:=((letter>='a') and (letter<='z')) or ((letter>='A') and (letter<='Z'));
  end;
  function IsDiggit(Diggit: Char): boolean;
  begin
    result:=(Diggit>='0') and (Diggit<='9');
  end;
  function AllLetters(AString:string): boolean;
  var i: integer;
  begin
    i:=1;
    while (i<=Length(AString)) and IsLetter(AString[i]) do Inc(i);
    result:=(i>Length(Astring));
  end;
  var functions: string;
  var operators: string;
  var c: AnsiChar;
begin
  functions:='';
  operators:='';
  {$REGION 'check white characters'}
  for c in TokenChars.WhiteChars do begin
    if IsLetter(Char(c)) or IsDiggit(Char(c)) then
      raise EInitParserException_InvalidDelimiter.Create('White characters ('+ TokenChars.FormulaSepChar + ') must not be a letter or digit.');
    operators:=operators+c;
  end;
  FWhiteChars:=TokenChars.WhiteChars;
  {$ENDREGION}
  {$REGION  'set decimal delimiter'}
  if IsLetter(TokenChars.DecimalChar) or IsDiggit(TokenChars.DecimalChar) then
    raise EInitParserException_InvalidDelimiter.Create('The decimal delimiter ('+ TokenChars.DecimalChar + ') must not be a letter or a digit.');
  if Pos(TokenChars.DecimalChar,operators)>0 then
    raise EInitParserException_DoubleDelimiter.Create('The formula delimiter (' + TokenChars.DecimalChar + ') is already in use.');
  FDecimalSep:=TokenChars.DecimalChar;
  operators:=operators + TokenChars.DecimalChar;
  {$ENDREGION}
  {$REGION  'check and register formula delimiter'}
  if IsLetter(TokenChars.FormulaSepChar) or IsDiggit(TokenChars.FormulaSepChar) then
    raise EInitParserException_InvalidDelimiter.Create('The formula delimiter ('+ TokenChars.FormulaSepChar + ') must not be a letter or digit.');
  if Pos(TokenChars.FormulaSepChar,operators)>0 then
    raise EInitParserException_DoubleDelimiter.Create('The formula delimiter (' + TokenChars.FormulaSepChar + ') is already in use.');
  RegisterToken(TokenChars.FormulaSepChar,TToken_FormulaSep);
  operators:=operators + UpperCase(TokenChars.FormulaSepChar);
  {$ENDREGION}
  {$REGION 'check and register parameter delimiter'}
  if IsLetter(TokenChars.ParaSepChar) or IsDiggit(TokenChars.ParaSepChar) then
    raise EInitParserException_InvalidDelimiter.Create('The parameter delimiter ('+ TokenChars.ParaSepChar + ') must not be a letter or digit.');
  if Pos(TokenChars.ParaSepChar,operators)>0 then
    raise EInitParserException_DoubleDelimiter.Create('The parameter delimiter (' + TokenChars.ParaSepChar + ') is already in use.');
  RegisterToken(TokenChars.ParaSepChar,TToken_ParaSep);
  operators:=operators + UpperCase(TokenChars.ParaSepChar);
  {$ENDREGION}
  {$REGION 'check and register opening brace'}
  if IsLetter(TokenChars.OpeningBraceChar) or IsDiggit(TokenChars.OpeningBraceChar) then
    raise EInitParserException_InvalidDelimiter.Create('The opening brace character ('+ TokenChars.OpeningBraceChar + ') must not be a letter or digit.');
  if Pos(TokenChars.OpeningBraceChar,operators)>0 then
    raise EInitParserException_DoubleDelimiter.Create('The opening brace character (' + TokenChars.OpeningBraceChar + ') is already in use.');
  RegisterToken(TokenChars.OpeningBraceChar,TToken_OpeningBrace);
  operators:=operators + UpperCase(TokenChars.OpeningBraceChar);
  {$ENDREGION}
  {$REGION 'check and register closing brace'}
  if IsLetter(TokenChars.ClosingBraceChar) or IsDiggit(TokenChars.ClosingBraceChar) then
    raise EInitParserException_InvalidDelimiter.Create('The closing brace character ('+ TokenChars.ClosingBraceChar + ') must not be a letter or digit.');
  if Pos(TokenChars.ClosingBraceChar,operators)>0then
    raise EInitParserException_DoubleDelimiter.Create('The closing brace character (' + TokenChars.ClosingBraceChar + ') is already in use.');
  RegisterToken(TokenChars.ClosingBraceChar,TToken_ClosingBrace);
  operators:=operators + UpperCase(TokenChars.ClosingBraceChar);
  {$ENDREGION}
  {$REGION 'check and register begin vector'}
  if IsLetter(TokenChars.BeginVecChar) or IsDiggit(TokenChars.BeginVecChar) then
    raise EInitParserException_InvalidDelimiter.Create('The begin vector character ('+ TokenChars.BeginVecChar + ') must not be a letter or digit.');
  if Pos(TokenChars.BeginVecChar,operators)>0 then
    raise EInitParserException_DoubleDelimiter.Create('The begin vector character (' + TokenChars.BeginVecChar + ') is already in use.');
  RegisterToken(TokenChars.BeginVecChar,TToken_BeginVec);
  operators:=operators + UpperCase(TokenChars.BeginVecChar);
  {$ENDREGION}
  {$REGION 'check and register end vector'}
  if IsLetter(TokenChars.EndVecChar) or IsDiggit(TokenChars.EndVecChar) then
    raise EInitParserException_InvalidDelimiter.Create('The end vector character ('+ TokenChars.EndVecChar + ') must not be a letter or digit.');
  if Pos(TokenChars.EndVecChar,operators)>0 then
    raise EInitParserException_DoubleDelimiter.Create('The end vector character (' + TokenChars.EndVecChar + ') is already in use.');
  RegisterToken(TokenChars.EndVecChar,TToken_EndVec);
  operators:=operators + UpperCase(TokenChars.EndVecChar);
  {$ENDREGION}
  {$REGION 'check and register plus'}
  if IsLetter(TokenChars.PlusChar) or IsDiggit(TokenChars.PlusChar) then
    raise EInitParserException_InvalidDelimiter.Create('The plus character ('+ TokenChars.PlusChar + ') must not be a letter or digit.');
  if Pos(TokenChars.PlusChar,operators)>0 then
    raise EInitParserException_DoubleDelimiter.Create('The plus character (' + TokenChars.PlusChar + ') is already in use.');
  RegisterToken(TokenChars.PlusChar,TToken_Plus);
  operators:=operators + UpperCase(TokenChars.PlusChar);
  {$ENDREGION}
  {$REGION 'check and register minus'}
  if IsLetter(TokenChars.MinusChar) or IsDiggit(TokenChars.MinusChar) then
    raise EInitParserException_InvalidDelimiter.Create('The minus character ('+ TokenChars.MinusChar + ') must not be a letter or digit.');
  if Pos(TokenChars.MinusChar,operators)>0 then
    raise EInitParserException_DoubleDelimiter.Create('The minus character (' + TokenChars.MinusChar + ') is already in use.');
  RegisterToken(TokenChars.MinusChar,TToken_Minus);
  operators:=operators + UpperCase(TokenChars.MinusChar);
  {$ENDREGION}
  {$REGION 'check and register multiplication'}
  if IsLetter(TokenChars.MulChar) or IsDiggit(TokenChars.MulChar) then
    raise EInitParserException_InvalidDelimiter.Create('The multiplication character ('+ TokenChars.MulChar + ') must not be a letter or digit.');
  if Pos(TokenChars.MulChar,operators)>0 then
    raise EInitParserException_DoubleDelimiter.Create('The multiplication character (' + TokenChars.MulChar + ') is already in use.');
  RegisterToken(TokenChars.MulChar,TToken_Mul);
  operators:=operators + UpperCase(TokenChars.MulChar);
  {$ENDREGION}
  {$REGION 'check and register division'}
  if IsLetter(TokenChars.DivChar) or IsDiggit(TokenChars.DivChar) then
    raise EInitParserException_InvalidDelimiter.Create('The division character ('+ TokenChars.DivChar + ') must not be a letter or digit.');
  if Pos(TokenChars.DivChar,operators)>0 then
    raise EInitParserException_DoubleDelimiter.Create('The division character (' + TokenChars.DivChar + ') is already in use.');
  RegisterToken(TokenChars.DivChar,TToken_Div);
  operators:=operators + UpperCase(TokenChars.DivChar);
  {$ENDREGION}
  {$REGION 'check and register power'}
  if IsLetter(TokenChars.PowerChar) or IsDiggit(TokenChars.PowerChar) then
    raise EInitParserException_InvalidDelimiter.Create('The power character ('+ TokenChars.PowerChar + ') must not be a letter or digit.');
  if Pos(TokenChars.PowerChar,operators)>0 then
    raise EInitParserException_DoubleDelimiter.Create('The power character (' + TokenChars.PowerChar + ') is already in use.');
  RegisterToken(TokenChars.PowerChar,TToken_Power);
  operators:=operators + UpperCase(TokenChars.PowerChar);
  {$ENDREGION}
  {$REGION 'check and register function cum'}
  if not AllLetters(TokenChars.FuncCumString) then
    raise EInitParserException_InvalidFunction.Create('The function '+ TokenChars.FuncCumString + ' must not contain other charecters be than letters.');
  if Pos(';' + TokenChars.FuncCumString + ';', functions)>0 then
    raise EInitParserException_DoubleFunction.Create('The function ' + TokenChars.FuncCumString + ' is already in use.');
  RegisterToken(TokenChars.FuncCumString,TToken_Cum);
  functions:=functions + UpperCase(TokenChars.FuncCumString) + ';';
  {$ENDREGION}
  {$REGION 'check and register function inc'}
  if not AllLetters(TokenChars.FuncIncString) then
    raise EInitParserException_InvalidFunction.Create('The function '+ TokenChars.FuncIncString + ' must not contain other charecters be than letters.');
  if Pos(';' + TokenChars.FuncIncString + ';', functions)>0 then
    raise EInitParserException_DoubleFunction.Create('The function ' + TokenChars.FuncIncString + ' is already in use.');
  RegisterToken(TokenChars.FuncIncString,TToken_Inc);
  functions:=functions + UpperCase(TokenChars.FuncIncString) + ';';
  {$ENDREGION}
  {$REGION 'check and register function max'}
  if not AllLetters(TokenChars.FuncMaxString) then
    raise EInitParserException_InvalidFunction.Create('The function '+ TokenChars.FuncMaxString + ' must not contain other charecters be than letters.');
  if Pos(';' + TokenChars.FuncMaxString + ';', functions)>0 then
    raise EInitParserException_DoubleFunction.Create('The function ' + TokenChars.FuncMaxString + ' is already in use.');
  RegisterToken(TokenChars.FuncMaxString,TToken_Max);
  functions:=functions + UpperCase(TokenChars.FuncMaxString) + ';';
  {$ENDREGION}
  {$REGION 'check and register function min'}
  if not AllLetters(TokenChars.FuncMinString) then
    raise EInitParserException_InvalidFunction.Create('The function '+ TokenChars.FuncMinString + ' must not contain other charecters be than letters.');
  if Pos(';' + TokenChars.FuncMinString + ';', functions)>0 then
    raise EInitParserException_DoubleFunction.Create('The function ' + TokenChars.FuncMinString + ' is already in use.');
  RegisterToken(TokenChars.FuncMinString,TToken_Min);
  functions:=functions + UpperCase(TokenChars.FuncMinString) + ';';
  {$ENDREGION}
  {$REGION 'check and register function Abs'}
  if not AllLetters(TokenChars.FuncAbsString) then
    raise EInitParserException_InvalidFunction.Create('The function '+ TokenChars.FuncAbsString + ' must not contain other charecters be than letters.');
  if Pos(';' + TokenChars.FuncAbsString + ';', functions)>0 then
    raise EInitParserException_DoubleFunction.Create('The function ' + TokenChars.FuncAbsString + ' is already in use.');
  RegisterToken(TokenChars.FuncAbsString,TToken_Abs);
  functions:=functions + UpperCase(TokenChars.FuncAbsString) + ';';
  {$ENDREGION}
  {$REGION 'check and register function Sqrt'}
  if not AllLetters(TokenChars.FuncSqrtString) then
    raise EInitParserException_InvalidFunction.Create('The function '+ TokenChars.FuncSqrtString + ' must not contain other charecters be than letters.');
  if Pos(';' + TokenChars.FuncSqrtString + ';', functions)>0 then
    raise EInitParserException_DoubleFunction.Create('The function ' + TokenChars.FuncSqrtString + ' is already in use.');
  RegisterToken(TokenChars.FuncSqrtString,TToken_Sqrt);
  functions:=functions + UpperCase(TokenChars.FuncSqrtString) + ';';
  {$ENDREGION}
  {$REGION 'check and register function ClaimsProperty'}
  if not AllLetters(TokenChars.ClaimPropString) then
    raise EInitParserException_InvalidFunction.Create('The function '+ TokenChars.ClaimPropString + ' must not contain other charecters be than letters.');
  if Pos(';' + TokenChars.ClaimPropString + ';', functions)>0 then
    raise EInitParserException_DoubleFunction.Create('The function ' + TokenChars.ClaimPropString + ' is already in use.');
  RegisterToken(TokenChars.ClaimPropString,TToken_ClaimProperty);
  functions:=functions + UpperCase(TokenChars.ClaimPropString) + ';';
  {$ENDREGION}
  {$REGION 'check and register function Column'}
  if not AllLetters(TokenChars.FuncColumnString) then
    raise EInitParserException_InvalidFunction.Create('The function '+ TokenChars.FuncColumnString + ' must not contain other charecters be than letters.');
  if Pos(';' + TokenChars.FuncColumnString + ';', functions)>0 then
    raise EInitParserException_DoubleFunction.Create('The function ' + TokenChars.FuncColumnString + ' is already in use.');
  RegisterToken(TokenChars.FuncColumnString,TToken_Column);
  functions:=functions + UpperCase(TokenChars.FuncColumnString) + ';';
  {$ENDREGION}
end;

constructor TFormulaParser.Create(TokenChars: TTokenChars);
begin
  FTokenList:= nil;
  FFormulas:=nil;
  try
    FTokenList:=TList<TTokenIdentifier>.Create(TTokenCompare.Create);
    FFormulas:=TObjectList<TToken>.Create(True);
    InitTokens(TokenChars);
  except
    FreeAndNil(FTokenList);
    FreeAndNil(FFormulas);
    raise;
  end;
end;

destructor TFormulaParser.Destroy;
begin
  FreeAndNil(FTokenList);
  FreeAndNil(FFormulas);
  inherited;
end;

function TFormulaParser.LastErrorMessage(AException: Exception): string;
  var E: EParseException;
begin
  if AException.InheritsFrom(EParseException) then begin
    E:=EParseException(AException);
    if E.InheritsFrom(EParseException_InvalidCharacter) then result:='Invalid character:' + sLineBreak + sLineBreak;
    if E.InheritsFrom(EParseException_SyntaxError) then result:='Syntax error:' + sLineBreak + sLineBreak;
    if E.InheritsFrom(EParseException_MissingFormula) then result:='Missing formula:' + sLineBreak + sLineBreak;
    if E.InheritsFrom(EParseException_MissingParameter) then result:='Missing parameter:' + sLineBreak + sLineBreak;
    if E.InheritsFrom(EParseException_MissingOpeningBrace) then result:='Missing openning brace:' + sLineBreak + sLineBreak;
    if E.InheritsFrom(EParseException_MissingClosingBrace) then result:='Missing closing brace:' + sLineBreak + sLineBreak;
    if E.InheritsFrom(EParseException_MissingParameterSeparator) then result:='Missing parameter separator:' + sLineBreak + sLineBreak;
    if E.InheritsFrom(EParseException_MissingFormulaSeparator) then result:='Formula separator expected:' + sLineBreak + sLineBreak;
    if E.InheritsFrom(EParseException_MissingBeginVector) then result:='Missing begin of vector:' + sLineBreak + sLineBreak;
    if E.InheritsFrom(EParseException_MissingEndVector) then result:='Missing end of vector:' + sLineBreak + sLineBreak;
    if E.InheritsFrom(EParseException_NumberExpected) then result:='Number expected:' + sLineBreak + sLineBreak;
    if E.InheritsFrom(EParseException_ClaimPropertyOutOfRange) then result:='Claim property is out of range:' + sLineBreak + sLineBreak;
    if E.InheritsFrom(EParseException_OnlyLinearAllowed) then result:='Only linear opertaions allowed:' + sLineBreak + sLineBreak;
    if (E.Start=-1) or (E.Start>length(FFormula)) then begin
      Result:=Result+FFormula;
      Result:=Result+sLineBreak+StringOfChar(' ',length(FFormula))+'_';
    end else if E.Start<2 then begin
      Result:=Result+'_' + sLineBreak + ' ' + FFormula;
    end else begin
      Result:=Result+LeftStr(FFormula,E.Start-1)+sLineBreak+StringOfChar(' ',E.Start-1)+RightStr(FFormula,Length(FFormula)-E.Start+1);
    end;
  end else Result:=AException.Message;
end;

procedure TFormulaParser.Parse(Formula: string; Paras: pointer);
var i, j, Level: integer;
    Tmps: string;
    TmpToken: TToken;
    CalcStack, TokenStack, WasteStack: TStack<TToken>;
    OPStack: TStack<TToken_Operator>;
begin
  CalcStack:=nil;
  OPStack:=nil;
  TmpToken:=nil;
  WasteStack:=nil;
  try
    {$REGION 'get token'}
    TokenStack:=TStack<TToken>.Create;
    // add a formula separator add the begining
    j:=0;
    while (j<FTokenList.Count) and not FTokenList[j].TokenClass.InheritsFrom(TToken_FormulaSep) do inc(j);
    TokenStack.Push(TToken_FormulaSep.Create(FTokenList[j].Name));
    TokenStack.Peek.Start:=0;
    if CaseSensitive then FFormula:=Formula else FFormula:=UpperCase(Formula);
    i:=1;
    Tmps:='';
    while i<=Length(FFormula) do begin
      // check if it is a number
      if CharInSet(FFormula[i], c_SetOfdigits) then begin
        // if Tmps<>'' then we have an unknown token
        if Tmps<>'' then begin
          TokenStack.Push(TToken_Unknown.Create(''));
          TmpToken:=TokenStack.Peek;
          with (TmpToken as TToken_Unknown) do begin
            WhiteChars:=Self.FWhiteChars;
            Name:=Tmps;
            Start:=i-length(Tmps);
          end;
        end;
        Tmps:=GetNumber(FFormula,i);
        TokenStack.Push(TToken_Number.Create(''));
        TmpToken:=TokenStack.Peek;
        TmpToken.Name:=Tmps;
        TmpToken.Start:=i;
        inc(i,length(Tmps));
        Tmps:='';
      end else begin
        // check if it is a registred token
        j:=FTokenList.Count-1;
        while (j>=0) and (FTokenList.Items[j].Name>MidStr(FFormula,i,length(FTokenList.Items[j].Name))) do dec(j);
        if (j>=0) and (FTokenList.Items[j].Name=MidStr(FFormula,i,length(FTokenList.Items[j].Name))) then begin
          // if Tmps<>'' then we have an unknown token
          if Tmps<>'' then begin
            TokenStack.Push(TToken_Unknown.Create(''));
            TmpToken:=TokenStack.Peek;
            with (TmpToken as TToken_Unknown) do begin
              WhiteChars:=Self.FWhiteChars;
              Name:=Tmps;
              Start:=i-length(Tmps);
            end;
            Tmps:='';
          end;
          TokenStack.Push(FTokenList.Items[j].TokenClass.Create(FTokenList.Items[j].Name));
          TmpToken:=TokenStack.Peek;
          TmpToken.Start:=i;
          inc(i,TmpToken.Len);
        end else begin
          // part of an unknown token
          Tmps:=Tmps+FFormula[i];
          inc(i);
        end;
      end;
    end;
    if Tmps<>'' then begin
      TokenStack.Push(TToken_Unknown.Create(''));
      TmpToken:=TokenStack.Peek;
      with (TmpToken as TToken_Unknown) do begin
        WhiteChars:=Self.FWhiteChars;
        Name:=Tmps;
        Start:=i-length(Tmps);
      end;
    end;
    {$ENDREGION}
    {$REGION 'parse token'}
      FFormulas.Clear;
      CalcStack:=TStack<TToken>.Create;
      WasteStack:=TStack<TToken>.Create;
      OPStack:=TStack<TToken_Operator>.Create;
      Level:=0;
      while TokenStack.Count>0 do begin
        TokenStack.Peek.Parse(TokenStack, CalcStack, WasteStack, OPStack, FFormulas, Level, Paras);
      end;
      if OPStack.Count>0 then OPStack.Peek.ProcessOPs(CalcStack,WasteStack,OPStack,Level,0,Paras);
      // check if the level equals zero (less than zero is not possible here, because this is handled by the opening brace and begin vector tag)
      if Level > 0 then begin
        // look for the next closing brace or end vector token within the calculation stack
        while (CalcStack.Count>0) and not (CalcStack.Peek.InheritsFrom(TToken_ClosingBrace) or CalcStack.Peek.InheritsFrom(TToken_EndVec)) do WasteStack.Push(CalcStack.Pop);
        if CalcStack.Count=0 then raise EParseException_SyntaxError.Create(0);
        if CalcStack.Peek.InheritsFrom(TToken_ClosingBrace) then raise EParseException_MissingOpeningBrace.Create(0);
        if CalcStack.Peek.InheritsFrom(TToken_EndVec) then raise EParseException_MissingBeginVector.Create(0);
      end;
    {$ENDREGION}
    {$REGION 'clean up'}
    if CalcStack<>nil then
      while CalcStack.Count>0 do begin
        TmpToken:=CalcStack.Pop;
        FreeAndNil(TmpToken);
      end;
    FreeAndNil(CalcStack);
    if OPStack<>nil then
      while OPStack.Count>0 do begin
        TmpToken:=OPStack.Pop;
        FreeAndNil(TmpToken);
      end;
    FreeAndNil(OpStack);
    if TokenStack<>nil then
      while TokenStack.Count>0 do begin
        TmpToken:=TokenStack.Pop;
        FreeAndNil(TmpToken);
      end;
    FreeAndNil(TokenStack);
    if WasteStack<>nil then
      while WasteStack.Count>0 do begin
        TmpToken:=WasteStack.Pop;
        FreeAndNil(TmpToken);
      end;
    FreeAndNil(WasteStack);
    {$ENDREGION}
  except
    if CalcStack<>nil then
      while CalcStack.Count>0 do begin
        TmpToken:=CalcStack.Pop;
        FreeAndNil(TmpToken);
      end;
    FreeAndNil(CalcStack);
    if OPStack<>nil then
      while OPStack.Count>0 do begin
        TmpToken:=OPStack.Pop;
        FreeAndNil(TmpToken);
      end;
    FreeAndNil(OpStack);
    if TokenStack<>nil then
      while TokenStack.Count>0 do begin
        TmpToken:=TokenStack.Pop;
        FreeAndNil(TmpToken);
      end;
    FreeAndNil(TokenStack);
    if WasteStack<>nil then
      while WasteStack.Count>0 do begin
        TmpToken:=WasteStack.Pop;
        FreeAndNil(TmpToken);
      end;
    FreeAndNil(WasteStack);
    raise;
  end;
end;

procedure TFormulaParser.RegisterToken(Name: string; TokenClass: TTokenClass);
  var TokenIdentifier: TTokenIdentifier;
begin
  if not CaseSensitive then Name:=UpperCase(Name);
  TokenIdentifier.Name:=Name;
  TokenIdentifier.TokenClass:=TokenClass;
  FTokenList.Add(TokenIdentifier);
  FTokenList.Sort;
end;

{$ENDREGION}

{$ENDREGION}

{$REGION 'TTokenCompare'}

function TTokenCompare.Compare(const Left, Right: TTokenIdentifier): integer;
begin
  result:=IComparer<string>(_LookupVtableInfo(giComparer, TypeInfo(string), SizeOf(string))).Compare(Left.Name,Right.Name);
end;

{$ENDREGION}

{$REGION 'TToken'}

{$REGION 'TToken protected'}

function TToken.GetEndPos: integer;
begin
  if Paras.Count>0 then result:=Paras[Paras.Count-1].EndPos+Paras[Paras.Count-1].Level
                   else result:=Start+Len;
end;

function TToken.GetLen: integer;
begin
  result:=Length(Name);
end;

function TToken.GetLevel: integer;
begin
  result:=FLevel;
end;

function TToken.GetLinear: boolean;
begin
  result:=FLinear;
end;

function TToken.GetName: string;
begin
  result:=FName;
end;

function TToken.GetNumberOfParas: integer;
begin
  result:=FNumberOfParas;
end;

function TToken.GetParas: TObjectList<TToken>;
begin
  result:=FParas;
end;

function TToken.GetStart: integer;
begin
  result:=FStart;
end;

procedure TToken.SetLevel(const Value: integer);
begin
  FLevel:=Value;
end;

procedure TToken.SetLinear(const Value: boolean);
begin
  FLinear:=Value;
end;

procedure TToken.SetName(const Value: string);
begin
  FName:=Value;
end;

procedure TToken.SetNumberOfParas(const Value: integer);
begin
  FNumberOfParas:=Value;
end;

procedure TToken.SetStart(const Value: integer);
begin
  FSTart:=Value;
end;

{$ENDREGION}

{$REGION 'TToken public'}

procedure TToken.Calc(Result, Paras: pointer);
begin
  raise Exception.Create('Cannot calculate this token.');
end;

procedure TToken.CheckNumberOfParas(i: integer);
begin
  // do nothing
end;

constructor TToken.Create(AName: String);
begin
  FName:=AName;
  FLinear:=False;
  FParas:=nil;
  try
    FParas:=TObjectList<TToken>.Create(True);
  except
    FreeAndNil(FParas);
  end;
end;

destructor TToken.Destroy;
begin
  FreeAndNil(FParas);
  inherited;
end;

procedure TToken.ProcessOPs(CalcStack, WasteStack: TStack<TToken>;
  OPStack: TStack<TToken_Operator>; Level, Priority: integer; Paras: Pointer);
var i: integer;
    OPToken: TToken_Operator;
begin
  // go trough all operators of the OPStack which priority is not smaller and which are on the same level
  while (OPStack.Count>0) and (OPStack.Peek.Level=Level) and (OPStack.Peek.Priority>Priority) do begin
    OPToken:=OPStack.Peek;
    OPToken.CheckNumberOfParas(CalcStack.Count);
    for i:=1 to OPToken.NumberOfParas  do begin
      OPToken.Paras.Add(CalcStack.Pop);
    end;
    // move the operator from the operator stack to the calculation stack
    CalcStack.Push(OPStack.Pop);
  end;
end;

{$ENDREGION}

{$ENDREGION}

{$REGION 'TToken_NotEvaluable'}

constructor TToken_NotEvaluable.Create;
begin
  inherited;
  Linear:=True;
end;

{$ENDREGION}

{$REGION 'TToken_Unnkown'}

{$REGION 'TToken_Unnkown protected'}

function TToken_Unknown.GetWhiteChars: TSysCharSet;
begin
  result:=FWhiteChars;
end;

procedure TToken_Unknown.SetWhiteChars(const Value: TSysCharSet);
begin
  FWhiteChars:=Value;
end;

{$ENDREGION}

{$REGION 'TToken_Unnkown public'}

constructor TToken_Unknown.Create(AName: string);
begin
  inherited;
end;

procedure TToken_Unknown.Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer);
var i: integer;
begin
  // raise an exception if the token contains non-white characters
  i:=1;
  while (i<=length(FName)) and CharInSet(FName[i],WhiteChars) do inc(i);
  if i<=length(FName) then raise EParseException_InvalidCharacter.Create(Start+i-1);
  // otherwise move the token from the token stack to the waste stack
  WasteStack.Push(TokenStack.Pop);
end;

{$ENDREGION}

{$ENDREGION}

{$REGION 'TToken_Number'}

{$REGION 'TToken_Number protected'}

function TToken_Number.GetValue: Double;
begin
  result:=FNumber;
end;

procedure TToken_Number.SetName(const Value: string);
begin
  inherited;
  FNumber:=StrToFloat(Value);
end;

{$ENDREGION}

{$REGION 'TToken_Number public'}

constructor TToken_Number.Create(AName: string);
begin
  inherited;
  Linear:=True;
end;

procedure TToken_Number.Calc(Result, Paras: pointer);
  var TmpDouble: PDouble;
      CalcParas: TCalcParas;
      i,k,l,j: integer;
begin
  TmpDouble:=PDouble(Result);
  CalcParas:=PCalcParas(Paras)^;
  if CalcParas.Mode=CalcGamma1 then begin
    for l := 0 to CalcParas.M do begin
      for k := 0 to CalcParas.J do begin
        for j := 0 to k do begin
          TmpDouble^:=Value;
          Inc(TmpDouble,1);
        end;
        Inc(TmpDouble,CalcParas.J-k);
      end;
    end;
  end else begin
    for i := 0 to CalcParas.I do begin
      for k := 0 to Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i) do begin
        TmpDouble^:=Value;
        Inc(TmpDouble,1);
      end;
      inc(TmpDouble,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
    end;
  end;
end;

procedure TToken_Number.Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer);
begin
  // move the number token from the token stack to the calculation stack
  CalcStack.Push(TokenStack.Pop);
end;

{$ENDREGION}

{$ENDREGION}

{$REGION 'Operators'}

{$REGION 'TToken_Operator'}

procedure TToken_Operator.CheckNumberOfParas(i: integer);
begin
  if NumberOfParas=1 then begin
    // only one parameter required
    if i<1 then raise EParseException_MissingParameter.Create(Start+Len-1);
  end else begin
    // two parameters required
    if i<1 then raise EParseException_MissingParameter.Create(Start)
    else if i<2 then raise EParseException_MissingParameter.Create(Start+Len-1);
  end;
end;

constructor TToken_Operator.Create(AName:string);
begin
  inherited;
  FNumberOfParas:=2;
end;

procedure TToken_Operator.Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer);
  var TmpOPToken: TToken;
begin
  // raise an exception if only linear tokens are allowed and the token itself is not linear
  if PParseParas(Paras)^.OnlyLinear and not Linear then raise EParseException_OnlyLinearAllowed.Create(Start);
  // set the level
  Self.Level:=Level;
  TmpOPToken:=nil;
  try
    // remove the operator token temporary from the tokenstack
    TmpOPToken:=TokenStack.Pop;
    // a term has ended, so process all corresponding operators
    ProcessOPs(CalcStack, WasteStack, OPStack, Level, Priority, Paras);
    // move the operator from the token stack to the operator stack
    OPStack.Push(TToken_Operator(TmpOpToken));
    TmpOPToken:=nil;
  except
    FreeAndNil(TmpOPToken);
    raise;
  end;
end;

function TToken_Operator.Priority: integer;
begin
  result:=0;
end;

{$ENDREGION}

{$REGION 'TToken_OP_VariableNumberOfParas'}

procedure TToken_OP_VariableNumberOfParas.Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer);
  var TmpOPToken: TToken;
begin
  // set the level
  Self.Level:=Level;
  TmpOPToken:=nil;
  try
    // remove the operator token temporary from the tokenstack
    TmpOPToken:=TokenStack.Pop;
    // if the next token, except for all unknown tokens that contain only white spaces, is evaluable but not an operator then we have a binary operator otherwise an unary operator
    while (TokenStack.Count>0) and TokenStack.Peek.InheritsFrom(TToken_Unknown) do TokenStack.Peek.Parse(TokenStack, CalcStack, WasteStack, OPStack, FormulaList, Level, Paras);
    if (TokenStack.Count>0) and TokenStack.Peek.InheritsFrom(TToken_Evaluable) and not TokenStack.Peek.InheritsFrom(TToken_Operator) then NumberOfParas:=2 else NumberOfParas:=1;
    // a term has ended, so process all corresponding operators
    ProcessOPs(CalcStack, WasteStack, OPStack, Level, Priority, Paras);
    // move the operator from the token stack to the operator stack
    OPStack.Push(TToken_Operator(TmpOpToken));
    TmpOPToken:=nil;
  except
    FreeAndNil(TmpOPToken);
    raise;
  end;
end;

{$ENDREGION}

{$REGION 'TToken_Plus'}

procedure TToken_Plus.Calc(Result, Paras: pointer);
  var TmpDouble,Tmp1Double: PDouble;
      CalcParas: TCalcParas;
      TmpResult: array of Double;
      i,k: integer;
begin
  Self.Paras[0].Calc(result, paras);
  if NumberOfParas=1 then exit;
  CalcParas:=PCalcParas(Paras)^;
  if CalcParas.Mode=CalcGamma1 then begin
    Setlength(TmpResult,(CalcParas.M+1)*(CalcParas.J+1)*(CalcParas.J+1));
    Self.Paras[1].Calc(TmpResult, paras);
    TmpDouble:=PDouble(Result);
    Tmp1Double:=@TmpResult[0];
    for i := low(TmpResult) to High(TmpResult) do begin
      TmpDouble^:=TmpDouble^+Tmp1Double^;
      Inc(TmpDouble,1);
      Inc(Tmp1Double,1);
    end;
  end else begin
    Setlength(TmpResult,(CalcParas.I+1)*(CalcParas.J+1));
    Self.Paras[1].Calc(TmpResult, paras);
    TmpDouble:=PDouble(Result);
    Tmp1Double:=@TmpResult[0];
    for i := 0 to CalcParas.I do begin
      for k := 0 to Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i) do begin
        TmpDouble^:=TmpDouble^+Tmp1Double^;
        Inc(TmpDouble,1);
        Inc(Tmp1Double,1);
      end;
      Inc(TmpDouble,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
      Inc(Tmp1Double,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
    end;
  end;
end;

constructor TToken_Plus.Create(AName: string);
begin
  inherited;
  Linear:=True;
end;

function TToken_Plus.Priority: integer;
begin
  case NumberOfParas of
    1: result:=C_Priority_Sign;
    else result:=C_Priority_Plus;
  end;
end;

{$ENDREGION}

{$REGION 'TToken_Minus'}

procedure TToken_Minus.Calc(Result, Paras: pointer);
  var TmpDouble,Tmp1Double: PDouble;
      CalcParas: TCalcParas;
      TmpResult: array of Double;
      i,k: integer;
begin
  Self.Paras[0].Calc(result, paras);
  CalcParas:=PCalcParas(Paras)^;
  TmpDouble:=PDouble(Result);
  if CalcParas.Mode=CalcGamma1 then begin
    if NumberOfParas=1 then begin
      for i := 0 to (CalcParas.M+1)*(CalcParas.J+1)*(CalcParas.J+1)-1 do begin
        TmpDouble^:=-TmpDouble^;
        Inc(TmpDouble,1);
      end;
    end else begin
      Setlength(TmpResult,(CalcParas.M+1)*(CalcParas.J+1)*(CalcParas.J+1));
      Self.Paras[1].Calc(TmpResult, paras);
      TmpDouble:=PDouble(Result);
      Tmp1Double:=@TmpResult[0];
      for i := low(TmpResult) to High(TmpResult) do begin
        TmpDouble^:=TmpDouble^-Tmp1Double^;
        Inc(TmpDouble,1);
        Inc(Tmp1Double,1);
      end;
    end;
  end else begin
    if NumberOfParas=1 then begin
      for i := 0 to CalcParas.I do begin
        for k := 0 to Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i) do begin
          TmpDouble^:=-TmpDouble^;
          Inc(TmpDouble,1);
        end;
        inc(TmpDouble,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
      end;
    end else begin
      Setlength(TmpResult,(CalcParas.I+1)*(CalcParas.J+1));
      Self.Paras[1].Calc(TmpResult, paras);
      Tmp1Double:=@TmpResult[0];
      for i := 0 to CalcParas.I do begin
        for k := 0 to Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i) do begin
          TmpDouble^:=TmpDouble^-Tmp1Double^;
          Inc(TmpDouble,1);
          Inc(Tmp1Double,1);
        end;
        inc(TmpDouble,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
        Inc(Tmp1Double,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
      end;
    end;
  end;
end;

constructor TToken_Minus.Create(AName: string);
begin
  inherited;
  Linear:= True;
end;

function TToken_Minus.Priority: integer;
begin
  case NumberOfParas of
    1: result:=C_Priority_Sign;
    else result:=C_Priority_Minus;
  end;
end;

{$ENDREGION}

{$REGION 'TToken_Mul'}

procedure TToken_Mul.Calc(Result, Paras: pointer);
  var TmpDouble,Tmp1Double: PDouble;
      CalcParas: TCalcParas;
      TmpResult: array of Double;
      i,k: integer;
begin
  Self.Paras[0].Calc(result, paras);
  CalcParas:=PCalcParas(Paras)^;
  if CalcParas.Mode=CalcGamma1 then begin
    Setlength(TmpResult,(CalcParas.M+1)*(CalcParas.J+1)*(CalcParas.J+1));
    Self.Paras[1].Calc(TmpResult, paras);
    TmpDouble:=PDouble(Result);
    Tmp1Double:=@TmpResult[0];
    for i := low(TmpResult) to High(TmpResult) do begin
      TmpDouble^:=TmpDouble^*Tmp1Double^;
      Inc(TmpDouble,1);
      Inc(Tmp1Double,1);
    end;
  end else begin
    Setlength(TmpResult,(CalcParas.I+1)*(CalcParas.J+1));
    Self.Paras[1].Calc(TmpResult, paras);
    TmpDouble:=PDouble(Result);
    Tmp1Double:=@TmpResult[0];
    for i := 0 to CalcParas.I do begin
      for k := 0 to Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i) do begin
        TmpDouble^:=TmpDouble^*Tmp1Double^;
        Inc(TmpDouble,1);
        Inc(Tmp1Double,1);
      end;
      inc(TmpDouble,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
      Inc(Tmp1Double,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
    end;
  end;
end;

constructor TToken_Mul.Create(AName: string);
begin
  inherited;
  Linear:= True;
end;

function TToken_Mul.Priority: integer;
begin
  result:=C_Priority_Mul;
end;

{$ENDREGION}

{$REGION 'TToken_Div'}

procedure TToken_Div.Calc(Result, Paras: pointer);
  var TmpDouble,Tmp1Double: PDouble;
      CalcParas: TCalcParas;
      TmpResult: array of Double;
      i,k: integer;
begin
  Self.Paras[0].Calc(result, paras);
  CalcParas:=PCalcParas(Paras)^;
  if CalcParas.Mode=CalcGamma1 then begin
    Setlength(TmpResult,(CalcParas.M+1)*(CalcParas.J+1)*(CalcParas.J+1));
    Self.Paras[1].Calc(TmpResult, paras);
    TmpDouble:=PDouble(Result);
    Tmp1Double:=@TmpResult[0];
    for i := low(TmpResult) to High(TmpResult) do begin
      if not((Tmp1Double^=0) and (TmpDouble^=0)) then TmpDouble^:=TmpDouble^/Tmp1Double^;
      Inc(TmpDouble,1);
      Inc(Tmp1Double,1);
    end;
  end else begin
    Setlength(TmpResult,(CalcParas.I+1)*(CalcParas.J+1));
    Self.Paras[1].Calc(TmpResult, paras);
    TmpDouble:=PDouble(Result);
    Tmp1Double:=@TmpResult[0];
    for i := 0 to CalcParas.I do begin
      for k := 0 to Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i) do begin
        if not((Tmp1Double^=0) and (TmpDouble^=0)) then TmpDouble^:=TmpDouble^/Tmp1Double^;
        Inc(TmpDouble,1);
        Inc(Tmp1Double,1);
      end;
      Inc(TmpDouble,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
      Inc(Tmp1Double,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
    end;
  end;
end;

constructor TToken_Div.Create(AName: string);
begin
  inherited;
  Linear:= True;
end;

function TToken_Div.Priority: integer;
begin
  result:=C_Priority_Div;
end;

{$ENDREGION}

{$REGION 'TToken_Power'}

procedure TToken_Power.Calc(Result, Paras: pointer);
  var TmpDouble,Tmp1Double: PDouble;
      TmpParas: TCalcParas;
      TmpResult: array of Double;
      i,k: integer;
begin
  Self.Paras[0].Calc(result, paras);
  TmpParas:=PCalcParas(Paras)^;
  Setlength(TmpResult,(TmpParas.I+1)*(TmpParas.J+1));
  Self.Paras[1].Calc(TmpResult, paras);
  TmpDouble:=PDouble(Result);
  Tmp1Double:=@TmpResult[0];
  for i := 0 to TmpParas.I do begin
    for k := 0 to Min(Min(TmpParas.J,TmpParas.LastColumn),TmpParas.LastDiag-i) do begin
      if TmpDouble^=0 then begin
        if Tmp1Double^=0 then TmpDouble^:=1
        else if Tmp1Double^<0 then TmpDouble^:=exp(ln(TmpDouble^)*Tmp1Double^);
      end else TmpDouble^:=exp(ln(TmpDouble^)*Tmp1Double^);
      Inc(TmpDouble,1);
      Inc(Tmp1Double,1);
    end;
    Inc(TmpDouble,TmpParas.J-Min(Min(TmpParas.J,TmpParas.LastColumn),TmpParas.LastDiag-i));
    Inc(TmpDouble,TmpParas.J-Min(Min(TmpParas.J,TmpParas.LastColumn),TmpParas.LastDiag-i));
  end;
end;

function TToken_Power.Priority: integer;
begin
  result:=C_Priority_Power;
end;

{$ENDREGION}

{$ENDREGION}

{$REGION 'Braces'}

{$REGION 'TToken_OpeningBrace'}

procedure TToken_OpeningBrace.Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer);
begin
  // a term has ended, so process all corresponding operators
  ProcessOPs(CalcStack, WasteStack, OPStack, Level, 0, Paras);
  // reduce the level (remember we parse backwards)
  dec(Level);
  // move the token from the token stack to the calculation stack
  CalcStack.Push(TokenStack.Pop);
  // check which function is behind this pair of braces
  if (TokenStack.Count>0) and TokenStack.Peek.InheritsFrom(TToken_Func) then begin
    // registred function found, do nothing
  end else begin
    // pair of braces = identity function
    TokenStack.Push(TToken_ID.Create(''));
    TokenStack.Peek.Level:=Level;
    TokenStack.Peek.Start:=Start;
  end;
end;

{$ENDREGION}

{$REGION 'TToken_ClosingBrace'}

procedure TToken_ClosingBrace.Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer);
begin
  // increase the level (remember we parse backwards)
  inc(Level);
  // set the Level
  Self.Level:=Level;
  // move the token from the token stack to the calculation stack
  CalcStack.Push(TokenStack.Pop);
end;

{$ENDREGION}

{$REGION 'TToken_BeginVec'}

procedure TToken_BeginVec.Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer);
begin
  // a term has ended, so process all corresponding operators
  ProcessOPs(CalcStack, WasteStack, OPStack, Level, 0, Paras);
  // reduce the level (remember we parse backwards)
  dec(Level);
  // move the token from the token stack to the calculation stack
  CalcStack.Push(TokenStack.Pop);
  // add a vector token to the token stack
  TokenStack.Push(TToken_Vector.Create(''));
  TokenStack.Peek.Level:=Level;
  TokenStack.Peek.Start:=Start;
  TokenStack.Peek.NumberOfParas:=PParseParas(Paras)^.J+1;
end;

{$ENDREGION}

{$REGION 'TToken_EndVec'}

procedure TToken_EndVec.Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer);
begin
  // increase the level (remember we parse backwards)
  inc(Level);
  // set the Level
  Self.Level:=Level;
  // move the token from the token stack to the calculation stack
  CalcStack.Push(TokenStack.Pop);
end;

{$ENDREGION}

{$ENDREGION}

{$REGION 'Separators'}

{$REGION 'TToken_FormulaSep'}

procedure TToken_FormulaSep.Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer);
begin
  // a term has ended, so process all corresponding operators
  ProcessOPs(CalcStack, WasteStack, OPStack, Level, 0, Paras);
  // check if the level equals zero (less than zero is not possible here, because this is handled by the opening brace and begin vector tag)
  if Level > 0 then begin
    // look for the next closing brace or end vector token within the calculation stack
    while (CalcStack.Count>0) and not (CalcStack.Peek.InheritsFrom(TToken_ClosingBrace) or CalcStack.Peek.InheritsFrom(TToken_EndVec)) do WasteStack.Push(CalcStack.Pop);
    if CalcStack.Count=0 then raise EParseException_SyntaxError.Create(Start+1);
    if CalcStack.Peek.InheritsFrom(TToken_ClosingBrace) then raise EParseException_MissingOpeningBrace.Create(Start+1);
    if CalcStack.Peek.InheritsFrom(TToken_EndVec) then raise EParseException_MissingBeginVector.Create(Start+1);
  end;
  // check if a complete formula exists
  if CalcStack.Count=0 then raise EParseException_MissingFormula.Create(Start+1);
  if CalcStack.Count>1 then raise EParseException_MissingFormulaSeparator.Create(CalcStack.Peek.EndPos);
  // get the formula
  FormulaList.Insert(0,CalcStack.Pop);
  // move the token from the token stack to the waste stack
  WasteStack.Push(TokenStack.Pop);
end;

{$ENDREGION}

{$REGION 'TToken_ParaSep'}

procedure TToken_ParaSep.Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer);
begin
  // a term has ended, so process all corresponding operators
  ProcessOPs(CalcStack, WasteStack, OPStack, Level, 0, Paras);
  // set the Level
  Self.Level:=Level;
  // move the token from the token stack to the calculation stack
  CalcStack.Push(TokenStack.Pop);
end;

{$ENDREGION}

{$ENDREGION}

{$REGION 'Functions'}

{$REGION 'TToken_Func'}

{$REGION 'TToken_Func protected'}

function TToken_Func.GetPara(i: integer): TToken;
begin
  result:=FParas.Items[i];
end;

procedure TToken_Func.SetPara(i: integer; const Value: TToken);
begin
  if i<FParas.Count then FParas.items[i]:=Value
                    else FParas.Add(Value);
end;

{$ENDREGION}

{$REGION 'TToken_Func public'}

procedure TToken_Func.Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>;
  OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer);
var i, j: integer;
begin
  // raise an exception if the token is not linear and only linear tokens are allowed
  if PParseParas(Paras)^.OnlyLinear and not Linear then raise EParseException_OnlyLinearAllowed.Create(Start);
  // set the level
  Self.Level:=Level;
  j:=Start+Len;
  // get the parameters (evaluable tokens separated by parameter separator token and enclosed by a pair of opening and closing braces)
  // check for opening brace
  if (CalcStack.Count<1) or not CalcStack.Peek.InheritsFrom(TToken_OpeningBrace) then raise EParseException_MissingOpeningBrace.Create(j);
  // move the brace token from the calculation stack to the waste stack
  j:=CalcStack.Peek.EndPos;
  WasteStack.Push(CalcStack.Pop);
  // get all parameters
  for i:=1 to NumberOfParas do begin
    if (CalcStack.Count<1) or not CalcStack.Peek.InheritsFrom(TToken_Evaluable) then raise EParseException_MissingParameter.Create(j);
    // get the parameter
    j:=CalcStack.Peek.EndPos;
    Self.Paras.Add(CalcStack.Pop);
    // look for a parameter separator or a closing brace
    if (i<NumberOfParas) and ((CalcStack.Count<1) or not CalcStack.Peek.InheritsFrom(TToken_ParaSep)) then raise EParseException_MissingParameterSeparator.Create(j);
    if (i=NumberOfParas) and ((CalcStack.Count<1) or not CalcStack.Peek.InheritsFrom(TToken_ClosingBrace)) then begin
      // get the previous token
      if CalcStack.Count>0 then
        j:=CalcStack.Peek.Start
      else
        j:=-1;
      raise EParseException_MissingClosingBrace.Create(j);
    end;
    // move the separator token or closing brace token from the calculation stack to the waste stack
    j:=CalcStack.Peek.EndPos;
    WasteStack.Push(CalcStack.Pop);
  end;
  // move the token from the token stack to the calculation stack
  CalcStack.Push(TokenStack.Pop);
end;

{$ENDREGION}

{$ENDREGION}

{$REGION 'TToken_Abs'}

procedure TToken_Abs.Calc(Result, Paras: pointer);
var i,k: integer;
    TmpDouble: PDouble;
    CalcParas: TCalcParas;
begin
  Self.Paras[0].Calc(result, paras);
  CalcParas:=PCalcParas(Paras)^;
  TmpDouble:=PDouble(Result);
  for i := 0 to CalcParas.I do begin
    for k := 0 to Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i) do begin
      TmpDouble^:=Abs(TmpDouble^);
      inc(TmpDouble,1);
    end;
    inc(TmpDouble,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
  end;
end;

constructor TToken_Abs.Create(AName: string);
begin
  inherited;
  NumberOfParas:=1;
end;

{$ENDREGION}

{$REGION 'TToken_ClaimProperty'}

constructor TToken_ClaimProperty.Create(AName: string);
begin
  inherited;
  NumberOfParas:=1;
  Linear:=True;
end;

procedure TToken_ClaimProperty.Parse(TokenStack, CalcStack,
  WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>;
  FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer);
begin
  inherited;
  if not Para[0].InheritsFrom(TToken_Number) then raise EParseException_NumberExpected.Create(Para[0].Start);
  if (TToken_Number(Para[0]).Value<0) or (TToken_Number(Para[0]).Value>PParseParas(Paras)^.M) then raise EParseException_ClaimPropertyOutOfRange.Create(Para[0].Start);
end;

procedure TToken_ClaimProperty.Calc(Result, Paras: pointer);
var l,i,k: integer;
    TmpDouble, SDouble, InfDouble: PDouble;
    CalcParas: TCalcParas;
begin
  CalcParas:=PCalcParas(Paras)^;
  if CalcParas.Mode=CalcGamma1 then begin
    TmpDouble:=PDouble(Result);
    l:=Round(TToken_Number(Para[0]).Value);
    CalcParas.Dep_ml.AddDependsOn(CalcParas.m1,l);
    inc(TmpDouble,l*(CalcParas.J+1)*(CalcParas.J+1));
    for k := 0 to CalcParas.J do begin
      TmpDouble^:=1;
      inc(TmpDouble,CalcParas.J+2);
    end;
  end else begin
    SDouble:=CalcParas.ClaimProperty.S;
    InfDouble:=CalcParas.ClaimProperty.V;
    if InfDouble=nil then begin
      // no inflation
      inc(SDouble,Round(TToken_Number(Para[0]).Value*(CalcParas.I+1)*(CalcParas.ClaimProperty.J+1)));
      TmpDouble:=PDouble(Result);
      for i := 0 to CalcParas.I do begin
        for k := 0 to Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i) do begin
          TmpDouble^:=SDouble^;
          inc(TmpDouble,1);
          Inc(SDouble,1);
        end;
        inc(TmpDouble,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
        inc(SDouble,CalcParas.ClaimProperty.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
      end;
    end else begin
      // inflation
      inc(SDouble,Round(TToken_Number(Para[0]).Value*(CalcParas.I+1)*(CalcParas.ClaimProperty.J+1)));
      inc(InfDouble,Round(TToken_Number(Para[0]).Value*(CalcParas.I+1)*(CalcParas.ClaimProperty.J+1)));
      TmpDouble:=PDouble(Result);
      for i := 0 to CalcParas.I do begin
        for k := 0 to Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i) do begin
          if SDouble^=0 then TmpDouble^:=0 else TmpDouble^:=SDouble^/InfDouble^;
          inc(TmpDouble,1);
          Inc(SDouble,1);
          Inc(InfDouble,1);
        end;
        inc(TmpDouble,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
        inc(SDouble,CalcParas.ClaimProperty.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
        inc(InfDouble,CalcParas.ClaimProperty.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
      end;
    end;
  end;
end;

{$ENDREGION}

{$REGION 'TToken_Column'}

procedure TToken_Column.Calc(Result, Paras: pointer);
var i,k,j,m,v: integer;
    TmpDouble: PDouble;
    CalcParas: TCalcParas;
begin
  Self.Paras[1].Calc(result, paras);
  CalcParas:=PCalcParas(Paras)^;
  TmpDouble:=PDouble(Result);
  j:=Round(TToken_Number(Para[0]).Value);
  if CalcParas.Mode=CalcGamma1 then begin
    for m := 0 to CalcParas.M do begin
      for k := 0 to j-1 do begin
        for v := 0 to CalcParas.J do begin
          TmpDouble^:=0;
          inc(TmpDouble,1);
        end;
      end;
      Inc(TmpDouble,CalcParas.J+1);
      for k := j+1 to CalcParas.J do begin
        for v := 0 to CalcParas.J do begin
          TmpDouble^:=0;
          inc(TmpDouble,1);
        end;
      end;
    end;
  end else begin
    for i := 0 to CalcParas.I do begin
      for k := 0 to Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i) do begin
        if j<>k then TmpDouble^:=0;
        inc(TmpDouble,1);
      end;
      inc(TmpDouble,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
    end;
  end;
end;

constructor TToken_Column.Create(AName: string);
begin
  inherited;
  NumberOfParas:=2;
  Linear:=True;
end;

procedure TToken_Column.Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>; OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer);
begin
  inherited;
  if Not Para[0].InheritsFrom(TToken_Number) then raise EParseException_NumberExpected.Create(Para[0].Start);
end;

{$ENDREGION}

{$REGION 'TToken_Cum'}

procedure TToken_Cum.Calc(Result, Paras: pointer);
var m,i,k,j: integer;
    TmpDouble, Tmp1Double: PDouble;
    CalcParas: TCalcParas;
begin
  Self.Paras[0].Calc(result, paras);
  CalcParas:=PCalcParas(Paras)^;
  if CalcParas.Mode=CalcGamma1 then begin
    for m := 0 to CalcParas.M do begin
      for j := 0 to CalcParas.J-1 do begin
        TmpDouble:=PDouble(Result);
        inc(TmpDouble,m*(CalcParas.J+1)*(CalcParas.J+1)+j);
        Tmp1Double:=TmpDouble;
        inc(Tmp1Double,CalcParas.J+1);
        for k := 0 to CalcParas.J-1 do begin
          Tmp1Double^:=TmpDouble^+Tmp1Double^;
          inc(TmpDouble,CalcParas.J+1);
          inc(Tmp1Double,CalcParas.J+1);
        end;
      end;
    end;
  end else begin
    TmpDouble:=PDouble(Result);
    Tmp1Double:=PDouble(Result);
    for i := 0 to CalcParas.I do begin
      inc(TmpDouble,1);
      for k := 1 to Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i) do begin
        TmpDouble^:=TmpDouble^+Tmp1Double^;
        inc(TmpDouble,1);
        inc(Tmp1Double,1);
      end;
      inc(TmpDouble,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
      inc(Tmp1Double,CalcParas.J+1-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
    end;
  end;
end;

constructor TToken_Cum.Create(AName: string);
begin
  inherited;
  NumberOfParas:=1;
  Linear:=True;
end;

{$ENDREGION}

{$REGION 'TToken_ID'}

procedure TToken_ID.Calc(Result, Paras: pointer);
begin
  Self.Paras[0].Calc(result, paras);
end;

constructor TToken_ID.Create(AName: string);
begin
  inherited;
  NumberOfParas:=1;
  Linear:=True;
end;

procedure TToken_ID.Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>;
  OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>;
  var Level: integer; Paras: Pointer);
begin
  inherited;
  // if only linear token are allowed then constants (number or vector) enclosed in brackets are forbidden
  if PParseParas(Paras)^.OnlyLinear
    and (Self.Paras[0].InheritsFrom(TToken_Number)
        or Self.Paras[0].InheritsFrom(TToken_Vector)) then raise EParseException_OnlyLinearAllowed.Create(Start);
end;

{$ENDREGION}

{$REGION 'TToken_Inc'}

procedure TToken_Inc.Calc(Result, Paras: pointer);
var m,i,k,j: integer;
    TmpDouble, Tmp1Double: PDouble;
    CalcParas: TCalcParas;
begin
  Self.Paras[0].Calc(result, paras);
  CalcParas:=PCalcParas(Paras)^;
  if CalcParas.Mode=CalcGamma1 then begin
    for m := 0 to CalcParas.M do begin
      for j := 0 to CalcParas.J-1 do begin
        TmpDouble:=PDouble(Result);
        inc(TmpDouble,(m*(CalcParas.J+1)+CalcParas.J)*(CalcParas.J+1)+j);
        Tmp1Double:=TmpDouble;
        dec(Tmp1Double,CalcParas.J+1);
        for k := CalcParas.J downto 1 do begin
          TmpDouble^:=TmpDouble^-Tmp1Double^;
          dec(TmpDouble,CalcParas.J+1);
          dec(Tmp1Double,CalcParas.J+1);
        end;
      end;
    end;
  end else begin
    TmpDouble:=PDouble(Result);
    // goto the last element
    Inc(TmpDouble,(CalcParas.J+1)*(CalcParas.I+1)-1);
    for i := CalcParas.I downto 0 do begin
      dec(TmpDouble,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
      Tmp1Double:=PDouble(TmpDouble);
      dec(Tmp1Double,1);
      for k := Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i) downto 1 do begin
        TmpDouble^:=TmpDouble^-Tmp1Double^;
        dec(TmpDouble,1);
        dec(Tmp1Double,1);
      end;
      dec(TmpDouble,1);
    end;
  end;
end;

constructor TToken_Inc.Create(AName: string);
begin
  inherited;
  NumberOfParas:=1;
  Linear:=True;
end;

{$ENDREGION}

{$REGION 'TToken_Max'}

procedure TToken_Max.Calc(Result, Paras: pointer);
  var TmpDouble,Tmp1Double: PDouble;
      CalcParas: TCalcParas;
      TmpResult: array of Double;
      i,k: integer;
begin
  Self.Paras[0].Calc(result, paras);
  CalcParas:=PCalcParas(Paras)^;
  Setlength(TmpResult,(CalcParas.I+1)*(CalcParas.J+1));
  Self.Paras[1].Calc(TmpResult, paras);
  TmpDouble:=PDouble(Result);
  Tmp1Double:=@TmpResult[0];
  for i := 0 to CalcParas.I do begin
    for k := 0 to Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i) do begin
      TmpDouble^:=Max(TmpDouble^,Tmp1Double^);
      Inc(TmpDouble,1);
      Inc(Tmp1Double,1);
    end;
    Inc(TmpDouble,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
    Inc(Tmp1Double,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
  end;
end;

constructor TToken_Max.Create(AName: string);
begin
  inherited;
  NumberOfParas:=2;
end;

{$ENDREGION}

{$REGION 'TToken_Min'}

procedure TToken_Min.Calc(Result, Paras: pointer);
  var TmpDouble,Tmp1Double: PDouble;
      CalcParas: TCalcParas;
      TmpResult: array of Double;
      i,k: integer;
begin
  Self.Paras[0].Calc(result, paras);
  CalcParas:=PCalcParas(Paras)^;
  Setlength(TmpResult,(CalcParas.I+1)*(CalcParas.J+1));
  Self.Paras[1].Calc(TmpResult, paras);
  TmpDouble:=PDouble(Result);
  Tmp1Double:=@TmpResult[0];
  for i := 0 to CalcParas.I do begin
    for k := 0 to Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i) do begin
      TmpDouble^:=Min(TmpDouble^,Tmp1Double^);
      Inc(TmpDouble,1);
      Inc(Tmp1Double,1);
    end;
    Inc(TmpDouble,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
    Inc(Tmp1Double,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
  end;
end;

constructor TToken_Min.Create(AName: string);
begin
  inherited;
  NumberOfParas:=2;
end;

{$ENDREGION}

{$REGION 'TToken_Sqrt'}

procedure TToken_Sqrt.Calc(Result, Paras: pointer);
var i,k: integer;
    TmpDouble: PDouble;
    CalcParas: TCalcParas;
begin
  Self.Paras[0].Calc(result, paras);
  CalcParas:=PCalcParas(Paras)^;
  TmpDouble:=PDouble(Result);
  for i := 0 to CalcParas.I do begin
    for k := 0 to Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i) do begin
      TmpDouble^:=SQRT(TmpDouble^);
      inc(TmpDouble,1);
    end;
    Inc(TmpDouble,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
  end;
end;

constructor TToken_Sqrt.Create(AName: string);
begin
  inherited;
  NumberOfParas:=1;
end;

{$ENDREGION}

{$ENDREGION}

{$REGION 'EParseException'}

constructor EParseException.Create(i: integer);
begin
  Start:=i;
end;

{$ENDREGION}

{$REGION 'TToken_Vector'}

procedure TToken_Vector.Calc(Result, Paras: pointer);
  var m, i, j, k: integer;
      TmpDouble: PDouble;
      CalcParas: TCalcParas;
begin
  TmpDouble:=Result;
  CalcParas:=PCalcParas(Paras)^;
  if CalcParas.Mode=CalcGamma1 then begin
    for m := 0 to CalcParas.M do begin
      for k := 0 to CalcParas.J do begin
        for j := 0 to CalcParas.J do begin
          TmpDouble^:=FValues[k];
          inc(TmpDouble,1);
        end;
      end;
    end;
  end else begin
    TmpDouble:=PDouble(Result);
    for i := 0 to CalcParas.I do begin
      for k := 0 to Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i) do begin
        TmpDouble^:=FValues[k];
        inc(TmpDouble,1);
      end;
      inc(TmpDouble,CalcParas.J-Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i));
    end;
  end;
//  TmpDouble:=Result;
//  CalcParas:=PCalcParas(Paras)^;
//  for i := 0 to PCalcParas(Paras)^.I do begin
//    Move(FValues[0],TmpDouble^,(Min(Min(CalcParas.J,CalcParas.LastColumn),CalcParas.LastDiag-i)+1)*SizeOf(Double));
//    inc(TmpDouble,CalcParas.I+1);
//  end;
end;

constructor TToken_Vector.Create(AName: string);
begin
  inherited;
  Linear:=True;
end;

procedure TToken_Vector.Parse(TokenStack, CalcStack, WasteStack: TStack<TToken>;
  OPStack: TStack<TToken_Operator>; FormulaList: TObjectList<TToken>; var Level: integer; Paras: Pointer);
var i, j: integer;
begin
  // set the level
  Self.Level:=Level;
  // get the parameters (evaluable tokens separated by parameter separator token and enclosed by a pair of begin and end vector token)
  // since a vector token is created by a begin vector token we know that there is a begin vector token on the calculation stack
  // move the begin vector token from the calculation stack to the waste stack
  WasteStack.Push(CalcStack.Pop);
  // get all parameters
  setlength(FValues,NumberOfParas);
  j:=Start+Len+1;
  for i:=1 to NumberOfParas do begin
    if (CalcStack.Count<1) then raise EParseException_MissingParameter.Create(j);
    if not CalcStack.Peek.InheritsFrom(TToken_Number) then raise EParseException_NumberExpected.Create(CalcStack.Peek.Start);
    j:=CalcStack.Peek.Start+CalcStack.Peek.Len;
    // get the parameter
    FValues[i-1]:=TToken_Number(CalcStack.Peek).Value;
    // move the parameter token from the calculation stack to the waste stack
    WasteStack.Push(CalcStack.Pop);
    // look for a parameter separator or a end vector token
    if (i<NumberOfParas) then
      if (CalcStack.Count<1) then
        raise EParseException_MissingParameterSeparator.Create(j)
      else if not CalcStack.Peek.InheritsFrom(TToken_ParaSep) then
        raise EParseException_MissingParameterSeparator.Create(CalcStack.Peek.Start);
    if (i=NumberOfParas) then
      if (CalcStack.Count<1) then
        raise EParseException_MissingEndVector.Create(j)
      else if not CalcStack.Peek.InheritsFrom(TToken_EndVec) then
        raise EParseException_MissingEndVector.Create(CalcStack.Peek.Start);
    j:=CalcStack.Peek.Start+CalcStack.Peek.Len;
    // move the separator token or end vector token from the calculation stack to the waste stack
    WasteStack.Push(CalcStack.Pop);
  end;
  // move the token from the token stack to the calculation stack
  CalcStack.Push(TokenStack.Pop);
end;

{$ENDREGION}

end.
