unit LSRM_ActiveX;

//Copyright 2013-2017 Ren Dahms (rene.dahms at math.ethz.ch)
//
//This file is part of LSRMTools.
//
//    LSRMTools is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    LSRMTools is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with LSRMTools.  If not, see <http://www.gnu.org/licenses/>.
//
//LSRMTools use some parts of ALGLIB which is a free software distributed
//under a GPL license - version 2 or (at your option) any later version.

{$WARN SYMBOL_PLATFORM OFF}

interface

uses
  ComObj, ActiveX, LSRMTools_TLB, StdVcl, Classes, LSRM;

{$REGION 'Description ...'}
// The ActiveX component, that handes all the communication regarding LSRMs to thze outside world.
// Before it can be used it has to be registered.
{$ENDREGION}
type
  TLSRM = class(TAutoObject, ILSRM)
  private {$REGION}
    // a list of LSRMs, addressed by their names
    FLSRMs:TStringList;
    // returns the LSRM with name LSRMName if it exists within the list FLSRMs
    function GetLSRM(LSRMName: OleVariant): TNormalLSRM;
    {$ENDREGION}
  protected {$REGION}
    {$REGION 'Description ...'}
    // returns a message about version, licence and author of the tool
    {$ENDREGION}
    function Get_About: OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the exposure R^m_ik for the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_R1(LSRMName: OleVariant; m, i, k: Integer): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the exposure R^m1m2_ik for the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_R2(LSRMName: OleVariant; m1, m2, i, k: Integer): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the claim property S^m_ik for the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_S(LSRMName: OleVariant; m, i, k: Integer): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the used weight w^m_ik for the LSRM with name LSRMName, where n means
    //   0: not normalised
    //   1: w^m,I_ik
    //   2: w^m,I+1_ik
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_w(LSRMName: OleVariant; n, m, i, k: Integer): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the used mixing alpha^m_i for the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_alpha(LSRMName: OleVariant; m, i: Integer): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns information about the initialisation step Step for the LSRM with name LSRMName, where InfoType
    // means:
    //   0: the short description (or name) of the initialisation step
    //   1: if or if not the step has been initialised
    //   2: how long the initialisation lasts
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_CalcInfo(LSRMName: OleVariant; Step, InfoType: Integer): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the one period uncertainty CDR for the LSRM with name LSRMName, where
    //   Proxy means:
    //     0: no proxy
    //     1: proxy
    //   Part means:
    //     0: process variance
    //     1: parameter error
    //     2: sum of both
    //   Tail mean:
    //     0: without tail
    //     1: with tail
    //     2: tail only
    //   if the specified accident period i is out of range the overall CDR will be returned
    //   if the claim property index m is out of range the alpha-mixed CDR will be returned
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_CDR(LSRMName: OleVariant; M, I, Proxy, Part, Tail: Integer): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the addend for indices m1, m2, i1 and i2 of the approximated calculation of the
    // one period uncertainty CDR for the LSRM with name LSRMName, where
    //   Proxy means:
    //     0: no proxy
    //     1: proxy
    //   Part means:
    //     0: process variance
    //     1: parameter error
    //   Tail mean:
    //     0: without tail
    //     1: with tail
    //     2: tail only
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_CDRDetail(LSRMName: OleVariant; m1, m2, i1, i2, Proxy, Part, Tail: Integer): OleVariant;
      safecall;
    {$REGION 'Description ...'}
    // returns the used development factors w^m_k for the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_f(LSRMName: OleVariant; m, k: Integer): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the default (i.e. calculated with the default weights) development factors w^m_k for the
    // LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_Defaultf(LSRMName: OleVariant; m, k: Integer): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the ultimate uncertainty MSEP for the LSRM with name LSRMName, where
    //   Proxy means:
    //     0: no proxy
    //     1: proxy
    //   Part means:
    //     0: process variance
    //     1: parameter error
    //     2: sum of both
    //   Tail mean:
    //     0: without tail
    //     1: with tail
    //     2: tail only
    //   if the specified accident period i is out of range the overall CDR will be returned
    //   if the claim property index m is out of range the alpha-mixed CDR will be returned
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_MSEP(LSRMName: OleVariant; M, I, Proxy, Part, Tail: Integer): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the addend for indices m1, m2, i1 and i2 of the approximated calculation of the
    // ultimate uncertainty MSEP for the LSRM with name LSRMName, where
    //   Proxy means:
    //     0: no proxy
    //     1: proxy
    //   Part means:
    //     0: process variance
    //     1: parameter error
    //   Tail mean:
    //     0: without tail
    //     1: with tail
    //     2: tail only
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_MSEPDetail(LSRMName: OleVariant; m1, m2, i1, i2, Proxy, Part, Tail: Integer): OleVariant;
      safecall;
    {$REGION 'Description ...'}
    // returns the used variance parameter sigma^m1m2_k for the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_sigma(LSRMName: OleVariant; m1, m2, k: Integer): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the estimated ultimate outcome for the LSRM with name LSRMName, where
    //   Tail mean:
    //     0: without tail
    //     1: with tail
    //   if the specified accident period i is out of range the overall estimated ultimate will be returned
    //   if the claim property index m is out of range the alpha-mixed estimated ultimate will be returned
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_Ultimate(LSRMName: OleVariant; M, I, Tail: Integer): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the default (which lead to the variance minimizing development factors) weight w^m_ik
    // for the LSRM with name LSRMName, where n means
    //   0: not normalised
    //   1: w^m,I_ik
    //   2: w^m,I+1_ik
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_Defaultw(LSRMName: OleVariant; n, m, i, k: Integer): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the deafult mixing alpha^m_i (defined as 1) for the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_Defaultalpha(LSRMName: OleVariant; M, I: Integer): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the number of accident periods for the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_I(LSRMName: OleVariant): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns which parameters have been initialised by procedure Init for the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_InitialisedParameters(LSRMName: OleVariant): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the number of development periods (without tail) for the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_J(LSRMName: OleVariant): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the number of claim properties for the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_M(LSRMName: OleVariant): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the default (i.e. calculated with the default weights) variance parameter sigma^m1m2_k
    // for the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_Defaultsigma(LSRMName: OleVariant; m1, m2, k: Integer): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns if the default mixing-alphas have been used for the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_DefaultalphaUsed(LSRMName: OleVariant): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns if the default development factors have been used for the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_DefaultfUsed(LSRMName: OleVariant): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns if the default variance parameter have been used for the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_DefaultsigmaUsed(LSRMName: OleVariant): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns if the default weights have been used for the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_DefaultwUsed(LSRMName: OleVariant): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the coordinates (F^m_ik)^l_hj of the operator F for the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_CoorF(LSRMName: OleVariant; m, i, k, l, h, j: Integer): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the number of future development periods for the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_K(LSRMName: OleVariant): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the type (normal, simple or simple with inflation) of the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_ModelType(LSRMName: OleVariant): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns with calculation flags have been set for the LSRM with name LSRMName, i.e.
    //   1: reduced memory usage
    //   2: no virtual process variance diversification
    //   4: no virtual parameter error diversification
    //   8: no virtual tail diversification
    // if no such LSRM exists within the list FLSRMs a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_CalculationFlags(LSRMName: OleVariant): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the m-th largest eigenvalue of the covariance matrix (sigma^m1m2_k*R^m1m2_ik)
    // for the LSRM with name LSRMName
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_Eigenvalue(LSRMName: OleVariant; i, k, m: Integer): OleVariant; safecall;
    function Get_Eigenvector(LSRMName: OleVariant; i, k, m, l: Integer): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the covariance matrix corresponding to the one period uncertainty CDR,
    // i.e. the sum over i1 and i2 of FCDRDetail or FCDRTailDetail
    //   Proxy means:
    //     0: without approximation
    //     1: with approximation
    //   Part means:
    //     0: process variance
    //     1: parameter error
    //     2: the sum of both
    //   Tail mean:
    //     0: without tail
    //     1: with tail
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_CDRCov(LSRMName: OleVariant; m1, m2, Proxy, Part, Tail: Integer): OleVariant; safecall;
    {$REGION 'Description ...'}
    // returns the covariance matrix corresponding to the ultimate uncertainty MSEP,
    // i.e. the sum over i1 and i2 of FMSEPDetail or FMSEPTailDetail
    //   Proxy means:
    //     0: without approximation
    //     1: with approximation
    //   Part means:
    //     0: process variance
    //     1: parameter error
    //     2: the sum of both
    //   Tail mean:
    //     0: without tail
    //     1: with tail
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_MSEPCov(LSRMName: OleVariant; m1, m2, Proxy, Part, Tail: Integer): OleVariant; safecall;
    {$REGION 'Description...'}
    // returns the m-th largest eigenvalue of the covariance matrix corresponding to the one period
    // uncertainty CDR, i.e. the sum over i1 and i2 of FCDRDetail or FCDRTailDetail
    //   Proxy means:
    //     0: without approximation
    //     1: with approximation
    //   Part means:
    //     0: process variance
    //     1: parameter error
    //     2: the sum of both
    //   Tail mean:
    //     0: without tail
    //     1: with tail
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_CDRCovEigenvalue(LSRMName: OleVariant; m, Proxy, Part, Tail: Integer): OleVariant; safecall;
    {$REGION 'Description...'}
    // returns l-th coordinate of the eigenvector corresponding to the m-th largest eigenvalue of the
    // covariance matrix corresponding to the one period uncertainty CDR,
    // i.e. the sum over i1 and i2 of FCDRDetail or FCDRTailDetail
    //   Proxy means:
    //     0: without approximation
    //     1: with approximation
    //   Part means:
    //     0: process variance
    //     1: parameter error
    //     2: the sum of both
    //   Tail mean:
    //     0: without tail
    //     1: with tail
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_CDRCovEigenvector(LSRMName: OleVariant; m, l, Proxy, Part, Tail: Integer): OleVariant;
      safecall;
    {$REGION 'Description...'}
    // returns the m-th largest eigenvalue of the covariance matrix corresponding to the ultimate
    // uncertainty MSEP, i.e. the sum over i1 and i2 of FCDRDetail or FCDRTailDetail
    //   Proxy means:
    //     0: without approximation
    //     1: with approximation
    //   Part means:
    //     0: process variance
    //     1: parameter error
    //     2: the sum of both
    //   Tail mean:
    //     0: without tail
    //     1: with tail
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_MSEPCovEigenvalue(LSRMName: OleVariant; m, Proxy, Part, Tail: Integer): OleVariant; safecall;
    {$REGION 'Description...'}
    // returns l-th coordinate of the eigenvector corresponding to the m-th largest eigenvalue of the
    // covariance matrix corresponding to the ultimate uncertainty MSEP,
    // i.e. the sum over i1 and i2 of FCDRDetail or FCDRTailDetail
    //   Proxy means:
    //     0: without approximation
    //     1: with approximation
    //   Part means:
    //     0: process variance
    //     1: parameter error
    //     2: the sum of both
    //   Tail mean:
    //     0: without tail
    //     1: with tail
    // if no such LSRM exists within the list FLSRMs or if at least one index is out of range a
    // corresponding error message will be returned
    {$ENDREGION}
    function Get_MSEPCovEigenvector(LSRMName: OleVariant; m, l, Proxy, Part, Tail: Integer): OleVariant;
      safecall;
    {$REGION 'Description...'}
    // in order to avoid numerical artefacts during the normalisation of the covariance parameters
    // we take sigma^m1,m_2:=0 if sigma^m_1,m_2>Abs(f^m1*f^m2)*MaxVarCoef1+MaxVarCoef2
    {$ENDREGION}
    function Get_MaxVarCoef1: Double; safecall;
    function Get_MaxVarCoef2: Double; safecall;
    procedure Set_MaxVarCoef1(Value: Double); safecall;
    procedure Set_MaxVarCoef2(Value: Double); safecall;
    {$REGION 'Descriptions...'}
    // gives control about the decimal delimiter, the formula delimiter and the parameter delimiter,
    // used to specify the formulas for exposures of LSRMs,
    // in terms of a string containing those in the specified order.
    {$ENDREGION}
    function Get_Delimiters: OleVariant; safecall;
    procedure Set_Delimiters(Value: OleVariant); safecall;
    {$ENDREGION}
  public {$REGION}
    // initialises the list
    procedure Initialize; override;
    // destroys the list and all contained LSRMs
    destructor Destroy; override;
    {$REGION 'Description ...'}
    // adds or reinitialises a LSRM with name LSRMName to the list
    // What specifies the objects that should be initialised:
    //   cLSRMInitGamma1:         exposures gamma^ml_ikhj, the corresponding values are specified within the
    //                            parameter Gamma1, which is either an array of the following columns
    //                              m, l, i, k, h, j, value of gamma^ml_ikhj,
    //                            where the columns i and h can be omitted in case of a simple LSRM,
    //                            or a string of semicolon separated formulas for each R^m_ik that
    //                            represents a triangle wise calculation of the exposure.
    //                            We support the following operations:
    //                              - Signs + and -
    //                              - operators +, -, *, / and ^ (power; not allowed for gamma1)
    //                              - constant real numbers
    //                              - vectors: a collection of FJ+FK (or FJ+FP) real numbers separated by
    //                                commas and enclosed within square brackets
    //                                this values are values for each column
    //                              - pais of round braces: priorisise the enclosed term to be calculated first
    //                              - Function:
    //                                - S(m): Claim property where m is an integer with 0<=m<=FM,
    //                                - Cum(triangle): cummulation of triangle
    //                                - Inc(triangle): incremental of triangle
    //                                - Abs(triangle): absolut value of each component of triangle (not allowed for gamma1)
    //                                - Sqrt(triangle): square root of each component of triangle (not allowed for gamma1)
    //                                - Column(k, triangle): empties all columns, except for column k, of the trinagle
    //                                - Max(A,B): takes the maximum of triangles A and B, componenten wise (not allowed for gamma1)
    //                                - Min(A,B): takes the mminimum of triangles A and B, componenten wise (not allowed for gamma1)
    //                            Examples for formulas of R^m_ik are:
    //                              - "S(0)": R^m_ik = S^0_ik
    //                              - "2*S(0)-S(1)": R^m_ik = 2*S^0_ik - S^1_ik
    //                              - "Cum(S(0))": R^m_ik = Cum(S^0_ik), i.e. the Chain Ladder exposure
    //                              - "Cum(S(1))-Cum(S(0))": R^m_ik = Cum(S^1_ik)-Cum(S^0_ik),
    //                                if S^1 are reported amounts and S^0 are payments then R^m_ik
    //                                represent the corresponding case reserves.
    //                              - "[1,0.5,0,0,0]*Cum(S(1))-[1,0.5,0,0,0]*Cum(S(0))+[0,0.5,1,1,1]*Cum(S(1))":
    //                                if the total number of development periods (FJ+FK+1) equals 5 and if S^1
    //                                are reported amounts and S^0 are payments then the formula represents a exposure which equals:
    //                                  - case reserves for the first development period
    //                                  - a 50%-50% mixture of case reserves and reported amounts for the second development period
    //                                  - reported amounts for the last three development periods (Chain-Ladder on reported amounts)
    //                              - Sqrt(Max(0,Cum(S(0)))*Max(0,Cum(S(1))))): a good choise for the covariance coupling of two Chain Ladders
    //                            NOTE: The used decimal delimiters (defined by the regional setting), the formula delimiter (;) and the
    //                                  parameter delimiter (,) can be changed by the property Delimiters.
    //   cLSRMInitGamma2:         exposures gamma^m1m2l_ikhj, the corresponding values are specified within
    //                            the parameter Gamma2, which is either an array of the following columns
    //                              m1, m2, l, i, k, h, j, value of gamma^m1m2l_ikhj,
    //                            where the columns i and h can be omitted in case of a simple LSRM,
    //                            or a string of semicolon separated formulas for each R^m1m2_ik, see Gamma1.
    //   cLSRMInitS:              claim properties S^m_ik, the corresponding values are specified within the
    //                            parameter S, which is a pointer to the first value of an array of real
    //                            numbers (Double) indexed by m, i and k.
    //   cLSRMInitw:              weights w^m_ik (for both periods I and I+1 without normalisation), the
    //                            corresponding values are specified within the
    //                            parameter w, which is a pointer to the first value of an array of real
    //                            numbers (Double) indexed by m, i and k.
    //   cLSRMUseDefaultw:        default weights should be used
    //   cLSRMInitf:              development factors f^m_k, the corresponding values are specified within the
    //                            parameter f, which is a pointer to the first value of an array of real
    //                            numbers (Double) indexed by m and k.
    //   cLSRMUseDefaultf:        default development factors should be used
    //   cLSRMInitsigma:          variance parameter sigma^m1m2_k, the corresponding values are specified
    //                            within the parameter sigma, which is a pointer to the first value of an
    //                            array of real numbers (Double) indexed by m, i and k.
    //   cLSRMUseDefaultsigma:    default variance parameter should be used
    //   cLSRMInitalpha:          mixing-alphas alpha^m_i, the corresponding values are specified within the
    //                            parameter alpha, which is a pointer to the first value of an array of real
    //                            numbers (Double) indexed by m and i.
    //   cLSRMUseDefaultalpha:    default mixing-alphas should be used
    //   cLSRMInitGeometry:       geometry, i.e. number of claim properties M, number of accident periods I
    //                            number of development periods J and number of future (or tail) periods K
    //   cLSRMInitInflationRate:  inflation rate, the corresponding values are specified within the
    //                            parameter S, which is a pointer to the first value of an array of real
    //                            numbers (Double) indexed by m, i and k.
    //   cLSRMReduceMemSize:      calculation should be restricted such that less memory is needed
    //   cLSRMOverwriteExistingLSRM:
    //                            if the LSRM with name LSRMName allready exists it should be overwritten
    //   cLSRMNoVirtualTailDiversification:
    //                            do not allow virtual diversification within the calculation of tail
    //                            uncertainties
    //   cLSRMNoVirtualProcVarDiversification:
    //                            do not allow virtual diversification within the calculation of
    //                            process variance uncertainties (without tails)
    //   cLSRMNoVirtualParaErrDiversification:
    //                            do not allow virtual diversification within the calculation of
    //                            parameter error uncertainties (without tails)
    //   cLSRMUseWeightsForDefaultSigma:
    //                            use weights in order to estimate the covariance parameters,
    //                            otherwise Mack's formula is taken
    {$ENDREGION}
    function Init(LSRMName: OleVariant; What, M, I, J, FuturePeriods: Integer; const Gamma1,
          Gamma2: OleVariant; var S, w, f, sigma, alpha, InflationRate: Double): OleVariant; safecall;
    // deletes the LSRM with name LSRMName
    // if LSRMName is an empty string the whole list of LSRMs will be emptied
    procedure Reset(LSRMName: OleVariant); safecall;
    {$ENDREGION}
end;

implementation

uses ComServ, SysUtils, Variants, LSRMTools_RS, Windows, Consts, Math, StrUtils;

{$REGION 'TLSRM'}

{$REGION '  TLSRM private'}

// returns the LSRM with name LSRMName if it exists within the list
function TLSRM.GetLSRM(LSRMName: OleVariant): TNormalLSRM;
var Index: integer;
begin
  Result:=nil;
  try
    Index:=FLSRMs.IndexOf(LSRMName);
    if Index<0 then Exit;
    Result:=TNormalLSRM(FLSRMs.Objects[Index]);
  except
  end;
end;

{$ENDREGION}

{$REGION '  TLSRM protected'}

function TLSRM.Get_About: OleVariant;
begin
  Result:=rsLSRMVersion + sLineBreak + sLineBreak + rsLSRMLicense;
end;

function TLSRM.Get_alpha(LSRMName: OleVariant; m, i: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the used alpha^m_i
  try
    Result:=LSRM.Usedalpha[m,i];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_CalcInfo(LSRMName: OleVariant; Step, InfoType: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the calculation information
  try
    case InfoType of
      0: begin
        Result:=LSRM.CalcInfoName[Step];
      end;
      1: Result:=LSRM.CalcInfo[Step];
      2: Result:=LSRM.CalcTime[Step];
      else
        Result:=rsLSRM_IndexOutOfRange;
    end;
  except
    on E:ELSRMRangeError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_CalculationFlags(LSRMName: OleVariant): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:=0;
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the set calculation flags
  if calc_ReduceMemSize in LSRM.CalculationFlags then Result:=Result+cLSRMReduceMemSize;
  if calc_VirtualProcVarDiversification in LSRM.CalculationFlags then Result:=Result+cLSRMVirtualProcVarDiversification;
  if calc_VirtualParaErrDiversification in LSRM.CalculationFlags then Result:=Result+cLSRMVirtualParaErrDiversification;
  if calc_VirtualTailDiversification in LSRM.CalculationFlags then Result:=Result+cLSRMVirtualTailDiversification;
  if calc_PreventPastExposureFromChangingSign in LSRM.CalculationFlags then Result:=Result+cLSRMPreventPastRFromChangingSign;
  if calc_PreventFutureExposureFromChangingSign in LSRM.CalculationFlags then Result:=Result+cLSRMPreventFutureRFromChangingSign;
  if calc_UseWeightsForDefaultSigma in LSRM.CalculationFlags then Result:=Result+cLSRMUseWeightsForDefaultSigma;
  if calc_UseMSEPifTailCDRisLarger in LSRM.CalculationFlags then Result:=Result+cLSRMUseMSEPifTailCDRisLarger;
end;

function TLSRM.Get_CDR(LSRMName: OleVariant; M, I, Proxy, Part, Tail: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the one period uncertainty
  try
    Result:=LSRM.CDR[m,i,Proxy,Part,Tail];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_CDRDetail(LSRMName: OleVariant; m1, m2, i1, i2, Proxy, Part, Tail: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the detail of the one period uncertainty
  try
    Result:=LSRM.CDRDetail[m1,m2,i1,i2,Proxy,Part,Tail];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_CDRCov(LSRMName: OleVariant; m1, m2, Proxy, Part, Tail: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return (m1, m2) entry of the coveriance matrix corresponding to the one period uncertainty
  try
    Result:=LSRM.CDRCov[m1,m2,Proxy,Part,Tail];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_CDRCovEigenvalue(LSRMName: OleVariant; m, Proxy, Part, Tail: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the m-th largest eigenvalue of the coveriance matrix corresponding to the one period uncertainty
  try
    Result:=LSRM.CDRCovEigenvalue[m,Proxy,Part,Tail];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_CDRCovEigenvector(LSRMName: OleVariant; m, l, Proxy, Part, Tail: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return l-th coordinate of the eigenvector corresponding to the m-th largest eigenvalue of the coveriance
  // matrix corresponding to the one period uncertainty
  try
    Result:=LSRM.CDRCovEigenvector[m,l,Proxy,Part,Tail];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_CoorF(LSRMName: OleVariant; m, i, k, l, h, j: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the coordinates (F^m_ik)^l_hj of the operator F
  try
    Result:=LSRM.CoorF[m,i,k,l,h,j];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_Defaultalpha(LSRMName: OleVariant; M, I: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the default alpha^m_i
  try
    Result:=LSRM.Defaultalpha[m,i];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_DefaultalphaUsed(LSRMName: OleVariant): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return if the LSRM uses the default alphas
  Result:=LSRM.DefaultalphaUsed;
end;

function TLSRM.Get_Defaultf(LSRMName: OleVariant; m, k: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the default development factor f^m_k
  try
    Result:=LSRM.Defaultf[m,k];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_DefaultfUsed(LSRMName: OleVariant): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return if the LSRM uses the default development factors
  Result:=LSRM.DefaultfUsed;
end;

function TLSRM.Get_Defaultsigma(LSRMName: OleVariant; m1, m2, k: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the default variance parameters sigma^m1m2_k
  try
    Result:=LSRM.Defaultsigma[m1,m2,k];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_DefaultsigmaUsed(LSRMName: OleVariant): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // returns if the LSRM uses the default variance parameter
  Result:=LSRM.DefaultsigmaUsed;
end;

function TLSRM.Get_Defaultw(LSRMName: OleVariant; n, m, i, k: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the default weights
  //   - not normalised, if n=0
  //   - for busines period I, if n=1
  //   - for business period I+1, if n=2
  try
    Result:=LSRM.Defaultw[n,m,i,k];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_DefaultwUsed(LSRMName: OleVariant): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return if the LSRM uses the default weights
  Result:=LSRM.DefaultwUsed;
end;

function TLSRM.Get_Eigenvalue(LSRMName: OleVariant; i, k, m: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the m-th largest eigenvalues of the covariance matrices (sigma^m1m2_k*R^m1m2_ik)_m1m2
  try
    Result:=LSRM.Eigenvalue[i,k,m];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_Eigenvector(LSRMName: OleVariant; i, k, m, l: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the l-th coordinate of the eigenvector of the m-th largest eigenvalues of the covariance matrices
  // (sigma^m1m2_k*R^m1m2_ik)_m1m2
  try
    Result:=LSRM.Eigenvector[i,k,m,l];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_f(LSRMName: OleVariant; m, k: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the used development factor f^m_k
  try
    Result:=LSRM.Usedf[m,k];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_I(LSRMName: OleVariant): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return I
  try
    Result:=LSRM.I;
  except
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_InitialisedParameters(LSRMName: OleVariant): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the parameters that have been initialised by the procedure Init
  try
    Result:=LSRM.InitialisedParameters;
  except
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_J(LSRMName: OleVariant): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return J
  try
    Result:=LSRM.J;
  except
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_K(LSRMName: OleVariant): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the number of furure development periods
  try
    Result:=LSRM.K;
  except
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_M(LSRMName: OleVariant): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return M
  try
    Result:=LSRM.M;
  except
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_ModelType(LSRMName: OleVariant): OleVariant;
var LSRM: TNormalLSRM;
begin
  // the standard result is a normal LSRM
  Result:=rsLSRM_NormalLSRM;
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the kind of the LSRM (normal, simple without inflation or simple with inflation)
  if LSRM.InheritsFrom(TSimpleInflationLSRM) then
    Result:=rsLSRM_InflationLSRM
  else if LSRM.InheritsFrom(TSimpleLSRM) then
    Result:=rsLSRM_SimpleLSRM;
end;

function TLSRM.Get_MSEP(LSRMName: OleVariant; M, I, Proxy, Part, Tail: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the ultimate uncertainty
  try
    Result:=LSRM.MSEP[m,i,Proxy,Part,Tail];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_MSEPDetail(LSRMName: OleVariant; m1, m2, i1, i2, Proxy, Part, Tail: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the detail of the ultimate uncertainty
  try
    Result:=LSRM.MSEPDetail[m1,m2,i1,i2,Proxy,Part,Tail];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_MSEPCov(LSRMName: OleVariant; m1, m2, Proxy, Part, Tail: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return (m1, m2) entry of the coveriance matrix corresponding to the ultimate uncertainty
  try
    Result:=LSRM.MSEPCov[m1,m2,Proxy,Part,Tail];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_MSEPCovEigenvalue(LSRMName: OleVariant; m, Proxy, Part, Tail: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the m-th largest eigenvalue of the coveriance matrix corresponding to the ultimate uncertainty
  try
    Result:=LSRM.MSEPCovEigenvalue[m,Proxy,Part,Tail];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_MSEPCovEigenvector(LSRMName: OleVariant; m, l, Proxy, Part, Tail: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return l-th coordinate of the eigenvector corresponding to the m-th largest eigenvalue of the coveriance
  // matrix corresponding to the ultimate uncertainty
  try
    Result:=LSRM.MSEPCovEigenvector[m,l,Proxy,Part,Tail];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_R1(LSRMName: OleVariant; m, i, k: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return R^m_ik
  try
    Result:=LSRM.R1[m,i,k];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_R2(LSRMName: OleVariant; m1, m2, i, k: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return R^m1m2_ik
  try
    Result:=LSRM.R2[m1,m2,i,k];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_S(LSRMName: OleVariant; m, i, k: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return S^m_ik
  try
    Result:=LSRM.ClaimProperty[m,i,k];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_sigma(LSRMName: OleVariant; m1, m2, k: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the used variance parameter sigme^m1m2_k
  try
    Result:=LSRM.Usedsigma[m1,m2,k];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_w(LSRMName: OleVariant; n, m, i, k: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // return the used weights
  //   - not normalised, if n=0
  //   - for busines period I, if n=1
  //   - for business period I+1, if n=2
  try
    Result:=LSRM.Usedw[n,m,i,k];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_Ultimate(LSRMName: OleVariant; M, I, Tail: Integer): OleVariant;
var LSRM: TNormalLSRM;
begin
  Result:='';
  {$REGION 'get the LSRM'}
  LSRM:=GetLSRM(LSRMName);
  if LSRM=nil then begin
    Result:=rsLSRM_LSRMNotFound;
    Exit;
  end;
  {$ENDREGION}
  // get the ultimate
  try
    Result:=LSRM.Ultimate[m,i,Tail];
  except
    on E:ELSRMRangeError do Result:=E.Message;
    on E:ELSRMInitError do Result:=E.Message;
    else raise;
  end;
end;

function TLSRM.Get_Delimiters: OleVariant;
begin
  Result:=TSimpleLSRM.Delimiters;
end;

function TLSRM.Get_MaxVarCoef1: Double;
begin
  Result:=TNormalLSRM.MaxVarCoef1;
end;

function TLSRM.Get_MaxVarCoef2: Double;
begin
  Result:=TNormalLSRM.MaxVarCoef2;
end;

procedure TLSRM.Set_Delimiters(Value: OleVariant);
begin
  TSimpleLSRM.Delimiters:=Value;
end;

procedure TLSRM.Set_MaxVarCoef1(Value: Double);
begin
  TNormalLSRM.MaxVarCoef1:=Value;
end;

procedure TLSRM.Set_MaxVarCoef2(Value: Double);
begin
  TNormalLSRM.MaxVarCoef2:=Value;
end;

{$ENDREGION}

{$REGION '  TLSRM public'}

destructor TLSRM.Destroy;
var i: integer;
begin
  // free all LSRMs and clear the list
  if FLSRMs<>nil then begin
    for i:=0 to FLSRMs.count-1 do
      if FLSRMs.Objects[i]<>nil then FLSRMs.Objects[i].Free;
    FreeAndNil(FLSRMs);
  end;
  inherited;
end;

function TLSRM.Init(LSRMName: OleVariant; What, M, I, J, FuturePeriods: Integer; const Gamma1,
          Gamma2: OleVariant; var S, w, f, sigma, alpha, InflationRate: Double): OleVariant;
var Index: integer;
    LSRM: TNormalLSRM;
    Gamma1LSRMType, Gamma2LSRMType, LSRMType: Integer; // (1=NormalLSRM, 2=SimpleLSRM, 3=both possible)
begin
  Result:='';
  {$REGION 'check the parameter Gamma1 and Gamma2'}
  // detect if the parameter Gamma1 and Gamma2 specify a normal (LSRMType=1) or a simple LSRM (LSRMType=2),
  // i.e.
  //  NormalLSRM:
  //   - gamma1 ia a two dimensional array of double with 7 columns in the second coordinate
  //   - gamma2 ia a two dimensional array of double with 8 columns in the second coordinate
  //  SimpleLSRM (formula based):
  //   - gamma 1 and gamm2 are a stringS containing a formula that defines the corresponding parameters
  //  SimpleLSRM (parameter based):
  //   - gamma1 is a string or a two dimensional array with 5 columns in the second coordinate
  //   - gamma2 is a string or a two dimensional array with 6 columns in the second coordinate
  // Finally, for simple LSRM's we have to distinuish between those
  //  with (TSimpleLSRM) and those without (TSimpleInflationLSRM) inflation
  Gamma1LSRMType:=3;
  Gamma2LSRMType:=3;
  if (cLSRMInitGamma1 and What)<>0 then begin
    if (((VarType(Gamma1) and VarArray) = 0)
            and ((VarType(Gamma1) and (VarOleStr + VarString + VarStrArg + VarUString + VarUStrArg))>0))
       or (((VarType(Gamma1) and (VarTypeMask+VarArray)) = VarArray + VarDouble)
          and (VarArrayDimCount(Gamma1)=2)
          and (VarArrayHighBound(Gamma1,2)-VarArrayLowBound(Gamma1,2)=4)) then
      Gamma1LSRMType:=2
    else if ((VarType(Gamma1) and (VarTypeMask+VarArray)) = VarArray + VarDouble)
            and (VarArrayDimCount(Gamma1)=2)
            and (VarArrayHighBound(Gamma1,2)-VarArrayLowBound(Gamma1,2)=6) then
      Gamma1LSRMType:=1
    else begin
      Result:='LSRM ''' + LSRMName + ''':' + rsLSRM_WrongGamma1;
      Exit;
    end;
  end;
  if (cLSRMInitGamma2 and What)<>0 then begin
    if (((VarType(Gamma2) and VarArray) = 0)
            and ((VarType(Gamma2) and (VarOleStr + VarString + VarStrArg + VarUString + VarUStrArg))>0))
       or (((VarType(Gamma2) and (VarTypeMask+VarArray)) = VarArray + VarDouble)
          and (VarArrayDimCount(Gamma2)=2)
          and (VarArrayHighBound(Gamma2,2)-VarArrayLowBound(Gamma2,2)=5)) then
      Gamma2LSRMType:=2
    else if ((VarType(Gamma2) and (VarTypeMask+VarArray)) = VarArray + VarDouble)
            and (VarArrayDimCount(Gamma2)=2)
            and (VarArrayHighBound(Gamma2,2)-VarArrayLowBound(Gamma2,2)=7) then
      Gamma2LSRMType:=1
    else begin
      Result:='LSRM ''' + LSRMName + ''':' + rsLSRM_WrongGamma2;
      Exit;
    end;
  end;
  // check if both parameters specify the same type
  LSRMType:=(Gamma1LSRMType and Gamma2LSRMType);
  if LSRMType=0 then begin
    Result:='LSRM ''' + LSRMName + ''':' + rsLSRM_Gamma1Gamma2Mismatch;
    Exit;
  end;
  {$ENDREGION}
  {$REGION 'get the LSRM and check if it is of the right type'}
  LSRM:=nil;
  Index:=FLSRMs.IndexOf(LSRMName);
  if Index>=0 then begin
    LSRM:=TNormalLSRM(FLSRMs.Objects[Index]);
    // if we should overwrite existing LSRMs we remove the found one, otherwise we
    // check if the LSRM is of the same type as defined by the parameters gamma1 and gamma2
    // and by inflation,
    // if the LSRM is not of the right type we remove it and create a new one later
    if (cLSRMOverwriteExistingLSRM and What)<>0 then begin
      FLSRMs.Delete(Index);
      FreeAndNil(LSRM);
    end else if ((LSRMType=2)
                  and ((((CLSRMInitInflationRate and What)<>0)
                          and not(LSRM.InheritsFrom(TSimpleInflationLSRM)))
                       or not(LSRM.InheritsFrom(TSimpleInflationLSRM))))
        or ((LSRMType=1) and LSRM.InheritsFrom(TSimpleLSRM)) then begin
      FreeAndNil(LSRM);
      FLSRMs.Delete(Index);
    end;
  end;
  {$ENDREGION}
  {$REGION 'create a new LSRM, if necessary'}
  if LSRM=nil then begin
    // create a new LSRM of the right type
    if LSRMType=1 then
      LSRM:=FLSRMs.Objects[FLSRMs.AddObject(LSRMName,TNormalLSRM.Create(LSRMName))] as TNormalLSRM
    else if (CLSRMInitInflationRate and What)<>0 then
      LSRM:=FLSRMs.Objects[FLSRMs.AddObject(LSRMName,TSimpleInflationLSRM.Create(LSRMName))]
                                                                                      as TSimpleInflationLSRM
    else
      LSRM:=FLSRMs.Objects[FLSRMs.AddObject(LSRMName,TSimpleLSRM.Create(LSRMName))] as TSimpleLSRM;
  end;
  {$ENDREGION}
  // initialise the LSRM
  Result:=LSRM.Init(What, M, I, J, FuturePeriods, Gamma1, Gamma2, S, w, f, sigma, alpha, InflationRate)
end;

procedure TLSRM.Initialize;
begin
  inherited;
  {$REGION 'initialise the list'}
  FLSRMs:=TStringList.Create;
  FLSRMs.Sorted:=True;
  FLSRMs.Duplicates:=dupError;
  FLSRMs.CaseSensitive:=False;
  {$ENDREGION}
end;

procedure TLSRM.Reset(LSRMName: OleVariant);
var i: Integer;
begin
  if (VarType(LSRMName) and VarString)<>0 then Exit;
  // if LSRMName is an empty string the whole list of LSRMs will be emptied
  if LSRMName='' then begin
    for i:=0 to FLSRMs.count-1 do
      if FLSRMs.Objects[i]<>nil then FLSRMs.Objects[i].Free;
    FLSRMs.Clear;
  end else begin
    // get the LSRM and delete it if possible
    i:=FLSRMs.IndexOf(LSRMName);
    if i>=0 then begin
      if FLSRMs.Objects[i]<>nil then FLSRMs.Objects[i].Free;
      FLSRMs.Delete(i);
    end;
  end;
end;

{$ENDREGION}

{$ENDREGION}

initialization
  TAutoObjectFactory.Create(ComServer, TLSRM, Class_LSRM, ciMultiInstance, tmApartment);
end.
