﻿unit LSRM;

//Copyright 2013-2017 René Dahms (rene.dahms at math.ethz.ch)
//
//This file is part of LSRMTools.
//
//    LSRMTools is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    LSRMTools is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with LSRMTools.  If not, see <http://www.gnu.org/licenses/>.
//
//LSRMTools use some parts of ALGLIB which is a free software distributed
//under a GPL license - version 2 or (at your option) any later version.

interface

uses Classes, evd, Ap, SysUtils, LSRMDependencies, FormulaParser;

type {$REGION 'some types'}
  // type to store the dependence structure of accident years and claim properties
  T2IntegerArray=array of array of Integer;
  // we use only one dimensional arrays of double, because of speed reasons
  T1Double = array of Double;
  PT1Double = ^T1Double;
  // type that is used for indices in T1Double instances
  IndexDouble = Integer;
  // type to store the eigenvalues of matrices
  TEigValArray = array of TReal1DArray;
  // type to store the eigenvectors of matrices
  TEigVecArray = array of TReal2DArray;
  {$REGION 'flags that control the kind of calculation'}
  TCalculationFlags=(calc_ReduceMemSize=13, // do only calculate proxies, i.e. no operator H
                     calc_VirtualProcVarDiversification=16, // no 'virtual diversification' in process variance
                     calc_VirtualParaErrDiversification=17, // no 'virtual diversification' in parameter error
                     calc_VirtualTailDiversification=15, // no 'virtual diversification' in tail uncertainty
                     // known exposures have the same sign for all development periods
                     calc_PreventPastExposureFromChangingSign=18,
                     // estimated exposures have the same sign for all development periods
                     calc_PreventFutureExposureFromChangingSign=19,
                     // use weights for default sigma
                     calc_UseWeightsForDefaultSigma=20,
                     // use the MSEP process variance in case the CDR is not well defined
                     calc_UseMSEPifTailCDRisLarger=21);
  TCalculationFlagSet=set of TCalculationFlags;
  {$ENDREGION}
  // type to store if a initialisation step of a LSRM has been calculated or not
  TCalcInfo=array[TLSRMInitialisedElement] of boolean;
  // type to store the calculation time for each initialisation step of a LSRM
  TCalcTime=array[TLSRMInitialisedElement] of double;
  // an exception type which will be raised within properties if some initialisation failed
  ELSRMInitError=class(Exception);
  // an exception type which will be raised within properties if some initialisation failed
  ELSRMRangeError=class(Exception);
  {$ENDREGION}

const
  {$REGION 'init functions that do need the operator H or that are only used in such functions'}
    cNonProxyInits=cInfluencedByrhoMSEP + cInfluencedByrhoCDR + cInfluencedBySxS0 + cInfluencedBySxSMSEP
                  + cInfluencedBySxSCDRParaErr + cInfluencedBySxSCDRProcVar;
    cInfluencedByTailDiversification = cInfluencedByMSEPProcVarTailDetail + cInfluencedByMSEPParaErrTailDetail
                          + cInfluencedByMSEPProcVarTailProxyDetail + cInfluencedByMSEPParaErrTailProxyDetail
                          + cInfluencedByCDRProcVarTailDetail + cInfluencedByCDRParaErrTailDetail
                          + cInfluencedByCDRProcVarTailProxyDetail + cInfluencedByCDRParaErrTailProxyDetail;
    cInfluencedByProcVarDiversification = cInfluencedByMSEPProcVarDetail
                          + cInfluencedByMSEPProcVarProxyDetail + cInfluencedByCDRProcVarDetail
                          + cInfluencedByCDRProcVarProxyDetail;
    cInfluencedByParaErrDiversification = cInfluencedByMSEPParaErrDetail
                          + cInfluencedByMSEPParaErrProxyDetail + cInfluencedByCDRParaErrDetail
                          + cInfluencedByCDRParaErrProxyDetail;
  {$ENDREGION}

{$REGION 'Description ...'}
// In order to speed up calculation we want to reduce the number of addends.
// Often we have gamma^m,l_i,k,h,j=0 for almost all l<>m and allmost all i<>h.
// Therefore we will introduce an array FmlDependency[0..M] with for each
// m (first dimension) returns an array [0..M] containing all l's with non trivial
// gamma's, whereby the last entry contains the number of such l's.
// And for i and h we will do the same. In order to construct such array we introduce
// the following class.
{$ENDREGION}
type TLSRMDependencies=class(TObject)
  private {$REGION}
    FDependsOn, FDoesInfluence:array of TBits;
    function GetDependsOn(Dependant, i: Integer): Boolean;
    function GetDoesInfluence(Influencer, i: Integer): Boolean;
    {$ENDREGION}
  public {$REGION}
    constructor Create(MaxInstance: Integer);
    destructor Destroy; override;
    procedure AddDependsOn(Dependant, i: Integer);
    procedure AddDoesInfluence(Influencer, i: Integer);
    procedure FillDependenceArray(var Dependencies: T2IntegerArray);
    property DependsOn[Dependant, i: Integer]: Boolean read GetDependsOn;
    property DoesInfluence[Influencer, i: Integer]: Boolean read GetDoesInfluence;
    {$ENDREGION}
end;

{$REGION 'parameters to parse and evaluate formulas for the definition of exposures'}
  TParseParas = record
    M: integer; // the index of the last claim properties
    J: integer; // the last development poeriods of the exposure (inclusive tail)
    OnlyLinear: boolean; // true for gamma1, false for gamma2
  end;
  PParseParas = ^TParseParas;

  TCalcMode = (CalcGamma1,CalcGamma2);
  TCalcParas = record
    M: integer; // the index of the last last claim property
    I: integer; // the last accident period
    J: integer; // the last development periods of the exposure (inclusive tail)
    m1: integer; // the index of the exposure (only used for gamma1 dependencies)
    LastDiag: integer; // the last diagonal that should be calculated (only for R2)
    LastColumn: integer; // the last column that should be calculated (only for R2)
    Mode: TCalcMode; // indicates if gamma1 or R2 should be calculated
    Dep_ml: TLSRMDependencies; // m,l-dependencies that should be filled (only for gamma1)
    ClaimProperty: record // only used for R2
      S: PDouble; // the pointer to the claim properties
      V: PDouble; // inflation pointer to the inflation (if nil then we have no inflation)
      J: integer; // the last development period of claim properties (inclusive tail)
    end;
  end;
  PCalcParas = ^TCalcParas;
{$ENDREGION}

{$REGION 'Description ...'}
// This is the main object for a LSRM
{$ENDREGION}
type TNormalLSRM=class(TObject)
  private {$REGION}
    // constants to avoid too huge sigmas caused by dividing something by almost zero
    // (abs(sigma)>ABS(f1*f2)*cMaxVarCoef1+cMaxVerCoef2 is forbidden)
    class var FMaxVarCoef1, FMaxVarCoef2: double;
    {$REGION 'variables'}
    var
    // In the following comments we specify the dimension of array by the typical (from the article) used
    // îndex character followed by its maximal value enclosed within round brackets.
    // Name of the LSRM
    FName: string;
    // Number of claim properties, accident periods, development periods with observed data
    // and future periods. For the uncertainty and other objects we only project one tail period and not
    // all K future periods. Therefore, we define FP=1 if FK>0 and FP=0 otherwise.
    FM, FI, FJ, FK, FP: Integer;
    // set of flags that indecate which part has already been initialised
    FInitialised: TLSRMInitialised;
    // Claim properties S^m_i,k: m(M) x i(I) x k(J+K)
    FS: T1Double;
    // model parameter gamma^m,l_i,k,h,j: m(M) x l(M) x i(I) x k(J - 1 + K) x h(I) x j(J - 1 + K)
    FGamma1:T1Double;
    // Inflation rate: m(M) x i(I) x j(J+K)
    //   gamma^m,l_i,k,h,j with inflation
    //  = (gamma^m,l_i,k,h,j without inflation) * FInflation^m_i,k / FInflation^l_h,j
    //   gamma^m1,m2,l_i,k,h,j with inflation
    //  = (gamma^m1,m2,l_i,k,h,j without inflation) * FInflation^m1_i,k * FInflation^m2_i,k / FInflation^l_h,j
    FInflation: T1Double;
    // Exposures R^m_i,k: m(M) x i(I) x k(J-1+K)
    FR1:T1Double;
    // manual specified weights w^n,m_i,k: n(3) x m(M) x i(I) x k(J-1+FP)
    // the first dimension of w means 0: not normalized, 1: time I and 2: time I+1
    Fw:T1Double;
    // default weights (lead to variance minimizing f^m_k): n(3) x m(M) x i(I) x k(J-1+FP)
    FDefaultw:T1Double;
    // Pointer on the used weights
    FUsedw:PT1Double;
    // manual specified development factors f^m_k: m(M) x k(J-1+K)
    Ff:T1Double;
    // default (using active weights) development factors: m(M) x k(J-1+K)
    FDefaultf:T1Double;
    // Pointer on the used development factors
    FUsedf:PT1Double;
    // model parameter gamma^m1,m2,l_i,k,h,j: m1(M) x m2(M) x l(M) x i(I) x k(J-1+FP) x h(I) x j(J-1+FP)
    FGamma2:T1Double;
    // Exposures R^m1,m2_i,k: m1(M) x m2(M) x i(I) x k(J-1+FP)
    FR2:T1Double;
    // manual specified variance parameters sigma^m1,m2_k: m1(M) x m2(M) x k(J-1+FP)
    Fsigma:T1Double;
    // default (using active weights) unbiased variance parameters sigma^m1,m2_k: m1(M) x m2(M) x k(J-1+FP)
    FDefaultsigma:T1Double;
    // Pointer on the used variance parameters
    FUsedsigma:PT1Double;
    // manual specified mixing parameters alpha^m_i: m(M) x i(I)
    Falpha: T1Double;
    // default (=1) mixing parameters alpha^m_i: m(M) x i(I)
    FDefaultalpha: T1Double;
    // Pointer on used mixing alphas
    FUsedalpha:PT1Double;
    // mixed ultimates: Part(1) x m(M+1) x i(I+1)
    // Part: 0 - without Tail
    //       1 - with Tail
    FUltimate:T1Double;
    // coordinates of the operator F, i.e.
    //  (F^n,m_i,k)^l_h,j: m(M) x i(I) x k(J-1+FP) [x n(I+J+FP)] x l(M) x h(I) x j(J+FP)
    // Since we only need those with n=h+j, we skip the dimension n.
    FCoorF: T1Double;
    // flags that specify how the program should calculate. For instance, the balance between extensive
    // memory useage and performance and if 'virtual diversification' should be avoided.
    FCalculationFlags: TCalculationFlagSet;
    // rho* without indizes i1 and i1: m1(M) x m2(M) x k(J-1+FP)
    FrhoStar: T1Double;
    // coordinates of rhos: Part(5) x m1(M) x m2(M) x i1(I) x i2(I) x k (J-1+FP)
    // part means:
    //    0: 0
    //    1: rho*
    //    2: rho*1
    //    3: rho*2
    //    4: rho*12
    //    5: rho bar
    Frho: T1Double;
    // the results of the operator H^(I+J+FP) applied to SxS for each rho in Frho:
    //   Part(5) x m1(M) x i1(I) x k1(J+FP) x m2(M) x i2(I) x k2(J+FP)
    FSxS: T1Double;
    // the addands for indizes m1, m2, i1, i2 in the sum of the MSEP estimation:
    //    Proxy(no=0 or yes=1) x Part(0=ProvVar, 1=ParaErr) x m1(M) x m2(M) x i1(I) x i2(I)
    FMSEPDetail: T1Double;
    // the addands for indizes m1, m2, i1, i2 in the sum of the MSEP Tail estimation:
    //    Proxy(no=0 or yes=1) x Part(0=ProvVar, 1=ParaErr) x m1(M) x m2(M) x i1(I) x i2(I)
    FMSEPTailDetail: T1Double;
    // the MSEP estimation: Proxy(no=0 or yes=1) x Part(0=ProvVar, 1=ParaErr, 2=both) x m(M+1) x i(I+1)
    // m=M+1 means alpha weighted sums and i=I+1 means total over all accident periods i
    FMSEP: T1Double;
    // the MSEP tail estimation: Proxy(no=0 or yes=1) x Part(0=ProvVar, 1=ParaErr, 2=both) x m(M+1) x i(I+1)
    // m=M+1 means alpha weighted sums and i=I+1 means total over all accident periods i
    FMSEPTail: T1Double;
    // the addands for indizes m1, m2, i1, i2 in the sum of the CDR estimation:
    //    Proxy(no=0 or yes=1) x Part(0=ProvVar, 1=ParaErr) x m1(M) x m2(M) x i1(I) x i2(I)
    FCDRDetail: T1Double;
    // the addands for indizes m1, m2, i1, i2 in the sum of the CDR Tail estimation:
    //    Proxy(no=0 or yes=1) x Part(0=ProvVar, 1=ParaErr) x m1(M) x m2(M) x i1(I) x i2(I)
    FCDRTailDetail: T1Double;
    // the CDR estimation: Proxy(no=0 or yes=1) x Part(0=ProvVar, 1=ParaErr, 2=both) x m(M+1) x i(I+1)
    // m=M+1 means alpha weighted sums and i=I+1 means total over all accident periods i
    FCDR: T1Double;
    // the CDR Tail estimation: Proxy(no=0 or yes=1) x Part(0=ProvVar, 1=ParaErr, 2=both) x m(M+1) x i(I+1)
    // m=M+1 means alpha weighted sums and i=I+1 means total over all accident periods i
    FCDRTail: T1Double;
    // Information which step has been calculated
    FCalcInfo: TCalcInfo;
    // information about how long the calculation of each step lasts
    FCalcTime: TCalcTime;
    // store which indize l (h) is needed for which indize m (i) in double sums
    FmlDependency,FihDependency: T2IntegerArray;
    // Eigenvalues and eigenvectors of the covariance assumptions: i (I+1) x k (J-1+FP)
    FEigenvalues: TEigValArray;
    FEigenvectors: TEigVecArray;
    // total CDR correlation:
    //   Proxy(no=0 or yes=1) x Part(0=ProvVar, 1=ParaErr, 2=both) x Tail(0=without, 1=with) x m1(M)) x m2(M)
    FCDRCov: T1Double;
    // Eigenvalues and eigenvectors of the one period uncertainty CDR: 
    //    Proxy(no=0 or yes=1) x Part(0=ProvVar, 1=ParaErr, 2=both) x Tail(0=without, 1=with)
    FCDRCovEigenvalues: TEigValArray;
    FCDRCovEigenvectors: TEigVecArray;
    // total MSEP correlation:
    //   Proxy(no=0 or yes=1) x Part(0=ProvVar, 1=ParaErr, 2=both) x Tail(0=without, 1=with) x m1(M) x m2(M)
    FMSEPCov: T1Double;
    // Eigenvalues and eigenvectors of the ultimate uncertainty MSEP
    //    Proxy(no=0 or yes=1) x Part(0=ProvVar, 1=ParaErr, 2=both) x Tail(0=without, 1=with)
    FMSEPCovEigenvalues: TEigValArray;
    FMSEPCovEigenvectors: TEigVecArray;
    {$ENDREGION}
    // free all memory
    procedure FreeAll;
    // normalisation of weights
    procedure NormalizeWeights(w: T1Double);
    // sets the initialisation flags according to dependencies
    procedure SetInitFlags(InitFlag: TLSRMInitialisedElement; UsedTime: Int64; UpdateTime: Boolean=True);
    // updates the field that contains the information about what has been calculated and how long it took
    procedure UpdateCalcInfo(InitFlag:TLSRMInitialisedElement; InfluencedFlags:TLSRMInitialised;
                             UsedTime: Int64);
    // if FRho is not kept within memory the following function is used to get its value
    // here n has the following meaning:
    //    0: 0
    //    1: rho*
    //    2: rho*1
    //    3: rho*2
    //    4: rho*12
    //    5: rho bar
    function GetRho(n, m1, m2, i1, i2, k: Integer): Double; virtual;
    class function GetMaxVarCoef1: double; static;
    class function GetMaxVarCoef2: double; static;
    class procedure SetMaxVarCoef1(const Value: double); static;
    class procedure SetMaxVarCoef2(const Value: double); static;
    {$ENDREGION}
  protected {$REGION}
    {$REGION 'property get functions}
    function GetCalcInfo(Step:Integer): Boolean; virtual;
    function GetCalcInfoName(Step: Integer): string; virtual;
    function GetCalcTime(Step: Integer): Double; virtual;
    function GetCalculationFlags: TCalculationFlagSet; virtual;
    function GetCDR(m, i, Proxy, Part, Tail: Integer): Double; virtual;
    function GetCDRCov(m1, m2, Proxy, Part, Tail: Integer): Double; virtual;
    function GetCDRCovEigenvalue(m, Proxy, Part, Tail: Integer): Double; virtual;
    function GetCDRCovEigenvector(m, l, Proxy, Part, Tail: Integer): Double; virtual;
    function GetCDRDetail(m1, m2, i1, i2, Proxy, Part, Tail: Integer): Double; virtual;
    function GetCoorF(m, i, k, l, h, j: Integer): Double; virtual;
    function GetDefaultalpha(m, i: integer): Double; virtual;
    function GetDefaultalphaUsed: Boolean; virtual;
    function GetDefaultf(m, k: integer): Double; virtual;
    function GetDefaultfUsed: Boolean; virtual;
    function GetDefaultsigma(m1, m2, k: integer): Double; virtual;
    function GetDefaultsigmaUsed: Boolean; virtual;
    function GetDefaultw(n, m, i, k: integer): Double; virtual;
    function GetDefaultwUsed: Boolean; virtual;
    function GetEigenvalue(i, k, m: integer): Double; virtual;
    function GetEigenvector(i, k, m, l: integer): Double; virtual;
    function GetI: integer; virtual;
    function GetInitialisedParameters: integer; virtual;
    function GetJ: integer; virtual;
    function GetK: integer; virtual;
    function GetM: integer; virtual;
    function GetMSEP(m, i, Proxy, Part, Tail: Integer): Double; virtual;
    function GetMSEPCov(m1, m2, Proxy, Part, Tail: Integer): Double; virtual;
    function GetMSEPCovEigenvalue(m, Proxy, Part, Tail: Integer): Double; virtual;
    function GetMSEPCovEigenvector(m, l, Proxy, Part, Tail: Integer): Double; virtual;
    function GetMSEPDetail(m1, m2, i1, i2, Proxy, Part, Tail: Integer): Double; virtual;
    function GetR1(m, i, k: integer): Double; virtual;
    function GetR2(m1, m2, i, k: integer): Double; virtual;
    function GetS(m, i, k: integer): Double; virtual;
    function GetUltimate(m, i, Tail: Integer): Double; virtual;
    function GetUsedalpha(m, i: Integer): Double; virtual;
    function GetUsedf(m, k: integer): Double; virtual;
    function GetUsedsigma(m1, m2, k: integer): Double; virtual;
    function GetUsedw(n, m, i, k: integer): Double; virtual;
    {$ENDREGION}
    {$REGION 'operators'}
    // the operator F^n
    function OP_F(n: integer; S: T1Double): string; virtual;
    // the operator Fn1<-n2
    function OP_Fn(n1, n2: integer; S: T1Double): string; virtual;
    // the opertor H^n_k(tau)
    function OP_H(n, k: integer; tau, SxS: PDouble): string; virtual;
    // the opertor H^n_J<-0(tau) (n=I in the article)
    function OP_Hn(n: Integer; tau, SxS: PDouble): string; virtual;
    {$ENDREGION}
    {$ENDREGION}
  public {$REGION}
    // (re)initialises the LSRM, 'What' is a sum of TLSRMInitFlags and spezifies what should be initialised
    function Init(What, M, I, J, K: Integer; const Gamma1, Gamma2: OleVariant;
      var S, w, f, sigma, alpha, InflationRate: Double): OleVariant; virtual;
    // create a LSRM
    constructor Create(Name: string); virtual;
    // destroy the LSRM
    destructor Destroy; override;
    {$REGION 'Init functions'}
    // initialise manual specified mixing alphas if necessary
    function Initalpha(var alpha: Double): String; virtual;
    {$REGION 'CDR'}
    // calculates the sum of process variance and parameter error of the CDR (without approximation)
    // the result is stored within FCDR(0,2)
    function InitCDR(): string; virtual;
    // calculates the sum of process variance and parameter error of the approximated CDR
    // the result is stored within FCDR(1,2)
    function InitCDRProxy(): string; virtual;
    // calculates the sum of the Tail process variance and the Tail parameter error of the of the CDR
    // (without approximation)
    // the result is stored within FCDRTail(1,2)
    function InitCDRTail(): string; virtual;
    // calculates the sum of the Tail process variance and the Tail parameter error of the of the
    // approximated CDR
    // the result is stored within FCDRTail(1,2)
    function InitCDRTailProxy(): string; virtual;
    {$REGION 'CDR correlations'}
    // calculate the correlation matrix of the total CDR without tail without proxy
    function InitCDRCov(): string; virtual;
    // calculate the correlation matrix of the total CDR without tail with proxy
    function InitCDRCovProxy(): string; virtual;
    // calculate the correlation matrix of the total CDR with tail without proxy
    function InitCDRCovTail(): string; virtual;
    // calculate the correlation matrix of the total CDR with tail with proxy
    function InitCDRCovTailProxy(): string; virtual;
    // calculate the correlation matrix of the CDR parameter error without tail without proxy
    function InitCDRCovParaErr(): string; virtual;
    // calculate the correlation matrix of the CDR parameter error without tail with proxy
    function InitCDRCovParaErrProxy(): string; virtual;
    // calculate the correlation matrix of the CDR parameter error with tail without proxy
    function InitCDRCovParaErrTail(): string; virtual;
    // calculate the correlation matrix of the CDR parameter error with tail with proxy
    function InitCDRCovParaErrTailProxy(): string; virtual;
    // calculate the correlation matrix of the CDR process variance without tail without proxy
    function InitCDRCovProcVar(): string; virtual;
    // calculate the correlation matrix of the CDR process variance without tail with proxy
    function InitCDRCovProcVarProxy(): string; virtual;
    // calculate the correlation matrix of the CDR process variance with tail without proxy
    function InitCDRCovProcVarTail(): string; virtual;
    // calculate the correlation matrix of the CDR process variance with tail with proxy
    function InitCDRCovProcVarTailProxy(): string; virtual;
    {$ENDREGION}
    {$REGION 'eigenvalues and eigenvectors of the CDR correlations'}
    // calculate the eigenvalues and eigenvectors the correlation matrix of the total CDR
    // without tail without proxy
    function InitCDRCovEigenvalue(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the total CDR
    // without tail with proxy
    function InitCDRCovEigenvalueProxy(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the total CDR
    // with tail without proxy
    function InitCDRCovEigenvalueTail(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the total CDR
    // with tail with proxy
    function InitCDRCovEigenvalueTailProxy(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the CDR parameter error
    // without tail without proxy
    function InitCDRCovEigenvalueParaErr(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the CDR parameter error
    // without tail with proxy
    function InitCDRCovEigenvalueParaErrProxy(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the CDR parameter error
    // with tail without proxy
    function InitCDRCovEigenvalueParaErrTail(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the CDR parameter error
    // with tail with proxy
    function InitCDRCovEigenvalueParaErrTailProxy(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the CDR process variance
    // without tail without proxy
    function InitCDRCovEigenvalueProcVar(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the CDR process variance
    // without tail with proxy
    function InitCDRCovEigenvalueProcVarProxy(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the CDR process variance
    // with tail without proxy
    function InitCDRCovEigenvalueProcVarTail(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the CDR process variance
    // with tail with proxy
    function InitCDRCovEigenvalueProcVarTailProxy(): string; virtual;
    {$ENDREGION}
    {$REGION 'CDR Parameter Error'}
    // calculates the parameter error (without approximation) of the CDR
    // the result is stored within FCDR(0,1)
    function InitCDRParaErr(): string; virtual;
    // calculates the approximated parameter error of the CDR
    // the result is stored within FCDR(1,1)
    function InitCDRParaErrProxy(): string; virtual;
    // calculates the addands m1, m2, i1, i2 of the parameter error (without approximation) of the CDR
    // the result is stored within FCDRDetail(0,1)
    function InitCDRParaErrDetail(): string; virtual;
    // calculates the addands m1, m2, i1, i2 of the approximated parameter error of the CDR
    // the result is stored within FCDRDetail(1,1)
    function InitCDRParaErrProxyDetail(): string; virtual;
    // calculates the Tail parameter error (without approximation) of the CDR
    // the result is stored within FCDRTail(0,1)
    function InitCDRParaErrTail(): string; virtual;
    // calculates the approximated Tail parameter error of the CDR
    // the result is stored within FCDRTail(1,1)
    function InitCDRParaErrTailProxy(): string; virtual;
    // calculates the addands m1, m2, i1, i2 of the of the Tail parameter error (without approximation) of
    // the CDR
    // the result is stored within FCDRTailDetail(0,1)
    function InitCDRParaErrTailDetail(): string; virtual;
    // calculates the addands m1, m2, i1, i2 of the of the approximated Tail parameter error of the CDR
    // the result is stored within FCDRTailDetail(1,1)
    function InitCDRParaErrTailProxyDetail(): string; virtual;
    {$ENDREGION}
    {$REGION 'CDR Process Variance'}
    // calculates the process variance (without approximation) of the CDR
    // the result is stored within FCDR(0,0)
    function InitCDRProcVar(): string; virtual;
    // calculates the approximated process variance of the CDR
    // the result is stored within FCDR(1,0)
    function InitCDRProcVarProxy(): string; virtual;
    // calculates the addands m1, m2, i1, i2 of the process variance (without approximation) of the CDR
    // the result is stored within FCDRDetail(0,0)
    function InitCDRProcVarDetail(): string; virtual;
    // calculates the addands m1, m2, i1, i2 of the approximated process variance of the CDR
    // the result is stored within FCDRDetail(1,0)
    function InitCDRProcVarProxyDetail(): string; virtual;
    // calculates of the Tail process variance (without approximation) of the CDR
    // the result is stored within FCDRTail(0,0)
    function InitCDRProcVarTail(): string; virtual;
    // calculates of the approximated Tail process variance of the CDR
    // the result is stored within FCDRTail(1,0)
    function InitCDRProcVarTailProxy(): string; virtual;
    // calculates the addands m1, m2, i1, i2 of the Tail process (withot approximation) variance of the CDR
    // the result is stored within FCDRTailDetail(0,0)
    function InitCDRProcVarTailDetail(): string; virtual;
    // calculates the addands m1, m2, i1, i2 of the approximated Tail process variance of the CDR
    // the result is stored within FCDRTailDetail(1,0)
    function InitCDRProcVarTailProxyDetail(): string; virtual;
    {$ENDREGION}
    {$ENDREGION}
    // calculates the coordinates of operator F
    function InitCoorF(): string; virtual;
    // initialises default mixing alphas if necessary
    function InitDefaultalpha(): String; virtual;
    // calculates default development factors
    function InitDefaultf(): String; virtual;
    // calculates the eigenvalues and eigenvectors of the covariance matrices
    function InitEigenvalues(): String; virtual;
    // initialises the inflation rate if necessary
    function InitInflationRate(var InflationRate: Double): String; virtual;
    {$REGION 'MSEP'}
    // calculates the sum of process variance and parameter error of the MSEP (without approximation)
    // the result is stored within FMSEP(0,2)
    function InitMSEP(): string; virtual;
    // calculates the sum of process variance and parameter error of the approximated MSEP
    // the result is stored within FMSEP(1,2)
    function InitMSEPProxy(): string; virtual;
    // calculates the sum of the Tail process variance and the Tail parameter error of the of the MSEP
    // (without approximation)
    // the result is stored within FMSEPTail(1,2)
    function InitMSEPTail(): string; virtual;
    // calculates the sum of the Tail process variance and the Tail parameter error of the of the
    // approximated MSEP
    // the result is stored within FMSEPTail(1,2)
    function InitMSEPTailProxy(): string; virtual;
    {$REGION 'MSEP correlations'}
    // calculate the correlation matrix of the total MSEP without tail without proxy
    function InitMSEPCov(): string; virtual;
    // calculate the correlation matrix of the total MSEP without tail with proxy
    function InitMSEPCovProxy(): string; virtual;
    // calculate the correlation matrix of the total MSEP with tail without proxy
    function InitMSEPCovTail(): string; virtual;
    // calculate the correlation matrix of the total MSEP with tail with proxy
    function InitMSEPCovTailProxy(): string; virtual;
    // calculate the correlation matrix of the MSEP parameter error without tail without proxy
    function InitMSEPCovParaErr(): string; virtual;
    // calculate the correlation matrix of the MSEP parameter error without tail with proxy
    function InitMSEPCovParaErrProxy(): string; virtual;
    // calculate the correlation matrix of the MSEP parameter error with tail without proxy
    function InitMSEPCovParaErrTail(): string; virtual;
    // calculate the correlation matrix of the MSEP parameter error with tail with proxy
    function InitMSEPCovParaErrTailProxy(): string; virtual;
    // calculate the correlation matrix of the MSEP process variance without tail without proxy
    function InitMSEPCovProcVar(): string; virtual;
    // calculate the correlation matrix of the MSEP process variance without tail with proxy
    function InitMSEPCovProcVarProxy(): string; virtual;
    // calculate the correlation matrix of the MSEP process variance with tail without proxy
    function InitMSEPCovProcVarTail(): string; virtual;
    // calculate the correlation matrix of the MSEP process variance with tail with proxy
    function InitMSEPCovProcVarTailProxy(): string; virtual;
    {$ENDREGION}
    {$REGION 'eigenvalues and eigenvectors of the MSEP correlations'}
    // calculate the eigenvalues and eigenvectors the correlation matrix of the total MSEP
    // without tail without proxy
    function InitMSEPCovEigenvalue(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the total MSEP
    // without tail with proxy
    function InitMSEPCovEigenvalueProxy(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the total MSEP
    // with tail without proxy
    function InitMSEPCovEigenvalueTail(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the total MSEP
    // with tail with proxy
    function InitMSEPCovEigenvalueTailProxy(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the MSEP parameter error
    // without tail without proxy
    function InitMSEPCovEigenvalueParaErr(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the MSEP parameter error
    // without tail with proxy
    function InitMSEPCovEigenvalueParaErrProxy(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the MSEP parameter error
    // with tail without proxy
    function InitMSEPCovEigenvalueParaErrTail(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the MSEP parameter error
    // with tail with proxy
    function InitMSEPCovEigenvalueParaErrTailProxy(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the MSEP process variance
    // without tail without proxy
    function InitMSEPCovEigenvalueProcVar(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the MSEP process variance
    // without tail with proxy
    function InitMSEPCovEigenvalueProcVarProxy(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the MSEP process variance
    // with tail without proxy
    function InitMSEPCovEigenvalueProcVarTail(): string; virtual;
    // calculate the eigenvalues and eigenvectors the correlation matrix of the MSEP process variance
    // with tail with proxy
    function InitMSEPCovEigenvalueProcVarTailProxy(): string; virtual;
    {$ENDREGION}
    {$REGION 'MSEP Parameter Error'}
    // calculates the parameter error (without approximation) of the MSEP
    // the result is stored within FMSEP(0,1)
    function InitMSEPParaErr(): string; virtual;
    // calculates the approximated parameter error of the MSEP
    // the result is stored within FMSEP(1,1)
    function InitMSEPParaErrProxy(): string; virtual;
    // calculates the addands m1, m2, i1, i2 of the parameter error (without approximation) of the MSEP
    // the result is stored within FMSEPDetail(0,1)
    function InitMSEPParaErrDetail(): string; virtual;
    // calculates the addands m1, m2, i1, i2 of the approximated parameter error of the MSEP
    // the result is stored within FMSEPDetail(1,1)
    function InitMSEPParaErrProxyDetail(): string; virtual;
    // calculates the Tail parameter error (without approximation) of the MSEP
    // the result is stored within FMSEPTail(0,1)
    function InitMSEPParaErrTail(): string; virtual;
    // calculates the approximated Tail parameter error of the MSEP
    // the result is stored within FMSEPTail(1,1)
    function InitMSEPParaErrTailProxy(): string; virtual;
    // calculates the addands m1, m2, i1, i2 of the of the Tail parameter error (without approximation)
    // of the MSEP
    // the result is stored within FMSEPTailDetail(0,1)
    function InitMSEPParaErrTailDetail(): string; virtual;
    // calculates the addands m1, m2, i1, i2 of the of the approximated Tail parameter error of the MSEP
    // the result is stored within FMSEPTailDetail(1,1)
    function InitMSEPParaErrTailProxyDetail(): string; virtual;
    {$ENDREGION}
    {$REGION 'MSEP Process Variance'}
    // calculates the process variance (without approximation) of the MSEP
    // the result is stored within FMSEP(0,0)
    function InitMSEPProcVar(): string; virtual;
    // calculates the approximated process variance of the MSEP
    // the result is stored within FMSEP(1,0)
    function InitMSEPProcVarProxy(): string; virtual;
    // calculates the addands m1, m2, i1, i2 of the process variance (without approximation) of the MSEP
    // the result is stored within FMSEPDetail(0,0)
    function InitMSEPProcVarDetail(): string; virtual;
    // calculates the addands m1, m2, i1, i2 of the approximated process variance of the MSEP
    // the result is stored within FMSEPDetail(1,0)
    function InitMSEPProcVarProxyDetail(): string; virtual;
    // calculates of the Tail process variance (without approximation) of the MSEP
    // the result is stored within FMSEPTail(0,0)
    function InitMSEPProcVarTail(): string; virtual;
    // calculates of the approximated Tail process variance of the MSEP
    // the result is stored within FMSEPTail(1,0)
    function InitMSEPProcVarTailProxy(): string; virtual;
    // calculates the addands m1, m2, i1, i2 of the Tail process (withot approximation) variance of the MSEP
    // the result is stored within FMSEPTailDetail(0,0)
    function InitMSEPProcVarTailDetail(): string; virtual;
    // calculates the addands m1, m2, i1, i2 of the approximated Tail process variance of the MSEP
    // the result is stored within FMSEPTailDetail(1,0)
    function InitMSEPProcVarTailProxyDetail(): string; virtual;
    {$ENDREGION}
    {$ENDREGION}
    // calculates default variance parameters
    function InitDefaultsigma(): String; virtual;
    // calculates default weights if necessary
    function InitDefaultw(): String; virtual;
    // initialises model parameter gamma^m,l_i,k,h,j
    function InitGamma1(const Gamma1: OleVariant; Hasgamma1, HasNewInflation: boolean;
                        var NewInflation: Double): String; virtual;
    // initialises model parameter gamma^m1,m2,l_i,k,h,j
    function InitGamma2(const Gamma2: OleVariant; Hasgamma2, HasNewInflation: boolean;
                        var NewInflation: Double): String; virtual;
    // initialises geometry (M, I, J and K) if necessary
    function InitGeometry(M, I, J, K: Integer): String; virtual;
    // initialises manual specified development factors if necessary
    function Initf(var f: Double): String; virtual;
    // calculates the past of exposures R^m_i,k
    function InitR1(): String; virtual;
    // calculate the past of exposure R^m1,m2_i,k
    function InitR2(): String; virtual;
    // calculate the future of exposure R^m1,m2_i,k
    function InitR2Full(): String; virtual;
    // calculates rhos needed for CDR, i.e. all
    function InitrhoCDR(): String; virtual;
    // calculates rhos needed for MSEP, i.e. only 0 and rho*
    function InitrhoMSEP(): String; virtual;
    // calculate rho* without indizes i1 and i2
    function InitrhoStar(): String; virtual;
    // initialises the past of claim properties if necessary
    function InitS(var S: Double): String; virtual;
    // initialises manual specifies variance parameters if necessary
    function Initsigma(var sigma: Double): String; virtual;
    // copies the past of SxS into SxS(0)
    function InitSxS0(): string; virtual;
    // calculates H(rhobar)SxS and stores result within SxS(5)
    function InitSxSCDRProcVar():string; virtual;
    // calculates H(rho*1)SxS, H(rho*2)SxS and H(rho*12)SxS and stores the result within SxS(2) ... SxS(4)
    function InitSxSCDRParaErr():string; virtual;
    // calculates H(rho*)SxS and stores the result within SxS(1)
    function InitSxSMSEP():string; virtual;
    // calculates the future of claim properties and exposures R^m_i,k
    function InitUltimateDetail(): string; virtual;
    // calculates the future of alpha-mixed ultimates
    function InitUltimate(): string; virtual;
    // initialises manual specified weights if necessary
    function Initw(var w: Double): String; virtual;
    {$ENDREGION}
    {$REGION 'properties to access the private variables'}
    property CalcInfo[Step:Integer]: Boolean read GetCalcInfo;
    property CalcInfoName[Step:Integer]: string read GetCalcInfoName;
    property CalcTime[Step:Integer]: Double read GetCalcTime;
    property CalculationFlags: TCalculationFlagSet read GetCalculationFlags;
    property CDR[m, i, Proxy, Part, Tail: Integer]: Double read GetCDR;
    property CDRCov[m1, m2, Proxy, Part, Tail: Integer]: Double read GetCDRCov;
    property CDRCovEigenvalue[m, Proxy, Part, Tail: Integer]: Double read GetCDRCovEigenvalue;
    property CDRCovEigenvector[m, l, Proxy, Part, Tail: Integer]: Double read GetCDRCovEigenvector;
    property CDRDetail[m1, m2, i1, i2, Proxy, Part, Tail: Integer]: Double read GetCDRDetail;
    property CoorF[m, i, k, l, h , j: Integer]: Double read GetCoorF;
    property Defaultalpha[m, i: integer]: Double read GetDefaultalpha;
    property DefaultalphaUsed: Boolean read GetDefaultalphaUsed;
    property Defaultf[m, k: integer]: Double read GetDefaultf;
    property DefaultfUsed: Boolean read GetDefaultfUsed;
    property Defaultsigma[m1, m2, k: integer]: Double read GetDefaultsigma;
    property DefaultsigmaUsed: Boolean read GetDefaultsigmaUsed;
    property Defaultw[n, m, i, k: integer]: Double read GetDefaultw;
    property DefaultwUsed: Boolean read GetDefaultwUsed;
    property Eigenvalue[i, k, m: integer]: Double read GetEigenvalue;
    property Eigenvector[i, k, m, l: integer]: Double read GetEigenvector;
    property I: integer read GetI;
    property InitialisedParameters: integer read GetInitialisedParameters;
    property J: integer read GetJ;
    property K: integer read GetK;
    property M: integer read GetM;
    property MSEP[m, i, Proxy, Part, Tail: Integer]: Double read GetMSEP;
    property MSEPCov[m1, m2, Proxy, Part, Tail: Integer]: Double read GetMSEPCov;
    property MSEPCovEigenvalue[m, Proxy, Part, Tail: Integer]: Double read GetMSEPCovEigenvalue;
    property MSEPCovEigenvector[m, l, Proxy, Part, Tail: Integer]: Double read GetMSEPCovEigenvector;
    property MSEPDetail[m1, m2, i1, i2, Proxy, Part, Tail: Integer]: Double read GetMSEPDetail;
    property R1[m, i, k: integer]: Double read GetR1;
    property R2[m1, m2, i, k: integer]: Double read GetR2;
    property ClaimProperty[m, i, k: integer]: Double read GetS;
    property Ultimate[m, i, Tail: Integer]: Double read GetUltimate;
    property Usedalpha[m, i: Integer]: Double read GetUsedalpha;
    property Usedf[m, k: integer]: Double read GetUsedf;
    property Usedsigma[m1, m2, k: integer]: Double read GetUsedsigma;
    property Usedw[n, m, i, k: integer]: Double read GetUsedw;
class  property MaxVarCoef1: double read GetMaxVarCoef1 write SetMaxVarCoef1;
class  property MaxVarCoef2: double read GetMaxVarCoef2 write SetMaxVarCoef2;
    {$ENDREGION}
    {$ENDREGION}
end;

{$REGION 'Description ...'}
// For TSimpleLSRM we assume that
//    gamma^ml_ikhj=0 for i<>h
//    gamma^ml_ikij=gamma^ml_hkhj for all i and h
//    gamma^m1m2l_ikhj=0 for i<>h
//    gamma^m1m2l_ikij=gamma^m1m2l_hkhj for all i and h
// Therefore, some of addents of sums can be skipt and we can reduce the memory for Fgamma1, Fgamma2, FCoorF.
// Moreover, we allow for a formula specification of the model parameters. This means, the corresponding
// exposures R^m_ik or R^m1,m2_ik can be determined triangle wise by formulas.
// Formulas mean that for each 0<=m<=M (or 0<=m1,m2<=M, here the order is first m1 then m2) we have a string
// (formula) constructed as described as follows and separated by semicolons.
// We support the following operations:
//  - Signs + and -
//  - operators +, -, *, / and ^ (power; not allowed for gamma1)
//  - constant real numbers
//  - vectors: a collection of FJ+FK (or FJ+FP) real numbers separated by semicolons and enclosed within square brackets
//             this values are values for each column
//  - pais of round braces: priorisise the enclosed term to be calculated first
//  - Function:
//    - S(m): Claim property where m is an integer with 0<=m<=FM,
//    - Cum(triangle): cummulation of triangle
//    - Inc(triangle): incremental of triangle
//    - Abs(triangle): absolut value of each component of triangle (not allowed for gamma1)
//    - Sqrt(triangle): square root of each component of triangle (not allowed for gamma1)
//    - Column(k, triangle): empties all columns, except for column k, of the trinagle
//    - Max(A,B): takes the maximum of triangles A and B, componenten wise (not allowed for gamma1)
//    - Min(A,B): takes the mminimum of triangles A and B, componenten wise (not allowed for gamma1)
// Examples for formulas of R^m_ik are:
//  - "S(0)": R^m_ik = S^0_ik
//  - "2*S(0)-S(1)": R^m_ik = 2*S^0_ik - S^1_ik
//  - "Cum(S(0))": R^m_ik = Cum(S^0_ik), i.e. the Chain Ladder exposure
//  - "Cum(S(1))-Cum(S(0))": R^m_ik = Cum(S^1_ik)-Cum(S^0_ik),
//    if S^1 are reported amounts and S^0 are payments then R^m_ik represent the corresponding case reserves.
//  - "[1;0.5;0;0;0]*Cum(S(1))-[1;0.5;0;0;0]*Cum(S(0))+[0;0.5;1;1;1]*Cum(S(1))":
//    if the total number of development periods (FJ+FK+1) equals 5 and if S^1 are reported amounts and S^0
//    are payments then the formula represents a exposure which equals:
//      - case reserves for the first development period
//      - a 50%-50% mixture of case reserves and reported amounts for the second development period
//      - reported amounts for the last three development periods (Chain-Ladder on reported amounts)
//  - Sqrt(Max(0,Cum(S(0)))*Max(0,Cum(S(1))))): a good choise for the covariance coupling of two Chain Ladders
{$ENDREGION}
type TSimpleLSRM=class(TNormalLSRM)
  private {$REGION}
    {$REGION 'variables'}
    class var
    // special charcter in formulas, i.e. white chars, formula seperator ...
    FTokenChars: TTokenChars;
    class function GetDelimiters: String; static;
    class procedure SetDelimiters(const Value: String); static;
    var
    // the parser for formula specified R2
    FFormulas: TFormulaParser;
    // Since gamma^ml_ikhj is special like specified above we have that the coordinates of the operator F
    //   (F^n,m_i,k)^l_h,j: m(M) x [i(I)] x k(J-1+FP) [x n(I+J+FP)] x l(M) x [h(I)] x j(J+FP)
    // are independend of i and h and since we only need those with n=h+j, we skip the dimensions i, n and h.
    // Since
    // model parameter gamma^m,l_i,k,h,j: m(M) x l(M) x [i(I)] x k(J - 1 + K) x [h(I)] x j(J - 1 + K) and
    // model parameter gamma^m1,m2,l_i,k,h,j: m1(M) x m2(M) x l(M) x [i(I)] x k(J-1+FP) x [h(I)] x j(J-1+FP)
    // do not depend on i and h we skip those dimensions, too.
    {$ENDREGION}
    {$ENDREGION}
  protected {$REGION}
    {$REGION 'property get functions'}
    function GetCoorF(m, i, k, l, h, j:integer): Double; override;
    {$ENDREGION}
    {$REGION 'operators'}
    // the operator F^n
    function OP_F(n: integer; S: T1Double): string; override;
    // the opertor H^n_k(tau)
    function OP_H(n, k: integer; tau, SxS: PDouble): string; override;
    {$ENDREGION}
  {$ENDREGION}
  public {$REGION}
    constructor Create(Name: string); override;
    destructor Destroy; override;
    {$REGION 'Init functions'}
    {$REGION 'CDR'}
    // calculates the addands m1, m2, i1, i2 of the approximated parameter error of the CDR
    // the result is stored within FCDRDetail(1,1)
    function InitCDRParaErrProxyDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the tail parameter error (without approximation) of the CDR
    // the result is stored within FCDRTailDetail(1,1)
    function InitCDRParaErrTailDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the approximated tail parameter error of the CDR
    // the result is stored within FCDRTailDetail(1,1)
    function InitCDRParaErrTailProxyDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the approximated process variance of the CDR
    // the result is stored within FCDR(1,0)
    function InitCDRProcVarProxyDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the  tail process variance (without approximation) of the CDR
    // the result is stored within FCDRTail(1,0)
    function InitCDRProcVarTailDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the approximated tail process variance of the CDR
    // the result is stored within FCDRTail(1,0)
    function InitCDRProcVarTailProxyDetail(): string; override;
    {$ENDREGION}
    // calculates the coordinates of operator F
    function InitCoorF(): string; override;
    {$REGION 'MSEP'}
    // calculates the addands m1, m2, i1, i2 of the approximated parameter error of the MSEP
    // the result is stored within FMSEPDetail(1,1)
    function InitMSEPParaErrProxyDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the tail parameter error (without approximation) of the MSEP
    // the result is stored within FMSEPTailDetail(0,1)
    function InitMSEPParaErrTailDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the approximated tail parameter error of the MSEP
    // the result is stored within FMSEPTailDetail(1,1)
    function InitMSEPParaErrTailProxyDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the process variance (without approximation) of the MSEP
    // the result is stored within FMSEPDetail(0,0)
    function InitMSEPProcVarDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the approximated process variance of the MSEP
    // the result is stored within FMSEPDetail(1,0)
    function InitMSEPProcVarProxyDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the tail process variance (without approximation) of the MSEP
    // the result is stored within FMSEPTailDetail(0,0)
    function InitMSEPProcVarTailDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the approximated tail process variance of the MSEP
    // the result is stored within and FMSEPTailDetail(1,0)
    function InitMSEPProcVarTailProxyDetail(): string; override;
    {$ENDREGION}
    // initialises model parameter gamma^m,l_i,k,h,j
    function InitGamma1(const Gamma1: OleVariant; Hasgamma1, HasNewInflation: boolean;
                        var NewInflation: Double): String; override;
    // initialises model parameter gamma^m1,m2,l_i,k,h,j
    function InitGamma2(const Gamma2: OleVariant; Hasgamma2, HasNewInflation: boolean;
                        var NewInflation: Double): String; override;
    // calculates the past of exposures R^m_i,k
    function InitR1(): String; override;
    // calculates the past of exposure R^m1,m2_i,k
    function InitR2(): String; override;
    // calculates the future of exposure R^m1,m2_i,k
    function InitR2Full(): String; override;
    // calculates the future of claim properties and exposures R^m_i,k
    function InitUltimateDetail(): string; override;
    {$ENDREGION}
    {$REGION 'properties'}
class  property Delimiters: String read GetDelimiters write SetDelimiters;
    {$ENDREGION}
end;

{$REGION 'Description ...'}
// If we have inflation we will still have (F^n,m_i,k)^l_h,j^= 0 for i<>j, but
// those coordinates will depend on i.
{$ENDREGION}
type TSimpleInflationLSRM=class(TSimpleLSRM)
  private {$REGION}
    {$REGION 'variables'}
    // The coordinates of the operator F
    // (F^n,m_i,k)^l_h,j: m(M) x i(I) x k(J-1+FP) [x n(I+J+FP)] x l(M) x [h(I)] x j(J+FP)
    // are zero for h<>i and since we only need those with n=h+j, we skip the dimensions n and h
    // (not i like for simple LSRM's without inflation).
    {$ENDREGION}
    // sets the initialisation flags according to dependencies
    // since gammas and inflation are handled separately gammas do not depend on inflation
    procedure SetInitFlags(InitFlag: TLSRMInitialisedElement; UsedTime: Int64; UpdateTime: Boolean=True);
    {$ENDREGION}
  protected {$REGION}
    {$REGION 'property get functions'}
    function GetCoorF(m, i, k, l, h, j:integer): Double; override;
    {$ENDREGION}
    {$REGION 'operators'}
    // the operator F^n
    function OP_F(n: integer; S: T1Double): string; override;
    // the opertor H^n_k(tau)
    function OP_H(n, k: integer; tau, SxS: PDouble): string; override;
    {$ENDREGION}
  {$ENDREGION}
  public {$REGION}
    {$REGION ' Init functions'}
    {$REGION 'CDR'}
    // calculates the addands m1, m2, i1, i2 of the approximated parameter error of the CDR
    // the result is stored within FCDRDetail(1,1)
    function InitCDRParaErrProxyDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the tail parameter error (without approximation) of the CDR
    // the result is stored within FCDRTailDetail(1,1)
    function InitCDRParaErrTailDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the approximated tail parameter error of the CDR
    // the result is stored within FCDRTailDetail(1,1)
    function InitCDRParaErrTailProxyDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the approximated process variance of the CDR
    // the result is stored within FCDR(1,0)
    function InitCDRProcVarProxyDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the  tail process variance (without approximation) of the CDR
    // the result is stored within FCDRTail(1,0)
    function InitCDRProcVarTailDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the approximated tail process variance of the CDR
    // the result is stored within FCDRTail(1,0)
    function InitCDRProcVarTailProxyDetail(): string; override;
    {$ENDREGION}
    // initialises coordinages of operator F if necessary
    function InitCoorF(): string; override;
    {$REGION 'MSEP'}
    // calculates the addands m1, m2, i1, i2 of the approximated parameter error of the MSEP
    // the result is stored within FMSEPDetail(1,1)
    function InitMSEPParaErrProxyDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the tail parameter error (without approximation) of the MSEP
    // the result is stored within FMSEPTailDetail(1,1)
    function InitMSEPParaErrTailDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the approximated tail parameter error of the MSEP
    // the result is stored within FMSEPTailDetail(1,1)
    function InitMSEPParaErrTailProxyDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the process variance (without approximation) of the MSEP
    // the result is stored within FMSEPDetail(0,0)
    function InitMSEPProcVarDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the approximated process variance of the MSEP
    // the result is stored within FFMSEPDetail(1,0)
    function InitMSEPProcVarProxyDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the tail process variance (without approximation) of the MSEP
    // the result is stored within FMSEPTailDetail(0,0)
    function InitMSEPProcVarTailDetail(): string; override;
    // calculates the addands m1, m2, i1, i2 of the approximated tail process variance of the MSEP
    // the result is stored within FMSEPTailDetail(1,0)
    function InitMSEPProcVarTailProxyDetail(): string; override;
    {$ENDREGION}
    // calculates the past of the exposures R^m_i,k
    function InitR1(): String; override;
    // calculate the past of the exposure R^m1,m2_i,k
    function InitR2(): String; override;
    // calculate the future of the exposure R^m1,m2_i,k
    function InitR2Full(): String; override;
    // calculates the future of claim properties and exposures R^m_i,k
    function InitUltimateDetail(): string; override;
    {$ENDREGION}
    {$ENDREGION}
end;

implementation

uses Variants, Windows, Consts, Math, StrUtils, LSRMTools_TLB, LSRMTools_RS;

var {$REGION 'two global variables, that will be initialised within initialization'}
  // this variable is used to convert time measured by the frequency of the high-resolution performance
  // counter into miliseconds
  // it will be initialised within initialization
  FFrequence: Int64;
  // the short descriptions of each initialisation step of a LSRM
  FCalcInfoNames:array[TLSRMInitialisedElement] of string;
  {$ENDREGION}

{$REGION 'TNormalLSRM'}

{$REGION '  TNormalLSRM private'}

procedure TNormalLSRM.FreeAll;
begin
  FM:=-1;
  FI:=-1;
  FJ:=-1;
  FK:=0;
  FP:=0;
  SetLength(FS,0);
  SetLength(FInflation,0);
  SetLength(FR1,0);
  SetLength(FR2,0);
  SetLength(Fw,0);
  SetLength(FDefaultw,0);
  FUsedw:=nil;
  SetLength(Ff,0);
  SetLength(FDefaultf,0);
  FUsedf:=nil;
  SetLength(Fsigma,0);
  SetLength(FDefaultsigma,0);
  FUsedsigma:=nil;
  SetLength(Falpha,0);
  SetLength(FDefaultalpha,0);
  FUsedalpha:=nil;
  SetLength(FGamma1,0);
  SetLength(FGamma2,0);
  SetLength(FUltimate,0);
  SetLength(FCoorF,0);
  SetLength(FrhoStar,0);
  SetLength(Frho,0);
  Setlength(FSxS,0);
  SetLength(FMSEPDetail,0);
  SetLength(FMSEPTailDetail,0);
  SetLength(FMSEP,0);
  SetLength(FMSEPTail,0);
  SetLength(FCDRDetail,0);
  SetLength(FCDRTailDetail,0);
  SetLength(FCDR,0);
  SetLength(FCDRTail,0);
  Setlength(FEigenvalues,0);
  Setlength(FEigenvectors,0);
  Setlength(FCDRCov,0);
  Setlength(FMSEPCov,0);
end;

procedure TNormalLSRM.NormalizeWeights(w: T1Double);
var i, k, m: integer;
    TmpSum: Double;
    PSourceDouble, PDestDouble, PExposureDouble: IndexDouble;
begin
  for m := 0 to FM do begin
    For k:=0 to FJ-1+FP do begin
      // normalise weights for estimation period I
      // a weight is set to zero if the corresponding exposure equals zero
      TmpSum:=0;
      PSourceDouble:=m*(FI+1)*(FJ+FP)+k;
      PExposureDouble:=m*(FI+1)*(FJ+FK)+k; // Exposure is R1
      for i:=0 to FI-k-1 do begin
        if FR1[PExposureDouble]<>0 then TmpSum:=TmpSum+w[PSourceDouble];
        Inc(PSourceDouble,FJ+FP);
        Inc(PExposureDouble,FJ+FK);
      end;
      if TmpSum>0 then begin
        PSourceDouble:=m*(FI+1)*(FJ+FP)+k;
        PExposureDouble:=m*(FI+1)*(FJ+FK)+k;
        PDestDouble:=((FM+1)+m)*(FI+1)*(FJ+FP)+k;
        for i:=0 to FI-k-1 do begin
          if FR1[PExposureDouble]<>0 then
            w[PDestDouble]:=w[PSourceDouble]/TmpSum
          else
            w[PDestDouble]:=0;
          Inc(PSourceDouble,FJ+FP);
          Inc(PExposureDouble,FJ+FK);
          Inc(PDestDouble,FJ+FP);
        end;
      end else begin
        PDestDouble:=((FM+1)+m)*(FI+1)*(FJ+FP)+k;
        ZeroMemory(@w[PDestDouble],(FI-k)*SizeOf(Double));
      end;
      // normalise weights for estimation period I+1
      // a weight is set to zero if the corresponding exposure equals zero
      if FR1[PExposureDouble]<>0 then TmpSum:=TmpSum+w[PSourceDouble];
      if TmpSum>0 then begin
        PSourceDouble:=m*(FI+1)*(FJ+FP)+k;
        PExposureDouble:=m*(FI+1)*(FJ+FK)+k;
        PDestDouble:=(2*(FM+1)+m)*(FI+1)*(FJ+FP)+k;
        for i:=0 to FI-k do begin
          if FR1[PExposureDouble]<>0 then
            w[PDestDouble]:=w[PSourceDouble]/TmpSum
          else
            w[PDestDouble]:=0;
          Inc(PSourceDouble,FJ+FP);
          Inc(PExposureDouble,FJ+FK);
          Inc(PDestDouble,FJ+FP);
        end;
      end else begin
        PDestDouble:=(2*(FM+1)+m)*(FI+1)*(FJ+FP)+k;
        ZeroMemory(@w[PDestDouble],(FI-k+1)*SizeOf(Double));
      end;
    end;
  end;
end;

procedure TNormalLSRM.SetInitFlags(InitFlag: TLSRMInitialisedElement; UsedTime: Int64;
  UpdateTime: Boolean=True);
var NonDefaultExceptions: TLSRMInitialised;
begin
  include(FInitialised,InitFlag);
  case InitFlag of
    Init_Geometry: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByGeometry;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByGeometry, UsedTime);
    end; {$ENDREGION}
    Init_S: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByS;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByS, UsedTime);
    end; {$ENDREGION}
    Init_InflationRate: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByInflationRate;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByInflationRate, UsedTime);
    end; {$ENDREGION}
    Init_Gamma1: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByGamma1;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByGamma1, UsedTime);
    end; {$ENDREGION}
    Init_Gamma2: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByGamma2;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByGamma2, UsedTime);
    end; {$ENDREGION}
    Init_R1: begin {$REGION}
      // Init_CoorF depends on R1 only via default f
      NonDefaultExceptions:=[];
      if FUsedf<>@FDefaultf then begin
        NonDefaultExceptions:=NonDefaultExceptions+[Init_CoorF];
      end;
      FInitialised:=FInitialised-(cInfluencedByR1-NonDefaultExceptions);
      if UpdateTime then UpdateCalcInfo(InitFlag, (cInfluencedByR1-NonDefaultExceptions), UsedTime);
    end; {$ENDREGION}
    Init_R2: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByR2;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByR2, UsedTime);
    end; {$ENDREGION}
    Init_w: begin {$REGION}
      // only if manual weights are used some other objects have to be changed
      if FUsedw=@Fw then begin
        // Init_CoorF, Init_R2Full, Init_SxS0, Init_Ultimate and Init_UltimateDetail
        // depend on w only via default f
        NonDefaultExceptions:=[];
        if FUsedf<>@FDefaultf then begin
          NonDefaultExceptions:=NonDefaultExceptions
                                +[Init_CoorF, Init_R2Full, Init_SxS0, Init_Ultimate, Init_UltimateDetail];
        end;
        // Init_rhoMSEP and Init_rhoStar depend on w only via default sigma
        if FUsedsigma<>@FDefaultsigma then begin
          NonDefaultExceptions:=NonDefaultExceptions+[Init_rhoMSEP, Init_rhoStar];
        end;
        FInitialised:=FInitialised-(cInfluencedByw-NonDefaultExceptions);
        if UpdateTime then UpdateCalcInfo(InitFlag, (cInfluencedByw-NonDefaultExceptions), UsedTime);
      end else
        if UpdateTime then UpdateCalcInfo(InitFlag, [], UsedTime);
    end; {$ENDREGION}
    Init_Defaultw: begin {$REGION}
      // only if default weights are used some other objects have to be changed
      if FUsedw=@FDefaultw then begin
        // Init_CoorF, Init_R2Full, Init_SxS0, Init_Ultimate and Init_UltimateDetail
        // depend on w only via default f
        NonDefaultExceptions:=[];
        if FUsedf<>@FDefaultf then begin
          NonDefaultExceptions:=NonDefaultExceptions
                                +[Init_CoorF, Init_R2Full, Init_SxS0, Init_Ultimate, Init_UltimateDetail];
        end;
        // Init_rhoMSEP and Init_rhoStar depend on w only via default sigma
        if FUsedsigma<>@FDefaultsigma then begin
          NonDefaultExceptions:=NonDefaultExceptions+[Init_rhoMSEP, Init_rhoStar];
        end;
        FInitialised:=FInitialised-(cInfluencedByDefaultw-NonDefaultExceptions);
        if UpdateTime then UpdateCalcInfo(InitFlag, (cInfluencedByDefaultw-NonDefaultExceptions), UsedTime);
      end else
        if UpdateTime then UpdateCalcInfo(InitFlag, [], UsedTime);
    end; {$ENDREGION}
    Init_f: begin {$REGION}
      // only if manual development factors are used some other objects have to be changed
      if FUsedf=@Ff then begin
        // Init_rhoMSEP and Init_rhoStar depend on f only via default sigma
        NonDefaultExceptions:=[];
        if FUsedsigma<>@FDefaultsigma then begin
          NonDefaultExceptions:=NonDefaultExceptions+[Init_rhoMSEP, Init_rhoStar];
        end;
        FInitialised:=FInitialised-(cInfluencedByf-NonDefaultExceptions);
        if UpdateTime then UpdateCalcInfo(InitFlag, (cInfluencedByf-NonDefaultExceptions), UsedTime);
      end else
        if UpdateTime then UpdateCalcInfo(InitFlag, [], UsedTime);
    end; {$ENDREGION}
    Init_Defaultf: begin {$REGION}
      // only if default development factors are used some other objects have to be changed
      if FUsedf=@FDefaultf then begin
        // Init_rhoMSEP and Init_rhoStar depend on f only via default sigma
        NonDefaultExceptions:=[];
        if FUsedsigma<>@FDefaultsigma then begin
          NonDefaultExceptions:=NonDefaultExceptions+[Init_rhoMSEP, Init_rhoStar];
        end;
        FInitialised:=FInitialised-(cInfluencedByDefaultf-NonDefaultExceptions);
        if UpdateTime then UpdateCalcInfo(InitFlag, (cInfluencedByDefaultf-NonDefaultExceptions), UsedTime);
      end else
        if UpdateTime then UpdateCalcInfo(InitFlag, [], UsedTime);
    end; {$ENDREGION}
    Init_sigma: begin {$REGION}
      if FUsedsigma=@Fsigma then begin
        FInitialised:=FInitialised-cInfluencedBysigma;
        if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedBysigma, UsedTime);
      end else // only if manual variance parameter are used some other objects may change
        if UpdateTime then UpdateCalcInfo(InitFlag, [], UsedTime);
    end; {$ENDREGION}
    Init_Defaultsigma: begin {$REGION}
      if FUsedsigma=@FDefaultsigma then begin
        FInitialised:=FInitialised-cInfluencedByDefaultsigma;
        if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByDefaultsigma, UsedTime);
      end else // only if default variance parameter are used some other objects may change
        if UpdateTime then UpdateCalcInfo(InitFlag, [], UsedTime);
    end;  {$ENDREGION}
    Init_alpha: begin {$REGION}
      if FUsedalpha=@Falpha then begin
        FInitialised:=FInitialised-cInfluencedByalpha;
        if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByalpha, UsedTime);
      end else // only if manual mixing-alphas are used some other objects may change
        if UpdateTime then UpdateCalcInfo(InitFlag, [], UsedTime);
    end; {$ENDREGION}
    Init_Defaultalpha: begin {$REGION}
      if FUsedalpha=@FDefaultalpha then begin
        FInitialised:=FInitialised-cInfluencedByDefaultalpha;
        if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByDefaultalpha, UsedTime);
      end else // only if default mixing-alphas are used some other objects may change
        if UpdateTime then UpdateCalcInfo(InitFlag, [], UsedTime);
    end; {$ENDREGION}
    Init_UltimateDetail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByUltimateDetail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByUltimateDetail, UsedTime);
    end; {$ENDREGION}
    Init_R2Full: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByR2Full;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByR2Full, UsedTime);
    end; {$ENDREGION}
    Init_Ultimate: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByUltimate;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByUltimate, UsedTime);
    end; {$ENDREGION}
    Init_CoorF: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCoorF;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCoorF, UsedTime);
    end; {$ENDREGION}
    Init_rhoStar: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByrhoStar;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByrhoStar, UsedTime);
    end; {$ENDREGION}
    Init_rhoMSEP: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByrhoMSEP;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByrhoMSEP, UsedTime);
    end; {$ENDREGION}
    Init_rhoCDR: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByrhoCDR;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByrhoCDR, UsedTime);
    end; {$ENDREGION}
    Init_SxS0: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedBySxS0;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedBySxS0, UsedTime);
    end; {$ENDREGION}
    Init_SxSMSEP: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedBySxSMSEP;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedBySxSMSEP, UsedTime);
    end; {$ENDREGION}
    Init_MSEPProcVarDetail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPProcVarDetail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPProcVarDetail, UsedTime);
    end; {$ENDREGION}
    Init_MSEPProcVarProxyDetail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPProcVarProxyDetail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPProcVarProxyDetail, UsedTime);
    end; {$ENDREGION}
    Init_MSEPProcVar: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPProcVar;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPProcVar, UsedTime);
    end; {$ENDREGION}
    Init_MSEPProcVarProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPProcVarProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPProcVarProxy, UsedTime);
    end; {$ENDREGION}
    Init_MSEPProcVarTailDetail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPProcVarTailDetail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPProcVarTailDetail, UsedTime);
    end; {$ENDREGION}
    Init_MSEPProcVarTailProxyDetail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPProcVarTailProxyDetail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPProcVarTailProxyDetail, UsedTime);
    end; {$ENDREGION}
    Init_MSEPProcVarTail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPProcVarTail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPProcVarTail, UsedTime);
    end; {$ENDREGION}
    Init_MSEPProcVarTailProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPProcVarTailProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPProcVarTailProxy, UsedTime);
    end; {$ENDREGION}
    Init_MSEPParaErrDetail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPParaErrDetail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPParaErrDetail, UsedTime);
    end; {$ENDREGION}
    Init_MSEPParaErrProxyDetail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPParaErrProxyDetail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPParaErrProxyDetail, UsedTime);
    end; {$ENDREGION}
    Init_MSEPParaErr: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPParaErr;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPParaErr, UsedTime);
    end; {$ENDREGION}
    Init_MSEPParaErrProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPParaErrProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPParaErrProxy, UsedTime);
    end; {$ENDREGION}
    Init_MSEPParaErrTailDetail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPParaErrTailDetail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPParaErrTailDetail, UsedTime);
    end; {$ENDREGION}
    Init_MSEPParaErrTailProxyDetail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPParaErrTailProxyDetail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPParaErrTailProxyDetail, UsedTime);
    end; {$ENDREGION}
    Init_MSEPParaErrTail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPParaErrTail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPParaErrTail, UsedTime);
    end; {$ENDREGION}
    Init_MSEPParaErrTailProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPParaErrTailProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPParaErrTailProxy, UsedTime);
    end; {$ENDREGION}
    Init_MSEP: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEP;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEP, UsedTime);
    end; {$ENDREGION}
    Init_MSEPProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPProxy, UsedTime);
    end; {$ENDREGION}
    Init_MSEPTail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPTail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPTail, UsedTime);
    end; {$ENDREGION}
    Init_MSEPTailProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPTailProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPTailProxy, UsedTime);
    end; {$ENDREGION}
    Init_SXSCDRProcVar: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedBySxSCDRProcVar;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedBySxSCDRProcVar, UsedTime);
    end; {$ENDREGION}
    Init_SXSCDRParaErr: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedBySxSCDRParaErr;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedBySxSCDRParaErr, UsedTime);
    end; {$ENDREGION}
    Init_CDRProcVarDetail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRProcVarDetail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRProcVarDetail, UsedTime);
    end; {$ENDREGION}
    Init_CDRProcVarProxyDetail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRProcVarProxyDetail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRProcVarProxyDetail, UsedTime);
    end; {$ENDREGION}
    Init_CDRProcVar: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRProcVar;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRProcVar, UsedTime);
    end; {$ENDREGION}
    Init_CDRProcVarProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRProcVarProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRProcVarProxy, UsedTime);
    end; {$ENDREGION}
    Init_CDRProcVarTailDetail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRProcVarTailDetail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRProcVarTailDetail, UsedTime);
    end; {$ENDREGION}
    Init_CDRProcVarTailProxyDetail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRProcVarTailProxyDetail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRProcVarTailProxyDetail, UsedTime);
    end; {$ENDREGION}
    Init_CDRProcVarTail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRProcVarTail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRProcVarTail, UsedTime);
    end; {$ENDREGION}
    Init_CDRProcVarTailProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRProcVarTailProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRProcVarTailProxy, UsedTime);
    end; {$ENDREGION}
    Init_CDRParaErrDetail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRParaErrDetail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRParaErrDetail, UsedTime);
    end; {$ENDREGION}
    Init_CDRParaErrProxyDetail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRParaErrProxyDetail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRParaErrProxyDetail, UsedTime);
    end; {$ENDREGION}
    Init_CDRParaErr: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRParaErr;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRParaErr, UsedTime);
    end; {$ENDREGION}
    Init_CDRParaErrProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRParaErrProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRParaErrProxy, UsedTime);
    end; {$ENDREGION}
    Init_CDRParaErrTailDetail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRParaErrTailDetail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRParaErrTailDetail, UsedTime);
    end; {$ENDREGION}
    Init_CDRParaErrTailProxyDetail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRParaErrTailProxyDetail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRParaErrTailProxyDetail, UsedTime);
    end; {$ENDREGION}
    Init_CDRParaErrTail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRParaErrTail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRParaErrTail, UsedTime);
    end; {$ENDREGION}
    Init_CDRParaErrTailProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRParaErrTailProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRParaErrTailProxy, UsedTime);
    end; {$ENDREGION}
    Init_CDR: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDR;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDR, UsedTime);
    end; {$ENDREGION}
    Init_CDRProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRProxy, UsedTime);
    end; {$ENDREGION}
    Init_CDRTail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRTail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRTail, UsedTime);
    end; {$ENDREGION}
    Init_CDRTailProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRTailProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRTailProxy, UsedTime);
    end; {$ENDREGION}
    Init_Eigenvalues: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByEigenvalues;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByEigenvalues, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCov: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCov;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCov, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovParaErr: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovParaErr;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovParaErr, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovProcVar: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovProcVar;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovProcVar, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovTail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovTail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovTail, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovParaErrTail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovParaErrTail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovParaErrTail, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovProcVarTail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovProcVarTail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovProcVarTail, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovProxy, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovParaErrProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovParaErrProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovParaErrProxy, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovProcVarProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovProcVarProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovProcVarProxy, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovTailProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovTailProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovTailProxy, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovParaErrTailProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovParaErrTailProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovParaErrTailProxy, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovProcVarTailProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovProcVarTailProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovProcVarTailProxy, UsedTime);
    end; {$ENDREGION}
    Init_CDRCov: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCov;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCov, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovParaErr: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovParaErr;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovParaErr, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovProcVar: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovProcVar;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovProcVar, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovTail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovTail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovTail, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovParaErrTail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovParaErrTail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovParaErrTail, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovProcVarTail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovProcVarTail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovProcVarTail, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovProxy, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovParaErrProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovParaErrProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovParaErrProxy, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovProcVarProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovProcVarProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovProcVarProxy, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovTailProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovTailProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovTailProxy, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovParaErrTailProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovParaErrTailProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovParaErrTailProxy, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovProcVarTailProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovProcVarTailProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovProcVarTailProxy, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovEigenvalue: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovEigenvalue;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovEigenvalue, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovEigenvalueParaErr: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovEigenvalueParaErr;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovEigenvalueParaErr, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovEigenvalueProcVar: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovEigenvalueProcVar;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovEigenvalueProcVar, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovEigenvalueTail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovEigenvalueTail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovEigenvalueTail, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovEigenvalueParaErrTail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovEigenvalueParaErrTail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovEigenvalueParaErrTail, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovEigenvalueProcVarTail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovEigenvalueProcVarTail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovEigenvalueProcVarTail, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovEigenvalueProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovEigenvalueProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovEigenvalueProxy, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovEigenvalueParaErrProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovEigenvalueParaErrProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovEigenvalueParaErrProxy, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovEigenvalueProcVarProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovEigenvalueProcVarProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovEigenvalueProcVarProxy, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovEigenvalueTailProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovEigenvalueTailProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovEigenvalueTailProxy, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovEigenvalueParaErrTailProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovEigenvalueParaErrTailProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovEigenvalueParaErrTailProxy, UsedTime);
    end; {$ENDREGION}
    Init_MSEPCovEigenvalueProcVarTailProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByMSEPCovEigenvalueProcVarTailProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByMSEPCovEigenvalueProcVarTailProxy, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovEigenvalue: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovEigenvalue;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovEigenvalue, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovEigenvalueParaErr: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovEigenvalueParaErr;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovEigenvalueParaErr, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovEigenvalueProcVar: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovEigenvalueProcVar;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovEigenvalueProcVar, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovEigenvalueTail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovEigenvalueTail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovEigenvalueTail, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovEigenvalueParaErrTail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovEigenvalueParaErrTail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovEigenvalueParaErrTail, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovEigenvalueProcVarTail: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovEigenvalueProcVarTail;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovEigenvalueProcVarTail, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovEigenvalueProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovEigenvalueProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovEigenvalueProxy, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovEigenvalueParaErrProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovEigenvalueParaErrProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovEigenvalueParaErrProxy, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovEigenvalueProcVarProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovEigenvalueProcVarProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovEigenvalueProcVarProxy, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovEigenvalueTailProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovEigenvalueTailProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovEigenvalueTailProxy, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovEigenvalueParaErrTailProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovEigenvalueParaErrTailProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovEigenvalueParaErrTailProxy, UsedTime);
    end; {$ENDREGION}
    Init_CDRCovEigenvalueProcVarTailProxy: begin {$REGION}
      FInitialised:=FInitialised-cInfluencedByCDRCovEigenvalueProcVarTailProxy;
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByCDRCovEigenvalueProcVarTailProxy, UsedTime);
    end; {$ENDREGION}
  end;
end;

class function TNormalLSRM.GetMaxVarCoef1: double;
begin
  result:=FMaxVarCoef1;
end;

class function TNormalLSRM.GetMaxVarCoef2: double;
begin
  result:=FMaxVarCoef2;
end;

class procedure TNormalLSRM.SetMaxVarCoef1(const Value: double);
begin
  FMaxVarCoef1:=Value;
end;

class procedure TNormalLSRM.SetMaxVarCoef2(const Value: double);
begin
  FMaxVarCoef2:=Value;
end;

procedure TNormalLSRM.UpdateCalcInfo(InitFlag: TLSRMInitialisedElement;
  InfluencedFlags: TLSRMInitialised; UsedTime: Int64);
var TmpFlag:TLSRMInitialisedElement;
begin
  FCalcInfo[InitFlag]:=True;
  FCalcTime[InitFlag]:=1000*UsedTime/FFrequence; // in milliseconds, FFrequence has to be initialised before!
  for TmpFlag in InfluencedFlags do begin
    FCalcInfo[TmpFlag]:=False;
    FCalcTime[TmpFlag]:=0;
  end;
end;

function TNormalLSRM.GetRho(n, m1, m2, i1, i2, k: Integer): Double;
begin
  Result:=0;
  case n of
    1: begin // rho*
      Result:= FRhoStar[(m1*(FM+1)+m2)*(FJ+FP)+k];
    end;
    2: begin // rho*1, same as in the article, accept for the tail k>=FJ
      If ((i1 + k > FI) And (i2 + k >= FI)) or (k>=FJ) then
        Result:=(1-FUsedw^[((2*(FM+1)+m1)*(FI+1)+FI-k)*(FJ+FP)+k])
                *Frhostar[(m1*(FM+1)+m2)*(FJ+FP)+k];
    end;
    3: begin // rho*2, same as in the article, accept for the tail k>=FJ
      If ((i1 + k >= FI) And (i2 + k > FI)) or (k>=FJ) Then
        Result:=(1-FUsedw^[((2*(FM+1)+m2)*(FI+1)+FI-k)*(FJ+FP)+k])
                *Frhostar[(m1*(FM+1)+m2)*(FJ+FP)+k];
    end;
    4: begin // rho*12, same as in the article, accept for the tail k>=FJ
      If ((i1 + k > FI) And (i2 + k > FI)) or (k>=FJ) Then
        Result:=(1-FUsedw^[((2*(FM+1)+m1)*(FI+1)+FI-k)*(FJ+FP)+k])
                *(1-FUsedw^[((2*(FM+1)+m2)*(FI+1)+FI-k)*(FJ+FP)+k])
                *Frhostar[(m1*(FM+1)+m2)*(FJ+FP)+k];
    end;
    5: begin // rho bar, same as in the article, accept for the tail k>=FJ
      Result:=0;
      if (FR1[(m1*(FI+1)+FI-k)*(FJ+FK)+k] <>0) and (FR1[(m2*(FI+1)+FI-k)*(FJ+FK)+k]<>0)  then begin
        if ((i1 > FI - k) or (k>=FJ)) and ((i2 > FI - k) or (k>=FJ)) then
          Result:=FUsedw^[((2*(FM+1)+m1)*(FI+1)+FI-k)*(FJ+FP)+k]
                  *FUsedw^[((2*(FM+1)+m2)*(FI+1)+FI-k)*(FJ+FP)+k]
                  *FUsedsigma^[(m1*(FM+1)+m2)*(FJ+FP)+k]
                  *FR2[((m1*(FM+1)+m2)*(FI+1)+FI-k)*(FJ+FP)+k]
                  /FR1[(m1*(FI+1)+FI-k)*(FJ+FK)+k]
                  /FR1[(m2*(FI+1)+FI-k)*(FJ+FK)+k]
        else if (i1 = FI - k) and ((i2 > FI - k) or (k>=FJ)) then
          Result:=FUsedw^[((2*(FM+1)+m2)*(FI+1)+FI-k)*(FJ+FP)+k]
                  *FUsedsigma^[(m1*(FM+1)+m2)*(FJ+FP)+k]
                  *FR2[((m1*(FM+1)+m2)*(FI+1)+FI-k)*(FJ+FP)+k]
                  /FR1[(m1*(FI+1)+FI-k)*(FJ+FK)+k]
                  /FR1[(m2*(FI+1)+FI-k)*(FJ+FK)+k]
        else if ((i1 > FI - k) or (k>=FJ)) and (i2 = FI - k) then
          Result:=FUsedw^[((2*(FM+1)+m1)*(FI+1)+FI-k)*(FJ+FP)+k]
                  *FUsedsigma^[(m1*(FM+1)+m2)*(FJ+FP)+k]
                  *FR2[((m1*(FM+1)+m2)*(FI+1)+FI-k)*(FJ+FP)+k]
                  /FR1[(m1*(FI+1)+FI-k)*(FJ+FK)+k]
                  /FR1[(m2*(FI+1)+FI-k)*(FJ+FK)+k]
        else if (i1 = FI - k) and (i2 = FI - k) then
          Result:=FUsedsigma^[(m1*(FM+1)+m2)*(FJ+FP)+k]
                  *FR2[((m1*(FM+1)+m2)*(FI+1)+FI-k)*(FJ+FP)+k]
                  /FR1[(m1*(FI+1)+FI-k)*(FJ+FK)+k]
                  /FR1[(m2*(FI+1)+FI-k)*(FJ+FK)+k]
        else
          Result:=0;
      end;
    end;
  end;
end;

{$ENDREGION}

{$REGION '  TNormalLSRM protected'}

{$REGION '    property get functions'}

function TNormalLSRM.GetCalcInfo(Step:Integer): Boolean;
begin
  {$REGION 'check indices'}
  if (Step<Ord(Low(TLSRMInitialisedElement))) or (Step>Ord(High(TLSRMInitialisedElement))) then
    raise ELSRMRangeError.Create(rsLSRM_NoSuchCalcStep);
  {$ENDREGION}
  Result:=FCalcInfo[TLSRMInitialisedElement(Step)];
end;

function TNormalLSRM.GetCalcInfoName(Step: Integer): string;
begin
  Result:='';
  {$REGION 'check indices'}
  if (Step<Ord(Low(TLSRMInitialisedElement))) or (Step>Ord(High(TLSRMInitialisedElement))) then
    raise ELSRMRangeError.Create(rsLSRM_NoSuchCalcStep);
  {$ENDREGION}
  Result:=FCalcInfoNames[TLSRMInitialisedElement(Step)];
end;

function TNormalLSRM.GetCalcTime(Step: Integer): Double;
begin
  {$REGION 'check indices'}
  if (Step<Ord(Low(TLSRMInitialisedElement))) or (Step>Ord(High(TLSRMInitialisedElement))) then
    raise ELSRMRangeError.Create(rsLSRM_NoSuchCalcStep);
  {$ENDREGION}
  Result:=FCalcTime[TLSRMInitialisedElement(Step)];
end;

function TNormalLSRM.GetCalculationFlags: TCalculationFlagSet;
begin
  Result:=FCalculationFlags;
end;

function TNormalLSRM.GetCDR(m, i, Proxy, Part, Tail: Integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  Result:=0;
  {$REGION 'test indecies'}
  if (Proxy<0) or (Proxy>1) then raise ELSRMRangeError.Create(rsLSRM_NoSuchProxy);
  if (Part<0) or (Part>2) then raise ELSRMRangeError.Create(rsLSRM_NoSuchPart);
  if (Tail<0) or (Tail>2) then raise ELSRMRangeError.Create(rsLSRM_NoSuchTail);
  {$ENDREGION}
  // if m is out of range the alpha-mixed CDR will be returned
  if (m<0) or (m>FM) then m:=FM+1;
  // if i is out of range the total CDR over all accident periods will be returned
  if (i<0) or (i>FI) then i:=FI+1;
  if Proxy=0 then begin
    {$REGION 'without proxy'}
    {$REGION 'no tail'}
    if (Tail=0) or ((Tail=1) and (FK=0)) then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_CDRProcVar in FInitialised) then ErrorMsg:=InitCDRProcVar();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_CDRParaErr in FInitialised) then ErrorMsg:=InitCDRParaErr();
      // calculate the sum of both uncertainties, if necessary
      if (Part=2) and not(Init_CDR in FInitialised) then ErrorMsg:=InitCDR();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the one period uncertainty without tail and without approximation, i.e. FCDR(Proxy,Part,m,i)
      Result:=FCDR[((Proxy*3+Part)*(FM+2)+m)*(FI+2)+i];
    end;
    {$ENDREGION}
    {$REGION 'tail only'}
    if Tail=2 then begin
      // calculate the tail process variance, if necessary
      if (Part=0) and not(Init_CDRProcVarTail in FInitialised) then ErrorMsg:=InitCDRProcVarTail();
      // calculate the tail parameter error, if necessary
      if (Part=1) and not(Init_CDRParaErrTail in FInitialised) then ErrorMsg:=InitCDRParaErrTail();
      // calculate the sum of both tail uncertainties, if necessary
      if (Part=2) and not(Init_CDRTail in FInitialised) then ErrorMsg:=InitCDRTail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the one period tail uncertainty without approximation, i.e. FCDRTail(Proxy,Part,m,i)
      Result:=FCDRTail[((Proxy*3+Part)*(FM+2)+m)*(FI+2)+i];
    end;
    {$ENDREGION}
    {$REGION 'with tail'}
    if (Tail=1) and (FK>0) then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_CDRProcVar in FInitialised) then ErrorMsg:=InitCDRProcVar();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_CDRParaErr in FInitialised) then ErrorMsg:=InitCDRParaErr();
      // calculate the sum of both uncertainties, if necessary
      if (Part=2) and not(Init_CDR in FInitialised) then ErrorMsg:=InitCDR();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // calculate the tail process variance, if necessary
      if (Part=0) and not(Init_CDRProcVarTail in FInitialised) then ErrorMsg:=InitCDRProcVarTail();
      // calculate the tail parameter error, if necessary
      if (Part=1) and not(Init_CDRParaErrTail in FInitialised) then ErrorMsg:=InitCDRParaErrTail();
      // calculate the sum of both tail uncertainties, if necessary
      if (Part=2) and not(Init_CDRTail in FInitialised) then ErrorMsg:=InitCDRTail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the one period uncertainty with tail but without approximation,
      // i.e. FCDR(Proxy,Part,m,i) + FCDRTail(Proxy,Part,m,i)
      Result:=FCDR[((Proxy*3+Part)*(FM+2)+m)*(FI+2)+i]
              +FCDRTail[((Proxy*3+Part)*(FM+2)+m)*(FI+2)+i];
    end;
    {$ENDREGION}
    {$ENDREGION}
  end else begin
    {$REGION 'proxy'}
    {$REGION 'no tail'}
    if (Tail=0) or ((Tail=1) and (FK=0)) then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_CDRProcVarProxy in FInitialised) then ErrorMsg:=InitCDRProcVarProxy();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_CDRParaErrProxy in FInitialised) then ErrorMsg:=InitCDRParaErrProxy();
      // calculate the sum of both uncertainties, if necessary
      if (Part=2) and not(Init_CDRProxy in FInitialised) then ErrorMsg:=InitCDRProxy();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the one period uncertainty without tail and without approximation, i.e. FCDR(Proxy,Part,m,i)
      Result:=FCDR[((Proxy*3+Part)*(FM+2)+m)*(FI+2)+i];
    end;
    {$ENDREGION}
    {$REGION 'tail only'}
    if Tail=2 then begin
      // calculate the tail process variance, if necessary
      if (Part=0) and not(Init_CDRProcVarTailProxy in FInitialised) then ErrorMsg:=InitCDRProcVarTailProxy();
      // calculate the tail parameter error, if necessary
      if (Part=1) and not(Init_CDRParaErrTailProxy in FInitialised) then ErrorMsg:=InitCDRParaErrTailProxy();
      // calculate the sum of both tail uncertainties, if necessary
      if (Part=2) and not(Init_CDRTailProxy in FInitialised) then ErrorMsg:=InitCDRTailProxy();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the one period tail uncertainty without approximation, i.e. FCDRTail(Proxy,Part,m,i)
      Result:=FCDRTail[((Proxy*3+Part)*(FM+2)+m)*(FI+2)+i];
    end;
    {$ENDREGION}
    {$REGION 'with tail'}
    if (Tail=1) and (FK>0) then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_CDRProcVarProxy in FInitialised) then ErrorMsg:=InitCDRProcVarProxy();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_CDRParaErrProxy in FInitialised) then ErrorMsg:=InitCDRParaErrProxy();
      // calculate the sum of both uncertainties, if necessary
      if (Part=2) and not(Init_CDRProxy in FInitialised) then ErrorMsg:=InitCDRProxy();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // calculate the tail process variance, if necessary
      if (Part=0) and not(Init_CDRProcVarTailProxy in FInitialised) then ErrorMsg:=InitCDRProcVarTailProxy();
      // calculate the tail parameter error, if necessary
      if (Part=1) and not(Init_CDRParaErrTailProxy in FInitialised) then ErrorMsg:=InitCDRParaErrTailProxy();
      // calculate the sum of both tail uncertainties, if necessary
      if (Part=2) and not(Init_CDRTailProxy in FInitialised) then ErrorMsg:=InitCDRTailProxy();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the one period uncertainty with tail but without approximation,
      // i.e. FCDR(Proxy,Part,m,i) + FCDRTail(Proxy,Part,m,i)
      Result:=FCDR[((Proxy*3+Part)*(FM+2)+m)*(FI+2)+i]
              +FCDRTail[((Proxy*3+Part)*(FM+2)+m)*(FI+2)+i];
    end;
    {$ENDREGION}
    {$ENDREGION}
  end;
end;

function TNormalLSRM.GetCDRCov(m1, m2, Proxy, Part, Tail: Integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  {$REGION 'test indecies'}
  if (m1<0) or (m1>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (m2<0) or (m2>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (Proxy<0) or (Proxy>1) then raise ELSRMRangeError.Create(rsLSRM_NoSuchProxy);
  if (Part<0) or (Part>2) then raise ELSRMRangeError.Create(rsLSRM_NoSuchPart);
  if (calc_ReduceMemSize in FCalculationFlags) and (Proxy=0) then
    raise ELSRMInitError.Create(rsLSRM_ProxyOnly);
  if (Tail<0) or (Tail>1) then raise ELSRMRangeError.Create(rsLSRM_NoSuchTail);
  {$ENDREGION}
  // calculate the coveriance matrix corresponding to the one period uncertainty, if necessary
  if Proxy=0 then begin
    case Part of {$REGION}
      0: begin
        if Tail=0 then begin
          if not(Init_CDRCovProcVar in FInitialised) then ErrorMsg:=InitCDRCovProcVar();
        end else begin
          if not(Init_CDRCovProcVarTail in FInitialised) then
            ErrorMsg:=InitCDRCovProcVarTail();
        end;
      end;
      1: begin
        if Tail=0 then begin
          if not(Init_CDRCovParaErr in FInitialised) then ErrorMsg:=InitCDRCovParaErr();
        end else begin
          if not(Init_CDRCovParaErrTail in FInitialised) then
            ErrorMsg:=InitCDRCovParaErrTail();
        end;
      end;
      2: begin
        if Tail=0 then begin
          if not(Init_CDRCov in FInitialised) then ErrorMsg:=InitCDRCov();
        end else begin
          if not(Init_CDRCovTail in FInitialised) then ErrorMsg:=InitCDRCovTail();
        end;
      end;
    end; {$ENDREGION}
  end else begin
    case Part of {$REGION}
      0: begin
        if Tail=0 then begin
          if not(Init_CDRCovProcVarProxy in FInitialised) then
            ErrorMsg:=InitCDRCovProcVarProxy();
        end else begin
          if not(Init_CDRCovProcVarTailProxy in FInitialised) then
            ErrorMsg:=InitCDRCovProcVarTailProxy();
        end;
      end;
      1: begin
        if Tail=0 then begin
          if not(Init_CDRCovParaErrProxy in FInitialised) then
            ErrorMsg:=InitCDRCovParaErrProxy();
        end else begin
          if not(Init_CDRCovParaErrTailProxy in FInitialised) then
            ErrorMsg:=InitCDRCovParaErrTailProxy();
        end;
      end;
      2: begin
        if Tail=0 then begin
          if not(Init_CDRCovProxy in FInitialised) then ErrorMsg:=InitCDRCovProxy();
        end else begin
          if not(Init_CDRCovTailProxy in FInitialised) then
            ErrorMsg:=InitCDRCovTailProxy();
        end;
      end;
    end; {$ENDREGION}
  end;
  if ErrorMsg<>'' then raise ELSRMRangeError.Create(ErrorMsg);
  // return (m1, m2) entry of the coveriance matrix corresponding to the one period uncertainty
  Result:=FCDRCov[((((Proxy*3+Part)*2+Tail)*(FM+1))+m1)*(FM+1)+m2];
end;

function TNormalLSRM.GetCDRCovEigenvalue(m, Proxy, Part, Tail: Integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  {$REGION 'test indecies'}
  if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchEigenvalue);
  if (Proxy<0) or (Proxy>1) then raise ELSRMRangeError.Create(rsLSRM_NoSuchProxy);
  if (Part<0) or (Part>2) then raise ELSRMRangeError.Create(rsLSRM_NoSuchPart);
  if (calc_ReduceMemSize in FCalculationFlags) and (Proxy=0) then
    raise ELSRMInitError.Create(rsLSRM_ProxyOnly);
  if (Tail<0) or (Tail>1) then raise ELSRMRangeError.Create(rsLSRM_NoSuchTail);
  {$ENDREGION}
  // calculate eigenvalues and vectors of the covariance matrices corresponding to the one period uncertainty
  if Proxy=0 then begin
    case Part of {$REGION}
      0: begin
        if Tail=0 then begin
          if not(Init_CDRCovEigenvalueProcVar in FInitialised) then ErrorMsg:=InitCDRCovEigenvalueProcVar();
        end else begin
          if not(Init_CDRCovEigenvalueProcVarTail in FInitialised) then
            ErrorMsg:=InitCDRCovEigenvalueProcVarTail();
        end;
      end;
      1: begin
        if Tail=0 then begin
          if not(Init_CDRCovEigenvalueParaErr in FInitialised) then ErrorMsg:=InitCDRCovEigenvalueParaErr();
        end else begin
          if not(Init_CDRCovEigenvalueParaErrTail in FInitialised) then
            ErrorMsg:=InitCDRCovEigenvalueParaErrTail();
        end;
      end;
      2: begin
        if Tail=0 then begin
          if not(Init_CDRCovEigenvalue in FInitialised) then ErrorMsg:=InitCDRCovEigenvalue();
        end else begin
          if not(Init_CDRCovEigenvalueTail in FInitialised) then ErrorMsg:=InitCDRCovEigenvalueTail();
        end;
      end;
    end; {$ENDREGION}
  end else begin
    case Part of {$REGION}
      0: begin
        if Tail=0 then begin
          if not(Init_CDRCovEigenvalueProcVarProxy in FInitialised) then
            ErrorMsg:=InitCDRCovEigenvalueProcVarProxy();
        end else begin
          if not(Init_CDRCovEigenvalueProcVarTailProxy in FInitialised) then
            ErrorMsg:=InitCDRCovEigenvalueProcVarTailProxy();
        end;
      end;
      1: begin
        if Tail=0 then begin
          if not(Init_CDRCovEigenvalueParaErrProxy in FInitialised) then
            ErrorMsg:=InitCDRCovEigenvalueParaErrProxy();
        end else begin
          if not(Init_CDRCovEigenvalueParaErrTailProxy in FInitialised) then
            ErrorMsg:=InitCDRCovEigenvalueParaErrTailProxy();
        end;
      end;
      2: begin
        if Tail=0 then begin
          if not(Init_CDRCovEigenvalueProxy in FInitialised) then ErrorMsg:=InitCDRCovEigenvalueProxy();
        end else begin
          if not(Init_CDRCovEigenvalueTailProxy in FInitialised) then
            ErrorMsg:=InitCDRCovEigenvalueTailProxy();
        end;
      end;
    end; {$ENDREGION}
  end;
  if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
  // return the m-th largest eigenvalues of the covariance matrix corresponding to the one period uncertainty
  Result:=FCDRCovEigenvalues[(Proxy*3+Part)*2+Tail][m];
end;

function TNormalLSRM.GetCDRCovEigenvector(m, l, Proxy, Part, Tail: Integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  {$REGION 'test indecies'}
  if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchEigenvalue);
  if (l<0) or (l>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchCoordinate);
  if (Proxy<0) or (Proxy>1) then raise ELSRMRangeError.Create(rsLSRM_NoSuchProxy);
  if (Part<0) or (Part>2) then raise ELSRMRangeError.Create(rsLSRM_NoSuchPart);
  if (calc_ReduceMemSize in FCalculationFlags) and (Proxy=0) then
    raise ELSRMInitError.Create(rsLSRM_ProxyOnly);
  if (Tail<0) or (Tail>1) then raise ELSRMRangeError.Create(rsLSRM_NoSuchTail);
  {$ENDREGION}
  // calculate eigenvalues and vectors of the covariance matrices corresponding to the one period uncertainty
  if Proxy=0 then begin
    case Part of {$REGION}
      0: begin
        if Tail=0 then begin
          if not(Init_CDRCovEigenvalueProcVar in FInitialised) then ErrorMsg:=InitCDRCovEigenvalueProcVar();
        end else begin
          if not(Init_CDRCovEigenvalueProcVarTail in FInitialised) then
            ErrorMsg:=InitCDRCovEigenvalueProcVarTail();
        end;
      end;
      1: begin
        if Tail=0 then begin
          if not(Init_CDRCovEigenvalueParaErr in FInitialised) then ErrorMsg:=InitCDRCovEigenvalueParaErr();
        end else begin
          if not(Init_CDRCovEigenvalueParaErrTail in FInitialised) then
            ErrorMsg:=InitCDRCovEigenvalueParaErrTail();
        end;
      end;
      2: begin
        if Tail=0 then begin
          if not(Init_CDRCovEigenvalue in FInitialised) then ErrorMsg:=InitCDRCovEigenvalue();
        end else begin
          if not(Init_CDRCovEigenvalueTail in FInitialised) then ErrorMsg:=InitCDRCovEigenvalueTail();
        end;
      end;
    end; {$ENDREGION}
  end else begin
    case Part of {$REGION}
      0: begin
        if Tail=0 then begin
          if not(Init_CDRCovEigenvalueProcVarProxy in FInitialised) then
            ErrorMsg:=InitCDRCovEigenvalueProcVarProxy();
        end else begin
          if not(Init_CDRCovEigenvalueProcVarTailProxy in FInitialised) then
            ErrorMsg:=InitCDRCovEigenvalueProcVarTailProxy();
        end;
      end;
      1: begin
        if Tail=0 then begin
          if not(Init_CDRCovEigenvalueParaErrProxy in FInitialised) then
            ErrorMsg:=InitCDRCovEigenvalueParaErrProxy();
        end else begin
          if not(Init_CDRCovEigenvalueParaErrTailProxy in FInitialised) then
            ErrorMsg:=InitCDRCovEigenvalueParaErrTailProxy();
        end;
      end;
      2: begin
        if Tail=0 then begin
          if not(Init_CDRCovEigenvalueProxy in FInitialised) then ErrorMsg:=InitCDRCovEigenvalueProxy();
        end else begin
          if not(Init_CDRCovEigenvalueTailProxy in FInitialised) then
            ErrorMsg:=InitCDRCovEigenvalueTailProxy();
        end;
      end;
    end; {$ENDREGION}
  end;
  if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
  // return l-th coordinate of the eigenvector of the m-th largest eigenvalues of the covariance matrix
  // corresponding to the one period uncertainty
  Result:=FCDRCovEigenvectors[(Proxy*3+Part)*2+Tail][l,m];
end;

function TNormalLSRM.GetCDRDetail(m1, m2, i1, i2, Proxy, Part, Tail: Integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  Result:=0;
  {$REGION 'test indecies'}
  if (m1<0) or (m1>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (i1<0) or (i1>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
  if (m2<0) or (m2>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (i2<0) or (i2>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
  if (Proxy<0) or (Proxy>1) then raise ELSRMRangeError.Create(rsLSRM_NoSuchProxy);
  if (Part<0) or (Part>1) then raise ELSRMRangeError.Create(rsLSRM_NoSuchPart);
  if (Tail<0) or (Tail>2) then raise ELSRMRangeError.Create(rsLSRM_NoSuchTail);
  {$ENDREGION}
  if Proxy=0 then begin
    {$REGION 'without proxy'}
    {$REGION 'without tail'}
    if (Tail=0) or ((Tail=1) and (FK=0)) then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_CDRProcVarDetail in FInitialised) then ErrorMsg:=InitCDRProcVarDetail();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_CDRParaErrDetail in FInitialised) then ErrorMsg:=InitCDRParaErrDetail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the addends for the indices m1, m2, i1, i2 of the one period uncertainty without tail
      // and without approximation, i.e. FCDRDetail(Proxy,Part,m1,m2,i1,i2)
      Result:=FCDRDetail[((((Proxy*2+Part)*(FM+1)+m1)*(FM+1)+m2)*(FI+1)+i1)*(FI+1)+i2];
    end;
    {$ENDREGION}
    {$REGION 'tail only'}
    if Tail=2 then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_CDRProcVarTailDetail in FInitialised) then
        ErrorMsg:=InitCDRProcVarTailDetail();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_CDRParaErrTailDetail in FInitialised) then
        ErrorMsg:=InitCDRParaErrTailDetail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the addends for the indices m1, m2, i1, i2 of the one period tail uncertainty
      // without approximation, i.e. FCDRTailDetail(Proxy,Part,m1,m2,i1,i2)
      Result:=FCDRTailDetail[((((Proxy*2+Part)*(FM+1)+m1)*(FM+1)+m2)*(FI+1)+i1)*(FI+1)+i2];
    end;
    {$ENDREGION}
    {$REGION 'with tail'}
    if (Tail=1) and (FK>0) then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_CDRProcVarDetail in FInitialised) then ErrorMsg:=InitCDRProcVarDetail();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_CDRParaErrDetail in FInitialised) then ErrorMsg:=InitCDRParaErrDetail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_CDRProcVarTailDetail in FInitialised) then
        ErrorMsg:=InitCDRProcVarTailDetail();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_CDRParaErrTailDetail in FInitialised) then
        ErrorMsg:=InitCDRParaErrTailDetail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the addends for the indices m1, m2, i1, i2 of the one period uncertainty with tail but without
      // approximation, i.e. FCDRDetail(Proxy,Part,m1,m2,i1,i2) + FCDRTailDetail(Proxy,Part,m1,m2,i1,i2)
      Result:=FCDRDetail[((((Proxy*2+Part)*(FM+1)+m1)*(FM+1)+m2)*(FI+1)+i1)*(FI+1)+i2]
              +FCDRTailDetail[((((Proxy*2+Part)*(FM+1)+m1)*(FM+1)+m2)*(FI+1)+i1)*(FI+1)+i2];
    end;
    {$ENDREGION}
  {$ENDREGION}
  end else begin
    {$REGION 'proxy'}
    {$REGION 'without tail'}
    if (Tail=0) or ((Tail=1) and (FK=0)) then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_CDRProcVarProxyDetail in FInitialised) then
        ErrorMsg:=InitCDRProcVarProxyDetail();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_CDRParaErrProxyDetail in FInitialised) then
        ErrorMsg:=InitCDRParaErrProxyDetail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the addends for the indices m1, m2, i1, i2 of the one period uncertainty without tail
      // and without approximation, i.e. FCDRDetail(Proxy,Part,m1,m2,i1,i2)
      Result:=FCDRDetail[((((Proxy*2+Part)*(FM+1)+m1)*(FM+1)+m2)*(FI+1)+i1)*(FI+1)+i2];
    end;
    {$ENDREGION}
    {$REGION 'tail only'}
    if Tail=2 then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_CDRProcVarTailProxyDetail in FInitialised) then
        ErrorMsg:=InitCDRProcVarTailProxyDetail();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_CDRParaErrTailProxyDetail in FInitialised) then
        ErrorMsg:=InitCDRParaErrTailProxyDetail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the addends for the indices m1, m2, i1, i2 of the one period tail uncertainty
      // without approximation, i.e. FCDRTailDetail(Proxy,Part,m1,m2,i1,i2)
      Result:=FCDRTailDetail[((((Proxy*2+Part)*(FM+1)+m1)*(FM+1)+m2)*(FI+1)+i1)*(FI+1)+i2];
    end;
    {$ENDREGION}
    {$REGION 'with tail'}
    if (Tail=1) and (FK>0) then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_CDRProcVarProxyDetail in FInitialised) then
        ErrorMsg:=InitCDRProcVarProxyDetail();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_CDRParaErrProxyDetail in FInitialised) then
        ErrorMsg:=InitCDRParaErrProxyDetail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_CDRProcVarTailProxyDetail in FInitialised) then
        ErrorMsg:=InitCDRProcVarTailProxyDetail();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_CDRParaErrTailProxyDetail in FInitialised) then
        ErrorMsg:=InitCDRParaErrTailProxyDetail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the addends for the indices m1, m2, i1, i2 of the one period uncertainty with tail but without
      // approximation, i.e. FCDRDetail(Proxy,Part,m1,m2,i1,i2) + FCDRTailDetail(Proxy,Part,m1,m2,i1,i2)
      Result:=FCDRDetail[((((Proxy*2+Part)*(FM+1)+m1)*(FM+1)+m2)*(FI+1)+i1)*(FI+1)+i2]
              +FCDRTailDetail[((((Proxy*2+Part)*(FM+1)+m1)*(FM+1)+m2)*(FI+1)+i1)*(FI+1)+i2];
    end;
    {$ENDREGION}
    {$ENDREGION}
  end;
end;

function TNormalLSRM.GetCoorF(m, i, k, l, h, j:integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  {$REGION 'test indecies'}
  if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (i<0) or (i>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
  if (k<0) or (k>FJ-1+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
  if (l<0) or (l>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (h<0) or (h>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
  if (j<0) or (j>FJ+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
  {$ENDREGION}
  // calculate the coordinates of the operator F, if necessary
  if not(Init_CoorF in FInitialised) then ErrorMsg:=InitCoorF();
  if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
  // return the coordinates (F^m_ik)^l_hj of the operator F
  Result:=FCoorF[((((m*(FI+1)+i)*(FJ+FP)+k)*(FM+1)+l)*(FI+1)+h)*(FJ+1+FP)+j];
end;

function TNormalLSRM.GetDefaultalpha(m, i: integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  {$REGION 'test indecies'}
  if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (i<0) or (i>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
  {$ENDREGION}
  // calculate default alphas, if necessary
  if not(Init_Defaultalpha in FInitialised) then ErrorMsg:=InitDefaultalpha();
  if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
  // return the default alpha alpha^m_i
  Result:=FDefaultalpha[m*(FI+1)+i];
end;

function TNormalLSRM.GetDefaultalphaUsed: Boolean;
begin
  Result:=(FUsedalpha = @FDefaultalpha);
end;

function TNormalLSRM.GetDefaultf(m, k: integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  {$REGION 'test indecies'}
  if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (k<0) or (k>FJ-1+FK) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
  {$ENDREGION}
  // calculate the default development factors, if necessary
  if not(Init_Defaultf in FInitialised) then ErrorMsg:=InitDefaultf();
  if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
  // return the default development factor f^m_k
  Result:=FDefaultf[m*(FJ+FK)+k];
end;

function TNormalLSRM.GetDefaultfUsed: Boolean;
begin
  Result:=(FUsedf = @FDefaultf);
end;

function TNormalLSRM.GetDefaultsigma(m1, m2, k: integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  {$REGION 'test indecies'}
  if (m1<0) or (m1>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (m2<0) or (m2>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (k<0) or (k>FJ-1+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
  {$ENDREGION}
  // calculate the default variance parameters, if necessary
  if not(Init_Defaultsigma in FInitialised) then ErrorMsg:=InitDefaultsigma();
  if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
  // return the default variance parameters sigma^m1m2_k
  Result:=FDefaultsigma[(m1*(FM+1)+m2)*(FJ+FP)+k];
end;

function TNormalLSRM.GetDefaultsigmaUsed: Boolean;
begin
  Result:=(FUsedsigma = @FDefaultsigma);
end;

function TNormalLSRM.GetDefaultw(n, m, i, k: integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  {$REGION 'test indecies'}
  if (n<0) or (n>2) then raise ELSRMRangeError.Create(rsLSRM_NoSuchTypeofWeights);
  if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (i<0) or (i>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
  if (k<0) or (k>FJ-1+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
  {$ENDREGION}
  // calculate the default weights, if necessary
  if not(Init_Defaultw in FInitialised) then ErrorMsg:=InitDefaultw();
  if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
  // return the default weights
  //   - not normalised, if n=0
  //   - for busines period I, if n=1
  //   - for business period I+1, if n=2
  Result:=FDefaultw[((n*(FM+1)+m)*(FI+1)+i)*(FJ+FP)+k];
end;

function TNormalLSRM.GetDefaultwUsed: Boolean;
begin
  Result:=(FUsedw = @FDefaultw);
end;

function TNormalLSRM.GetEigenvalue(i, k, m: integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  {$REGION 'test indecies'}
  if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchEigenvalue);
  if (i<0) or (i>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
  if (k<0) or (k>FJ-1+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
  {$ENDREGION}
  // calculate the eigenvalues of the covariance matrices (sigma^m1m2_k*R^m1m2_ik)_m1m2
  if not(Init_Eigenvalues in FInitialised) then ErrorMsg:=InitEigenvalues();
  if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
  // return the m-th largest eigenvalues of the covariance matrices (sigma^m1m2_k*R^m1m2_ik)_m1m2
  Result:=FEigenvalues[i*(FJ+FP)+k][m];
end;

function TNormalLSRM.GetEigenvector(i, k, m, l: integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  {$REGION 'test indecies'}
  if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchEigenvalue);
  if (l<0) or (l>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchCoordinate);
  if (i<0) or (i>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
  if (k<0) or (k>FJ-1+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
  {$ENDREGION}
  // calculate the eigenvectors (and eigenvalues) of the covariance matrices (sigma^m1m2_k*R^m1m2_ik)_m1m2
  if not(Init_Eigenvalues in FInitialised) then ErrorMsg:=InitEigenvalues();
  if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
  // return the l-th coordinate of the eigenvector of the m-th largest eigenvalues of the covariance matrices
  // (sigma^m1m2_k*R^m1m2_ik)_m1m2
  Result:=FEigenvectors[i*(FJ+FP)+k][l,m];
end;

function TNormalLSRM.GetI: integer;
begin
  // if the geometry of the LSRM has not been initialised an error message will be returned
  if not(Init_Geometry in FInitialised) then raise ELSRMInitError.Create(rsLSRM_NotInitialisedGeometry);
  // return I
  Result:=FI;
end;

function TNormalLSRM.GetInitialisedParameters:integer;
begin
  result:=0;
  // check if gamma1 has been initialised
  if Init_Gamma1 in FInitialised then Result:=Result or cLSRMInitGamma1;
  // check if gamma2 has been initialised
  if Init_Gamma2 in FInitialised then Result:=Result or cLSRMInitGamma2;
  // check if S has been initialised
  if Init_S in FInitialised then Result:=Result or cLSRMInitS;
  // check if w has been initialised
  if Init_w in FInitialised then Result:=Result or cLSRMInitw;
  // check if f has been initialised
  if Init_f in FInitialised then Result:=Result or cLSRMInitf;
  // check if sigma has been initialised
  if Init_sigma in FInitialised then Result:=Result or cLSRMInitsigma;
  // check if alpha has been initialised
  if Init_alpha in FInitialised then Result:=Result or cLSRMInitalpha;
  // check if geometry has been initialised
  if Init_Geometry in FInitialised then Result:=Result or cLSRMInitGeometry;
  // check if Inflation has been initialised
  if Init_InflationRate in FInitialised then Result:=Result or cLSRMInitInflationRate;
end;

function TNormalLSRM.GetJ: integer;
begin
  // if the geometry of the LSRM hes not been initialised an error message will be returned
  if not(Init_Geometry in FInitialised) then raise ELSRMInitError.Create(rsLSRM_NotInitialisedGeometry);
  // return J
  Result:=FJ;
end;

function TNormalLSRM.GetK: integer;
begin
  // if the geometry of the LSRM hes not been initialised an error message will be returned
  if not(Init_Geometry in FInitialised) then raise ELSRMInitError.Create(rsLSRM_NotInitialisedGeometry);
  // return the number of future development periods
  Result:=FK;
end;

function TNormalLSRM.GetM: integer;
begin
  // if the geometry of the LSRM hes not been initialised an error message will be returned
  if not(Init_Geometry in FInitialised) then raise ELSRMInitError.Create(rsLSRM_NotInitialisedGeometry);
  // return M
  Result:=FM;
end;

function TNormalLSRM.GetMSEP(m, i, Proxy, Part, Tail: Integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  Result:=0;
  {$REGION 'test indecies'}
  if (Proxy<0) or (Proxy>1) then raise ELSRMRangeError.Create(rsLSRM_NoSuchProxy);
  if (Part<0) or (Part>2) then raise ELSRMRangeError.Create(rsLSRM_NoSuchPart);
  if (Tail<0) or (Tail>2) then raise ELSRMRangeError.Create(rsLSRM_NoSuchTail);
  {$ENDREGION}
  // if m is out of range the alpha-mixed MSEP will be returned
  if (m<0) or (m>FM) then m:=FM+1;
  // if i is out of range the total MSEP over all accident periods will be returned
  if (i<0) or (i>FI) then i:=FI+1;
  if Proxy=0 then begin
    {$REGION 'without proxy'}
    {$REGION 'no tail'}
    if (Tail=0) or ((Tail=1) and (FK=0)) then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_MSEPProcVar in FInitialised) then ErrorMsg:=InitMSEPProcVar();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_MSEPParaErr in FInitialised) then ErrorMsg:=InitMSEPParaErr();
      // calculate the sum of both uncertainties, if necessary
      if (Part=2) and not(Init_MSEP in FInitialised) then ErrorMsg:=InitMSEP();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the ultimate uncertainty without tail and without approximation, i.e. FMSEP(Proxy,Part,m,i)
      Result:=FMSEP[((Proxy*3+Part)*(FM+2)+m)*(FI+2)+i];
    end;
    {$ENDREGION}
    {$REGION 'tail only'}
    if Tail=2 then begin
      // calculate the tail process variance, if necessary
      if (Part=0) and not(Init_MSEPProcVarTail in FInitialised) then ErrorMsg:=InitMSEPProcVarTail();
      // calculate the tail parameter error, if necessary
      if (Part=1) and not(Init_MSEPParaErrTail in FInitialised) then ErrorMsg:=InitMSEPParaErrTail();
      // calculate the sum of both tail uncertainties, if necessary
      if (Part=2) and not(Init_MSEPTail in FInitialised) then ErrorMsg:=InitMSEPTail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the ultimate tail uncertainty without approximation, i.e. FMSEPTail(Proxy,Part,m,i)
      Result:=FMSEPTail[((Proxy*3+Part)*(FM+2)+m)*(FI+2)+i];
    end;
    {$ENDREGION}
    {$REGION 'with tail'}
    if (Tail=1) and (FK>0) then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_MSEPProcVar in FInitialised) then ErrorMsg:=InitMSEPProcVar();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_MSEPParaErr in FInitialised) then ErrorMsg:=InitMSEPParaErr();
      // calculate the sum of both uncertainties, if necessary
      if (Part=2) and not(Init_MSEP in FInitialised) then ErrorMsg:=InitMSEP();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // calculate the tail process variance, if necessary
      if (Part=0) and not(Init_MSEPProcVarTail in FInitialised) then ErrorMsg:=InitMSEPProcVarTail();
      // calculate the tail parameter error, if necessary
      if (Part=1) and not(Init_MSEPParaErrTail in FInitialised) then ErrorMsg:=InitMSEPParaErrTail();
      // calculate the sum of both tail uncertainties, if necessary
      if (Part=2) and not(Init_MSEPTail in FInitialised) then ErrorMsg:=InitMSEPTail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the ultimate uncertainty with tail but without approximation,
      // i.e. FMSEP(Proxy,Part,m,i) + FMSEPTail(Proxy,Part,m,i)
      Result:=FMSEP[((Proxy*3+Part)*(FM+2)+m)*(FI+2)+i]
              +FMSEPTail[((Proxy*3+Part)*(FM+2)+m)*(FI+2)+i];
    end;
    {$ENDREGION}
    {$ENDREGION}
  end else begin
    {$REGION 'proxy'}
    {$REGION 'no tail'}
    if (Tail=0) or ((Tail=1) and (FK=0)) then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_MSEPProcVarProxy in FInitialised) then ErrorMsg:=InitMSEPProcVarProxy();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_MSEPParaErrProxy in FInitialised) then ErrorMsg:=InitMSEPParaErrProxy();
      // calculate the sum of both uncertainties, if necessary
      if (Part=2) and not(Init_MSEPProxy in FInitialised) then ErrorMsg:=InitMSEPProxy();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the ultimate uncertainty without tail and without approximation, i.e. FMSEP(Proxy,Part,m,i)
      Result:=FMSEP[((Proxy*3+Part)*(FM+2)+m)*(FI+2)+i];
    end;
    {$ENDREGION}
    {$REGION 'tail only'}
    if Tail=2 then begin
      // calculate the tail process variance, if necessary
      if (Part=0) and not(Init_MSEPProcVarTailProxy in FInitialised) then
        ErrorMsg:=InitMSEPProcVarTailProxy();
      // calculate the tail parameter error, if necessary
      if (Part=1) and not(Init_MSEPParaErrTailProxy in FInitialised) then
        ErrorMsg:=InitMSEPParaErrTailProxy();
      // calculate the sum of both tail uncertainties, if necessary
      if (Part=2) and not(Init_MSEPTailProxy in FInitialised) then ErrorMsg:=InitMSEPTailProxy();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the ultimate tail uncertainty without approximation, i.e. FMSEPTail(Proxy,Part,m,i)
      Result:=FMSEPTail[((Proxy*3+Part)*(FM+2)+m)*(FI+2)+i];
    end;
    {$ENDREGION}
    {$REGION 'with tail'}
    if (Tail=1) and (FK>0) then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_MSEPProcVarProxy in FInitialised) then ErrorMsg:=InitMSEPProcVarProxy();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_MSEPParaErrProxy in FInitialised) then ErrorMsg:=InitMSEPParaErrProxy();
      // calculate the sum of both uncertainties, if necessary
      if (Part=2) and not(Init_MSEPProxy in FInitialised) then ErrorMsg:=InitMSEPProxy();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // calculate the tail process variance, if necessary
      if (Part=0) and not(Init_MSEPProcVarTailProxy in FInitialised) then
        ErrorMsg:=InitMSEPProcVarTailProxy();
      // calculate the tail parameter error, if necessary
      if (Part=1) and not(Init_MSEPParaErrTailProxy in FInitialised) then
        ErrorMsg:=InitMSEPParaErrTailProxy();
      // calculate the sum of both tail uncertainties, if necessary
      if (Part=2) and not(Init_MSEPTailProxy in FInitialised) then ErrorMsg:=InitMSEPTailProxy();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the ultimate uncertainty with tail but without approximation,
      // i.e. FMSEP(Proxy,Part,m,i) + FMSEPTail(Proxy,Part,m,i)
      Result:=FMSEP[((Proxy*3+Part)*(FM+2)+m)*(FI+2)+i]
              +FMSEPTail[((Proxy*3+Part)*(FM+2)+m)*(FI+2)+i];
    end;
    {$ENDREGION}
    {$ENDREGION}
  end;
end;

function TNormalLSRM.GetMSEPCov(m1, m2, Proxy, Part, Tail: Integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  {$REGION 'test indecies'}
  if (m1<0) or (m1>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (m2<0) or (m2>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (Proxy<0) or (Proxy>1) then raise ELSRMRangeError.Create(rsLSRM_NoSuchProxy);
  if (Part<0) or (Part>2) then raise ELSRMRangeError.Create(rsLSRM_NoSuchPart);
  if (calc_ReduceMemSize in FCalculationFlags) and (Proxy=0) then
    raise ELSRMInitError.Create(rsLSRM_ProxyOnly);
  if (Tail<0) or (Tail>1) then raise ELSRMRangeError.Create(rsLSRM_NoSuchTail);
  {$ENDREGION}
  // calculate the coveriance matrix corresponding to the ultimate uncertainty, if necessary
  if Proxy=0 then begin
    case Part of {$REGION}
      0: begin
        if Tail=0 then begin
          if not(Init_MSEPCovProcVar in FInitialised) then ErrorMsg:=InitMSEPCovProcVar();
        end else begin
          if not(Init_MSEPCovProcVarTail in FInitialised) then
            ErrorMsg:=InitMSEPCovProcVarTail();
        end;
      end;
      1: begin
        if Tail=0 then begin
          if not(Init_MSEPCovParaErr in FInitialised) then ErrorMsg:=InitMSEPCovParaErr();
        end else begin
          if not(Init_MSEPCovParaErrTail in FInitialised) then
            ErrorMsg:=InitMSEPCovParaErrTail();
        end;
      end;
      2: begin
        if Tail=0 then begin
          if not(Init_MSEPCov in FInitialised) then ErrorMsg:=InitMSEPCov();
        end else begin
          if not(Init_MSEPCovTail in FInitialised) then ErrorMsg:=InitMSEPCovTail();
        end;
      end;
    end; {$ENDREGION}
  end else begin
    case Part of {$REGION}
      0: begin
        if Tail=0 then begin
          if not(Init_MSEPCovProcVarProxy in FInitialised) then
            ErrorMsg:=InitMSEPCovProcVarProxy();
        end else begin
          if not(Init_MSEPCovProcVarTailProxy in FInitialised) then
            ErrorMsg:=InitMSEPCovProcVarTailProxy();
        end;
      end;
      1: begin
        if Tail=0 then begin
          if not(Init_MSEPCovParaErrProxy in FInitialised) then
            ErrorMsg:=InitMSEPCovParaErrProxy();
        end else begin
          if not(Init_MSEPCovParaErrTailProxy in FInitialised) then
            ErrorMsg:=InitMSEPCovParaErrTailProxy();
        end;
      end;
      2: begin
        if Tail=0 then begin
          if not(Init_MSEPCovProxy in FInitialised) then ErrorMsg:=InitMSEPCovProxy();
        end else begin
          if not(Init_MSEPCovTailProxy in FInitialised) then
            ErrorMsg:=InitMSEPCovTailProxy();
        end;
      end;
    end; {$ENDREGION}
  end;
  if ErrorMsg<>'' then raise ELSRMRangeError.Create(ErrorMsg);
  // return (m1, m2) entry of the coveriance matrix corresponding to the ultimate uncertainty
  Result:=FMSEPCov[((((Proxy*3+Part)*2+Tail)*(FM+1))+m1)*(FM+1)+m2];
end;

function TNormalLSRM.GetMSEPCovEigenvalue(m, Proxy, Part, Tail: Integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  {$REGION 'test indecies'}
  if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchEigenvalue);
  if (Proxy<0) or (Proxy>1) then raise ELSRMRangeError.Create(rsLSRM_NoSuchProxy);
  if (Part<0) or (Part>2) then raise ELSRMRangeError.Create(rsLSRM_NoSuchPart);
  if (calc_ReduceMemSize in FCalculationFlags) and (Proxy=0) then
    raise ELSRMInitError.Create(rsLSRM_ProxyOnly);
  if (Tail<0) or (Tail>1) then raise ELSRMRangeError.Create(rsLSRM_NoSuchTail);
  {$ENDREGION}
  // calculate eigenvalues and vectors of the covariance matrices corresponding to the ultimate uncertainty
  if Proxy=0 then begin
    case Part of {$REGION}
      0: begin
        if Tail=0 then begin
          if not(Init_MSEPCovEigenvalueProcVar in FInitialised) then ErrorMsg:=InitMSEPCovEigenvalueProcVar();
        end else begin
          if not(Init_MSEPCovEigenvalueProcVarTail in FInitialised) then
            ErrorMsg:=InitMSEPCovEigenvalueProcVarTail();
        end;
      end;
      1: begin
        if Tail=0 then begin
          if not(Init_MSEPCovEigenvalueParaErr in FInitialised) then ErrorMsg:=InitMSEPCovEigenvalueParaErr();
        end else begin
          if not(Init_MSEPCovEigenvalueParaErrTail in FInitialised) then
            ErrorMsg:=InitMSEPCovEigenvalueParaErrTail();
        end;
      end;
      2: begin
        if Tail=0 then begin
          if not(Init_MSEPCovEigenvalue in FInitialised) then ErrorMsg:=InitMSEPCovEigenvalue();
        end else begin
          if not(Init_MSEPCovEigenvalueTail in FInitialised) then ErrorMsg:=InitMSEPCovEigenvalueTail();
        end;
      end;
    end; {$ENDREGION}
  end else begin
    case Part of {$REGION}
      0: begin
        if Tail=0 then begin
          if not(Init_MSEPCovEigenvalueProcVarProxy in FInitialised) then
            ErrorMsg:=InitMSEPCovEigenvalueProcVarProxy();
        end else begin
          if not(Init_MSEPCovEigenvalueProcVarTailProxy in FInitialised) then
            ErrorMsg:=InitMSEPCovEigenvalueProcVarTailProxy();
        end;
      end;
      1: begin
        if Tail=0 then begin
          if not(Init_MSEPCovEigenvalueParaErrProxy in FInitialised) then
            ErrorMsg:=InitMSEPCovEigenvalueParaErrProxy();
        end else begin
          if not(Init_MSEPCovEigenvalueParaErrTailProxy in FInitialised) then
            ErrorMsg:=InitMSEPCovEigenvalueParaErrTailProxy();
        end;
      end;
      2: begin
        if Tail=0 then begin
          if not(Init_MSEPCovEigenvalueProxy in FInitialised) then ErrorMsg:=InitMSEPCovEigenvalueProxy();
        end else begin
          if not(Init_MSEPCovEigenvalueTailProxy in FInitialised) then
            ErrorMsg:=InitMSEPCovEigenvalueTailProxy();
        end;
      end;
    end; {$ENDREGION}
  end;
  if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
  // return the m-th largest eigenvalues of the covariance matrix corresponding to the ultimate uncertainty
  Result:=FMSEPCovEigenvalues[(Proxy*3+Part)*2+Tail][m];
end;

function TNormalLSRM.GetMSEPCovEigenvector(m, l, Proxy, Part, Tail: Integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  {$REGION 'test indecies'}
  if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchEigenvalue);
  if (l<0) or (l>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchCoordinate);
  if (Proxy<0) or (Proxy>1) then raise ELSRMRangeError.Create(rsLSRM_NoSuchProxy);
  if (Part<0) or (Part>2) then raise ELSRMRangeError.Create(rsLSRM_NoSuchPart);
  if (calc_ReduceMemSize in FCalculationFlags) and (Proxy=0) then
    raise ELSRMInitError.Create(rsLSRM_ProxyOnly);
  if (Tail<0) or (Tail>1) then raise ELSRMRangeError.Create(rsLSRM_NoSuchTail);
  {$ENDREGION}
  // calculate eigenvalues and vectors of the covariance matrices corresponding to the ultimate uncertainty
  if Proxy=0 then begin
    case Part of {$REGION}
      0: begin
        if Tail=0 then begin
          if not(Init_MSEPCovEigenvalueProcVar in FInitialised) then ErrorMsg:=InitMSEPCovEigenvalueProcVar();
        end else begin
          if not(Init_MSEPCovEigenvalueProcVarTail in FInitialised) then
            ErrorMsg:=InitMSEPCovEigenvalueProcVarTail();
        end;
      end;
      1: begin
        if Tail=0 then begin
          if not(Init_MSEPCovEigenvalueParaErr in FInitialised) then ErrorMsg:=InitMSEPCovEigenvalueParaErr();
        end else begin
          if not(Init_MSEPCovEigenvalueParaErrTail in FInitialised) then
            ErrorMsg:=InitMSEPCovEigenvalueParaErrTail();
        end;
      end;
      2: begin
        if Tail=0 then begin
          if not(Init_MSEPCovEigenvalue in FInitialised) then ErrorMsg:=InitMSEPCovEigenvalue();
        end else begin
          if not(Init_MSEPCovEigenvalueTail in FInitialised) then ErrorMsg:=InitMSEPCovEigenvalueTail();
        end;
      end;
    end; {$ENDREGION}
  end else begin
    case Part of {$REGION}
      0: begin
        if Tail=0 then begin
          if not(Init_MSEPCovEigenvalueProcVarProxy in FInitialised) then
            ErrorMsg:=InitMSEPCovEigenvalueProcVarProxy();
        end else begin
          if not(Init_MSEPCovEigenvalueProcVarTailProxy in FInitialised) then
            ErrorMsg:=InitMSEPCovEigenvalueProcVarTailProxy();
        end;
      end;
      1: begin
        if Tail=0 then begin
          if not(Init_MSEPCovEigenvalueParaErrProxy in FInitialised) then
            ErrorMsg:=InitMSEPCovEigenvalueParaErrProxy();
        end else begin
          if not(Init_MSEPCovEigenvalueParaErrTailProxy in FInitialised) then
            ErrorMsg:=InitMSEPCovEigenvalueParaErrTailProxy();
        end;
      end;
      2: begin
        if Tail=0 then begin
          if not(Init_MSEPCovEigenvalueProxy in FInitialised) then ErrorMsg:=InitMSEPCovEigenvalueProxy();
        end else begin
          if not(Init_MSEPCovEigenvalueTailProxy in FInitialised) then
            ErrorMsg:=InitMSEPCovEigenvalueTailProxy();
        end;
      end;
    end; {$ENDREGION}
  end;
  if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
  // return l-th coordinate of the eigenvector of the m-th largest eigenvalues of the covariance matrix
  // corresponding to the ultimate uncertainty
  Result:=FMSEPCovEigenvectors[(Proxy*3+Part)*2+Tail][l,m];
end;

function TNormalLSRM.GetMSEPDetail(m1, m2, i1, i2, Proxy, Part, Tail: Integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  Result:=0;
  {$REGION 'test indecies'}
  if (m1<0) or (m1>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (i1<0) or (i1>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
  if (m2<0) or (m2>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (i2<0) or (i2>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
  if (Proxy<0) or (Proxy>1) then raise ELSRMRangeError.Create(rsLSRM_NoSuchProxy);
  if (Part<0) or (Part>1) then raise ELSRMRangeError.Create(rsLSRM_NoSuchPart);
  if (Tail<0) or (Tail>2) then raise ELSRMRangeError.Create(rsLSRM_NoSuchTail);
  {$ENDREGION}
  if Proxy=0 then begin
    {$REGION 'without proxy'}
    {$REGION 'without tail'}
    if (Tail=0) or ((Tail=1) and (FK=0)) then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_MSEPProcVarDetail in FInitialised) then ErrorMsg:=InitMSEPProcVarDetail();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_MSEPParaErrDetail in FInitialised) then ErrorMsg:=InitMSEPParaErrDetail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the addends for the indices m1, m2, i1, i2 of the ultimate uncertainty without tail
      // and without approximation, i.e. FMSEPDetail(Proxy,Part,m1,m2,i1,i2)
      Result:=FMSEPDetail[((((Proxy*2+Part)*(FM+1)+m1)*(FM+1)+m2)*(FI+1)+i1)*(FI+1)+i2];
    end;
    {$ENDREGION}
    {$REGION 'tail only'}
    if Tail=2 then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_MSEPProcVarTailDetail in FInitialised) then
        ErrorMsg:=InitMSEPProcVarTailDetail();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_MSEPParaErrTailDetail in FInitialised) then
        ErrorMsg:=InitMSEPParaErrTailDetail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the addends for the indices m1, m2, i1, i2 of the ultimate tail uncertainty
      // without approximation, i.e. FMSEPTailDetail(Proxy,Part,m1,m2,i1,i2)
      Result:=FMSEPTailDetail[((((Proxy*2+Part)*(FM+1)+m1)*(FM+1)+m2)*(FI+1)+i1)*(FI+1)+i2];
    end;
    {$ENDREGION}
    {$REGION 'with tail'}
    if (Tail=1) and (FK>0) then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_MSEPProcVarDetail in FInitialised) then ErrorMsg:=InitMSEPProcVarDetail();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_MSEPParaErrDetail in FInitialised) then ErrorMsg:=InitMSEPParaErrDetail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_MSEPProcVarTailDetail in FInitialised) then
        ErrorMsg:=InitMSEPProcVarTailDetail();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_MSEPParaErrTailDetail in FInitialised) then
        ErrorMsg:=InitMSEPParaErrTailDetail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the addends for the indices m1, m2, i1, i2 of the ultimate uncertainty with tail but without
      // approximation, i.e. FMSEPDetail(Proxy,Part,m1,m2,i1,i2) + FMSEPTailDetail(Proxy,Part,m1,m2,i1,i2)
      Result:=FMSEPDetail[((((Proxy*2+Part)*(FM+1)+m1)*(FM+1)+m2)*(FI+1)+i1)*(FI+1)+i2]
              +FMSEPTailDetail[((((Proxy*2+Part)*(FM+1)+m1)*(FM+1)+m2)*(FI+1)+i1)*(FI+1)+i2];
    end;
    {$ENDREGION}
    {$ENDREGION}
  end else begin
    {$REGION 'proxy'}
    {$REGION 'without tail'}
    if (Tail=0) or ((Tail=1) and (FK=0)) then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_MSEPProcVarProxyDetail in FInitialised) then
        ErrorMsg:=InitMSEPProcVarProxyDetail();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_MSEPParaErrProxyDetail in FInitialised) then
        ErrorMsg:=InitMSEPParaErrProxyDetail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the addends for the indices m1, m2, i1, i2 of the ultimate uncertainty without tail
      // and without approximation, i.e. FMSEPDetail(Proxy,Part,m1,m2,i1,i2)
      Result:=FMSEPDetail[((((Proxy*2+Part)*(FM+1)+m1)*(FM+1)+m2)*(FI+1)+i1)*(FI+1)+i2];
    end;
    {$ENDREGION}
    {$REGION 'tail only'}
    if Tail=2 then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_MSEPProcVarTailProxyDetail in FInitialised) then
        ErrorMsg:=InitMSEPProcVarTailProxyDetail();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_MSEPParaErrTailProxyDetail in FInitialised) then
        ErrorMsg:=InitMSEPParaErrTailProxyDetail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the addends for the indices m1, m2, i1, i2 of the ultimate tail uncertainty
      // without approximation, i.e. FMSEPTailDetail(Proxy,Part,m1,m2,i1,i2)
      Result:=FMSEPTailDetail[((((Proxy*2+Part)*(FM+1)+m1)*(FM+1)+m2)*(FI+1)+i1)*(FI+1)+i2];
    end;
    {$ENDREGION}
    {$REGION 'with tail'}
    if (Tail=1) and (FK>0) then begin
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_MSEPProcVarProxyDetail in FInitialised) then
        ErrorMsg:=InitMSEPProcVarProxyDetail();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_MSEPParaErrProxyDetail in FInitialised) then
        ErrorMsg:=InitMSEPParaErrProxyDetail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // calculate the process variance, if necessary
      if (Part=0) and not(Init_MSEPProcVarTailProxyDetail in FInitialised) then
        ErrorMsg:=InitMSEPProcVarTailProxyDetail();
      // calculate the parameter error, if necessary
      if (Part=1) and not(Init_MSEPParaErrTailProxyDetail in FInitialised) then
        ErrorMsg:=InitMSEPParaErrTailProxyDetail();
      if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
      // return the addends for the indices m1, m2, i1, i2 of the ultimate uncertainty with tail but without
      // approximation, i.e. FMSEPDetail(Proxy,Part,m1,m2,i1,i2) + FMSEPTailDetail(Proxy,Part,m1,m2,i1,i2)
      Result:=FMSEPDetail[((((Proxy*2+Part)*(FM+1)+m1)*(FM+1)+m2)*(FI+1)+i1)*(FI+1)+i2]
              +FMSEPTailDetail[((((Proxy*2+Part)*(FM+1)+m1)*(FM+1)+m2)*(FI+1)+i1)*(FI+1)+i2];
    end;
    {$ENDREGION}
    {$ENDREGION}
  end;
end;

function TNormalLSRM.GetR1(m, i, k: integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  {$REGION 'test indecies'}
  if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (i<0) or (i>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
  if (k<0) or (k>FJ-1+FK) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
  {$ENDREGION}
  // calculate the past of the exposure R^m_ik, if necessary
  if (i+k<=FI) and (k<=FJ-1) and not(Init_R1 in FInitialised) then ErrorMsg:=InitR1();
  // calculate the futurae of the exposure R^m_ik, if necessary
  if ((i+k>FI) or (k>FJ-1)) and not(Init_UltimateDetail in FInitialised) then ErrorMsg:=InitUltimateDetail();
  if ErrorMsg<>'' then raise ELSRMRangeError.Create(ErrorMsg);
  // return the exposure R^m_ik
  Result:=FR1[(m*(FI+1)+i)*(FJ+FK)+k];
end;

function TNormalLSRM.GetR2(m1, m2, i, k: integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  {$REGION 'test indecies'}
  if (m1<0) or (m1>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (m2<0) or (m2>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (i<0) or (i>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
  if (k<0) or (k>FJ-1+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
  {$ENDREGION}
  // calculate the past of the exposure R^m1m2_ik, if necessary
  if (i+k<=FI) and (k<=FJ-1) and not(Init_R2 in FInitialised) then ErrorMsg:=InitR2();
  // calculate the future of the exposure R^m1m2_ik, if necessary
  if ((i+k>FI) or (k>FJ-1)) and not(Init_R2Full in FInitialised) then ErrorMsg:=InitR2Full();
  if ErrorMsg<>'' then raise ELSRMRangeError.Create(ErrorMsg);
  // return the exposure R^m1m2_ik
  Result:=FR2[((m1*(FM+1)+m2)*(FI+1)+i)*(FJ+FP)+k];
end;

function TNormalLSRM.GetS(m, i, k: integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  {$REGION 'test indecies'}
  if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (i<0) or (i>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
  if (k<0) or (k>FJ+FK) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
  {$ENDREGION}
  // if no claim properties have been initialised an error message will be returned
  if not(Init_S in FInitialised) then raise ELSRMInitError.Create(rsLSRM_NotInitialisedS);
  // calculate the future of claim propertoes, if necessary
  if ((i+k>FI) or (k>FJ)) and not(Init_UltimateDetail in FInitialised) then ErrorMsg:=InitUltimateDetail();
  if ErrorMsg<>'' then raise ELSRMRangeError.Create(ErrorMsg);
  // return the claim property S^m_ik
  Result:=FS[(m*(FI+1)+i)*(FJ+1+FK)+k];
end;

function TNormalLSRM.GetUltimate(m, i, Tail: Integer): Double;
  var ErrorMsg: string;
begin
  ErrorMsg:='';
  {$REGION 'test indices'}
  if (Tail<0) or (Tail>1) then raise ELSRMRangeError.Create(rsLSRM_NoSuchTail);
  {$ENDREGION}
  // if m is out of range the alpha-mixed ultimate will be returned
  if (m<0) or (m>FM) then m:=FM+1;
  // if i is out of range the total ultimate over all accident periods will be returned
  if (i<0) or (i>FI) then i:=FI+1;
  // calculate the ultimate
  if (m<>FM+1) and not(Init_UltimateDetail in FInitialised) then ErrorMsg:=InitUltimateDetail();
  if (m=FM+1) and not(Init_Ultimate in FInitialised) then ErrorMsg:=InitUltimate();
  if ErrorMsg<>'' then raise ELSRMRangeError.Create(ErrorMsg);
  // return the ultimate
  Result:=FUltimate[(Tail*(FM+2)+m)*(FI+2)+i];
end;

function TNormalLSRM.GetUsedalpha(m, i: Integer): Double;
begin
  {$REGION 'test indecies'}
  if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (i<0) or (i>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
  {$ENDREGION}
  Result:=FUsedalpha^[m*(FI+1)+i];
end;

function TNormalLSRM.GetUsedf(m, k: integer): Double;
begin
  {$REGION 'test indecies'}
  if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (k<0) or (k>FJ-1+FK) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
  {$ENDREGION}
  // return the used development factor f^m_k
  Result:=FUsedf^[m*(FJ+FK)+k];
end;

function TNormalLSRM.GetUsedsigma(m1, m2, k: integer): Double;
begin
  {$REGION 'test indecies'}
  if (m1<0) or (m1>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (m2<0) or (m2>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (k<0) or (k>FJ-1+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
  {$ENDREGION}
  // return the used variance parameter sigma^m1m2_k
  Result:=FUsedsigma^[(m1*(FM+1)+m2)*(FJ+FP)+k];
end;

function TNormalLSRM.GetUsedw(n, m, i, k: integer): Double;
begin
  {$REGION 'test indecies'}
  if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (i<0) or (i>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
  if (k<0) or (k>FJ-1+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
  {$ENDREGION}
  // return the used weights
  //   - not normalised, if n=0
  //   - for busines period I, if n=1
  //   - for business period I+1, if n=2
  Result:=FUsedw^[((n*(FM+1)+m)*(FI+1)+i)*(FJ+FP)+k];
end;

{$ENDREGION}

{$REGION '    Operators'}

function TNormalLSRM.OP_F(n: integer; S: T1Double): string;
var m, i, cl, l, ch, h, j: Integer;
    TmpDestDouble, TmpSDouble, Tmpgamma1Double, TmpfDouble: IndexDouble;
begin
  Result:='';
  try
    For m := 0 To FM do begin
      For i := Max(0, n + 1 - FJ - FP) To Min(FI, n) do begin
        // k = n-i+1
        TmpDestDouble:=(m*(FI+1)+i)*(FJ+1+FP)+n-i+1;
        S[TmpDestDouble]:=0;
        TmpfDouble:=m*(FJ+FK)+n-i;
        for cl:=0 to FmlDependency[m,FM] do begin
          l:=FmlDependency[m,cl];
          for ch:=0 to FihDependency[i,FI] do begin
            h:=FihDependency[i,ch];
            TmpSDouble:=(l*(FI+1)+h)*(FJ+1+FP);
            Tmpgamma1Double:=((((m*(FM+1)+l)*(FI+1)+i)*(FJ+FK)+(n-i))*(FI+1)+h)*(FJ+FK);
            For j := 0 To Min(n - h, n - i) do begin
              S[TmpDestDouble]:=S[TmpDestDouble] + FGamma1[TmpGamma1Double] * S[TmpSDouble];
              Inc(TmpSDouble,1);
              Inc(Tmpgamma1Double,1);
            end;
          end;
        end;
        S[TmpdestDouble]:=S[TmpDestDouble] * FUsedf^[TmpfDouble];
      end;
    end;
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorOP_F + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorOP_F + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.OP_Fn(n1, n2: integer; S: T1Double): string;
var n: integer;
begin
  Result:='';
  for n:=n1 to n2 do begin
    Result:=OP_F(n,S);
    if Result<>'' then Exit;
  end;
end;

function TNormalLSRM.OP_H(n, k: integer; tau, SxS: PDouble): string;
var m1, m2, i1, i2, k1, k2, cl1, cl2, l1, l2, ch1, ch2, h1, h2, j1, j2: Integer;
    PDestDouble, PSxSDouble, PTauDouble: PDouble;
    PGamma11Double, PGamma12Double, Pf1Double, Pf2Double: IndexDouble;
begin
  Result:='';
  try
    Pf1Double:=k;
    For m1 := 0 To FM do begin
      Pf2Double:=k;
      For m2 := 0 To FM do begin
        {$REGION 'k1=k+1 and k2=k+1 and i1+k1>n and i2+k2>n'}
        For i1 := n - k To FI do begin
          For i2 := n - k To FI do begin
            PDestDouble:=SxS;
            Inc(PDestDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k+1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k+1);
            PDestDouble^:=0;
            for cl1:=0 to FmlDependency[m1,FM] do begin
              l1:=FmlDependency[m1,cl1];
              for cl2:=0 to FmlDependency[m2,FM] do begin
                l2:=FmlDependency[m2,cl2];
                for ch1:=0 to FihDependency[i1,FI] do begin
                  h1:=FihDependency[i1,ch1];
                  for ch2:=0 to FihDependency[i2,FI] do begin
                    h2:=FihDependency[i2,ch2];
                    For j1 := 0 To Min(i1 + k - h1, k) do begin
                      PGamma11Double:=((((m1*(FM+1)+l1)*(FI+1)+i1)*(FJ+FK)+k)*(FI+1)+h1)*(FJ+FK)+j1;
                      For j2 := 0 To Min(i2 + k - h2, k) do begin
                        PGamma12Double:=((((m2*(FM+1)+l2)*(FI+1)+i2)*(FJ+FK)+k)*(FI+1)+h2)*(FJ+FK)+j2;
                        PSxSDouble:=SxS;
                        Inc(PSxSDouble,((((l1*(FI+1)+h1)*(FJ+1+FP)+j1)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j2);
                        PDestDouble^:=
                              PDestDouble^ + FGamma1[PGamma11Double] * FGamma1[PGamma12Double] * PSxSDouble^;
                      end;
                    end;
                  end;
                end;
              end;
            end;
            PtauDouble:=tau;
            Inc(PtauDouble,(((m1*(FM+1)+m2)*(FI+1)+i1)*(FI+1)+i2)*(FJ+FP)+k);
            PDestDouble^:=(FUsedf^[Pf1Double] * FUsedf^[Pf2Double]  + PTauDouble^) * PDestDouble^;
          end;
        end;{$ENDREGION}
        {$REGION 'k1=k+1 and i1+k1>n and k2<k+1'}
        For i1 := n - k To FI do begin
          For i2 := 0 To FI do begin
            For k2 := 0 To k do begin
              PDestDouble:=SxS;
              Inc(PDestDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k+1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k2);
              PDestDouble^:=0;
              for cl1:=0 to FmlDependency[m1,FM] do begin
                l1:=FmlDependency[m1,cl1];
                for ch1:=0 to FihDependency[i1,FI] do begin
                  h1:=FihDependency[i1,ch1];
                  For j1 := 0 To Min(i1 + k - h1, k) do begin
                    PGamma11Double:=((((m1*(FM+1)+l1)*(FI+1)+i1)*(FJ+FK)+k)*(FI+1)+h1)*(FJ+FK)+j1;
                    PSxSDouble:=SxS;
                    Inc(PSxSDouble,((((l1*(FI+1)+h1)*(FJ+1+FP)+j1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k2);
                    PDestDouble^:=PDestDouble^ + FGamma1[PGamma11Double] * PSxSDouble^;
                  end;
                end;
              end;
              PDestDouble^:=FUsedf^[Pf1Double] * PDestDouble^;
            end;
          end;
        end; {$ENDREGION}
        {$REGION 'k1=k+1 and i1+k1>n and k2=k+1 and i2+k2<=n'}
        For i1 := n - k To FI do begin
          For i2 := 0 To n - k - 1 do begin
            PDestDouble:=SxS;
            Inc(PDestDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k+1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k+1);
            PDestDouble^:=0;
            for cl1:=0 to FmlDependency[m1,FM] do begin
              l1:=FmlDependency[m1,cl1];
              for ch1:=0 to FihDependency[i1,FI] do begin
                h1:=FihDependency[i1,ch1];
                For j1 := 0 To Min(i1 + k - h1, k) do begin
                  PGamma11Double:=((((m1*(FM+1)+l1)*(FI+1)+i1)*(FJ+FK)+k)*(FI+1)+h1)*(FJ+FK)+j1;
                  PSxSDouble:=SxS;
                  Inc(PSxSDouble,((((l1*(FI+1)+h1)*(FJ+1+FP)+j1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k+1);
                  PDestDouble^:=PDestDouble^ + FGamma1[PGamma11Double] * PSxSDouble^;
                end;
              end;
            end;
            PDestDouble^:=FUsedf^[Pf1Double] * PDestDouble^;
          end;
        end; {$ENDREGION}
        {$REGION 'k1<k+1 and k2=k+1 and i2+k2>n'}
        For i1 := 0 To FI do begin
          For i2 := n - k To FI do begin
            For k1 := 0 To k do begin
              PDestDouble:=SxS;
              Inc(PDestDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k+1);
              PDestDouble^:=0;
              for cl2:=0 to FmlDependency[m2,FM] do begin
                l2:=FmlDependency[m2,cl2];
                for ch2:=0 to FihDependency[i2,FI] do begin
                  h2:=FihDependency[i2,ch2];
                  For j2 := 0 To Min(i2 + k - h2, k) do begin
                    PGamma12Double:=((((m2*(FM+1)+l2)*(FI+1)+i2)*(FJ+FK)+k)*(FI+1)+h2)*(FJ+FK)+j2;
                    PSxSDouble:=SxS;
                    Inc(PSxSDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k1)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j2);
                    PDestDouble^:=PDestDouble^ + FGamma1[PGamma12Double] * PSxSDouble^;
                  end;
                end;
              end;
              PDestDouble^:=FUsedf^[Pf2Double] * PDestDouble^;
            end;
          end;
        end;{$ENDREGION}
        {$REGION 'k1=k+1 and i1+k1<=n and k2=k+1 and i2+k2>n'}
        For i1 := 0 To n - k - 1 do begin
          For i2 := n - k To FI do begin
            PDestDouble:=SxS;
            Inc(PDestDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k+1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k+1);
            PDestDouble^:=0;
            for cl2:=0 to FmlDependency[m2,FM] do begin
              l2:=FmlDependency[m2,cl2];
              for ch2:=0 to FihDependency[i2,FI] do begin
                h2:=FihDependency[i2,ch2];
                For j2 := 0 To Min(i2 + k - h2, k) do begin
                  PGamma12Double:=((((m2*(FM+1)+l2)*(FI+1)+i2)*(FJ+FK)+k)*(FI+1)+h2)*(FJ+FK)+j2;
                  PSxSDouble:=SxS;
                  Inc(PSxSDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k+1)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j2);
                  PDestDouble^:=PDestDouble^ + FGamma1[PGamma12Double] * PSxSDouble^;
                end;
              end;
            end;
            PDestDouble^:=FUsedf^[Pf2Double] * PDestDouble^;
          end;
        end;{$ENDREGION}
        Inc(Pf2Double,FJ+FK);
      end;
      Inc(Pf1Double,FJ+FK);
    end;
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorOP_H + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorOP_H + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.OP_Hn(n: integer; tau, SxS: PDouble): string;
var k: integer;
begin
  Result:='';
  for k := 0 To FJ - 1 + FP do begin
    Result:=OP_H(n, k, tau, SxS);
    if Result<>'' then Exit;
  end;
end;

{$ENDREGION}

{$ENDREGION}

{$REGION '  TNormalLSRM public'}

constructor TNormalLSRM.Create(Name: String);
begin
  inherited Create;
  FName:=Name;
end;

destructor TNormalLSRM.Destroy;
begin
  FreeAll;
  inherited;
end;

function TNormalLSRM.Init(What, M, I, J, K: Integer; const Gamma1,
  Gamma2: OleVariant; var S, w, f, sigma, alpha, InflationRate: Double): OleVariant;
var Tmps: string;
    TmpWhat: integer;
begin
  Result:='';
  {$REGION 'check for special performance flags'}
  if (What and cLSRMReduceMemSize)<>0 then begin
    if not(calc_ReduceMemSize in FCalculationFlags) then begin
      Include(FCalculationFlags,calc_ReduceMemSize);
      FInitialised:=FInitialised-cNonProxyInits;
    end;
  end else begin
    if calc_ReduceMemSize in FCalculationFlags then begin
      Exclude(FCalculationFlags,calc_ReduceMemSize);
      FInitialised:=FInitialised-cNonProxyInits;
    end;
  end;
  {$ENDREGION}
  {$REGION 'check for calculation flags regarding virtual diversification'}
  // Note, in order to avoid double negative in the interface we ask for Virtual...Diversification
  if (What and cLSRMVirtualProcVarDiversification)<>0 then begin
    if not(calc_VirtualProcVarDiversification in FCalculationFlags) then begin
      FInitialised:=FInitialised-cInfluencedByProcVarDiversification;
      Include(FCalculationFlags,calc_VirtualProcVarDiversification);
    end;
  end else begin
    if calc_VirtualProcVarDiversification in FCalculationFlags then begin
      FInitialised:=FInitialised-cInfluencedByProcVarDiversification;
      Exclude(FCalculationFlags,calc_VirtualProcVarDiversification);
    end;
  end;
  if (What and cLSRMVirtualParaErrDiversification)<>0 then begin
    if not(calc_VirtualParaErrDiversification in FCalculationFlags) then begin
      FInitialised:=FInitialised-cInfluencedByParaErrDiversification;
      Include(FCalculationFlags,calc_VirtualParaErrDiversification);
    end;
  end else begin
    if calc_VirtualParaErrDiversification in FCalculationFlags then begin
      FInitialised:=FInitialised-cInfluencedByParaErrDiversification;
      Exclude(FCalculationFlags,calc_VirtualParaErrDiversification);
    end;
  end;
  if (What and cLSRMVirtualTailDiversification)<>0 then begin
    if not(calc_VirtualTailDiversification in FCalculationFlags) then begin
      FInitialised:=FInitialised-cInfluencedByTailDiversification;
      Include(FCalculationFlags,calc_VirtualTailDiversification);
    end;
  end else begin
    if calc_VirtualTailDiversification in FCalculationFlags then begin
      FInitialised:=FInitialised-cInfluencedByTailDiversification;
      Exclude(FCalculationFlags,calc_VirtualTailDiversification);
    end;
  end;
  {$ENDREGION}
  {$REGION 'check for calculation flags regarding exposures'}
  // Note, in order to avoid estimation artefacts concerning almost zero exposures with the wrong sign
  // we introduce two flags that will prevent exposures from changing signs, i.e. set to zero instead
  if (What and cLSRMPreventPastRfromChangingSign)=0 then begin
    if calc_PreventPastExposureFromChangingSign in FCalculationFlags then begin
      FInitialised:=FInitialised-cInfluencedByR1-cInfluencedByR2-[Init_R1,Init_R2];
      Exclude(FCalculationFlags,calc_PreventPastExposureFromChangingSign);
    end;
  end else begin
    if not(calc_PreventPastExposureFromChangingSign in FCalculationFlags) then begin
      FInitialised:=FInitialised-cInfluencedByR1-cInfluencedByR2-[Init_R1,Init_R2];
      Include(FCalculationFlags,calc_PreventPastExposureFromChangingSign);
    end;
  end;
  if (What and cLSRMPreventFutureRfromChangingSign)=0 then begin
    if calc_PreventFutureExposureFromChangingSign in FCalculationFlags then begin
      FInitialised:=FInitialised-cInfluencedByR1-cInfluencedByR2-[Init_R1,Init_R2];
      Exclude(FCalculationFlags,calc_PreventFutureExposureFromChangingSign);
    end;
  end else begin
    if not(calc_PreventFutureExposureFromChangingSign in FCalculationFlags) then begin
      FInitialised:=FInitialised-cInfluencedByR1-cInfluencedByR2-[Init_R1,Init_R2];
      Include(FCalculationFlags,calc_PreventFutureExposureFromChangingSign);
    end;
  end;
  {$ENDREGION}
  {$REGION 'check for calculation flags regarding Tail CDR'}
  // If there are to few observation the process variance of the CDR cannot be calculated.
  // If cLSRMUseMSEPifCDRisZero is set we use the MSEP instead of the CDR for such cases,
  // otherwise we take zero.
  if (What and cLSRMUseMSEPifTailCDRisLarger)=0 then begin
    if calc_UseMSEPifTailCDRisLarger in FCalculationFlags then begin
      FInitialised:=FInitialised-cInfluencedByCDRProcVarTailProxyDetail-cInfluencedByCDRProcVarTailDetail;
      Exclude(FCalculationFlags,calc_UseMSEPifTailCDRisLarger);
    end;
  end else begin
    if not(calc_UseMSEPifTailCDRisLarger in FCalculationFlags) then begin
      FInitialised:=FInitialised-cInfluencedByCDRProcVarTailProxyDetail-cInfluencedByCDRProcVarTailDetail;
      Include(FCalculationFlags,calc_UseMSEPifTailCDRisLarger);
    end;
  end;
  {$ENDREGION}
  {$REGION 'check if weights should be used for default sigme'}
  if (What and cLSRMUseWeightsForDefaultSigma)=0 then begin
    if calc_UseWeightsForDefaultSigma in FCalculationFlags then begin
      Exclude(FCalculationFlags,calc_UseWeightsForDefaultSigma);
      // make sure dependent objects will be recalculated
      Exclude(FInitialised,Init_DefaultSigma);
      if FUsedsigma=@FDefaultsigma then begin
        SetInitFlags(Init_sigma,0,False);
        what:=what or cLSRMUseDefaultsigma;
      end;
    end;
  end else begin
    if not (calc_UseWeightsForDefaultSigma in FCalculationFlags) then begin
      Include(FCalculationFLags,calc_UseWeightsForDefaultSigma);
      // make sure dependent objects will be recalculated
      Exclude(FInitialised,Init_DefaultSigma);
      if FUsedsigma=@FDefaultsigma then begin
        SetInitFlags(Init_sigma,0,False);
        what:=what or cLSRMUseDefaultsigma;
      end;
    end;
  end;
  {$ENDREGION}
  {$REGION 'initialise geometry, if necessary'}
  if (What and cLSRMInitGeometry)<>0 then begin
    Tmps:=InitGeometry(M,I,J,K);
    if Tmps<>'' then
      if Result<>'' then
        Result:=Result + sLineBreak + Tmps
      else
        Result:=Tmps;
  end;
  {$ENDREGION}
  {$REGION 'take default values, if required'}
  {$REGION 'if no weigths are specified or initialised take defaults'}
  if (((What and cLSRMInitw)=0) and not(Init_w in FInitialised)) or ((What and cLSRMUseDefaultw)<>0) then begin
    what:=what or cLSRMUseDefaultw;
    // make sure dependent objects will be recalculated if manual has been changed to default
    if FUsedw<>@FDefaultw then SetInitFlags(Init_Defaultw,0,False);
    FUsedw:=@FDefaultw;
  end else begin
    // make sure dependent objects will be recalculated if deafult has been changed to manual
    if FUsedw<>@Fw then SetInitFlags(Init_w,0,False);
    FUsedw:=@Fw;
  end;
  {$ENDREGION}
  {$REGION 'if no development factors are specified or initialised take defaults'}
  if (((what and cLSRMInitf)=0) and not(Init_f in FInitialised)) or ((what and cLSRMUseDefaultf)<>0) then begin
    what:=what or cLSRMUseDefaultf;
    // make sure dependent objects will be recalculated if manual has been changed to default
    if FUsedf<>@FDefaultf then SetInitFlags(Init_Defaultf,0,False);
    FUsedf:=@FDefaultf;
  end else begin
    // make sure dependent objects will be recalculated if deafult has been changed to manual
    if FUsedf<>@Ff then SetInitFlags(Init_f,0,False);
    FUsedf:=@Ff;
  end;
  {$ENDREGION}
  {$REGION 'if no variance parameter are specified or initialised take defaults'}
  if (((what and cLSRMInitsigma)=0) and not(Init_sigma in FINitialised)) or ((what and cLSRMUseDefaultsigma)<>0) then begin
    what:=what or cLSRMUseDefaultsigma;
    // make sure dependent objects will be recalculated if manual has been changed to default
    if FUsedsigma<>@FDefaultsigma then SetInitFlags(Init_Defaultsigma,0,False);
    FUsedsigma:=@FDefaultsigma;
  end else begin
    // make sure dependent objects will be recalculated if deafult has been changed to manual
    if FUsedsigma<>@Fsigma then SetInitFlags(Init_sigma,0,False);
    FUsedsigma:=@Fsigma;
  end;
  {$ENDREGION}
  {$REGION 'if no mixing parameter are specified or initialised take defaults'}
  if (((what and cLSRMInitalpha)=0) and not(Init_alpha in FInitialised)) or ((what and cLSRMUseDefaultalpha)<>0) then begin
    what:=what or cLSRMUseDefaultalpha;
    FUsedalpha:=@FDefaultalpha;
  end else
    FUsedalpha:=@Falpha;
  {$ENDREGION}
  {$ENDREGION}
  {$REGION 'initialise S, if necessary'}
  if (What and cLSRMInitS)<>0 then begin
    Tmps:=InitS(S);
    if Tmps<>'' then
      if Result<>'' then
        Result:=Result + sLineBreak + Tmps
      else
        Result:=Tmps;
  end;
  {$ENDREGION}
  {$REGION 'solve dependence of inflation rate and gammas'}
  TmpWhat:=0;
  if (What and cLSRMInitInflationRate)<>0 then begin
    if Init_gamma1 in FInitialised then TmpWhat:=(TmpWhat or cLSRMInitgamma1);
    if Init_gamma2 in FInitialised then TmpWhat:=(TmpWhat or cLSRMInitgamma2);
  end;
  {$ENDREGION}
  {$REGION 'initialise Gamma1, if necessary'}
  if ((What or TmpWhat) and cLSRMInitGamma1)<>0 then begin
    Tmps:=InitGamma1(Gamma1,(What and cLSRMInitgamma1)<>0, (What and cLSRMInitInflationRate)<>0,InflationRate);
    if Tmps<>'' then
      if Result<>'' then
        Result:=Result + sLineBreak + Tmps
      else
        Result:=Tmps;
  end;
  {$ENDREGION}
  {$REGION 'initialise Gamma2, if necessary'}
  if ((What or TmpWhat) and cLSRMInitGamma2)<>0 then begin
    Tmps:=InitGamma2(Gamma2,(What and cLSRMInitgamma2)<>0,(What and cLSRMInitInflationRate)<>0,InflationRate);
    if Tmps<>'' then
      if Result<>'' then
        Result:=Result + sLineBreak + Tmps
      else
        Result:=Tmps;
  end;
  {$ENDREGION}
  {$REGION 'initialise inflation rate, if necessary'}
  if (What and cLSRMInitInflationRate)<>0 then begin
    Tmps:=InitInflationRate(InflationRate);
    // correct FInitialised if necessary
    if Tmps<>'' then
      if Result<>'' then
        Result:=Result + sLineBreak + Tmps
      else
        Result:=Tmps;
  end;
  {$ENDREGION}
  {$REGION 'initialise manual specified w, if necessary'}
  if (What and cLSRMInitw)<>0 then begin
    Tmps:=Initw(w);
    if Tmps<>'' then
      if Result<>'' then
        Result:=Result + sLineBreak + Tmps
      else
        Result:=Tmps;
  end;
  {$ENDREGION}
  {$REGION 'initialise default w, if necessary'}
  if ((What and cLSRMUseDefaultw)<>0) then begin
    Tmps:=InitDefaultw();
    if Tmps<>'' then
      if Result<>'' then
        Result:=Result + sLineBreak + Tmps
      else
        Result:=Tmps;
  end;
  {$ENDREGION}
  {$REGION 'initialise manual specified f, if necessary'}
  if (What and cLSRMInitf)<>0 then begin
    Tmps:=Initf(f);
    if Tmps<>'' then
      if Result<>'' then
        Result:=Result + sLineBreak + Tmps
      else
        Result:=Tmps;
  end;
  {$ENDREGION}
  {$REGION 'initialise default f, if necessary'}
  if (What and cLSRMUseDefaultf)<>0 then begin
    Tmps:=InitDefaultf();
    if Tmps<>'' then
      if Result<>'' then
        Result:=Result + sLineBreak + Tmps
      else
        Result:=Tmps;
  end;
  {$ENDREGION}
  {$REGION 'initialise manual specified sigma, if necessary'}
  if (What and cLSRMInitsigma)<>0 then begin
    Tmps:=Initsigma(sigma);
    if Tmps<>'' then
      if Result<>'' then
        Result:=Result + sLineBreak + Tmps
      else
        Result:=Tmps;
  end;
  {$ENDREGION}
  {$REGION 'initialise default sigma, if necessary'}
  if (What and cLSRMUseDefaultsigma)<>0 then begin
    Tmps:=InitDefaultsigma();
    if Tmps<>'' then
      if Result<>'' then
        Result:=Result + sLineBreak + Tmps
      else
        Result:=Tmps;
  end;
  {$ENDREGION}
  {$REGION 'initialise manual specified alpha, if necessary'}
  if (What and cLSRMInitalpha)<>0 then begin
    Tmps:=Initalpha(alpha);
    if Tmps<>'' then
      if Result<>'' then
        Result:=Result + sLineBreak + Tmps
      else
        Result:=Tmps;
  end;
  {$ENDREGION}
  {$REGION 'initialise default alpha, if necessary'}
  if (What and cLSRMUseDefaultalpha)<>0 then begin
    Tmps:=InitDefaultalpha();
    if Tmps<>'' then
      if Result<>'' then
        Result:=Result + sLineBreak + Tmps
      else
        Result:=Tmps;
  end;
  {$ENDREGION}
end;

{$REGION '    Init functions'}

function TNormalLSRM.Initalpha(var alpha: Double): String;
var HasChanged: Boolean;
    StartTime, EndTime: Int64;
begin
  Result:='';
  HasChanged:=False;
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:= 'LSRM ''' + FName + ''':' + rsLSRM_ErrorInitalpha + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the array, if necessary
    if Length(Falpha)<=0 then begin
      SetLength(Falpha,(FM+1)*(FI+1));
      HasChanged:=True;
    end;
    // check if the alphas are different from the one we possible already have, if yes take the new one and
    // set all dependent objects to not initialised
    if HasChanged or not(Init_alpha in FInitialised)
        or not(CompareMem(@Falpha[0],@alpha,(FM+1)*(FI+1)*SizeOf(Double))) then begin
      Move(alpha,Falpha[0],(FM+1)*(FI+1)*SizeOf(Double));
      QueryPerformanceCounter(EndTime);
      SetInitFlags(Init_alpha,EndTime-StartTime);
    end;
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitalpha + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitalpha + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDR: string;
var StartTime, EndTime: Int64;
    PProcVarDouble, PParaErrDouble, PTotalDouble: IndexDouble;
    m, i: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  if calc_ReduceMemSize in FCalculationFlags then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDR + sLineBreak + rsLSRM_ProxyOnly;
    Exit;
  end;
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDR + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRParaErr in FInitialised) then Result:=InitCDRParaErr();
  if Result<>'' then Exit;
  If not(Init_CDRProcVar in FInitialised) then Result:=InitCDRProcVar();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // calculate the sum of process variance and parameter error (without approximation
    // and store result within FCDR(0,2)
    PProcVarDouble:=0;
    PParaErrDouble:=PProcVarDouble;
    Inc(PParaErrDouble,(FM+2)*(FI+2));
    PTotalDouble:=PParaErrDouble;
    Inc(PTotalDouble,(FM+2)*(FI+2));
    for m := 0 to FM+1 do begin
      for i := 0 to FI+1 do begin
        FCDR[PTotalDouble]:=FCDR[PProcVarDouble] + FCDR[PParaErrDouble];
        Inc(PTotalDouble,1);
        Inc(PProcVarDouble,1);
        Inc(PParaErrDouble,1);
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDR,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDR + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDR + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCov: string;
var StartTime, EndTime: Int64;
    PDestDouble, PParaErrDouble, PProcVarDouble: IndexDouble;
    m1, m2, Proxy, Part: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCov + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRCovParaErr in FInitialised) then Result:=InitCDRCovParaErr();
  if Result<>'' then Exit;
  If not(Init_CDRCovProcVar in FInitialised) then Result:=InitCDRCovProcVar();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=0;
    Part:=2;
    // initialise array or clear the memory
    if Length(FCDRCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FCDRCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // without tail
      ZeroMemory(@FCDRCov[(Proxy*3+Part)*2*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    PDestDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    PProcVarDouble:=(Proxy*3+0)*2*(FM+1)*(FM+1);
    PParaErrDouble:=(Proxy*3+1)*2*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        FCDRCov[PDestDouble]:=FCDRCov[PProcVarDouble]+FCDRCov[PParaErrDouble];
        Inc(PDestDouble,1);
        Inc(PProcVarDouble,1);
        Inc(PParaErrDouble,1);
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCov,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCov + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCov + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovEigenvalue: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalue + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRCov in FInitialised) then Result:=InitCDRCov();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FCDRCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FCDRCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 0, Part = 2, Tail = 0
    Proxy:=0;
    Part:=2;
    Tail:=0;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FCDRCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FCDRCovEigenvalues[PDestDouble],FCDRCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalue + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovEigenvalue,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalue + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalue + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovEigenvalueParaErr: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueParaErr + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRCovParaErr in FInitialised) then Result:=InitCDRCovParaErr();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FCDRCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FCDRCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 0, Part = 1, Tail = 0
    Proxy:=0;
    Part:=1;
    Tail:=0;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FCDRCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FCDRCovEigenvalues[PDestDouble],FCDRCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueParaErr + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovEigenvalueParaErr,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueParaErr + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueParaErr + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovEigenvalueParaErrProxy: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueParaErrProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRCovParaErrProxy in FInitialised) then Result:=InitCDRCovParaErrProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FCDRCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FCDRCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 1, Part = 1, Tail = 0
    Proxy:=1;
    Part:=1;
    Tail:=0;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FCDRCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FCDRCovEigenvalues[PDestDouble],FCDRCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueParaErrProxy + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovEigenvalueParaErrProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueParaErrProxy + sLineBreak 
              + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueParaErrProxy + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovEigenvalueParaErrTail: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueParaErrTail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRCovParaErrTail in FInitialised) then Result:=InitCDRCovParaErrTail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FCDRCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FCDRCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 0, Part = 1, Tail = 1
    Proxy:=0;
    Part:=1;
    Tail:=1;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FCDRCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FCDRCovEigenvalues[PDestDouble],FCDRCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueParaErrTail + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovEigenvalueParaErrTail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueParaErrTail + sLineBreak 
              + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueParaErrTail + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovEigenvalueParaErrTailProxy: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueParaErrTailProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRCovParaErrTailProxy in FInitialised) then Result:=InitCDRCovParaErrTailProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FCDRCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FCDRCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 1, Part = 1, Tail = 1
    Proxy:=1;
    Part:=1;
    Tail:=1;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FCDRCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FCDRCovEigenvalues[PDestDouble],FCDRCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueParaErrTailProxy + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovEigenvalueParaErrTailProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueParaErrTailProxy + sLineBreak 
              + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueParaErrTailProxy + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovEigenvalueProcVar: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProcVar + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRCovProcVar in FInitialised) then Result:=InitCDRCovProcVar();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FCDRCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FCDRCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 0, Part = 0, Tail = 0
    Proxy:=0;
    Part:=0;
    Tail:=0;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FCDRCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FCDRCovEigenvalues[PDestDouble],FCDRCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProcVar + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovEigenvalueProcVar,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProcVar + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProcVar + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovEigenvalueProcVarProxy: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProcVarProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRCovProcVarProxy in FInitialised) then Result:=InitCDRCovProcVarProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FCDRCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FCDRCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 1, Part = 0, Tail = 0
    Proxy:=1;
    Part:=0;
    Tail:=0;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FCDRCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FCDRCovEigenvalues[PDestDouble],FCDRCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProcVarProxy + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovEigenvalueProcVarProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProcVarProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProcVarProxy + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovEigenvalueProcVarTail: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProcVarTail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRCovProcVarTail in FInitialised) then Result:=InitCDRCovProcVarTail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FCDRCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FCDRCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 0, Part = 0, Tail = 1
    Proxy:=0;
    Part:=0;
    Tail:=1;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FCDRCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FCDRCovEigenvalues[PDestDouble],FCDRCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProcVarTail + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovEigenvalueProcVarTail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProcVarTail + sLineBreak 
              + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProcVarTail + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovEigenvalueProcVarTailProxy: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProcVarTailProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRCovProcVarTailProxy in FInitialised) then Result:=InitCDRCovProcVarTailProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FCDRCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FCDRCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 1, Part = 0, Tail = 1
    Proxy:=1;
    Part:=0;
    Tail:=1;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FCDRCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FCDRCovEigenvalues[PDestDouble],FCDRCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProcVarTailProxy + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovEigenvalueProcVarTailProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProcVarTailProxy + sLineBreak 
              + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProcVarTailProxy + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovEigenvalueProxy: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRCovProxy in FInitialised) then Result:=InitCDRCovProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FCDRCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FCDRCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 1, Part = 2, Tail = 0
    Proxy:=1;
    Part:=2;
    Tail:=0;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FCDRCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FCDRCovEigenvalues[PDestDouble],FCDRCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProxy + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovEigenvalueProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueProxy + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovEigenvalueTail: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueTail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRCovTail in FInitialised) then Result:=InitCDRCovTail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FCDRCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FCDRCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 0, Part = 2, Tail = 1
    Proxy:=0;
    Part:=2;
    Tail:=1;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FCDRCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FCDRCovEigenvalues[PDestDouble],FCDRCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueTail + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovEigenvalueTail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueTail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueTail + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovEigenvalueTailProxy: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueTailProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRCovTailProxy in FInitialised) then Result:=InitCDRCovTailProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FCDRCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FCDRCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FCDRCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 0, Part = 1, Tail = 0
    Proxy:=1;
    Part:=2;
    Tail:=1;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FCDRCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FCDRCovEigenvalues[PDestDouble],FCDRCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueTailProxy + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovEigenvalueTailProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueTailProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovEigenvalueTailProxy + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovParaErr: string;
var StartTime, EndTime: Int64;
    PDestDouble, PCDRDetailDouble, Palpha1Double, Palpha2Double: IndexDouble;
    m1, m2, i1, i2, Proxy, Part: Integer;
    TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovParaErr + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovParaErr + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_CDRParaErrDetail in FInitialised) then Result:=InitCDRParaErrDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=0;
    Part:=1;
    // initialise array or clear the memory
    if Length(FCDRCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FCDRCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // without tail
      ZeroMemory(@FCDRCov[(Proxy*3+Part)*2*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    {$REGION 'calculate'}
    PCDRDetailDouble:=(Proxy*2+Part)*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PDestDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          for i2 := 0 to FI do begin
            TmpValue:=FUsedalpha^[Palpha1Double]
                      *FUsedalpha^[Palpha2Double]
                      *FCDRDetail[PCDRDetailDouble];
            FCDRCov[PDestDouble]:=FCDRCov[PDestDouble]+TmpValue;
            Inc(PCDRDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
        Inc(PDestDouble,1);
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovParaErr,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovParaErr + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovParaErr + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovParaErrProxy: string;
var StartTime, EndTime: Int64;
    PDestDouble, PCDRDetailDouble, Palpha1Double, Palpha2Double: IndexDouble;
    m1, m2, i1, i2, Proxy, Part: Integer;
    TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovParaErrProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovParaErrProxy + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_CDRParaErrProxyDetail in FInitialised) then Result:=InitCDRParaErrProxyDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=1;
    Part:=1;
    // initialise array or clear the memory
    if Length(FCDRCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FCDRCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // without tail
      ZeroMemory(@FCDRCov[(Proxy*3+Part)*2*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    {$REGION 'calculate'}
    PCDRDetailDouble:=(Proxy*2+Part)*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PDestDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          for i2 := 0 to FI do begin
            TmpValue:=FUsedalpha^[Palpha1Double]
                      *FUsedalpha^[Palpha2Double]
                      *FCDRDetail[PCDRDetailDouble];
            FCDRCov[PDestDouble]:=FCDRCov[PDestDouble]+TmpValue;
            Inc(PCDRDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
        Inc(PDestDouble,1);
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovParaErrProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovParaErrProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovParaErrProxy + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovParaErrTail: string;
var StartTime, EndTime: Int64;
    PNonTailDouble, PDestDouble, PCDRDetailDouble, Palpha1Double, Palpha2Double: IndexDouble;
    m1, m2, i1, i2, Proxy, Part: Integer;
    TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovParaErrTail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovParaErrTail + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_CDRCovParaErr in FInitialised) then Result:=InitCDRCovParaErr();
  if Result<>'' then Exit;
  If not(Init_CDRParaErrTailDetail in FInitialised) then Result:=InitCDRParaErrTailDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=0;
    Part:=1;
    // initialise array or clear the memory
    if Length(FCDRCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FCDRCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // with tail only
      ZeroMemory(@FCDRCov[((Proxy*3+Part)*2+1)*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    {$REGION 'calculate'}
    PCDRDetailDouble:=(Proxy*2+Part)*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PNonTailDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    PDestDouble:=((Proxy*3+Part)*2+1)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        // get the non tail part
        FCDRCov[PDestDouble]:=FCDRCov[PNonTailDouble];
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          for i2 := 0 to FI do begin
            // tail
            TmpValue:=FUsedalpha^[Palpha1Double]
                      *FUsedalpha^[Palpha2Double]
                      *FCDRTailDetail[PCDRDetailDouble];
            FCDRCov[PDestDouble]:=FCDRCov[PDestDouble]+TmpValue;
            Inc(PCDRDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
        Inc(PNonTailDouble,1);
        Inc(PDestDouble,1);
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovParaErrTail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovParaErrTail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovParaErrTail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovParaErrTailProxy: string;
var StartTime, EndTime: Int64;
    PNonTailDouble, PDestDouble, PCDRDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2, Proxy, Part: Integer;
    TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovParaErrTailProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovParaErrTailProxy + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_CDRCovParaErrProxy in FInitialised) then Result:=InitCDRCovParaErrProxy();
  if Result<>'' then Exit;
  If not(Init_CDRParaErrTailProxyDetail in FInitialised) then Result:=InitCDRParaErrTailProxyDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=1;
    Part:=1;
    // initialise array or clear the memory
    if Length(FCDRCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FCDRCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // tail only
      ZeroMemory(@FCDRCov[((Proxy*3+Part)*2+1)*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    {$REGION 'calculate'}
    PCDRDetailDouble:=(Proxy*2+Part)*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PNonTailDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    PDestDouble:=((Proxy*3+Part)*2+1)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        // get the non tail part
        FCDRCov[PDestDouble]:=FCDRCov[PNonTailDouble];
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          for i2 := 0 to FI do begin
            // tail
            TmpValue:=FUsedalpha^[Palpha1Double]
                      *FUsedalpha^[Palpha2Double]
                      *FCDRTailDetail[PCDRDetailDouble];
            FCDRCov[PDestDouble]:=FCDRCov[PDestDouble]+TmpValue;
            Inc(PCDRDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
        Inc(PNonTailDouble,1);
        Inc(PDestDouble,1);
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovParaErrTailProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovParaErrTailProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovParaErrTailProxy + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovProcVar: string;
var StartTime, EndTime: Int64;
    PDestDouble, PCDRDetailDouble, Palpha1Double, Palpha2Double: IndexDouble;
    m1, m2, i1, i2, Proxy, Part: Integer;
    TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovProcVar + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovProcVar + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_CDRProcVarDetail in FInitialised) then Result:=InitCDRProcVarDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=0;
    Part:=0;
    // initialise array or clear the memory
    if Length(FCDRCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FCDRCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // without tail
      ZeroMemory(@FCDRCov[(Proxy*3+Part)*2*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    {$REGION 'calculate'}
    PCDRDetailDouble:=(Proxy*2+Part)*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PDestDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          for i2 := 0 to FI do begin
            TmpValue:=FUsedalpha^[Palpha1Double]
                      *FUsedalpha^[Palpha2Double]
                      *FCDRDetail[PCDRDetailDouble];
            FCDRCov[PDestDouble]:=FCDRCov[PDestDouble]+TmpValue;
            Inc(PCDRDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
        Inc(PDestDouble,1);
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovProcVar,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovProcVar + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovProcVar + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovProcVarProxy: string;
var StartTime, EndTime: Int64;
    PDestDouble, PCDRDetailDouble, Palpha1Double, Palpha2Double: IndexDouble;
    m1, m2, i1, i2, Proxy, Part: Integer;
    TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovProcVarProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovProcVarProxy + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_CDRProcVarProxyDetail in FInitialised) then Result:=InitCDRProcVarProxyDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=1;
    Part:=0;
    // initialise array or clear the memory
    if Length(FCDRCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FCDRCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // without tail
      ZeroMemory(@FCDRCov[(Proxy*3+Part)*2*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    {$REGION 'calculate'}
    PCDRDetailDouble:=(Proxy*2+Part)*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PDestDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          for i2 := 0 to FI do begin
            TmpValue:=FUsedalpha^[Palpha1Double]
                      *FUsedalpha^[Palpha2Double]
                      *FCDRDetail[PCDRDetailDouble];
            FCDRCov[PDestDouble]:=FCDRCov[PDestDouble]+TmpValue;
            Inc(PCDRDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
        Inc(PDestDouble,1);
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovProcVarProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovProcVarProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovProcVarProxy + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovProcVarTail: string;
var StartTime, EndTime: Int64;
    PNonTailDouble, PDestDouble, PCDRDetailDouble, Palpha1Double, Palpha2Double: IndexDouble;
    m1, m2, i1, i2, Proxy, Part: Integer;
    TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovProcVarTail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovProcVarTail + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_CDRCovProcVar in FInitialised) then Result:=InitCDRCovProcVar();
  if Result<>'' then Exit;
  If not(Init_CDRProcVarTailDetail in FInitialised) then Result:=InitCDRProcVarTailDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=0;
    Part:=0;
    // initialise array or clear the memory
    if Length(FCDRCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FCDRCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // tail only
      ZeroMemory(@FCDRCov[((Proxy*3+Part)*2+1)*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    {$REGION 'calculate'}
    PCDRDetailDouble:=(Proxy*2+Part)*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PNonTailDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    PDestDouble:=((Proxy*3+Part)*2+1)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        // get the non tail part
        FCDRCov[PDestDouble]:=FCDRCov[PNonTailDouble];
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          for i2 := 0 to FI do begin
            // tail
            TmpValue:=FUsedalpha^[Palpha1Double]
                      *FUsedalpha^[Palpha2Double]
                      *FCDRTailDetail[PCDRDetailDouble];
            FCDRCov[PDestDouble]:=FCDRCov[PDestDouble]+TmpValue;
            Inc(PCDRDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
        Inc(PNonTailDouble,1);
        Inc(PDestDouble,1);
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovProcVarTail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovProcVarTail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovProcVarTail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovProcVarTailProxy: string;
var StartTime, EndTime: Int64;
    PNonTailDouble, PDestDouble, PCDRDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2, Proxy, Part: Integer;
    TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovProcVarTailProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovProcVarTailProxy + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_CDRCovProcVarProxy in FInitialised) then Result:=InitCDRCovProcVarProxy();
  if Result<>'' then Exit;
  If not(Init_CDRProcVarTailProxyDetail in FInitialised) then Result:=InitCDRProcVarTailProxyDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=1;
    Part:=0;
    // initialise array or clear the memory
    if Length(FCDRCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FCDRCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // tail only
      ZeroMemory(@FCDRCov[((Proxy*3+Part)*2+1)*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    {$REGION 'calculate'}
    PCDRDetailDouble:=(Proxy*2+Part)*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PNonTailDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    PDestDouble:=((Proxy*3+Part)*2+1)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        // get the non tail part
        FCDRCov[PDestDouble]:=FCDRCov[PNonTailDouble];
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          for i2 := 0 to FI do begin
            // tail
            TmpValue:=FUsedalpha^[Palpha1Double]
                      *FUsedalpha^[Palpha2Double]
                      *FCDRTailDetail[PCDRDetailDouble];
            FCDRCov[PDestDouble]:=FCDRCov[PDestDouble]+TmpValue;
            Inc(PCDRDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
        Inc(PNonTailDouble,1);
        Inc(PDestDouble,1);
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovProcVarTailProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovProcVarTailProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovProcVarTailProxy + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovProxy: string;
var StartTime, EndTime: Int64;
    PDestDouble, PParaErrDouble, PProcVarDouble: IndexDouble;
    m1, m2, Proxy, Part: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRCovParaErrProxy in FInitialised) then Result:=InitCDRCovParaErrProxy();
  if Result<>'' then Exit;
  If not(Init_CDRCovProcVarProxy in FInitialised) then Result:=InitCDRCovProcVarProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=1;
    Part:=2;
    // initialise array or clear the memory
    if Length(FCDRCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FCDRCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // without tail
      ZeroMemory(@FCDRCov[(Proxy*3+Part)*2*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    PDestDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    PProcVarDouble:=(Proxy*3+0)*2*(FM+1)*(FM+1);
    PParaErrDouble:=(Proxy*3+1)*2*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        FCDRCov[PDestDouble]:=FCDRCov[PProcVarDouble]+FCDRCov[PParaErrDouble];
        Inc(PDestDouble,1);
        Inc(PProcVarDouble,1);
        Inc(PParaErrDouble,1);
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovProxy + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovTail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PParaErrDouble, PProcVarDouble: IndexDouble;
    m1, m2, Proxy, Part: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovTail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRCovParaErrTail in FInitialised) then Result:=InitCDRCovParaErrTail();
  if Result<>'' then Exit;
  If not(Init_CDRCovProcVarTail in FInitialised) then Result:=InitCDRCovProcVarTail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=0;
    Part:=2;
    // initialise array or clear the memory
    if Length(FCDRCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FCDRCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      ZeroMemory(@FCDRCov[((Proxy*3+Part)*2+1)*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    PDestDouble:=((Proxy*3+Part)*2+1)*(FM+1)*(FM+1);
    PProcVarDouble:=((Proxy*3+0)*2+1)*(FM+1)*(FM+1);
    PParaErrDouble:=((Proxy*3+1)*2+1)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        FCDRCov[PDestDouble]:=FCDRCov[PProcVarDouble]+FCDRCov[PParaErrDouble];
        Inc(PDestDouble,1);
        Inc(PProcVarDouble,1);
        Inc(PParaErrDouble,1);
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovTail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovTail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovTail + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRCovTailProxy: string;
var StartTime, EndTime: Int64;
    PDestDouble, PParaErrDouble, PProcVarDouble: IndexDouble;
    m1, m2, Proxy, Part: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovTailProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRCovParaErrTailProxy in FInitialised) then Result:=InitCDRCovParaErrTailProxy();
  if Result<>'' then Exit;
  If not(Init_CDRCovProcVarTailProxy in FInitialised) then Result:=InitCDRCovProcVarTailProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=1;
    Part:=2;
    // initialise array or clear the memory
    if Length(FCDRCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FCDRCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      ZeroMemory(@FCDRCov[((Proxy*3+Part)*2+1)*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    PDestDouble:=((Proxy*3+Part)*2+1)*(FM+1)*(FM+1);
    PProcVarDouble:=((Proxy*3+0)*2+1)*(FM+1)*(FM+1);
    PParaErrDouble:=((Proxy*3+1)*2+1)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        FCDRCov[PDestDouble]:=FCDRCov[PProcVarDouble]+FCDRCov[PParaErrDouble];
        Inc(PDestDouble,1);
        Inc(PProcVarDouble,1);
        Inc(PParaErrDouble,1);
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRCovTailProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovTailProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRCovTailProxy + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRParaErr: string;
var StartTime, EndTime: Int64;
    PmmiiDouble, PTotalmDouble, PweightediDouble, PTotalweightedDouble, PCDRDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  if calc_ReduceMemSize in FCalculationFlags then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDR + sLineBreak + rsLSRM_ProxyOnly;
    Exit;
  end;
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErr + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErr + sLineBreak + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_CDRParaErrDetail in FInitialised) then Result:=InitCDRParaErrDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the array, if necessary
    if Length(FCDR)<=0 then begin
      //            Proxy Part    m      i
      Setlength(FCDR,2  *  3  *(FM+2)*(FI+2));
    end;
    {$REGION 'unweighted estimates'}
    // Proxy=0, part=1
    PCDRDetailDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PmmiiDouble:=(FM+2)*(FI+2);
    ZeroMemory(@FCDR[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
    PTotalmDouble:=PmmiiDouble;
    Inc(PTotalmDouble,FI+1);
    for m1:=0 to FM do begin
      for i1 := 0 to FI do begin
        // the unweighted estimate of claim property m and accident period i
        for i2 := 0 to i1-1 do begin
          FCDR[PTotalmDouble]:=FCDR[PTotalmDouble]+FCDRDetail[PCDRDetailDouble];
          Inc(PCDRDetailDouble,1);
        end;
        FCDR[PmmiiDouble]:=FCDRDetail[PCDRDetailDouble];
        Inc(PmmiiDouble,1);
        FCDR[PTotalmDouble]:=FCDR[PTotalmDouble]+FCDRDetail[PCDRDetailDouble];
        Inc(PCDRDetailDouble,1);
        for i2 := i1+1 to FI do begin
          FCDR[PTotalmDouble]:=FCDR[PTotalmDouble]+FCDRDetail[PCDRDetailDouble];
          Inc(PCDRDetailDouble,1);
        end;
      end;
      // next m (m1=m2=m)
      Inc(PCDRDetailDouble,(FM+1)*(FI+1)*(FI+1));
      Inc(PTotalmDouble,FI+2);
      Inc(PmmiiDouble,1);
    end;
    {$ENDREGION}
    {$REGION 'weighted estimates'}
      // Proxy=0, part=1
      PCDRDetailDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PTotalweightedDouble:=((FM+2)+(FM+1))*(FI+2)+FI+1;
      for m1 := 0 to FM do begin
        for m2 := 0 to FM do begin
          Palpha1Double:=m1*(FI+1);
          for i1 := 0 to FI do begin
            Palpha2Double:=m2*(FI+1);
            PweightediDouble:=((FM+2)+(FM+1))*(FI+2)+i1;
            for i2 := 0 to FI do begin
              if i1=i2 then begin
                // weighted estimate for accident period i1=i2
                FCDR[PweightediDouble]:=FCDR[PweightediDouble]
                                        +FUsedalpha^[Palpha1Double]
                                        *FUsedalpha^[Palpha2Double]
                                        *FCDRDetail[PCDRDetailDouble];
              end;
              // total (over all accident periods) weighted estimate
              FCDR[PTotalweightedDouble]:=FCDR[PTotalweightedDouble]
                                          +FUsedalpha^[Palpha1Double]
                                          *FUsedalpha^[Palpha2Double]
                                          *FCDRDetail[PCDRDetailDouble];
              Inc(PCDRDetailDouble,1);
              Inc(Palpha2Double,1);
            end;
            Inc(Palpha1Double,1);
          end;
        end;
      end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRParaErr,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErr + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErr + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRParaErrDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PSxSrhostarDouble, PSxSrhostar1Double, PSxSrhostar2Double, PSxSrhostar12Double: IndexDouble;
    m1, m2, i1, i2, k1, k2: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  if calc_ReduceMemSize in FCalculationFlags then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDR + sLineBreak + rsLSRM_ProxyOnly;
    Exit;
  end;
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_SxSCDRParaErr in FInitialised) then Result:=InitSxSCDRParaErr();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRDetail)<=0 then begin
      //                  Proxy Part    m1     m2     i1     i2
      SetLength(FCDRDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    // parameter error non proxy
    // store the results within FCDRDetail(0,1)
    PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
    ZeroMemory(@FCDRDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
    For m1 := 0 To FM do begin
      For m2 := 0 To FM do begin
        For i1 := 0 To FI do begin
          For i2 := 0 To FI do begin
            For k1 := FI - i1 To FJ - 1 do begin
              PSxSrhostarDouble:=
                          ((((((FM+1)+m1)*(FI+1)+i1)*(FJ+1+FP)+k1+1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+FI-i2+1;
              PSxSrhostar1Double:=PSxSrhostarDouble;
              Inc(PSxSrhostar1Double,(FM+1)*(FI+1)*(FJ+1+FP)*(FM+1)*(FI+1)*(FJ+1+FP));
              PSxSrhostar2Double:=PSxSrhostar1Double;
              Inc(PSxSrhostar2Double,(FM+1)*(FI+1)*(FJ+1+FP)*(FM+1)*(FI+1)*(FJ+1+FP));
              PSxSrhostar12Double:=PSxSrhostar2Double;
              Inc(PSxSrhostar12Double,(FM+1)*(FI+1)*(FJ+1+FP)*(FM+1)*(FI+1)*(FJ+1+FP));
              For k2 := FI - i2 To FJ - 1 do begin
                FCDRDetail[PDestDouble]:=FCDRDetail[PDestDouble]
                                        +FSxS[PSXSrhostarDouble]
                                        -FSxS[PSXSrhostar1Double]
                                        -FSxS[PSXSrhostar2Double]
                                        +FSxS[PSXSrhostar12Double];
                Inc(PSxSrhostarDouble,1);
                Inc(PSxSrhostar1Double,1);
                Inc(PSxSrhostar2Double,1);
                Inc(PSxSrhostar12Double,1);
              end;
            end;
            Inc(PDestDouble,1);
          end;
        end;
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRParaErrDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrDetail + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRParaErrProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhostarDouble, Prhostar1Double, Prhostar2Double, Prhostar12Double,
      PCoorF1Double, PCoorF2Double: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements are calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoMSEP in FInitialised) then
    Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoCDR in FInitialised) then
    Result:=InitrhoCDR();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRDetail)<=0 then begin
      //                  Proxy Part    m1     m2     i1     i2
      SetLength(FCDRDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    // parameter error proxy
    // store the results within FCDRDetail(1,1)
    PDestDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    ZeroMemory(@FCDRDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
    if calc_ReduceMemSize in FCalculationFlags then begin
      {$REGION 'calculate rhos each time'}
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                    For ch1 := 0 To FihDependency[i1,FI] do begin
                      h1 := FihDependency[i1,ch1];
                      If h1 >= FI - j Then begin
                        For ch2 := 0 To FihDependency[i2,FI] do begin
                          h2 := FihDependency[i2,ch2];
                          If h2 >= FI - j Then begin
                            For cl1 := 0 To FmlDependency[m1,FM] do begin
                              l1 := FmlDependency[m1,cl1];
                              For cl2 := 0 To FmlDependency[m2,FM] do begin
                                l2 := FmlDependency[m2,cl2];
                                PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                Rho:=Getrho(1,l1,l2,h1,h2,j)
                                    -Getrho(2,l1,l2,h1,h2,j)
                                    -Getrho(3,l1,l2,h1,h2,j)
                                    +Getrho(4,l1,l2,h1,h2,j);
                                TmpValue:=Rho
                                          *FR1[PR11Double]
                                          *FR1[PR12Double]
                                          *FCoorF[PCoorF1Double]
                                          *FCoorF[PCoorF2Double];
                                // diversification only if rho is negative
                                if (calc_VirtualParaErrDiversification in FCalculationFlags)
                                        or (Rho<=0) or (TmpValue>=0) then
                                  FCDRDetail[PDestDouble]:=FCDRDetail[PDestDouble]+TmpValue;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end else begin
      {$REGION 'take rhos from memory'}
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                    For ch1 := 0 To FihDependency[i1,FI] do begin
                      h1 := FihDependency[i1,ch1];
                      If h1 >= FI - j Then begin
                        For ch2 := 0 To FihDependency[i2,FI] do begin
                          h2 := FihDependency[i2,ch2];
                          If h2 >= FI - j Then begin
                            For cl1 := 0 To FmlDependency[m1,FM] do begin
                              l1 := FmlDependency[m1,cl1];
                              For cl2 := 0 To FmlDependency[m2,FM] do begin
                                l2 := FmlDependency[m2,cl2];
                                PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                Prhostar1Double:=PrhostarDouble;
                                Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                Prhostar2Double:=Prhostar1Double;
                                Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                Prhostar12Double:=Prhostar2Double;
                                Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                Rho:=Frho[PrhostarDouble]
                                    -Frho[Prhostar1Double]
                                    -Frho[Prhostar2Double]
                                    +Frho[Prhostar12Double];
                                TmpValue:=Rho
                                          *FR1[PR11Double]
                                          *FR1[PR12Double]
                                          *FCoorF[PCoorF1Double]
                                          *FCoorF[PCoorF2Double];
                                // diversification only if rho is negative
                                if (calc_VirtualParaErrDiversification in FCalculationFlags)
                                        or (Rho<=0) or (TmpValue>=0) then
                                  FCDRDetail[PDestDouble]:=FCDRDetail[PDestDouble]+TmpValue;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRParaErrProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRParaErrTailProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhostarDouble, Prhostar1Double, Prhostar2Double, Prhostar12Double,
      PCoorF1Double, PCoorF2Double, PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoMSEP in FInitialised) then
    Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoCDR in FInitialised) then
    Result:=InitrhoCDR();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // update FCDRTailDetail if necessary
    if Length(FCDRTailDetail)<=0 then begin
      //                  Proxy Part    m1     m2     i1     i2
      SetLength(FCDRTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_CDRParaErrTailDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PProxyDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FCDRTailDetail[PProxyDouble],@FCDRTailDetail[PDestDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the CDR tail parameter error proxy'}
      // store the results within FCDRTailDetail(1,1)
      PDestDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FCDRTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      if calc_ReduceMemSize in FCalculationFlags then begin
        {$REGION 'calculate rhos each time'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To min(Min(k1, k2),FJ-1) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        // here we have to take all old accident periods into account, too
                        // therefore, the condition "h1 >= FI - j" has to be ignored
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            // here we have to take all old accident periods into account, too
                            // therefore, the condition "h2 >= FI - j" has to be ignored
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                          end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else begin
        {$REGION 'take rhos from memory'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        // here we have to take all old accident periods into account, too
                        // therefore, the condition "h1 >= FI - j" has to be ignored
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            // here we have to take all old accident periods into account, too
                            // therefore, the condition "h2 >= FI - j" has to be ignored
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                          end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRParaErrTailProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRParaErrProxy: string;
var StartTime, EndTime: Int64;
    PmmiiDouble, PTotalmDouble, PweightediDouble, PTotalweightedDouble, PCDRDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrProxy + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_CDRParaErrProxyDetail in FInitialised) then Result:=InitCDRParaErrProxyDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDR)<=0 then begin
      //            Proxy Part    m      i
      Setlength(FCDR,2  *  3  *(FM+2)*(FI+2));
    end;
    {$REGION 'unweighted estimates'}
    // Proxy=1, part=1
    PCDRDetailDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PmmiiDouble:=4*(FM+2)*(FI+2);
    ZeroMemory(@FCDR[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
    PTotalmDouble:=PmmiiDouble;
    Inc(PTotalmDouble,FI+1);
    for m1:=0 to FM do begin
      for i1 := 0 to FI do begin
        // the unweighted estimate of claim property m and accident period i
        for i2 := 0 to i1-1 do begin
          FCDR[PTotalmDouble]:=FCDR[PTotalmDouble]+FCDRDetail[PCDRDetailDouble];
          Inc(PCDRDetailDouble,1);
        end;
        FCDR[PmmiiDouble]:=FCDRDetail[PCDRDetailDouble];
        Inc(PmmiiDouble,1);
        FCDR[PTotalmDouble]:=FCDR[PTotalmDouble]+FCDRDetail[PCDRDetailDouble];
        Inc(PCDRDetailDouble,1);
        for i2 := i1+1 to FI do begin
          FCDR[PTotalmDouble]:=FCDR[PTotalmDouble]+FCDRDetail[PCDRDetailDouble];
          Inc(PCDRDetailDouble,1);
        end;
      end;
      // next m (m1=m2=m)
      Inc(PCDRDetailDouble,(FM+1)*(FI+1)*(FI+1));
      Inc(PTotalmDouble,FI+2);
      Inc(PmmiiDouble,1);
    end;
    {$ENDREGION}
    {$REGION 'weighted estimates'}
    // Proxy=1, part=1
    PCDRDetailDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PTotalweightedDouble:=(4*(FM+2)+(FM+1))*(FI+2)+FI+1;
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          PweightediDouble:=(4*(FM+2)+(FM+1))*(FI+2)+i1;
          for i2 := 0 to FI do begin
            if i1=i2 then begin
              // weighted estimate for accident period i1=i2
              FCDR[PweightediDouble]:=FCDR[PweightediDouble]
                                      +FUsedalpha^[Palpha1Double]
                                      *FUsedalpha^[Palpha2Double]
                                      *FCDRDetail[PCDRDetailDouble];
            end;
            // total (over all accident periods) weighted estimate
            FCDR[PTotalweightedDouble]:=FCDR[PTotalweightedDouble]
                                        +FUsedalpha^[Palpha1Double]
                                        *FUsedalpha^[Palpha2Double]
                                        *FCDRDetail[PCDRDetailDouble];
            Inc(PCDRDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRParaErrProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrProxy + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRParaErrTail: string;
var StartTime, EndTime: Int64;
    PmmiiDouble, PTotalmDouble, PweightediDouble, PTotalweightedDouble, PCDRDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRParaErrTailDetail in FInitialised) then Result:=InitCDRParaErrTailDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRTail)<=0 then begin
      //            Proxy Part    m      i
      Setlength(FCDRTail,2  *  3  *(FM+2)*(FI+2));
    end;
    if (Init_CDRParaErrTailProxy in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PmmiiDouble:=4*(FM+2)*(FI+2);
      CopyMemory(@FCDRTail[(FM+2)*(FI+2)],@FCDRTail[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the CDR tail parameter error'}
      {$REGION 'unweighted estimates'}
      // Proxy=0, part=1
      PCDRDetailDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PmmiiDouble:=(FM+2)*(FI+2);
      ZeroMemory(@FCDRTail[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
      PTotalmDouble:=PmmiiDouble;
      Inc(PTotalmDouble,FI+1);
      for m1:=0 to FM do begin
        for i1 := 0 to FI do begin
          // the unweighted estimate of claim property m and accident period i
          for i2 := 0 to i1-1 do begin
            FCDRTail[PTotalmDouble]:=FCDRTail[PTotalmDouble]+FCDRTailDetail[PCDRDetailDouble];
            Inc(PCDRDetailDouble,1);
          end;
          FCDRTail[PmmiiDouble]:=FCDRTailDetail[PCDRDetailDouble];
          Inc(PmmiiDouble,1);
          FCDRTail[PTotalmDouble]:=FCDRTail[PTotalmDouble]+FCDRTailDetail[PCDRDetailDouble];
          Inc(PCDRDetailDouble,1);
          for i2 := i1+1 to FI do begin
            FCDRTail[PTotalmDouble]:=FCDRTail[PTotalmDouble]+FCDRTailDetail[PCDRDetailDouble];
            Inc(PCDRDetailDouble,1);
          end;
        end;
        // next m (m1=m2=m)
        Inc(PCDRDetailDouble,(FM+1)*(FI+1)*(FI+1));
        Inc(PTotalmDouble,FI+2);
        Inc(PmmiiDouble,1);
      end;
      {$ENDREGION}
      {$REGION 'weighted estimates'}
      // Proxy=0, part=1
      PCDRDetailDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PTotalweightedDouble:=((FM+2)+(FM+1))*(FI+2)+FI+1;
      for m1 := 0 to FM do begin
        for m2 := 0 to FM do begin
          Palpha1Double:=m1*(FI+1);
          for i1 := 0 to FI do begin
            Palpha2Double:=m2*(FI+1);
            PweightediDouble:=((FM+2)+(FM+1))*(FI+2)+i1;
            for i2 := 0 to FI do begin
              if i1=i2 then begin
                // weighted estimate for accident period i1=i2
                FCDRTail[PweightediDouble]:=FCDRTail[PweightediDouble]
                                            +FUsedalpha^[Palpha1Double]
                                            *FUsedalpha^[Palpha2Double]
                                            *FCDRTailDetail[PCDRDetailDouble];
              end;
              // total (over all accident periods) weighted estimate
              FCDRTail[PTotalweightedDouble]:=FCDRTail[PTotalweightedDouble]
                                              +FUsedalpha^[Palpha1Double]
                                              *FUsedalpha^[Palpha2Double]
                                              *FCDRTailDetail[PCDRDetailDouble];
              Inc(PCDRDetailDouble,1);
              Inc(Palpha2Double,1);
            end;
            Inc(Palpha1Double,1);
          end;
        end;
      end;
      {$ENDREGION}
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRParaErrTail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTail + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRParaErrTailDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhostarDouble, Prhostar1Double, Prhostar2Double, Prhostar12Double,
      PCoorF1Double, PCoorF2Double, PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoMSEP in FInitialised) then
    Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoCDR in FInitialised) then
    Result:=InitrhoCDR();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRTailDetail)<=0 then begin
      //                     Proxy Part    m1     m2     i1     i2
      SetLength(FCDRTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_CDRParaErrTailProxyDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PProxyDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FCDRTailDetail[PDestDouble],@FCDRTailDetail[PProxyDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the CDR tail parameter error'}
      // store the results within FCDRTailDetail(0,1)
      PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FCDRTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      if calc_ReduceMemSize in FCalculationFlags then begin
        {$REGION 'calculate rhos each time'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error non proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2), FJ-1) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        // here we have to take all old accident periods into account, too
                        // therefore, the condition "h1 >= FI - j" has to be ignored
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            // here we have to take all old accident periods into account, too
                            // therefore, the condition "h2 >= FI - j" has to be ignored
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                          end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else begin
        {$REGION 'take rhos from memory'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error non proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        // here we have to take all old accident periods into account, too
                        // therefore, the condition "h1 >= FI - j" has to be ignored
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            // here we have to take all old accident periods into account, too
                            // therefore, the condition "h2 >= FI - j" has to be ignored
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                          end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRParaErrTailDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRParaErrTailProxy: string;
var StartTime, EndTime: Int64;
    PmmiiDouble, PTotalmDouble, PweightediDouble, PTotalweightedDouble, PCDRDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailProxy + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_CDRParaErrTailProxyDetail in FInitialised) then Result:=InitCDRParaErrTailProxyDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRTail)<=0 then begin
      //                Proxy Part    m      i
      Setlength(FCDRTail,2  *  3  *(FM+2)*(FI+2));
    end;
    if (Init_CDRParaErrTail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PmmiiDouble:=4*(FM+2)*(FI+2);
      CopyMemory(@FCDRTail[PmmiiDouble],@FCDRTail[(FM+2)*(FI+2)],(FM+2)*(FI+2)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the CDR tail parameter error proxy'}
      {$REGION 'unweighted estimates'}
      // Proxy=1, part=1
      PCDRDetailDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PmmiiDouble:=4*(FM+2)*(FI+2);
      ZeroMemory(@FCDRTail[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
      PTotalmDouble:=PmmiiDouble;
      Inc(PTotalmDouble,FI+1);
      for m1:=0 to FM do begin
        for i1 := 0 to FI do begin
          // the unweighted estimate of claim property m and accident period i
          for i2 := 0 to i1-1 do begin
            FCDRTail[PTotalmDouble]:=FCDRTail[PTotalmDouble]+FCDRTailDetail[PCDRDetailDouble];
            Inc(PCDRDetailDouble,1);
          end;
          FCDRTail[PmmiiDouble]:=FCDRTailDetail[PCDRDetailDouble];
          Inc(PmmiiDouble,1);
          FCDRTail[PTotalmDouble]:=FCDRTail[PTotalmDouble]+FCDRTailDetail[PCDRDetailDouble];
          Inc(PCDRDetailDouble,1);
          for i2 := i1+1 to FI do begin
            FCDRTail[PTotalmDouble]:=FCDRTail[PTotalmDouble]+FCDRTailDetail[PCDRDetailDouble];
            Inc(PCDRDetailDouble,1);
          end;
        end;
        // next m (m1=m2=m)
        Inc(PCDRDetailDouble,(FM+1)*(FI+1)*(FI+1));
        Inc(PTotalmDouble,FI+2);
        Inc(PmmiiDouble,1);
      end;
      {$ENDREGION}
      {$REGION 'weighted estimates'}
      // Proxy=1, part=1
      PCDRDetailDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PTotalweightedDouble:=(4*(FM+2)+(FM+1))*(FI+2)+FI+1;
      for m1 := 0 to FM do begin
        for m2 := 0 to FM do begin
          Palpha1Double:=m1*(FI+1);
          for i1 := 0 to FI do begin
            Palpha2Double:=m2*(FI+1);
            PweightediDouble:=(4*(FM+2)+(FM+1))*(FI+2)+i1;
            for i2 := 0 to FI do begin
              if i1=i2 then begin
                // weighted estimate for accident period i1=i2
                FCDRTail[PweightediDouble]:=FCDRTail[PweightediDouble]
                                            +FUsedalpha^[Palpha1Double]
                                            *FUsedalpha^[Palpha2Double]
                                            *FCDRTailDetail[PCDRDetailDouble];
              end;
              // total (over all accident periods) weighted estimate
              FCDRTail[PTotalweightedDouble]:=FCDRTail[PTotalweightedDouble]
                                              +FUsedalpha^[Palpha1Double]
                                              *FUsedalpha^[Palpha2Double]
                                              *FCDRTailDetail[PCDRDetailDouble];
              Inc(PCDRDetailDouble,1);
              Inc(Palpha2Double,1);
            end;
            Inc(Palpha1Double,1);
          end;
        end;
      end;
      {$ENDREGION}
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRParaErrTailProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailProxy + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRProcVar: string;
var StartTime, EndTime: Int64;
    PmmiiDouble, PTotalmDouble, PweightediDouble, PTotalweightedDouble, PCDRDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  if calc_ReduceMemSize in FCalculationFlags then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDR + sLineBreak + rsLSRM_ProxyOnly;
    Exit;
  end;
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVar + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVar + sLineBreak + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_CDRProcVarDetail in FInitialised) then Result:=InitCDRProcVarDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDR)<=0 then begin
      //            Proxy Part    m      i
      Setlength(FCDR,2  *  3  *(FM+2)*(FI+2));
    end;
    {$REGION 'unweighted estimates'}
    // Proxy=0, part=0
    PCDRDetailDouble:=0;
    PmmiiDouble:=0;
    ZeroMemory(@FCDR[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
    PTotalmDouble:=PmmiiDouble;
    Inc(PTotalmDouble,FI+1);
    for m1:=0 to FM do begin
      for i1 := 0 to FI do begin
        // the unweighted estimate of claim property m and accident period i
        for i2 := 0 to i1-1 do begin
          FCDR[PTotalmDouble]:=FCDR[PTotalmDouble]+FCDRDetail[PCDRDetailDouble];
          Inc(PCDRDetailDouble,1);
        end;
        FCDR[PmmiiDouble]:=FCDRDetail[PCDRDetailDouble];
        Inc(PmmiiDouble,1);
        FCDR[PTotalmDouble]:=FCDR[PTotalmDouble]+FCDRDetail[PCDRDetailDouble];
        Inc(PCDRDetailDouble,1);
        for i2 := i1+1 to FI do begin
          FCDR[PTotalmDouble]:=FCDR[PTotalmDouble]+FCDRDetail[PCDRDetailDouble];
          Inc(PCDRDetailDouble,1);
        end;
      end;
      // next m (m1=m2=m)
      Inc(PCDRDetailDouble,(FM+1)*(FI+1)*(FI+1));
      Inc(PTotalmDouble,FI+2);
      Inc(PmmiiDouble,1);
    end;
    {$ENDREGION}
    {$REGION 'weighted estimates'}
    // Proxy=0, part=0
    PCDRDetailDouble:=0;
    PTotalweightedDouble:=(FM+1)*(FI+2)+FI+1;
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
          Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          PweightediDouble:=(FM+1)*(FI+2)+i1;
          for i2 := 0 to FI do begin
            if i1=i2 then begin
              // weighted estimate for accident period i1=i2
              FCDR[PweightediDouble]:=FCDR[PweightediDouble]
                                      +FUsedalpha^[Palpha1Double]
                                      *FUsedalpha^[Palpha2Double]
                                      *FCDRDetail[PCDRDetailDouble];
            end;
            // total (over all accident periods) weighted estimate
            FCDR[PTotalweightedDouble]:=FCDR[PTotalweightedDouble]
                                        +FUsedalpha^[Palpha1Double]
                                        *FUsedalpha^[Palpha2Double]
                                        *FCDRDetail[PCDRDetailDouble];
            Inc(PCDRDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRProcVar,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVar + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVar + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRProcVarDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PSxS0Double, PSxSrhobarDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  if calc_ReduceMemSize in FCalculationFlags then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDR + sLineBreak + rsLSRM_ProxyOnly;
    Exit;
  end;
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_SxSCDRProcVar in FInitialised) then Result:=InitSxSCDRProcVar();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRDetail)<=0 then begin
      //                  Proxy Part    m1     m2     i1     i2
      SetLength(FCDRDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    // process variance non proxy
    // store the results within FCDRDetail(0,0)
    PDestDouble:=0;
    ZeroMemory(@FCDRDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
    For m1 := 0 To FM do begin
      For m2 := 0 To FM do begin
        For i1 := 0 To FI do begin
          For i2 := 0 To FI do begin
            For k1 := FI - i1 To FJ - 1 do begin
              PSxS0Double:=((((m1*(FI+1)+i1)*(FJ+1+FP)+k1+1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+FI-i2+1;
              PSxSrhobarDouble:=PSxS0Double;
              Inc(PSxSrhobarDouble,5*(FM+1)*(FI+1)*(FJ+1+FP)*(FM+1)*(FI+1)*(FJ+1+FP));
              For k2 := FI - i2 To FJ - 1 do begin
                FCDRDetail[PDestDouble]:=FCDRDetail[PDestDouble]
                                        +FSxS[PSXSrhobarDouble]
                                        -FSxS[PSXS0Double];
                Inc(PSxS0Double,1);
                Inc(PSxSrhobarDouble,1);
              end;
            end;
            Inc(PDestDouble,1);
          end;
        end;
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRProcVarDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarDetail + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRProcVarProxy: string;
var StartTime, EndTime: Int64;
    PmmiiDouble, PTotalmDouble, PweightediDouble, PTotalweightedDouble, PCDRDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarProxy + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_CDRProcVarProxyDetail in FInitialised) then Result:=InitCDRProcVarProxyDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDR)<=0 then begin
      //            Proxy Part    m      i
      Setlength(FCDR,2  *  3  *(FM+2)*(FI+2));
    end;
    {$REGION 'unweighted estimates'}
    // Proxy=1, part=0
    PCDRDetailDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PmmiiDouble:=3*(FM+2)*(FI+2);
    ZeroMemory(@FCDR[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
    PTotalmDouble:=PmmiiDouble;
    Inc(PTotalmDouble,FI+1);
    for m1:=0 to FM do begin
      for i1 := 0 to FI do begin
        // the unweighted estimate of claim property m and accident period i
        for i2 := 0 to i1-1 do begin
          FCDR[PTotalmDouble]:=FCDR[PTotalmDouble]+FCDRDetail[PCDRDetailDouble];
          Inc(PCDRDetailDouble,1);
        end;
        FCDR[PmmiiDouble]:=FCDRDetail[PCDRDetailDouble];
        Inc(PmmiiDouble,1);
        FCDR[PTotalmDouble]:=FCDR[PTotalmDouble]+FCDRDetail[PCDRDetailDouble];
        Inc(PCDRDetailDouble,1);
        for i2 := i1+1 to FI do begin
          FCDR[PTotalmDouble]:=FCDR[PTotalmDouble]+FCDRDetail[PCDRDetailDouble];
          Inc(PCDRDetailDouble,1);
        end;
      end;
      // next m (m1=m2=m)
      Inc(PCDRDetailDouble,(FM+1)*(FI+1)*(FI+1));
      Inc(PTotalmDouble,FI+2);
      INc(PmmiiDouble,1);
    end;
    {$ENDREGION}
    {$REGION 'weighted estimates'}
    // Proxy=1, part=0
    PCDRDetailDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PTotalweightedDouble:=(3*(FM+2)+(FM+1))*(FI+2)+FI+1;
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          PweightediDouble:=(3*(FM+2)+(FM+1))*(FI+2)+i1;
          for i2 := 0 to FI do begin
            if i1=i2 then begin
              // weighted estimate for accident period i1=i2
              FCDR[PweightediDouble]:=FCDR[PweightediDouble]
                                      +FUsedalpha^[Palpha1Double]
                                      *FUsedalpha^[Palpha2Double]
                                      *FCDRDetail[PCDRDetailDouble];
            end;
            // total (over all accident periods) weighted estimate
            FCDR[PTotalweightedDouble]:=FCDR[PTotalweightedDouble]
                                        +FUsedalpha^[Palpha1Double]
                                        *FUsedalpha^[Palpha2Double]
                                        *FCDRDetail[PCDRDetailDouble];
            Inc(PCDRDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRProcVarProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarproxy + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRProcVarTail: string;
var StartTime, EndTime: Int64;
    PmmiiDouble, PTotalmDouble, PweightediDouble, PTotalweightedDouble, PCDRDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTail + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_CDRProcVarTailDetail in FInitialised) then Result:=InitCDRProcVarTailDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRTail)<=0 then begin
      //               Proxy Part    m      i
      Setlength(FCDRTail,2  *  3  *(FM+2)*(FI+2));
    end;
    if (Init_CDRProcVarTailProxy in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PmmiiDouble:=3*(FM+2)*(FI+2);
      CopyMemory(@FCDRTail[0],@FCDRTail[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the CDR tail process variance'}
      {$REGION 'unweighted estimates'}
      // Proxy=0, part=0
      PCDRDetailDouble:=0;
      PmmiiDouble:=0;
      ZeroMemory(@FCDRTail[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
      PTotalmDouble:=PmmiiDouble;
      Inc(PTotalmDouble,FI+1);
      for m1:=0 to FM do begin
        for i1 := 0 to FI do begin
          // the unweighted estimate of claim property m and accident period i
          for i2 := 0 to i1-1 do begin
            FCDRTail[PTotalmDouble]:=FCDRTail[PTotalmDouble]+FCDRTailDetail[PCDRDetailDouble];
            Inc(PCDRDetailDouble,1);
          end;
          FCDRTail[PmmiiDouble]:=FCDRTailDetail[PCDRDetailDouble];
          Inc(PmmiiDouble,1);
          FCDRTail[PTotalmDouble]:=FCDRTail[PTotalmDouble]+FCDRTailDetail[PCDRDetailDouble];
          Inc(PCDRDetailDouble,1);
          for i2 := i1+1 to FI do begin
            FCDRTail[PTotalmDouble]:=FCDRTail[PTotalmDouble]+FCDRTailDetail[PCDRDetailDouble];
            Inc(PCDRDetailDouble,1);
          end;
        end;
        // next m (m1=m2=m)
        Inc(PCDRDetailDouble,(FM+1)*(FI+1)*(FI+1));
        Inc(PTotalmDouble,FI+2);
        INc(PmmiiDouble,1);
      end;
      {$ENDREGION}
      {$REGION 'weighted estimates'}
      // Proxy=0, part=0
      PCDRDetailDouble:=0;
      PTotalweightedDouble:=(FM+1)*(FI+2)+FI+1;
      for m1 := 0 to FM do begin
        for m2 := 0 to FM do begin
          Palpha1Double:=m1*(FI+1);
          for i1 := 0 to FI do begin
            Palpha2Double:=m2*(FI+1);
            PweightediDouble:=(FM+1)*(FI+2)+i1;
            for i2 := 0 to FI do begin
              if i1=i2 then begin
                // weighted estimate for accident period i1=i2
                FCDRTail[PweightediDouble]:=FCDRTail[PweightediDouble]
                                            +FUsedalpha^[Palpha1Double]
                                            *FUsedalpha^[Palpha2Double]
                                            *FCDRTailDetail[PCDRDetailDouble];
              end;
              // total (over all accident periods) weighted estimate
              FCDRTail[PTotalweightedDouble]:=FCDRTail[PTotalweightedDouble]
                                              +FUsedalpha^[Palpha1Double]
                                              *FUsedalpha^[Palpha2Double]
                                              *FCDRTailDetail[PCDRDetailDouble];
              Inc(PCDRDetailDouble,1);
              Inc(Palpha2Double,1);
            end;
            Inc(Palpha1Double,1);
          end;
        end;
      end;
      {$ENDREGION}
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRProcVarTail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTail + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRProcVarTailDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhobarDouble, PCoorF1Double, PCoorF2Double,
    PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoCDR in FInitialised) then
    Result:=InitrhoCDR();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  // if calc_UseMAEPifTailCDRisLarger is set the we may need tail MSEP
  if calc_UseMSEPifTailCDRisLarger in FCalculationFlags then begin
    If not(Init_MSEPProcVarTailDetail in FInitialised) then Result:=InitMSEPProcVarTailDetail();
    if Result<>'' then Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRTailDetail)<=0 then begin
      //                     Proxy Part    m1     m2     i1     i2
      SetLength(FCDRTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_CDRProcVarTailProxyDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=0;
      PProxyDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FCDRTailDetail[PDestDouble],@FCDRTailDetail[PProxyDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the CDR tail process variance'}
      // store the results within FCDRTailDetail(0,0)
      PDestDouble:=0;
      ZeroMemory(@FCDRTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      if calc_ReduceMemSize in FCalculationFlags then begin
        {$REGION 'calculate rhos each time'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error non proxy: k1 tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k2 tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1 and k2 tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1, k2 and j tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        // here we have to take all old accident periods into account, too
                        // therefore, the condition "h1 >= FI - j" has to be ignored
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            // here we have to take all old accident periods into account, too
                            // therefore, the condition "h2 >= FI - j" has to be ignored
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                            end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                if (calc_UseMSEPifTailCDRisLarger in FCalculationFlags)
                    and (FCDRTailDetail[PDestDouble]>FMSEPTailDetail[PDestDouble]) then
                  FCDRTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble];
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else begin
        {$REGION 'take rhos from memory'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'process variance non proxy: k1 tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance non proxy: k2 tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance non proxy: k1 and k2 tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance non proxy: k1, k2 and j tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        // here we have to take all old accident periods into account, too
                        // therefore, the condition "h1 >= FI - j" has to be ignored
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            // here we have to take all old accident periods into account, too
                            // therefore, the condition "h2 >= FI - j" has to be ignored
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                          end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                if (calc_UseMSEPifTailCDRisLarger in FCalculationFlags)
                    and (FCDRTailDetail[PDestDouble]>FMSEPTailDetail[PDestDouble]) then
                  FCDRTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble];
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRProcVarTailDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRProcVarTailProxy: string;
var StartTime, EndTime: Int64;
    PmmiiDouble, PTotalmDouble, PweightediDouble, PTotalweightedDouble, PCDRDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailProxy + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_CDRProcVarTailProxyDetail in FInitialised) then Result:=InitCDRProcVarTailProxyDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRTail)<=0 then begin
      //               Proxy Part    m      i
      Setlength(FCDRTail,2  *  3  *(FM+2)*(FI+2));
    end;
    if (Init_CDRProcVarTail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PmmiiDouble:=3*(FM+2)*(FI+2);
      CopyMemory(@FCDRTail[PmmiiDouble],@FCDRTail[0],(FM+2)*(FI+2)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the CDR tail process variance proxy'}
      {$REGION 'unweighted estimates'}
      // Proxy=1, part=0
      PCDRDetailDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PmmiiDouble:=3*(FM+2)*(FI+2);
      ZeroMemory(@FCDRTail[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
      PTotalmDouble:=PmmiiDouble;
      Inc(PTotalmDouble,FI+1);
      for m1:=0 to FM do begin
        for i1 := 0 to FI do begin
          // the unweighted estimate of claim property m and accident period i
          for i2 := 0 to i1-1 do begin
            FCDRTail[PTotalmDouble]:=FCDRTail[PTotalmDouble]+FCDRTailDetail[PCDRDetailDouble];
            Inc(PCDRDetailDouble,1);
          end;
          FCDRTail[PmmiiDouble]:=FCDRTailDetail[PCDRDetailDouble];
          Inc(PmmiiDouble,1);
          FCDRTail[PTotalmDouble]:=FCDRTail[PTotalmDouble]+FCDRTailDetail[PCDRDetailDouble];
          Inc(PCDRDetailDouble,1);
          for i2 := i1+1 to FI do begin
            FCDRTail[PTotalmDouble]:=FCDRTail[PTotalmDouble]+FCDRTailDetail[PCDRDetailDouble];
            Inc(PCDRDetailDouble,1);
          end;
        end;
        // next m (m1=m2=m)
        Inc(PCDRDetailDouble,(FM+1)*(FI+1)*(FI+1));
        Inc(PTotalmDouble,FI+2);
        INc(PmmiiDouble,1);
      end;
      {$ENDREGION}
      {$REGION 'weighted estimates'}
      // Proxy=1, part=0
      PCDRDetailDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PTotalweightedDouble:=(3*(FM+2)+(FM+1))*(FI+2)+FI+1;
      for m1 := 0 to FM do begin
        for m2 := 0 to FM do begin
          Palpha1Double:=m1*(FI+1);
          for i1 := 0 to FI do begin
            Palpha2Double:=m2*(FI+1);
            PweightediDouble:=(3*(FM+2)+(FM+1))*(FI+2)+i1;
            for i2 := 0 to FI do begin
              if i1=i2 then begin
                // weighted estimate for accident period i1=i2
                FCDRTail[PweightediDouble]:=FCDRTail[PweightediDouble]
                                            +FUsedalpha^[Palpha1Double]
                                            *FUsedalpha^[Palpha2Double]
                                            *FCDRTailDetail[PCDRDetailDouble];
              end;
              // total (over all accident periods) weighted estimate
              FCDRTail[PTotalweightedDouble]:=FCDRTail[PTotalweightedDouble]
                                              +FUsedalpha^[Palpha1Double]
                                              *FUsedalpha^[Palpha2Double]
                                              *FCDRTailDetail[PCDRDetailDouble];
              Inc(PCDRDetailDouble,1);
              Inc(Palpha2Double,1);
            end;
            Inc(Palpha1Double,1);
          end;
        end;
      end;
      {$ENDREGION}
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRProcVarTailProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailProxy + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRProcVarProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhobarDouble, PCoorF1Double, PCoorF2Double: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoCDR in FInitialised) then
    Result:=InitrhoCDR();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRDetail)<=0 then begin
      //                  Proxy Part    m1     m2     i1     i2
      SetLength(FCDRDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    // calculate the details for the CDR process variance proxy
    // store the results within FCDRDetail(1,0)
    PDestDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    ZeroMemory(@FCDRDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
    if calc_ReduceMemSize in FCalculationFlags then begin
      {$REGION 'calculate rhos each time'}
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                    For ch1 := 0 To FihDependency[i1,FI] do begin
                      h1 := FihDependency[i1,ch1];
                      If h1 >= FI - j Then begin
                        For ch2 := 0 To FihDependency[i2,FI] do begin
                          h2 := FihDependency[i2,ch2];
                          If h2 >= FI - j Then begin
                            For cl1 := 0 To FmlDependency[m1,FM] do begin
                              l1 := FmlDependency[m1,cl1];
                              For cl2 := 0 To FmlDependency[m2,FM] do begin
                                l2 := FmlDependency[m2,cl2];
                                PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                Rho:=Getrho(5,l1,l2,h1,h2,j);
                                TmpValue:=Rho
                                          *FR1[PR11Double]
                                          *FR1[PR12Double]
                                          *FCoorF[PCoorF1Double]
                                          *FCoorF[PCoorF2Double];
                                // diversification only if rho is negative
                                if (calc_VirtualProcVarDiversification in FCalculationFlags)
                                        or (Rho<=0) or (TmpValue>=0) then
                                  FCDRDetail[PDestDouble] := FCDRDetail[PDestDouble] + TmpValue;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end else begin
      {$REGION 'take rhos from memory'}
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                    For ch1 := 0 To FihDependency[i1,FI] do begin
                      h1 := FihDependency[i1,ch1];
                      If h1 >= FI - j Then begin
                        For ch2 := 0 To FihDependency[i2,FI] do begin
                          h2 := FihDependency[i2,ch2];
                          If h2 >= FI - j Then begin
                            For cl1 := 0 To FmlDependency[m1,FM] do begin
                              l1 := FmlDependency[m1,cl1];
                              For cl2 := 0 To FmlDependency[m2,FM] do begin
                                l2 := FmlDependency[m2,cl2];
                                PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                Rho:=Frho[PrhobarDouble];
                                TmpValue:=Rho
                                          *FR1[PR11Double]
                                          *FR1[PR12Double]
                                          *FCoorF[PCoorF1Double]
                                          *FCoorF[PCoorF2Double];
                                // diversification only if rho is negative
                                if (calc_VirtualProcVarDiversification in FCalculationFlags)
                                        or (Rho<=0) or (TmpValue>=0) then
                                  FCDRDetail[PDestDouble] := FCDRDetail[PDestDouble] + TmpValue;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRProcVarProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRProcVarTailProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhobarDouble, PCoorF1Double, PCoorF2Double,
      PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoCDR in FInitialised) then
    Result:=InitrhoCDR();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  // if calc_UseMAEPifTailCDRisLarger is set the we may need tail MSEP
  if calc_UseMSEPifTailCDRisLarger in FCalculationFlags then begin
    If not(Init_MSEPProcVarTailDetail in FInitialised) then Result:=InitMSEPProcVarTailProxyDetail();
    if Result<>'' then Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRTailDetail)<=0 then begin
      //                     Proxy Part    m1     m2     i1     i2
      SetLength(FCDRTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_CDRProcVarTailDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=0;
      PProxyDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FCDRTailDetail[PProxyDouble],@FCDRTailDetail[PDestDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the CDR tail process variance proxy'}
      // store the results within FCDRTailDetail(1,0)
      PDestDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FCDRTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      if calc_ReduceMemSize in FCalculationFlags then begin
        {$REGION 'calculate rhos each time'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'process variance proxy: k1 tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance proxy: k2 tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance proxy: k1 and k2 tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance proxy: k1, k2 and j tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        // here we have to take all old accident periods into account, too
                        // therefore, the condition "h1 >= FI - j" has to be ignored
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            // here we have to take all old accident periods into account, too
                            // therefore, the condition "h2 >= FI - j" has to be ignored
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                          end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                if (calc_UseMSEPifTailCDRisLarger in FCalculationFlags)
                    and (FCDRTailDetail[PDestDouble]>FMSEPTailDetail[PDestDouble]) then
                  FCDRTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble];
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else begin
        {$REGION 'take rhos from memory'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'process variance proxy: k1 tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance proxy: k2 tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance proxy: k1 and k2 tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance proxy: k1, k2 and j tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        // here we have to take all old accident periods into account, too
                        // therefore, the condition "h1 >= FI - j" has to be ignored
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            // here we have to take all old accident periods into account, too
                            // therefore, the condition "h2 >= FI - j" has to be ignored
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                          end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                if (calc_UseMSEPifTailCDRisLarger in FCalculationFlags)
                    and (FCDRTailDetail[PDestDouble]>FMSEPTailDetail[PDestDouble]) then
                  FCDRTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble];
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRProcVarTailProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRProxy: string;
var StartTime, EndTime: Int64;
    PProcVarDouble, PParaErrDouble, PTotalDouble: IndexDouble;
    m, i: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProxy + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRParaErrProxy in FInitialised) then Result:=InitCDRParaErrProxy();
  if Result<>'' then Exit;
  If not(Init_CDRProcVarProxy in FInitialised) then Result:=InitCDRProcVarProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // calculate the sum of the process variance and the parameter error (with proxy)
    // store the result into FCDR(1,2)
    PProcVarDouble:=3*(FM+2)*(FI+2);
    PParaErrDouble:=PProcVarDouble;
    Inc(PParaErrDouble,(FM+2)*(FI+2));
    PTotalDouble:=PParaErrDouble;
    Inc(PTotalDouble,(FM+2)*(FI+2));
    for m := 0 to FM+1 do begin
      for i := 0 to FI+1 do begin
        FCDR[PTotalDouble]:=FCDR[PProcVarDouble]+FCDR[PParaErrDouble];
        Inc(PTotalDouble,1);
        Inc(PProcVarDouble,1);
        Inc(PParaErrDouble,1);
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProxy + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRTail: string;
var StartTime, EndTime: Int64;
    PProcVarDouble, PParaErrDouble, PTotalDouble: IndexDouble;
    m, i: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRTail + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRParaErrTail in FInitialised) then Result:=InitCDRParaErrTail();
  if Result<>'' then Exit;
  If not(Init_CDRProcVarTail in FInitialised) then Result:=InitCDRProcVarTail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // calculate the sum of the tail process variance and the tail parameter error (without proxy)
    // store the result into FCDRTail(0,2)
    PProcVarDouble:=0;
    PParaErrDouble:=PProcVarDouble;
    Inc(PParaErrDouble,(FM+2)*(FI+2));
    PTotalDouble:=PParaErrDouble;
    Inc(PTotalDouble,(FM+2)*(FI+2));
    for m := 0 to FM+1 do begin
      for i := 0 to FI+1 do begin
        FCDRTail[PTotalDouble]:=FCDRTail[PProcVarDouble]+FCDRTail[PParaErrDouble];
        Inc(PTotalDouble,1);
        Inc(PProcVarDouble,1);
        Inc(PParaErrDouble,1);
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRTail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRTail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRTail + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCDRTailProxy: string;
var StartTime, EndTime: Int64;
    PProcVarDouble, PParaErrDouble, PTotalDouble: IndexDouble;
    m, i: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRTailProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CDRParaErrTailProxy in FInitialised) then Result:=InitCDRParaErrTailProxy();
  if Result<>'' then Exit;
  If not(Init_CDRProcVarTailProxy in FInitialised) then Result:=InitCDRProcVarTailProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // calculate the sum of the tail process variance and the tail parameter error (with proxy)
    // store the result into FCDRTail(1,2)
    PProcVarDouble:=3*(FM+2)*(FI+2);
    PParaErrDouble:=PProcVarDouble;
    Inc(PParaErrDouble,(FM+2)*(FI+2));
    PTotalDouble:=PParaErrDouble;
    Inc(PTotalDouble,(FM+2)*(FI+2));
    for m := 0 to FM+1 do begin
      for i := 0 to FI+1 do begin
        FCDRTail[PTotalDouble]:=FCDRTail[PProcVarDouble]+FCDRTail[PParaErrDouble];
        Inc(PTotalDouble,1);
        Inc(PProcVarDouble,1);
        Inc(PParaErrDouble,1);
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRTailProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRTailProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRTailProxy + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitCoorF: string;
var m, k, i, cl, l, ch, h, j: Integer;
    PDestDouble, PElhjDouble, PTmp1Double: IndexDouble;
    StartTime, EndTime: Int64;
    Ta:T1Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCoorF + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_Gamma1 in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCoorF + sLineBreak + rsLSRM_NotInitialisedGamma1;
    Exit;
  end;
  if ([Init_f, Init_Defaultf]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitDefaultsigma + sLineBreak + rsLSRM_NotInitialisedf;
    Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCoorF)<=0 then begin
      //                 m       i    k        l     h       j
      Setlength(FCoorF,(FM+1)*(FI+1)*(FJ+FP)*(FM+1)*(FI+1)*(FJ+1+FP));
    end;
    // initialise temporary arry
    SetLength(TA,(FM+1)*(FI+1)*(FJ+1+FP));
    For m := 0 To FM do begin
      For i := 0 To FI do begin
        For k := 0 To FJ - 1 + FP do begin
          PDestDouble:=((((m*(FI+1)+i)*(FJ+FP)+k)*(FM+1)+m)*(FI+1)+i)*(FJ+1+FP)+k+1;
          // for n=h+j>i+k we have (F^n,m_i,k x)^l_h,j = 1 iff l=m,h=i, j=k+1
          FCoorF[PDestDouble]:=1;
          for cl:=0 to FmlDependency[m,FM] do begin
            l:=FmlDependency[m,cl];
            for ch:=0 to FihDependency[i,FI] do begin
              h:=FihDependency[i,ch];
              PElhjDouble:=(l*(FI+1)+h)*(FJ+1+FP);
              For j := 0 To Min(FJ - 1 + FP, i + k - h) do begin
                // set the (l, h, j)-th coordinate of the temporary array to 1, all other coordinates to zero
                ZeroMemory(@TA[0],(FM+1)*(FI+1)*(FJ+1+FP)*SizeOf(Double));
                Ta[PElhjDouble]:=1;
                // apply the operator F
                Result:=OP_Fn(h+j,i+k,@Ta[0]);
                if Result<>'' then Exit;
                // get the coordinate of F
                PDestDouble:=((((m*(FI+1)+i)*(FJ+FP)+k)*(FM+1)+l)*(FI+1)+h)*(FJ+1+FP)+j;
                PTmp1Double:=(m*(FI+1)+i)*(FJ+1+FP)+k+1;
                FCoorF[PDestDouble]:=TA[PTmp1Double];
                Inc(PElhjDouble,1);
              end;
            end;
          end;
        end;
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CoorF,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCoorF + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCoorF + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitDefaultalpha: String;
var StartTime, EndTime: Int64;
    m, i:Integer;
    PDestDouble:IndexDouble;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitDefaultalpha + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise default alpha, if necessary
    if Length(FDefaultalpha)<=0 then begin
      //                         m      i
      SetLength(FDefaultalpha,(FM+1)*(FI+1));
      // set all values to 1
      PDestDouble:=0;
      for m:=0 to FM do begin
        for i:=0 to FI do begin
          FDefaultalpha[PDestDouble]:=1;
          Inc(PDestDouble);
        end;
      end;
      QueryPerformanceCounter(EndTime);
      SetInitFlags(Init_Defaultalpha,EndTime-StartTime);
    end;
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitDefaultalpha + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitDefaultalpha + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitDefaultf: String;
var m, k, i: Integer;
    PDestDouble, PwDouble, PSDouble, PR1Double: IndexDouble;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitDefaultf + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_w, Init_Defaultw]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitDefaultf + sLineBreak + rsLSRM_NotInitialisedw;
    Exit;
  end;
  If not(Init_R1 in FInitialised) then Result:=InitR1();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the array or clear the corresponding memory
    if Length(FDefaultf)<=0 then begin
      //                    m      k
      SetLength(FDefaultf,(FM+1)*(FJ+FK));
    end else begin
      ZeroMemory(@FDefaultf[0],(FM+1)*(FJ+FK)*SizeOf(Double));
    end;
    {$REGION 'calculate default f, for future periods (k>=FJ) we take zero'}
    PDestDouble:=0;
    for m:=0 to FM do begin
      For k := 0 To FJ - 1 do begin
        PSDouble:=m*(FI+1)*(FJ+1+FK)+k+1;
        // take n=1
        PwDouble:=((FM+1)*(FI+1)+m*(FI+1))*(FJ+FP)+k;
        PR1Double:=m*(FI+1)*(FJ+FK)+k;
        For i := 0 To FI - k - 1 do begin
          if FR1[PR1Double]<>0 then
            FDefaultf[PDestDouble]:=FDefaultF[PDestDouble]+FUsedw^[PwDouble]*FS[PSDouble]/FR1[PR1Double];
          Inc(PSDouble,FJ+1+FK);
          Inc(PwDouble,FJ+FP);
          Inc(PR1Double,FJ+FK);
        end;
        Inc(PDestDouble,1);
      end;
      Inc(PDestDouble,FK);
    end;
    {$ENDREGION}
    {$REGION 'if manual f have been specified we take those as default for future periods (k>=FJ)'}
    if (FK>0) and (Init_f in FInitialised) then begin
      // take manual values as default in the tail
      for m := 0 to FM do begin
        for k := FJ to FJ+FK-1 do begin
          fDefaultf[m*(FJ+FK)+k]:=Ff[m*(FJ+FK)+k];
        end;
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_Defaultf,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitDefaultf + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitDefaultf + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitDefaultsigma: String;
var m1, m2, k, i, NrExposureNeQ0: Integer;
    Psigma1, PSigma2, PDestDouble, Pw1Double, Pw2Double, PS1Double, PS2Double, Pf1Double, Pf2Double,
      PR11Double, PR12Double, PR2Double: IndexDouble;
    TmpSum, TmpZ1, TmpZ2, TmpZ, MaxExposure1: Double;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitDefaultsigma + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_S in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitDefaultsigma + sLineBreak + rsLSRM_NotInitialisedS;
    Exit;
  end;
  if ([Init_w, Init_Defaultw]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitDefaultsigma + sLineBreak + rsLSRM_NotInitialisedw;
    Exit;
  end;
  if ([Init_f, Init_Defaultf]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitDefaultsigma + sLineBreak + rsLSRM_NotInitialisedf;
    Exit;
  end;
  If not(Init_R1 in FInitialised) then Result:=InitR1();
  if Result<>'' then Exit;
  If not(Init_R2 in FInitialised) then Result:=InitR2();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array or clear the corresponding memory
    if Length(FDefaultsigma)<=0 then begin
      //                        m1     m2    k
      SetLength(FDefaultsigma,(FM+1)*(FM+1)*(FJ+FP));
    end else begin
      ZeroMemory(@FDefaultsigma[0],(FM+1)*(FM+1)*(FJ+FP)*SizeOf(Double));
    end;
    // calculate
    For k := 0 To FJ - 1 do begin
      {$REGION 'first all variances'}
      for m1:=0 to FM do begin
        m2:=m1;
        PDestDouble:=(m1*(FM+1)+m2)*(FJ+FP)+k;
        PF1Double:=m1*(FJ+FK)+k;
        PF2Double:=m2*(FJ+FK)+k;
        PS1Double:=m1*(FI+1)*(FJ+1+FK)+k+1;
        PS2Double:=m2*(FI+1)*(FJ+1+FK)+k+1;
        Pw1Double:=((FM+1)+m1)*(FI+1)*(FJ+FP)+k;
        Pw2Double:=((FM+1)+m2)*(FI+1)*(FJ+FP)+k;
        PR11Double:=m1*(FI+1)*(FJ+FK)+k;
        PR12Double:=m2*(FI+1)*(FJ+FK)+k;
        PR2Double:=(m1*(FM+1)+m2)*(FI+1)*(FJ+FP)+k;
        TmpSum:=0;
        TmpZ1:=0;
        TmpZ2:=0;
        NrExposureNeQ0:=0;
        MaxExposure1:=0;
        {$REGION 'without normalisation'}
        if not(calc_UseWeightsForDefaultSigma in FCalculationFlags) then begin
          {$REGION 'do not use weights'}
          For i := 0 To FI - k - 1 do begin
            if (FR1[PR12Double]<>0) and (FR1[PR11Double]<>0) and (FR2[PR2Double]<>0) then begin
              FDefaultsigma[PDestDouble]:=FDefaultsigma[PDestDouble]
                                          +FR1[PR11Double]
                                          *FR1[PR12Double]
                                          /FR2[PR2Double]
                                          *(FS[PS1Double]/FR1[PR11Double]-FUsedf^[Pf1Double])
                                          *(FS[PS2Double]/FR1[PR12Double]-Fusedf^[Pf2Double]);
              TmpSum:=TmpSum
                      +FR1[PR11Double]
                      *FR1[PR12Double]
                      /FR2[PR2Double];
              TmpZ1:=TmpZ1 + 1 - FUsedw^[Pw1Double] - FUsedw^[Pw2Double];
              TmpZ2:=TmpZ2
                    +FUsedw^[Pw1Double]
                    *FUsedw^[Pw2Double]
                    *FR2[PR2Double]
                    /FR1[PR11Double]
                    /FR1[PR12Double];
              Inc(NrExposureneq0);
            end else if FR2[PR2Double] <> 0 then begin
              TmpZ1:=TmpZ1+1;
              Inc(NrExposureneq0);
            end;
            Inc(PS1Double,FJ+1+FK);
            Inc(PS2Double,FJ+1+FK);
            Inc(Pw1Double,FJ+FP);
            Inc(Pw2Double,FJ+FP);
            Inc(PR11Double,FJ+FK);
            Inc(PR12Double,FJ+FK);
            Inc(PR2Double,FJ+FP);
          end;
          {$ENDREGION}
        end else begin
          {$REGION 'use weights'}
          For i := 0 To FI - k - 1 do begin
            if min(FUsedw^[Pw1Double],FUsedw^[Pw2Double])>0 then begin
              if (FR1[PR12Double]<>0) and (FR1[PR11Double]<>0) and (FR2[PR2Double]<>0) then begin
                FDefaultsigma[PDestDouble]:=FDefaultsigma[PDestDouble]+FUsedw^[Pw1Double]*FUsedw^[Pw2Double]
                    /FR2[PR2Double]
                    *(FS[PS1Double]/FR1[PR11Double]-FUsedf^[Pf1Double])
                    *(FS[PS2Double]/FR1[PR12Double]-Fusedf^[Pf2Double]);
                TmpSum:=TmpSum
                      +FUsedw^[Pw1Double]
                      *FUsedw^[Pw2Double]
                      /FR2[PR2Double];
                TmpZ1:=TmpZ1 + FUsedw^[Pw1Double]*FUsedw^[Pw2Double]/FR1[PR11Double]/FR1[PR12Double]
                             - FUsedw^[Pw1Double]*FUsedw^[Pw1Double]*FUsedw^[Pw2Double]/FR1[PR11Double]/FR1[PR12Double]
                             -FUsedw^[Pw1Double]*FUsedw^[Pw2Double]*FUsedw^[Pw2Double]/FR1[PR11Double]/FR1[PR12Double];
                TmpZ2:=TmpZ2
                      +FUsedw^[Pw1Double]
                      *FUsedw^[Pw2Double]
                      *FR2[PR2Double]
                      /FR1[PR11Double]
                      /FR1[PR12Double];
                MaxExposure1:=Max(MaxExposure1,FR1[PR11Double]);
                Inc(NrExposureneq0);
              end else if FR2[PR2Double] <> 0 then begin
                TmpZ1:=TmpZ1+1;
                Inc(NrExposureneq0);
              end;
            end;
            Inc(PS1Double,FJ+1+FK);
            Inc(PS2Double,FJ+1+FK);
            Inc(Pw1Double,FJ+FP);
            Inc(Pw2Double,FJ+FP);
            Inc(PR11Double,FJ+FK);
            Inc(PR12Double,FJ+FK);
            Inc(PR2Double,FJ+FP);
          end;
          {$ENDREGION}
        end;
        TmpZ:=TmpZ1+TmpZ2*TmpSum;
        {$ENDREGION}
        {$REGION 'normalise, we have to take care of false accuracy of computer calculations around zero'}
        // take only sigmas wich are not too large (avoid numerical side affects)
        if (NrExposureneq0>1) and (TmpZ > 0) and (FDefaultsigma[PDestDouble]>0)
              and ((ABS(FDefaultsigma[PDestDouble]/TmpZ)
                    <ABS(FUsedf^[Pf1Double] * FUsedf^[Pf2Double]*MaxExposure1)*FMaxVarCoef1+FMaxVarCoef2)) then begin
          FDefaultsigma[PDestDouble]:=FDefaultsigma[PDestDouble]/TmpZ;
        end else if (NrExposureneq0<=1) then begin
          // too few observations in order to estimate sigma with the standard formula, take Mack's suggestion
          Psigma1:=PDestDouble;
          Dec(Psigma1);
          Psigma2:=Psigma1;
          Dec(Psigma2);
          FDefaultsigma[PDestDouble]:=Min(FDefaultsigma[Psigma1],FDefaultsigma[Psigma2]);
          if FDefaultsigma[Psigma2]<>0 then
            FDefaultsigma[PDestDouble]:=Min(FDefaultsigma[PDestDouble],
                                            FDefaultsigma[Psigma1]
                                            *FDefaultsigma[Psigma1]/FDefaultsigma[Psigma2]);
        end else FDefaultsigma[PDestDouble]:=0; // avoid negative variances, which can only be the result of numerical issues
        {$ENDREGION}
      end;
      {$ENDREGION}
      {$REGION 'now all covariance terms'}
      for m1 := 0 to FM do begin
        for m2:=m1+1 to FM do begin
          PDestDouble:=(m1*(FM+1)+m2)*(FJ+FP)+k;
          PF1Double:=m1*(FJ+FK)+k;
          PF2Double:=m2*(FJ+FK)+k;
          PS1Double:=m1*(FI+1)*(FJ+1+FK)+k+1;
          PS2Double:=m2*(FI+1)*(FJ+1+FK)+k+1;
          // take n=1
          Pw1Double:=((FM+1)+m1)*(FI+1)*(FJ+FP)+k;
          Pw2Double:=((FM+1)+m2)*(FI+1)*(FJ+FP)+k;
          PR11Double:=m1*(FI+1)*(FJ+FK)+k;
          PR12Double:=m2*(FI+1)*(FJ+FK)+k;
          PR2Double:=(m1*(FM+1)+m2)*(FI+1)*(FJ+FP)+k;
          TmpSum:=0;
          TmpZ1:=0;
          TmpZ2:=0;
          NrExposureneq0:=0;
          {$REGION 'without normalisation'}
          if not(calc_UseWeightsForDefaultSigma in FCalculationFlags) then begin
            {$REGION 'do not use weights'}
            For i := 0 To FI - k - 1 do begin
              if (FR1[PR12Double]<>0) and (FR1[PR11Double]<>0) and (FR2[PR2Double]<>0) then begin
                FDefaultsigma[PDestDouble]:=FDefaultsigma[PDestDouble]+FR1[PR11Double]*FR1[PR12Double]/FR2[PR2Double]*(FS[PS1Double]/FR1[PR11Double]-FUsedf^[Pf1Double])*(FS[PS2Double]/FR1[PR12Double]-Fusedf^[Pf2Double]);
                TmpSum := TmpSum + FR1[PR11Double] * FR1[PR12Double] / FR2[PR2Double];
                TmpZ1 := TmpZ1 + 1 - FUsedw^[Pw1Double] - FUsedw^[Pw2Double];
                TmpZ2 := TmpZ2 + FUsedw^[Pw1Double] * FUsedw^[Pw2Double] * FR2[PR2Double] / FR1[PR11Double] / FR1[PR12Double];
                Inc(NrExposureneq0);
              end else if FR2[PR2Double] <> 0 then begin
                TmpZ1:=TmpZ1+1;
                Inc(NrExposureneq0);
              end;
              Inc(PS1Double,FJ+1+FK);
              Inc(PS2Double,FJ+1+FK);
              Inc(Pw1Double,FJ+FP);
              Inc(Pw2Double,FJ+FP);
              Inc(PR11Double,FJ+FK);
              Inc(PR12Double,FJ+FK);
              Inc(PR2Double,FJ+FP);
            end;
            {$ENDREGION}
          end else begin
            {$REGION 'use weights'}
            For i := 0 To FI - k - 1 do begin
              if min(FUsedw^[Pw1Double],FUsedw^[Pw2Double])>0 then begin
                if (FR1[PR12Double]<>0) and (FR1[PR11Double]<>0) and (FR2[PR2Double]<>0) then begin
                  FDefaultsigma[PDestDouble]:=FDefaultsigma[PDestDouble]
                                              +FUsedw^[Pw1Double]
                                              *FUsedw^[Pw2Double]
                                              /FR2[PR2Double]
                                              *(FS[PS1Double]/FR1[PR11Double]-FUsedf^[Pf1Double])
                                              *(FS[PS2Double]/FR1[PR12Double]-Fusedf^[Pf2Double]);
                  TmpSum:=TmpSum
                        +FUsedw^[Pw1Double]
                        *FUsedw^[Pw2Double]
                        /FR2[PR2Double];
                  TmpZ1:=TmpZ1 + FUsedw^[Pw1Double]*FUsedw^[Pw2Double]/FR1[PR11Double]/FR1[PR12Double]
                               - FUsedw^[Pw1Double]*FUsedw^[Pw1Double]*FUsedw^[Pw2Double]/FR1[PR11Double]/FR1[PR12Double]
                               -FUsedw^[Pw1Double]*FUsedw^[Pw2Double]*FUsedw^[Pw2Double]/FR1[PR11Double]/FR1[PR12Double];
                  TmpZ2:=TmpZ2
                        +FUsedw^[Pw1Double]
                        *FUsedw^[Pw2Double]
                        *FR2[PR2Double]
                        /FR1[PR11Double]
                        /FR1[PR12Double];
                  Inc(NrExposureneq0);
                end else if FR2[PR2Double] <> 0 then begin
                  TmpZ1:=TmpZ1+1;
                  Inc(NrExposureneq0);
                end;
              end;
              Inc(PS1Double,FJ+1+FK);
              Inc(PS2Double,FJ+1+FK);
              Inc(Pw1Double,FJ+FP);
              Inc(Pw2Double,FJ+FP);
              Inc(PR11Double,FJ+FK);
              Inc(PR12Double,FJ+FK);
              Inc(PR2Double,FJ+FP);
            end;
            {$ENDREGION}
          end;
          TmpZ:=TmpZ1+TmpZ2*TmpSum;
          {$ENDREGION}
          {$REGION 'normalise, we have to take care of false accuracy of computer calculations around zero'}
          // take only sigmas wich are not too large (avoid numerical side affects)
          if (NrExposureneq0>1) and (TmpZ <> 0) and (FDefaultsigma[PDestDouble]<>0)
              and ((ABS(FDefaultsigma[PDestDouble]/TmpZ)
                    <ABS(FUsedf^[Pf1Double] * FUsedf^[Pf2Double])*FMaxVarCoef1+FMaxVarCoef2)) then begin
            FDefaultsigma[PDestDouble]:=FDefaultsigma[PDestDouble]/TmpZ;
          end else begin
            // too few observations in order to estimate sigma with the standard formula
            If NrExposureneq0<=1 Then begin
              // take the same correlation matrix as for k-1
              Psigma1:=PDestDouble;
              Dec(Psigma1,1);
              FDefaultsigma[PDestDouble]:=FDefaultsigma[Psigma1];
              Psigma1:=0;
              Inc(Psigma1,(m1*(FM+1)+m1)*(FJ+FP)+k-1);
              Psigma2:=0;
              Inc(Psigma2,(m2*(FM+1)+m2)*(FJ+FP)+k-1);
              if (FDefaultsigma[Psigma1]>0) and (FDefaultsigma[Psigma2]>0) then begin
                FDefaultsigma[PDestDouble]:=FDefaultsigma[PDestDouble]
                                            /sqrt(FDefaultsigma[Psigma1])
                                            /sqrt(FDefaultsigma[Psigma2]);
                Psigma1:=0;
                Inc(Psigma1,(m1*(FM+1)+m1)*(FJ+FP)+k);
                Psigma2:=0;
                Inc(Psigma2,(m2*(FM+1)+m2)*(FJ+FP)+k);
                FDefaultsigma[PDestDouble]:=FDefaultsigma[PDestDouble]
                                            *sqrt(FDefaultsigma[Psigma1])
                                            *sqrt(FDefaultsigma[Psigma2]);
              end else begin
                FDefaultsigma[PDestDouble]:=0;
              end;
            end else begin
              FDefaultsigma[PDestDouble]:=0;
            end;
          end;
          {$ENDREGION}
          // set the same value for m1 and m2 exchanged
          Psigma1:=(m2*(FM+1)+m1)*(FJ+FP)+k;
          FDefaultsigma[Psigma1]:=FDefaultsigma[PDestDouble];
        end;
      end;
      {$ENDREGION}
    end;
    {$REGION 'tail sigma still missing'}
    if (FK>0) then begin
      // we take the same values as for k=FJ-1
      for m1 := 0 to FM do begin
        FDefaultsigma[(m1*(FM+1)+m1)*(FJ+FP)+FJ]:=FDefaultsigma[(m1*(FM+1)+m1)*(FJ+FP)+FJ-1];
        for m2 := m1+1 to FM do begin
          FDefaultsigma[(m1*(FM+1)+m2)*(FJ+FP)+FJ]:=FDefaultsigma[(m1*(FM+1)+m2)*(FJ+FP)+FJ-1];
          FDefaultsigma[(m2*(FM+1)+m1)*(FJ+FP)+FJ]:=FDefaultsigma[(m2*(FM+1)+m1)*(FJ+FP)+FJ-1];
        end;
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_Defaultsigma,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitDefaultsigma + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitDefaultsigma + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitDefaultw(): String;
var HasChanged: Boolean;
    StartTime, EndTime: Int64;
    PDestDouble, PR1Double, PR2Double: IndexDouble;
    k, i: integer;
begin
  Result:='';
  HasChanged:=False;
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitDefaultw + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_R1 in FInitialised) then Result:=InitR1();
  if Result<>'' then Exit;
  If not(Init_R2 in FInitialised) then Result:=InitR2();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FDefaultw)<=0 then begin
      //                  n    m     i      k
      SetLength(FDefaultw,3*(FM+1)*(FI+1)*(FJ+FP));
      HasChanged:=True;
    end;
    // default (not normalized) weights are equal to R1^2/r2
    {$REGION 'first check if something has changed'}
    HasChanged:=HasChanged;
    PDestDouble:=0;
    PR1Double:=0;
    PR2Double:=0;
    k:=0;
    i:=0;
    while not(HasChanged) and (PDestDouble<(FM+1)*(FI+1)*(FJ+FP)) do begin
      HasChanged:=
            (((FR2[PR2Double]<>0) and (FDefaultw[PDestDouble]<>FR1[PR1Double]*FR1[PR1Double]/FR2[PR2Double]))
              or ((FR2[PR2Double]=0) and (FDefaultw[PDestDouble]=0)));
      if not HasChanged then begin
        if k=Min(FI-i,FJ-1+FP) then begin
          k:=0;
          Inc(PDestDouble,FJ+FP-Min(FI-i,FJ-1+FP));
          Inc(PR1Double,FJ+FK-Min(FI-i,FJ-1+FP));
          Inc(PR2Double,FJ+FK-Min(FI-i,FJ-1+FP));
          if i=FI then begin
            i:=0;
            // m1=m2
            Inc(PR2Double,(FM+1)*(FI+1)*(FJ+FP));
          end else Inc(i,1);
        end else begin
          Inc(k,1);
          Inc(PDestDouble,1);
          Inc(PR1Double,1);
          Inc(PR2Double,1);
        end;
      end;
    end;
    {$ENDREGION}
    {$REGION 'copy FR1^2/FR2 to the not normalised weights w(0), if necessary'}
    if HasChanged then begin
      while PDestDouble<(FM+1)*(FI+1)*(FJ+FP) do begin
        if FR2[PR2Double]<>0 then
          FDefaultw[PDestDouble]:=FR1[PR1Double]*FR1[PR1Double]/FR2[PR2Double]
        else
          FDefaultw[PDestDouble]:=0;
        if k=Min(FI-i,FJ-1+FP) then begin
          k:=0;
          Inc(PDestDouble,FJ+FP-Min(FI-i,FJ-1+FP));
          Inc(PR1Double,FJ+FK-Min(FI-i,FJ-1+FP));
          Inc(PR2Double,FJ+FP-Min(FI-i,FJ-1+FP));
          if i=FI then begin
            i:=0;
            // m1=m2
            Inc(PR2Double,(FM+1)*(FI+1)*(FJ+FP));
          end else Inc(i,1);
        end else begin
          Inc(k,1);
          Inc(PDestDouble,1);
          Inc(PR1Double,1);
          Inc(PR2Double,1);
        end;
      end;
      // normalize weights
      NormalizeWeights(FDefaultw);
      QueryPerformanceCounter(EndTime);
      SetInitFlags(Init_Defaultw,EndTime-StartTime);
    end;
    {$ENDREGION}
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitDefaultw + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitDefaultw + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitEigenvalues: String;
var m1, m2, k, i: Integer;
    PDestDouble, PsigmaDouble, PR2Double: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitEigenvalues + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_sigma, Init_Defaultsigma]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitEigenvalues + sLineBreak + rsLSRM_NotInitialisedsigma;
    Exit;
  end;
  If not(Init_R2Full in FInitialised) then Result:=InitR2Full();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FEigenvalues)<=0 then begin
      //                        i      k
      SetLength(FEigenvalues,(FI+1)*(FJ+FP));
    end;
    if Length(FEigenvectors)<=0 then begin
      //                        i       k
      SetLength(FEigenvectors,(FI+1)*(FJ+FP));
    end;
    // calculate eigenvalues and eigenvectors for each covariance matrix
    setlength(TmpArray,FM+1,FM+1);
    PDestDouble:=0;
    for i := 0 to FI do begin
      for k := 0 to FJ-1+FP do begin
        PsigmaDouble:=k;
        PR2Double:=i*(FJ+FP)+k;
        // fill the covariance matrix
        for m1 := 0 to FM do begin
          Inc(PsigmaDouble,m1*(FJ+FP));
          Inc(PR2Double,m1*(FI+1)*(FJ+FP));
          for m2 := m1 to FM do begin
            TmpArray[m1,m2]:=FUsedsigma^[PsigmaDouble]*FR2[PR2Double];
            Inc(PsigmaDouble,(FJ+FP));
            Inc(PR2Double,(FI+1)*(FJ+FP));
          end;
        end;
        // get eigenvalues and eigenvectors
        if not SMatrixEVD(TmpArray,FM+1,1,true,FEigenvalues[PDestDouble],FEigenvectors[PDestDouble]) then begin
          Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitEigenvalues + sLineBreak
                  + rsLSRM_AlgorithmDoesNotConverge;
          Exit;
        end;
        Inc(PDestDouble,1);
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_Eigenvalues,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitEigenvalues + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitEigenvalues + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.Initf(var f: Double): String;
var HasChanged: Boolean;
    StartTime, EndTime: Int64;
    m, k: Integer;
begin
  Result:='';
  HasChanged:=False;
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitf + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(Ff)<=0 then begin
      //              m      k
      SetLength(Ff,(FM+1)*(FJ+FK));
      HasChanged:=True;
    end;
    if HasChanged or not(Init_f in FInitialised)
        or not(CompareMem(@Ff[0],@f,(FM+1)*(FJ+FK)*SizeOf(Double))) then begin
      // copy the specified development factors
      Move(f,Ff[0],(FM+1)*(FJ+FK)*SizeOf(Double));
      if (FK>0) and (Init_Defaultf in FInitialised) then begin
        // take manual values for the default development factors in the tail
        for m := 0 to FM do begin
          for k := FJ to FJ+FK-1 do begin
            fDefaultf[m*(FJ+FK)+k]:=Ff[m*(FJ+FK)+k];
          end;
        end;
      end;
      QueryPerformanceCounter(EndTime);
      SetInitFlags(Init_f,EndTime-StartTime);
    end;
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitf + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitf + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitGamma1(const Gamma1: OleVariant; Hasgamma1, HasNewInflation: boolean;
                                var NewInflation: Double): String;
var row, m, l, i, k, h, j, ch, cl: Integer;
    StartTime, EndTime: Int64;
    Dep_ih, Dep_ml: TLSRMDependencies; // store which l (h) is needed for which m (i) in double sums
    PNewInf1Double, PNewInf2Double: PDouble;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitGamma1 + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  {$ENDREGION}
  Dep_ml:=nil;
  Dep_ih:=nil;
  try
    QueryPerformanceCounter(StartTime);
    Dep_ml:=TLSRMDependencies.Create(FM);
    Dep_ih:=TLSRMDependencies.Create(FI);
    // independend of Gamma1 (operator F is the identity in special cases) we have the
    // dependencies m,m and i,i
    For i:=0 to FI do Dep_ih.AddDependsOn(i,i);
    For m:=0 to FM do Dep_ml.AddDependsOn(m,m);
    // initialise array or clear corresponding memory
    if Length(FGamma1)<=0 then begin
      //                   m     l      i      k        h      j
      SetLength(FGamma1,(FM+1)*(FM+1)*(FI+1)*(FJ+FK)*(FI+1)*(FJ+FK));
    end else if Hasgamma1 then begin
      ZeroMemory(@FGamma1[0],(FM+1)*(FM+1)*(FI+1)*(FJ+FK)*(FI+1)*(FJ+FK)*SizeOf(Double));
    end;
    if HasNewInflation then begin
      if Hasgamma1 then begin
        {$REGION 'initialise gamma1 with new inflation'}
        for row:=VarArrayLowBound(Gamma1,1) to VarArrayHighBound(Gamma1,1) do begin
          // get values
          m:=Gamma1[row,0];
          if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
          l:=Gamma1[row,1];
          if (l<0) or (l>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
          i:=Gamma1[row,2];
          if (i<0) or (i>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
          k:=Gamma1[row,3];
          if (k<0) or (k>FJ-1+FK) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
          h:=Gamma1[row,4];
          if (h<0) or (h>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
          j:=Gamma1[row,5];
          if (j<0) or (j>FJ-1+FK) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
          PNewInf1Double:=@NewInflation;
          Inc(PNewInf1Double,(m*(FI+1)+i)*(FJ+1+FK)+k+1);
          PNewInf2Double:=@NewInflation;
          Inc(PNewInf2Double,(l*(FI+1)+h)*(FJ+1+FK)+j);
          if PNewInf2Double^<=0 then raise ELSRMInitError.Create(rsLSRM_NonPositiveInflation);
          FGamma1[((((m*(FM+1)+l)*(FI+1)+i)*(FJ+FK)+k)*(FI+1)+h)*(FJ+FK)+j]
              :=Gamma1[row,6]
                *PNewInf1Double^/PNewInf2Double^;
          // add dependencies
          Dep_ih.AddDependsOn(Gamma1[row,2],Gamma1[row,4]);
          Dep_ml.AddDependsOn(Gamma1[row,0],Gamma1[row,1]);
        end;
        // fill the corresponding dependency arrays
        Dep_ml.FillDependenceArray(FmlDependency);
        Dep_ih.FillDependenceArray(FihDependency);
        {$ENDREGION}
      end else if (Init_gamma1 in FInitialised) and (Init_InflationRate in FInitialised) then begin
        {$REGION 'correct inflation for already initialised gamma1 with inflation'}
        for m := 0 to FM do begin
          For cl := 0 To FmlDependency[m,FM] do begin
            l := FmlDependency[m,cl];
            for i := 0 to FI do begin
              for k := 0 to FJ-1+FK do begin
                For ch := 0 To FihDependency[i,FI] do begin
                  h := FihDependency[i,ch];
                  for j := 0 to FJ-1+FK do begin
                    PNewInf1Double:=@NewInflation;
                    Inc(PNewInf1Double,(m*(FI+1)+i)*(FJ+1+FK)+k+1);
                    PNewInf2Double:=@NewInflation;
                    Inc(PNewInf2Double,(l*(FI+1)+h)*(FJ+1+FK)+j);
                    if PNewInf2Double^<=0 then raise ELSRMInitError.Create(rsLSRM_NonPositiveInflation);
                    FGamma1[((((m*(FM+1)+l)*(FI+1)+i)*(FJ+FK)+k)*(FI+1)+h)*(FJ+FK)+j]
                        :=FGamma1[((((m*(FM+1)+l)*(FI+1)+i)*(FJ+FK)+k)*(FI+1)+h)*(FJ+FK)+j]
                          /FInflation[(m*(FI+1)+i)*(FJ+1+FK)+k+1]*FInflation[(l*(FI+1)+h)*(FJ+1+FK)+j]
                          *PNewInf1Double^/PNewInf2Double^;

                  end;
                end;
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else if (Init_gamma1 in FInitialised) then begin
        {$REGION 'correct inflation for already initialised gamma1 without inflation'}
        for m := 0 to FM do begin
          For cl := 0 To FmlDependency[m,FM] do begin
            l := FmlDependency[m,cl];
            for i := 0 to FI do begin
              for k := 0 to FJ-1+FK do begin
                For ch := 0 To FihDependency[i,FI] do begin
                  h := FihDependency[i,ch];
                  for j := 0 to FJ-1+FK do begin
                    PNewInf1Double:=@NewInflation;
                    Inc(PNewInf1Double,(m*(FI+1)+i)*(FJ+1+FK)+k+1);
                    PNewInf2Double:=@NewInflation;
                    Inc(PNewInf2Double,(l*(FI+1)+h)*(FJ+1+FK)+j);
                    if PNewInf2Double^<=0 then raise ELSRMInitError.Create(rsLSRM_NonPositiveInflation);
                    FGamma1[((((m*(FM+1)+l)*(FI+1)+i)*(FJ+FK)+k)*(FI+1)+h)*(FJ+FK)+j]
                        :=FGamma1[((((m*(FM+1)+l)*(FI+1)+i)*(FJ+FK)+k)*(FI+1)+h)*(FJ+FK)+j]
                          *PNewInf1Double^/PNewInf2Double^;

                  end;
                end;
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else raise ELSRMInitError.Create(rsLSRM_WrongInternalParameter);
    end else if Init_InflationRate in FInitialised then begin
      {$REGION 'initialise gamma1 with old inflation'}
      for row:=VarArrayLowBound(Gamma1,1) to VarArrayHighBound(Gamma1,1) do begin
        // get values
        m:=Gamma1[row,0];
        if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
        l:=Gamma1[row,1];
        if (l<0) or (l>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
        i:=Gamma1[row,2];
        if (i<0) or (i>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
        k:=Gamma1[row,3];
        if (k<0) or (k>FJ-1+FK) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
        h:=Gamma1[row,4];
        if (h<0) or (h>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
        j:=Gamma1[row,5];
        if (j<0) or (j>FJ-1+FK) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
        FGamma1[((((m*(FM+1)+l)*(FI+1)+i)*(FJ+FK)+k)*(FI+1)+h)*(FJ+FK)+j]:=Gamma1[row,6]
                   /FInflation[(m*(FI+1)+i)*(FJ+1+FK)+k+1]*FInflation[(l*(FI+1)+h)*(FJ+1+FK)+j];
        // add dependencies
        Dep_ih.AddDependsOn(Gamma1[row,2],Gamma1[row,4]);
        Dep_ml.AddDependsOn(Gamma1[row,0],Gamma1[row,1]);
      end;
      // fill the corresponding dependency arrays
      Dep_ml.FillDependenceArray(FmlDependency);
      Dep_ih.FillDependenceArray(FihDependency);
      {$ENDREGION}
    end else begin
      {$REGION 'initialise gamma1 without inflation'}
      for row:=VarArrayLowBound(Gamma1,1) to VarArrayHighBound(Gamma1,1) do begin
        // get values
        m:=Gamma1[row,0];
        if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
        l:=Gamma1[row,1];
        if (l<0) or (l>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
        i:=Gamma1[row,2];
        if (i<0) or (i>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
        k:=Gamma1[row,3];
        if (k<0) or (k>FJ-1+FK) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
        h:=Gamma1[row,4];
        if (h<0) or (h>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
        j:=Gamma1[row,5];
        if (j<0) or (j>FJ-1+FK) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
        FGamma1[((((m*(FM+1)+l)*(FI+1)+i)*(FJ+FK)+k)*(FI+1)+h)*(FJ+FK)+j]:=Gamma1[row,6];
        // add dependencies
        Dep_ih.AddDependsOn(Gamma1[row,2],Gamma1[row,4]);
        Dep_ml.AddDependsOn(Gamma1[row,0],Gamma1[row,1]);
      end;
      // fill the corresponding dependency arrays
      Dep_ml.FillDependenceArray(FmlDependency);
      Dep_ih.FillDependenceArray(FihDependency);
      {$ENDREGION}
    end;
    // clean up
    FreeAndNil(Dep_ml);
    FreeAndNil(Dep_ih);
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_Gamma1,EndTime-StartTime);
  except
    On E:Exception do begin
      FreeAndNil(Dep_ml);
      FreeAndNil(Dep_ih);
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitGamma1 + sLineBreak + E.Message;
    end else begin
      FreeAndNil(Dep_ml);
      FreeAndNil(Dep_ih);
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitGamma1 + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitGamma2(const Gamma2: OleVariant; Hasgamma2, HasNewInflation: boolean;
                                var NewInflation: Double): String;
var row, m1, m2, L, i, k, h, j: Integer;
    StartTime, EndTime: Int64;
    PNewInf1Double, PNewInf2Double, PNewInf3Double: PDouble;
begin
  // we do not try to reduce the complexity in sums over m1,m2,l and i,h, because
  // such sums appear only once in the calculation of R^m1,m2_i,k
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitGamma2 + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array or clear corresponding memory
    if Length(FGamma2)<=0 then begin
      //                   m1     m2     l     i       k       h      j
      SetLength(FGamma2,(FM+1)*(FM+1)*(FM+1)*(FI+1)*(FJ+FP)*(FI+1)*(FJ+FP));
    end else if Hasgamma2 then begin
      ZeroMemory(@FGamma2[0],(FM+1)*(FM+1)*(FM+1)*(FI+1)*(FJ+FP)*(FI+1)*(FJ+FP)*SizeOf(Double));
    end;
    if HasNewInflation then begin
      if Hasgamma2 then begin
        {$REGION 'initialise gamma2 with new inflation'}
        for row:=VarArrayLowBound(Gamma2,1) to VarArrayHighBound(Gamma2,1) do begin
          // get values
          m1:=Gamma2[row,0];
          if (m1<0) or (m1>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
          m2:=Gamma2[row,1];
          if (m2<0) or (m2>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
          l:=Gamma2[row,2];
          if (l<0) or (l>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
          i:=Gamma2[row,3];
          if (i<0) or (i>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
          k:=Gamma2[row,4];
          if (k<0) or (k>FJ-1+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
          h:=Gamma2[row,5];
          if (h<0) or (h>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
          j:=Gamma2[row,6];
          if (j<0) or (j>FJ-1+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
          PNewInf1Double:=@NewInflation;
          Inc(PNewInf1Double,(m1*(FI+1)+i)*(FJ+1+FK)+k+1);
          PNewInf2Double:=@NewInflation;
          Inc(PNewInf2Double,(m2*(FI+1)+i)*(FJ+1+FK)+k+1);
          PNewInf3Double:=@NewInflation;
          Inc(PNewInf3Double,(l*(FI+1)+h)*(FJ+1+FK)+j);
          if PNewInf3Double^<=0 then raise ELSRMInitError.Create(rsLSRM_NonPositiveInflation);
          FGamma2[(((((m1*(FM+1)+m2)*(FM+1)+l)*(FI+1)+i)*(FJ+FP)+k)*(FI+1)+h)*(FJ+FP)+j]
            :=Gamma2[row,7]
              *PNewInf1Double^
              *PNewInf2Double^
              /PNewInf3Double^;
        end;
        {$ENDREGION}
      end else if (Init_gamma2 in FInitialised) and (Init_InflationRate in FInitialised) then begin
        {$REGION 'correct inflation for already initialised gamma2 with inflation'}
        for m1 := 0 to FM do begin
          for m2 := 0 to FM do begin
            For l := 0 To FM do begin
              for i := 0 to FI do begin
                for k := 0 to FJ-1+FP do begin
                  For h := 0 To FI do begin
                    for j := 0 to FJ-1+FP do begin
                      PNewInf1Double:=@NewInflation;
                      Inc(PNewInf1Double,(m1*(FI+1)+i)*(FJ+1+FK)+k+1);
                      PNewInf2Double:=@NewInflation;
                      Inc(PNewInf2Double,(m2*(FI+1)+i)*(FJ+1+FK)+k+1);
                      PNewInf3Double:=@NewInflation;
                      Inc(PNewInf3Double,(l*(FI+1)+h)*(FJ+1+FK)+j);
                      if PNewInf3Double^<=0 then raise ELSRMInitError.Create(rsLSRM_NonPositiveInflation);
                        FGamma2[(((((m1*(FM+1)+m2)*(FM+1)+l)*(FI+1)+i)*(FJ+FP)+k)*(FI+1)+h)*(FJ+FP)+j]
                          :=FGamma2[(((((m1*(FM+1)+m2)*(FM+1)+l)*(FI+1)+i)*(FJ+FP)+k)*(FI+1)+h)*(FJ+FP)+j]
                            /FInflation[(m1*(FI+1)+i)*(FJ+1+FK)+k+1]
                            /FInflation[(m2*(FI+1)+i)*(FJ+1+FK)+k+1]
                            *FInflation[(l*(FI+1)+h)*(FJ+1+FK)+j]
                            *PNewInf1Double^
                            *PNewInf2Double^
                            /PNewInf3Double^;
                    end;
                  end;
                end;
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else if (Init_gamma2 in FInitialised) then begin
        {$REGION 'correct inflation for already initialised gamma2 without inflation'}
        for m1 := 0 to FM do begin
          for m2 := 0 to FM do begin
            For l := 0 To FM do begin
              for i := 0 to FI do begin
                for k := 0 to FJ-1+FP do begin
                  For h := 0 To FI do begin
                    for j := 0 to FJ-1+FP do begin
                      PNewInf1Double:=@NewInflation;
                      Inc(PNewInf1Double,(m1*(FI+1)+i)*(FJ+1+FK)+k+1);
                      PNewInf2Double:=@NewInflation;
                      Inc(PNewInf2Double,(m2*(FI+1)+i)*(FJ+1+FK)+k+1);
                      PNewInf3Double:=@NewInflation;
                      Inc(PNewInf3Double,(l*(FI+1)+h)*(FJ+1+FK)+j);
                      if PNewInf3Double^<=0 then raise ELSRMInitError.Create(rsLSRM_NonPositiveInflation);
                        FGamma2[(((((m1*(FM+1)+m2)*(FM+1)+l)*(FI+1)+i)*(FJ+FP)+k)*(FI+1)+h)*(FJ+FP)+j]
                          :=FGamma2[(((((m1*(FM+1)+m2)*(FM+1)+l)*(FI+1)+i)*(FJ+FP)+k)*(FI+1)+h)*(FJ+FP)+j]
                            *PNewInf1Double^
                            *PNewInf2Double^
                            /PNewInf3Double^;
                    end;
                  end;
                end;
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else raise ELSRMInitError.Create(rsLSRM_WrongInternalParameter);
    end else if Init_InflationRate in FInitialised then begin
      {$REGION 'initialise gamma2 with old inflation'}
      for row:=VarArrayLowBound(Gamma2,1) to VarArrayHighBound(Gamma2,1) do begin
        // get values
        m1:=Gamma2[row,0];
        if (m1<0) or (m1>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
        m2:=Gamma2[row,1];
        if (m2<0) or (m2>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
        l:=Gamma2[row,2];
        if (l<0) or (l>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
        i:=Gamma2[row,3];
        if (i<0) or (i>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
        k:=Gamma2[row,4];
        if (k<0) or (k>FJ-1+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
        h:=Gamma2[row,5];
        if (h<0) or (h>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
        j:=Gamma2[row,6];
        if (j<0) or (j>FJ-1+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
        FGamma2[(((((m1*(FM+1)+m2)*(FM+1)+l)*(FI+1)+i)*(FJ+FP)+k)*(FI+1)+h)*(FJ+FP)+j]
          :=Gamma2[row,7]
            *FInflation[(m1*(FI+1)+i)*(FJ+1+FK)+k+1]
            *FInflation[(m2*(FI+1)+i)*(FJ+1+FK)+k+1]
            /FInflation[(l*(FI+1)+h)*(FJ+1+FK)+j];
      end;
      {$ENDREGION}
    end else begin
      {$REGION 'initialise gamma2 without inflation'}
      for row:=VarArrayLowBound(Gamma2,1) to VarArrayHighBound(Gamma2,1) do begin
        // get values
        m1:=Gamma2[row,0];
        if (m1<0) or (m1>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
        m2:=Gamma2[row,1];
        if (m2<0) or (m2>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
        l:=Gamma2[row,2];
        if (l<0) or (l>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
        i:=Gamma2[row,3];
        if (i<0) or (i>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
        k:=Gamma2[row,4];
        if (k<0) or (k>FJ-1+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
        h:=Gamma2[row,5];
        if (h<0) or (h>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
        j:=Gamma2[row,6];
        if (j<0) or (j>FJ-1+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
        FGamma2[(((((m1*(FM+1)+m2)*(FM+1)+l)*(FI+1)+i)*(FJ+FP)+k)*(FI+1)+h)*(FJ+FP)+j]:=Gamma2[row,7];
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_Gamma2,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitGamma2 + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitGamma2 + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitGeometry(M, I, J, K: Integer): String;
var StartTime, EndTime: Int64;
begin
  Result:='';
  // check geometry
  if (M<0) then begin
    Result:= 'LSRM ''' + FName + ''':' + rsLSRM_ErrorInitGeometry + sLineBreak + rsLSRM_WrongGeometryM;
    Exit;
  end;
  if (I<0) then begin
    Result:= 'LSRM ''' + FName + ''':' + rsLSRM_ErrorInitGeometry + sLineBreak + rsLSRM_WrongGeometryI;
    Exit;
  end;
  if (J<0) then begin
    Result:= 'LSRM ''' + FName + ''':' + rsLSRM_ErrorInitGeometry + sLineBreak + rsLSRM_WrongGeometryJ1;
    Exit;
  end;
  if (J>I) then begin
    Result:= 'LSRM ''' + FName + ''':' + rsLSRM_ErrorInitGeometry + sLineBreak + rsLSRM_WrongGeometryJ2;
    Exit;
  end;
  if (K<0) then begin
    Result:= 'LSRM ''' + FName + ''':' + rsLSRM_ErrorInitGeometry + sLineBreak + rsLSRM_WrongGeometryK;
    Exit;
  end;
  try
    QueryPerformanceCounter(StartTime);
    // update geometry, if necessary
    if (M<>FM) or (I<>FI) or (J<>FJ) or (K<>FK) then begin
      FreeAll;
      FM:=M;
      FI:=I;
      FJ:=J;
      FK:=K;
      if FK>0 then begin
        FP:=1
      end else begin
        FP:=0;
      end;
      QueryPerformanceCounter(EndTime);
      SetInitFlags(Init_Geometry,EndTime-StartTime);
    end;
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitGeometry + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitGeometry + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitInflationRate(var InflationRate: Double): String;
var HasChanged: boolean;
    StartTime, EndTime: Int64;
    m, i, k: Integer;
    PSource: ^Double;
    PDestDouble: IndexDouble;
begin
  Result:='';
  HasChanged:=False;
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitInflationRate + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FInflation)<=0 then begin
      //                      m      i      k
      SetLength(FInflation,(FM+1)*(FI+1)*(FJ+1+FK));
      HasChanged:=True;
    end;
    if HasChanged or not(Init_InflationRate in FInitialised)
        or not(CompareMem(@FInflation[0],@InflationRate,(FM+1)*(FI+1+FK)*(FJ+1)*SizeOf(Double))) then begin
      // cannot simply move the memory, because we have to check that the inflationrate is strictly positive
      PSource:=@InflationRate;
      PDestDouble:=0;
      for m := 0 to FM do begin
        for i := 0 to FI do begin
          for k := 0 to FJ+FK do begin
            if PSource^<=0 then begin
              Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitInflationRate + sLineBreak
                      + rsLSRM_NonPositiveInflation;
              Exit;
            end;
            FInflation[PDestDouble]:=PSource^;
            Inc(PDestDouble,1);
            Inc(PSource,1);
          end;
        end;
      end;
      QueryPerformanceCounter(EndTime);
      SetInitFlags(Init_InflationRate,EndTime-StartTime);
    end;
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitInflationRate + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitInflationRate + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEP: string;
var StartTime, EndTime: Int64;
    PProcVarDouble, PParaErrDouble, PTotalDouble: IndexDouble;
    m, i: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  if calc_ReduceMemSize in FCalculationFlags then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDR + sLineBreak + rsLSRM_ProxyOnly;
    Exit;
  end;
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEP + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPParaErr in FInitialised) then Result:=InitMSEPParaErr();
  if Result<>'' then Exit;
  If not(Init_MSEPProcVar in FInitialised) then Result:=InitMSEPProcVar();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // calculate the sum of the process variation and the parameter error (without proxy)
    // and store the result in FMSEP(0,2)
    PProcVarDouble:=0;
    PParaErrDouble:=PProcVarDouble;
    Inc(PParaErrDouble,(FM+2)*(FI+2));
    PTotalDouble:=PParaErrDouble;
    Inc(PTotalDouble,(FM+2)*(FI+2));
    for m := 0 to FM+1 do begin
      for i := 0 to FI+1 do begin
        FMSEP[PTotalDouble]:=FMSEP[PProcVarDouble]+FMSEP[PParaErrDouble];
        Inc(PTotalDouble,1);
        Inc(PProcVarDouble,1);
        Inc(PParaErrDouble,1);
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEP,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEP + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEP + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCov: string;
var StartTime, EndTime: Int64;
    PDestDouble, PParaErrDouble, PProcVarDouble: IndexDouble;
    m1, m2, Proxy, Part: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCov + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPCovParaErr in FInitialised) then Result:=InitMSEPCovParaErr();
  if Result<>'' then Exit;
  If not(Init_MSEPCovProcVar in FInitialised) then Result:=InitMSEPCovProcVar();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=0;
    Part:=2;
    // initialise array or clear the memory
    if Length(FMSEPCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FMSEPCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // without tail
      ZeroMemory(@FMSEPCov[(Proxy*3+Part)*2*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    PDestDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    PProcVarDouble:=(Proxy*3+0)*2*(FM+1)*(FM+1);
    PParaErrDouble:=(Proxy*3+1)*2*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        FMSEPCov[PDestDouble]:=FMSEPCov[PProcVarDouble]+FMSEPCov[PParaErrDouble];
        Inc(PDestDouble,1);
        Inc(PProcVarDouble,1);
        Inc(PParaErrDouble,1);
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCov,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCov + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCov + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovEigenvalue: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalue + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPCov in FInitialised) then Result:=InitMSEPCov();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FMSEPCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FMSEPCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 0, Part = 2, Tail = 0
    Proxy:=0;
    Part:=2;
    Tail:=0;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FMSEPCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FMSEPCovEigenvalues[PDestDouble],FMSEPCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalue + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovEigenvalue,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalue + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalue + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovEigenvalueParaErr: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueParaErr + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPCovParaErr in FInitialised) then Result:=InitMSEPCovParaErr();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FMSEPCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FMSEPCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 0, Part = 1, Tail = 0
    Proxy:=0;
    Part:=1;
    Tail:=0;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FMSEPCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FMSEPCovEigenvalues[PDestDouble],FMSEPCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueParaErr + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovEigenvalueParaErr,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueParaErr + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueParaErr + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovEigenvalueParaErrProxy: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueParaErrProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPCovParaErrProxy in FInitialised) then Result:=InitMSEPCovParaErrProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FMSEPCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FMSEPCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 1, Part = 1, Tail = 0
    Proxy:=1;
    Part:=1;
    Tail:=0;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FMSEPCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FMSEPCovEigenvalues[PDestDouble],FMSEPCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueParaErrProxy + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovEigenvalueParaErrProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueParaErrProxy + sLineBreak 
              + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueParaErrProxy + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovEigenvalueParaErrTail: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueParaErrTail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPCovParaErrTail in FInitialised) then Result:=InitMSEPCovParaErrTail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FMSEPCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FMSEPCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 0, Part = 1, Tail = 1
    Proxy:=0;
    Part:=1;
    Tail:=1;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FMSEPCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FMSEPCovEigenvalues[PDestDouble],FMSEPCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueParaErrTail + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovEigenvalueParaErrTail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueParaErrTail + sLineBreak 
              + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueParaErrTail + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovEigenvalueParaErrTailProxy: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueParaErrTailProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPCovParaErrTailProxy in FInitialised) then Result:=InitMSEPCovParaErrTailProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FMSEPCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FMSEPCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 1, Part = 1, Tail = 1
    Proxy:=1;
    Part:=1;
    Tail:=1;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FMSEPCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FMSEPCovEigenvalues[PDestDouble],FMSEPCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueParaErrTailProxy + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovEigenvalueParaErrTailProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueParaErrTailProxy + sLineBreak 
              + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueParaErrTailProxy + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovEigenvalueProcVar: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProcVar + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPCovProcVar in FInitialised) then Result:=InitMSEPCovProcVar();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FMSEPCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FMSEPCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 0, Part = 0, Tail = 0
    Proxy:=0;
    Part:=0;
    Tail:=0;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FMSEPCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FMSEPCovEigenvalues[PDestDouble],FMSEPCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProcVar + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovEigenvalueProcVar,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProcVar + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProcVar + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovEigenvalueProcVarProxy: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProcVarProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPCovProcVarProxy in FInitialised) then Result:=InitMSEPCovProcVarProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FMSEPCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FMSEPCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 1, Part = 0, Tail = 0
    Proxy:=1;
    Part:=0;
    Tail:=0;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FMSEPCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FMSEPCovEigenvalues[PDestDouble],FMSEPCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProcVarProxy + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovEigenvalueProcVarProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProcVarProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProcVarProxy + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovEigenvalueProcVarTail: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProcVarTail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPCovProcVarTail in FInitialised) then Result:=InitMSEPCovProcVarTail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FMSEPCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FMSEPCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 0, Part = 0, Tail = 1
    Proxy:=0;
    Part:=0;
    Tail:=1;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FMSEPCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FMSEPCovEigenvalues[PDestDouble],FMSEPCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProcVarTail + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovEigenvalueProcVarTail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProcVarTail + sLineBreak 
              + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProcVarTail + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovEigenvalueProcVarTailProxy: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProcVarTailProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPCovProcVarTailProxy in FInitialised) then Result:=InitMSEPCovProcVarTailProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FMSEPCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FMSEPCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 1, Part = 0, Tail = 1
    Proxy:=1;
    Part:=0;
    Tail:=1;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FMSEPCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FMSEPCovEigenvalues[PDestDouble],FMSEPCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProcVarTailProxy + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovEigenvalueProcVarTailProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProcVarTailProxy + sLineBreak 
              + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProcVarTailProxy + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovEigenvalueProxy: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPCovProxy in FInitialised) then Result:=InitMSEPCovProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FMSEPCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FMSEPCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 1, Part = 2, Tail = 0
    Proxy:=1;
    Part:=2;
    Tail:=0;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FMSEPCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FMSEPCovEigenvalues[PDestDouble],FMSEPCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProxy + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovEigenvalueProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueProxy + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovEigenvalueTail: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueTail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPCovTail in FInitialised) then Result:=InitMSEPCovTail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FMSEPCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FMSEPCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 0, Part = 2, Tail = 1
    Proxy:=0;
    Part:=2;
    Tail:=1;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FMSEPCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FMSEPCovEigenvalues[PDestDouble],FMSEPCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueTail + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovEigenvalueTail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueTail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueTail + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovEigenvalueTailProxy: string;
var Proxy, Part, Tail, m1, m2: Integer;
    PDestDouble: IndexDouble;
    TmpArray: TReal2DArray;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueTailProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPCovTailProxy in FInitialised) then Result:=InitMSEPCovTailProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the arrays for eigenvalues and eigenvectors
    if Length(FMSEPCovEigenvalues)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvalues,2   *   3   *   2);
    end;
    if Length(FMSEPCovEigenvectors)<=0 then begin
      //                          Proxy    Part    Tail 
      SetLength(FMSEPCovEigenvectors,2   *   3   *   2);
    end;
    // calculate eigenvalues and eigenvectors for the covariance matrix of the ultimate uncertainty for
    // Proxy = 0, Part = 1, Tail = 0
    Proxy:=1;
    Part:=2;
    Tail:=1;
    setlength(TmpArray,FM+1,FM+1);
    // fill the covariance matrix
    PDestDouble:=((Proxy*3+Part)*2+Tail)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        TmpArray[m1,m2]:=FMSEPCov[PDestDouble];
        Inc(PDestDouble,1);
      end;
    end;
    // get eigenvalues and eigenvectors
    PDestDouble:=(Proxy*3+Part)*2+Tail;
    if not SMatrixEVD(TmpArray,FM+1,1,true,FMSEPCovEigenvalues[PDestDouble],FMSEPCovEigenvectors[PDestDouble]) 
        then begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueTailProxy + sLineBreak
              + rsLSRM_AlgorithmDoesNotConverge;
      Exit;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovEigenvalueTailProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueTailProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovEigenvalueTailProxy + sLineBreak 
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovParaErr: string;
var StartTime, EndTime: Int64;
    PDestDouble, PMSEPDetailDouble, Palpha1Double, Palpha2Double: IndexDouble;
    m1, m2, i1, i2, Proxy, Part: Integer;
    TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovParaErr + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovParaErr + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_MSEPParaErrDetail in FInitialised) then Result:=InitMSEPParaErrDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=0;
    Part:=1;
    // initialise array or clear the memory
    if Length(FMSEPCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FMSEPCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // without tail
      ZeroMemory(@FMSEPCov[(Proxy*3+Part)*2*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    {$REGION 'calculate'}
    PMSEPDetailDouble:=(Proxy*2+Part)*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PDestDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          for i2 := 0 to FI do begin
            TmpValue:=FUsedalpha^[Palpha1Double]
                      *FUsedalpha^[Palpha2Double]
                      *FMSEPDetail[PMSEPDetailDouble];
            FMSEPCov[PDestDouble]:=FMSEPCov[PDestDouble]+TmpValue;
            Inc(PMSEPDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
        Inc(PDestDouble,1);
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovParaErr,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovParaErr + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovParaErr + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovParaErrProxy: string;
var StartTime, EndTime: Int64;
    PDestDouble, PMSEPDetailDouble, Palpha1Double, Palpha2Double: IndexDouble;
    m1, m2, i1, i2, Proxy, Part: Integer;
    TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovParaErrProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovParaErrProxy + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_MSEPParaErrProxyDetail in FInitialised) then Result:=InitMSEPParaErrProxyDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=1;
    Part:=1;
    // initialise array or clear the memory
    if Length(FMSEPCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FMSEPCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // without tail
      ZeroMemory(@FMSEPCov[(Proxy*3+Part)*2*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    {$REGION 'calculate'}
    PMSEPDetailDouble:=(Proxy*2+Part)*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PDestDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          for i2 := 0 to FI do begin
            TmpValue:=FUsedalpha^[Palpha1Double]
                      *FUsedalpha^[Palpha2Double]
                      *FMSEPDetail[PMSEPDetailDouble];
            FMSEPCov[PDestDouble]:=FMSEPCov[PDestDouble]+TmpValue;
            Inc(PMSEPDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
        Inc(PDestDouble,1);
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovParaErrProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovParaErrProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovParaErrProxy + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovParaErrTail: string;
var StartTime, EndTime: Int64;
    PNonTailDouble, PDestDouble, PMSEPDetailDouble, Palpha1Double, Palpha2Double: IndexDouble;
    m1, m2, i1, i2, Proxy, Part: Integer;
    TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovParaErrTail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovParaErrTail + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_MSEPCovParaErr in FInitialised) then Result:=InitMSEPCovParaErr();
  if Result<>'' then Exit;
  If not(Init_MSEPParaErrTailDetail in FInitialised) then Result:=InitMSEPParaErrTailDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=0;
    Part:=1;
    // initialise array or clear the memory
    if Length(FMSEPCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FMSEPCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // with tail only
      ZeroMemory(@FMSEPCov[((Proxy*3+Part)*2+1)*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    {$REGION 'calculate'}
    PMSEPDetailDouble:=(Proxy*2+Part)*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PNonTailDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    PDestDouble:=((Proxy*3+Part)*2+1)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        // get the non tail part
        FMSEPCov[PDestDouble]:=FMSEPCov[PNonTailDouble];
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          for i2 := 0 to FI do begin
            // tail
            TmpValue:=FUsedalpha^[Palpha1Double]
                      *FUsedalpha^[Palpha2Double]
                      *FMSEPTailDetail[PMSEPDetailDouble];
            FMSEPCov[PDestDouble]:=FMSEPCov[PDestDouble]+TmpValue;
            Inc(PMSEPDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
        Inc(PNonTailDouble,1);
        Inc(PDestDouble,1);
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovParaErrTail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovParaErrTail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovParaErrTail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovParaErrTailProxy: string;
var StartTime, EndTime: Int64;
    PNonTailDouble, PDestDouble, PMSEPDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2, Proxy, Part: Integer;
    TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovParaErrTailProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovParaErrTailProxy + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_MSEPCovParaErrProxy in FInitialised) then Result:=InitMSEPCovParaErrProxy();
  if Result<>'' then Exit;
  If not(Init_MSEPParaErrTailProxyDetail in FInitialised) then Result:=InitMSEPParaErrTailProxyDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=1;
    Part:=1;
    // initialise array or clear the memory
    if Length(FMSEPCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FMSEPCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // tail only
      ZeroMemory(@FMSEPCov[((Proxy*3+Part)*2+1)*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    {$REGION 'calculate'}
    PMSEPDetailDouble:=(Proxy*2+Part)*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PNonTailDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    PDestDouble:=((Proxy*3+Part)*2+1)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        // get the non tail part
        FMSEPCov[PDestDouble]:=FMSEPCov[PNonTailDouble];
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          for i2 := 0 to FI do begin
            // tail
            TmpValue:=FUsedalpha^[Palpha1Double]
                      *FUsedalpha^[Palpha2Double]
                      *FMSEPTailDetail[PMSEPDetailDouble];
            FMSEPCov[PDestDouble]:=FMSEPCov[PDestDouble]+TmpValue;
            Inc(PMSEPDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
        Inc(PNonTailDouble,1);
        Inc(PDestDouble,1);
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovParaErrTailProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovParaErrTailProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovParaErrTailProxy + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovProcVar: string;
var StartTime, EndTime: Int64;
    PDestDouble, PMSEPDetailDouble, Palpha1Double, Palpha2Double: IndexDouble;
    m1, m2, i1, i2, Proxy, Part: Integer;
    TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovProcVar + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovProcVar + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_MSEPProcVarDetail in FInitialised) then Result:=InitMSEPProcVarDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=0;
    Part:=0;
    // initialise array or clear the memory
    if Length(FMSEPCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FMSEPCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // without tail
      ZeroMemory(@FMSEPCov[(Proxy*3+Part)*2*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    {$REGION 'calculate'}
    PMSEPDetailDouble:=(Proxy*2+Part)*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PDestDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          for i2 := 0 to FI do begin
            TmpValue:=FUsedalpha^[Palpha1Double]
                      *FUsedalpha^[Palpha2Double]
                      *FMSEPDetail[PMSEPDetailDouble];
            FMSEPCov[PDestDouble]:=FMSEPCov[PDestDouble]+TmpValue;
            Inc(PMSEPDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
        Inc(PDestDouble,1);
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovProcVar,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovProcVar + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovProcVar + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovProcVarProxy: string;
var StartTime, EndTime: Int64;
    PDestDouble, PMSEPDetailDouble, Palpha1Double, Palpha2Double: IndexDouble;
    m1, m2, i1, i2, Proxy, Part: Integer;
    TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovProcVarProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovProcVarProxy + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_MSEPProcVarProxyDetail in FInitialised) then Result:=InitMSEPProcVarProxyDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=1;
    Part:=0;
    // initialise array or clear the memory
    if Length(FMSEPCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FMSEPCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // without tail
      ZeroMemory(@FMSEPCov[(Proxy*3+Part)*2*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    {$REGION 'calculate'}
    PMSEPDetailDouble:=(Proxy*2+Part)*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PDestDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          for i2 := 0 to FI do begin
            TmpValue:=FUsedalpha^[Palpha1Double]
                      *FUsedalpha^[Palpha2Double]
                      *FMSEPDetail[PMSEPDetailDouble];
            FMSEPCov[PDestDouble]:=FMSEPCov[PDestDouble]+TmpValue;
            Inc(PMSEPDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
        Inc(PDestDouble,1);
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovProcVarProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovProcVarProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovProcVarProxy + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovProcVarTail: string;
var StartTime, EndTime: Int64;
    PNonTailDouble, PDestDouble, PMSEPDetailDouble, Palpha1Double, Palpha2Double: IndexDouble;
    m1, m2, i1, i2, Proxy, Part: Integer;
    TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovProcVarTail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovProcVarTail + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_MSEPCovProcVar in FInitialised) then Result:=InitMSEPCovProcVar();
  if Result<>'' then Exit;
  If not(Init_MSEPProcVarTailDetail in FInitialised) then Result:=InitMSEPProcVarTailDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=0;
    Part:=0;
    // initialise array or clear the memory
    if Length(FMSEPCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FMSEPCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // tail only
      ZeroMemory(@FMSEPCov[((Proxy*3+Part)*2+1)*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    {$REGION 'calculate'}
    PMSEPDetailDouble:=(Proxy*2+Part)*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PNonTailDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    PDestDouble:=((Proxy*3+Part)*2+1)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        // get the non tail part
        FMSEPCov[PDestDouble]:=FMSEPCov[PNonTailDouble];
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          for i2 := 0 to FI do begin
            // tail
            TmpValue:=FUsedalpha^[Palpha1Double]
                      *FUsedalpha^[Palpha2Double]
                      *FMSEPTailDetail[PMSEPDetailDouble];
            FMSEPCov[PDestDouble]:=FMSEPCov[PDestDouble]+TmpValue;
            Inc(PMSEPDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
        Inc(PNonTailDouble,1);
        Inc(PDestDouble,1);
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovProcVarTail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovProcVarTail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovProcVarTail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovProcVarTailProxy: string;
var StartTime, EndTime: Int64;
    PNonTailDouble, PDestDouble, PMSEPDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2, Proxy, Part: Integer;
    TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovProcVarTailProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovProcVarTailProxy + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_MSEPCovProcVarProxy in FInitialised) then Result:=InitMSEPCovProcVarProxy();
  if Result<>'' then Exit;
  If not(Init_MSEPProcVarTailProxyDetail in FInitialised) then Result:=InitMSEPProcVarTailProxyDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=1;
    Part:=0;
    // initialise array or clear the memory
    if Length(FMSEPCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FMSEPCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // tail only
      ZeroMemory(@FMSEPCov[((Proxy*3+Part)*2+1)*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    {$REGION 'calculate'}
    PMSEPDetailDouble:=(Proxy*2+Part)*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PNonTailDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    PDestDouble:=((Proxy*3+Part)*2+1)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        // get the non tail part
        FMSEPCov[PDestDouble]:=FMSEPCov[PNonTailDouble];
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          for i2 := 0 to FI do begin
            // tail
            TmpValue:=FUsedalpha^[Palpha1Double]
                      *FUsedalpha^[Palpha2Double]
                      *FMSEPTailDetail[PMSEPDetailDouble];
            FMSEPCov[PDestDouble]:=FMSEPCov[PDestDouble]+TmpValue;
            Inc(PMSEPDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
        Inc(PNonTailDouble,1);
        Inc(PDestDouble,1);
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovProcVarTailProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovProcVarTailProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovProcVarTailProxy + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovProxy: string;
var StartTime, EndTime: Int64;
    PDestDouble, PParaErrDouble, PProcVarDouble: IndexDouble;
    m1, m2, Proxy, Part: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPCovParaErrProxy in FInitialised) then Result:=InitMSEPCovParaErrProxy();
  if Result<>'' then Exit;
  If not(Init_MSEPCovProcVarProxy in FInitialised) then Result:=InitMSEPCovProcVarProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=1;
    Part:=2;
    // initialise array or clear the memory
    if Length(FMSEPCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FMSEPCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      // without tail
      ZeroMemory(@FMSEPCov[(Proxy*3+Part)*2*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    PDestDouble:=(Proxy*3+Part)*2*(FM+1)*(FM+1);
    PProcVarDouble:=(Proxy*3+0)*2*(FM+1)*(FM+1);
    PParaErrDouble:=(Proxy*3+1)*2*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        FMSEPCov[PDestDouble]:=FMSEPCov[PProcVarDouble]+FMSEPCov[PParaErrDouble];
        Inc(PDestDouble,1);
        Inc(PProcVarDouble,1);
        Inc(PParaErrDouble,1);
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovProxy + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovTail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PParaErrDouble, PProcVarDouble: IndexDouble;
    m1, m2, Proxy, Part: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovTail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPCovParaErrTail in FInitialised) then Result:=InitMSEPCovParaErrTail();
  if Result<>'' then Exit;
  If not(Init_MSEPCovProcVarTail in FInitialised) then Result:=InitMSEPCovProcVarTail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=0;
    Part:=2;
    // initialise array or clear the memory
    if Length(FMSEPCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FMSEPCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      ZeroMemory(@FMSEPCov[((Proxy*3+Part)*2+1)*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    PDestDouble:=((Proxy*3+Part)*2+1)*(FM+1)*(FM+1);
    PProcVarDouble:=((Proxy*3+0)*2+1)*(FM+1)*(FM+1);
    PParaErrDouble:=((Proxy*3+1)*2+1)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        FMSEPCov[PDestDouble]:=FMSEPCov[PProcVarDouble]+FMSEPCov[PParaErrDouble];
        Inc(PDestDouble,1);
        Inc(PProcVarDouble,1);
        Inc(PParaErrDouble,1);
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovTail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovTail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovTail + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPCovTailProxy: string;
var StartTime, EndTime: Int64;
    PDestDouble, PParaErrDouble, PProcVarDouble: IndexDouble;
    m1, m2, Proxy, Part: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovTailProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPCovParaErrTailProxy in FInitialised) then Result:=InitMSEPCovParaErrTailProxy();
  if Result<>'' then Exit;
  If not(Init_MSEPCovProcVarTailProxy in FInitialised) then Result:=InitMSEPCovProcVarTailProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    Proxy:=1;
    Part:=2;
    // initialise array or clear the memory
    if Length(FMSEPCov)<=0 then begin
      //               Proxy Part  Tail   m1     m2
      SetLength(FMSEPCov,2  *  3  * 2 * (FM+1)*(FM+1));
    end else begin
      ZeroMemory(@FMSEPCov[((Proxy*3+Part)*2+1)*(FM+1)*(FM+1)],(FM+1)*(FM+1)*SizeOf(Double));
    end;
    PDestDouble:=((Proxy*3+Part)*2+1)*(FM+1)*(FM+1);
    PProcVarDouble:=((Proxy*3+0)*2+1)*(FM+1)*(FM+1);
    PParaErrDouble:=((Proxy*3+1)*2+1)*(FM+1)*(FM+1);
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        FMSEPCov[PDestDouble]:=FMSEPCov[PProcVarDouble]+FMSEPCov[PParaErrDouble];
        Inc(PDestDouble,1);
        Inc(PProcVarDouble,1);
        Inc(PParaErrDouble,1);
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPCovTailProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovTailProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPCovTailProxy + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPParaErr: string;
var StartTime, EndTime: Int64;
    PmmiiDouble, PTotalmDouble, PweightediDouble, PTotalweightedDouble, PMSEPDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  if calc_ReduceMemSize in FCalculationFlags then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDR + sLineBreak + rsLSRM_ProxyOnly;
    Exit;
  end;
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErr + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErr + sLineBreak + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_MSEPParaErrDetail in FInitialised) then Result:=InitMSEPParaErrDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise aeeay, if necessary
    if Length(FMSEP)<=0 then begin
      //            Proxy Part    m      i
      SetLength(FMSEP,2  *  3  *(FM+2)*(FI+2));
    end;
    {$REGION 'unweighted estimates'}
    // Proxy=0, part=1
    PMSEPDetailDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PmmiiDouble:=(FM+2)*(FI+2);
    ZeroMemory(@FMSEP[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
    PTotalmDouble:=PmmiiDouble;
    Inc(PTotalmDouble,FI+1);
    for m1:=0 to FM do begin
      for i1 := 0 to FI do begin
        // the unweighted estimate of claim property m and accident period i
        for i2 := 0 to i1-1 do begin
          FMSEP[PTotalmDouble]:=FMSEP[PTotalmDouble]+FMSEPDetail[PMSEPDetailDouble];
          Inc(PMSEPDetailDouble,1);
        end;
        FMSEP[PmmiiDouble]:=FMSEPDetail[PMSEPDetailDouble];
        Inc(PmmiiDouble,1);
        FMSEP[PTotalmDouble]:=FMSEP[PTotalmDouble]+FMSEPDetail[PMSEPDetailDouble];
        Inc(PMSEPDetailDouble,1);
        for i2 := i1+1 to FI do begin
          FMSEP[PTotalmDouble]:=FMSEP[PTotalmDouble]+FMSEPDetail[PMSEPDetailDouble];
          Inc(PMSEPDetailDouble,1);
        end;
      end;
      // next m (m1=m2=m)
      Inc(PMSEPDetailDouble,(FM+1)*(FI+1)*(FI+1));
      Inc(PTotalmDouble,FI+2);
      Inc(PmmiiDouble,1);
    end;
    {$ENDREGION}
    {$REGION 'weighted estimates'}
    // Proxy=0, part=1
    PMSEPDetailDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PTotalweightedDouble:=((FM+2)+(FM+1))*(FI+2)+FI+1;
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          PweightediDouble:=((FM+2)+(FM+1))*(FI+2)+i1;
          for i2 := 0 to FI do begin
            if i1=i2 then begin
              // weighted estimate for accident period i1=i2
              FMSEP[PweightediDouble]:=FMSEP[PweightediDouble]
                                      +Fusedalpha^[Palpha1Double]
                                      *FUsedalpha^[Palpha2Double]
                                      *FMSEPDetail[PMSEPDetailDouble];
            end;
            // total (over all accident periods) weighted estimate
            FMSEP[PTotalweightedDouble]:=FMSEP[PTotalweightedDouble]
                                        +Fusedalpha^[Palpha1Double]
                                        *FUsedalpha^[Palpha2Double]
                                        *FMSEPDetail[PMSEPDetailDouble];
            Inc(PMSEPDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPParaErr,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErr + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErr + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPParaErrDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PSxS0Double, PSxSrhostarDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  if calc_ReduceMemSize in FCalculationFlags then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDR + sLineBreak + rsLSRM_ProxyOnly;
    Exit;
  end;
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_SxSMSEP in FInitialised) then Result:=InitSxSMSEP();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPDetail)<=0 then begin
      //                  Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    // parameter error non proxy
    // store the results within FMSEPDetail(0,1)
    PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
    ZeroMemory(@FMSEPDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
    For m1 := 0 To FM do begin
      For m2 := 0 To FM do begin
        For i1 := 0 To FI do begin
          For i2 := 0 To FI do begin
            For k1 := FI - i1 To FJ - 1 do begin
              PSxS0Double:=((((m1*(FI+1)+i1)*(FJ+1+FP)+k1+1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+FI-i2+1;
              PSxSrhostarDouble:=
                          ((((((FM+1)+m1)*(FI+1)+i1)*(FJ+1+FP)+k1+1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+FI-i2+1;
              For k2 := FI - i2 To FJ - 1 do begin
                FMSEPDetail[PDestDouble]:=FMSEPDetail[PDestDouble]
                                          +FSxS[PSXSrhostarDouble]
                                          -FSxS[PSxS0Double];
                Inc(PSxS0Double,1);
                Inc(PSxSrhostarDouble,1);
              end;
            end;
            Inc(PDestDouble,1);
          end;
        end;
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPParaErrDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPParaErrProxy: string;
var StartTime, EndTime: Int64;
    PmmiiDouble, PTotalmDouble, PweightediDouble, PTotalweightedDouble, PMSEPDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrProxy + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_MSEPParaErrProxyDetail in FInitialised) then Result:=InitMSEPParaErrProxyDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEP)<=0 then begin
      //            Proxy Part    m      i
      SetLength(FMSEP,2  *  3  *(FM+2)*(FI+2));
    end;
    {$REGION 'unweighted estimates'}
    // Proxy=1, part=1
    PMSEPDetailDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PmmiiDouble:=4*(FM+2)*(FI+2);
    ZeroMemory(@FMSEP[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
    PTotalmDouble:=PmmiiDouble;
    Inc(PTotalmDouble,FI+1);
    for m1:=0 to FM do begin
      for i1 := 0 to FI do begin
        // the unweighted estimate of claim property m and accident period i
        for i2 := 0 to i1-1 do begin
          FMSEP[PTotalmDouble]:=FMSEP[PTotalmDouble]+FMSEPDetail[PMSEPDetailDouble];
          Inc(PMSEPDetailDouble,1);
        end;
        FMSEP[PmmiiDouble]:=FMSEPDetail[PMSEPDetailDouble];
        Inc(PmmiiDouble,1);
        FMSEP[PTotalmDouble]:=FMSEP[PTotalmDouble]+FMSEPDetail[PMSEPDetailDouble];
        Inc(PMSEPDetailDouble,1);
        for i2 := i1+1 to FI do begin
          FMSEP[PTotalmDouble]:=FMSEP[PTotalmDouble]+FMSEPDetail[PMSEPDetailDouble];
          Inc(PMSEPDetailDouble,1);
        end;
      end;
      // next m (m1=m2=m)
      Inc(PMSEPDetailDouble,(FM+1)*(FI+1)*(FI+1));
      Inc(PTotalmDouble,FI+2);
      Inc(PmmiiDouble,1);
    end;
    {$ENDREGION}
    {$REGION 'weighted estimates'}
    // Proxy=1, part=1
    PMSEPDetailDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    PTotalweightedDouble:=(4*(FM+2)+(FM+1))*(FI+2)+FI+1;
    for m1 := 0 to FM do begin
      for m2 := 0 to FM do begin
        Palpha1Double:=m1*(FI+1);
        for i1 := 0 to FI do begin
          Palpha2Double:=m2*(FI+1);
          PweightediDouble:=(4*(FM+2)+(FM+1))*(FI+2)+i1;
          for i2 := 0 to FI do begin
            if i1=i2 then begin
              // weighted estimate for accident period i1=i2
              FMSEP[PweightediDouble]:=FMSEP[PweightediDouble]
                                      +FUsedalpha^[Palpha1Double]
                                      *FUsedalpha^[Palpha2Double]
                                      *FMSEPDetail[PMSEPDetailDouble];
            end;
            // total (over all accident periods) weighted estimate
            FMSEP[PTotalweightedDouble]:=FMSEP[PTotalweightedDouble]
                                        +FUsedalpha^[Palpha1Double]
                                        *FUsedalpha^[Palpha2Double]
                                        *FMSEPDetail[PMSEPDetailDouble];
            Inc(PMSEPDetailDouble,1);
            Inc(Palpha2Double,1);
          end;
          Inc(Palpha1Double,1);
        end;
      end;
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPParaErrProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrProxy + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPParaErrTail: string;
var StartTime, EndTime: Int64;
    PmmiiDouble, PTotalmDouble, PweightediDouble, PTotalweightedDouble, PMSEPDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTail + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_MSEPParaErrTailDetail in FInitialised) then Result:=InitMSEPParaErrTailDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPTail)<=0 then begin
      //                Proxy Part    m      i
      SetLength(FMSEPTail,2  *  3  *(FM+2)*(FI+2));
    end;
    if (Init_MSEPParaErrTailProxy in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PmmiiDouble:=4*(FM+2)*(FI+2);
      CopyMemory(@FMSEPTail[(FM+2)*(FI+2)],@FMSEPTail[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the MSEP tail parameter error without proxy'}
      {$REGION 'unweighted estimates'}
      // Proxy=0, part=1
      PMSEPDetailDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PmmiiDouble:=(FM+2)*(FI+2);
      ZeroMemory(@FMSEPTail[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
      PTotalmDouble:=PmmiiDouble;
      Inc(PTotalmDouble,FI+1);
      for m1:=0 to FM do begin
        for i1 := 0 to FI do begin
          // the unweighted estimate of claim property m and accident period i
          for i2 := 0 to i1-1 do begin
            FMSEPTail[PTotalmDouble]:=FMSEPTail[PTotalmDouble]+FMSEPTailDetail[PMSEPDetailDouble];
            Inc(PMSEPDetailDouble,1);
          end;
          FMSEPTail[PmmiiDouble]:=FMSEPTailDetail[PMSEPDetailDouble];
          Inc(PmmiiDouble,1);
          FMSEPTail[PTotalmDouble]:=FMSEPTail[PTotalmDouble]+FMSEPTailDetail[PMSEPDetailDouble];
          Inc(PMSEPDetailDouble,1);
          for i2 := i1+1 to FI do begin
            FMSEPTail[PTotalmDouble]:=FMSEPTail[PTotalmDouble]+FMSEPTailDetail[PMSEPDetailDouble];
            Inc(PMSEPDetailDouble,1);
          end;
        end;
        // next m (m1=m2=m)
        Inc(PMSEPDetailDouble,(FM+1)*(FI+1)*(FI+1));
        Inc(PTotalmDouble,FI+2);
        Inc(PmmiiDouble,1);
      end;
      {$ENDREGION}
      {$REGION 'weighted estimates'}
      // Proxy=0, part=1
      PMSEPDetailDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PTotalweightedDouble:=((FM+2)+(FM+1))*(FI+2)+FI+1;
      for m1 := 0 to FM do begin
        for m2 := 0 to FM do begin
          Palpha1Double:=m1*(FI+1);
          for i1 := 0 to FI do begin
            Palpha2Double:=m2*(FI+1);
            PweightediDouble:=((FM+2)+(FM+1))*(FI+2)+i1;
            for i2 := 0 to FI do begin
              if i1=i2 then begin
                // weighted estimate for accident period i1=i2
                FMSEPTail[PweightediDouble]:=FMSEPTail[PweightediDouble]
                                            +FUsedalpha^[Palpha1Double]
                                            *FUsedalpha^[Palpha2Double]
                                            *FMSEPTailDetail[PMSEPDetailDouble];
              end;
              // total (over all accident periods) weighted estimate
              FMSEPTail[PTotalweightedDouble]:=FMSEPTail[PTotalweightedDouble]
                                              +FUsedalpha^[Palpha1Double]
                                              *FUsedalpha^[Palpha2Double]
                                              *FMSEPTailDetail[PMSEPDetailDouble];
              Inc(PMSEPDetailDouble,1);
              Inc(Palpha2Double,1);
            end;
            Inc(Palpha1Double,1);
          end;
        end;
      end;
      {$ENDREGION}
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPParaErrTail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTail + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPParaErrTailDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhostarDouble, PCoorF1Double, PCoorF2Double,
      PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoMSEP in FInitialised) then
    Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPTailDetail)<=0 then begin
      //                      Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_MSEPParaErrTailProxyDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PProxyDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FMSEPTailDetail[PDestDouble],@FMSEPTailDetail[PProxyDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the MSEP tail parameter error without proxy'}
      // strore the results into FMSEPTailDetail(0,1)
      PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FMSEPTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      if calc_ReduceMemSize in FCalculationFlags then begin
        {$REGION 'calculate rhos each time'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error non proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 +FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        // here we have to take all old accident periods into account, too
                        // therefore, the condition "h1 >= FI - j" has to be ignored
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            // here we have to take all old accident periods into account, too
                            // therefore, the condition "h2 >= FI - j" has to be ignored
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                          end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else begin
        {$REGION 'take rhos from memory'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error non proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        // here we have to take all old accident periods into account, too
                        // therefore, the condition "h1 >= FI - j" has to be ignored
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            // here we have to take all old accident periods into account, too
                            // therefore, the condition "h2 >= FI - j" has to be ignored
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                          end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPParaErrTailDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPParaErrTailProxy: string;
var StartTime, EndTime: Int64;
    PmmiiDouble, PTotalmDouble, PweightediDouble, PTotalweightedDouble, PMSEPDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailProxy + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_MSEPParaErrTailProxyDetail in FInitialised) then Result:=InitMSEPParaErrTailProxyDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPTail)<=0 then begin
      //                Proxy Part    m      i
      SetLength(FMSEPTail,2  *  3  *(FM+2)*(FI+2));
    end;
    if (Init_MSEPParaErrTail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PmmiiDouble:=4*(FM+2)*(FI+2);
      CopyMemory(@FMSEPTail[PmmiiDouble],@FMSEPTail[(FM+2)*(FI+2)],(FM+2)*(FI+2)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the MSEP tail parameter error proxy'}
      {$REGION 'unweighted estimates'}
      // Proxy=1, part=1
      PMSEPDetailDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PmmiiDouble:=4*(FM+2)*(FI+2);
      ZeroMemory(@FMSEPTail[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
      PTotalmDouble:=PmmiiDouble;
      Inc(PTotalmDouble,FI+1);
      for m1:=0 to FM do begin
        for i1 := 0 to FI do begin
          // the unweighted estimate of claim property m and accident period i
          for i2 := 0 to i1-1 do begin
            FMSEPTail[PTotalmDouble]:=FMSEPTail[PTotalmDouble]+FMSEPTailDetail[PMSEPDetailDouble];
            Inc(PMSEPDetailDouble,1);
          end;
          FMSEPTail[PmmiiDouble]:=FMSEPTailDetail[PMSEPDetailDouble];
          Inc(PmmiiDouble,1);
          FMSEPTail[PTotalmDouble]:=FMSEPTail[PTotalmDouble]+FMSEPTailDetail[PMSEPDetailDouble];
          Inc(PMSEPDetailDouble,1);
          for i2 := i1+1 to FI do begin
            FMSEPTail[PTotalmDouble]:=FMSEPTail[PTotalmDouble]+FMSEPTailDetail[PMSEPDetailDouble];
            Inc(PMSEPDetailDouble,1);
          end;
        end;
        // next m (m1=m2=m)
        Inc(PMSEPDetailDouble,(FM+1)*(FI+1)*(FI+1));
        Inc(PTotalmDouble,FI+2);
        Inc(PmmiiDouble,1);
      end;
      {$ENDREGION}
      {$REGION 'weighted estimates'}
      // Proxy=1, part=1
      PMSEPDetailDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PTotalweightedDouble:=(4*(FM+2)+(FM+1))*(FI+2)+FI+1;
      for m1 := 0 to FM do begin
        for m2 := 0 to FM do begin
          Palpha1Double:=m1*(FI+1);
          for i1 := 0 to FI do begin
            Palpha2Double:=m2*(FI+1);
            PweightediDouble:=(4*(FM+2)+(FM+1))*(FI+2)+i1;
            for i2 := 0 to FI do begin
              if i1=i2 then begin
                // weighted estimate for accident period i1=i2
                FMSEPTail[PweightediDouble]:=FMSEPTail[PweightediDouble]
                                            +FUsedalpha^[Palpha1Double]
                                            *FUsedalpha^[Palpha2Double]
                                            *FMSEPTailDetail[PMSEPDetailDouble];
              end;
              // total (over all accident periods) weighted estimate
              FMSEPTail[PTotalweightedDouble]:=FMSEPTail[PTotalweightedDouble]
                                              +FUsedalpha^[Palpha1Double]
                                              *FUsedalpha^[Palpha2Double]
                                              *FMSEPTailDetail[PMSEPDetailDouble];
              Inc(PMSEPDetailDouble,1);
              Inc(Palpha2Double,1);
            end;
            Inc(Palpha1Double,1);
          end;
        end;
      end;
      {$ENDREGION}
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPParaErrTailProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailProxy + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPParaErrProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhostarDouble, PCoorF1Double, PCoorF2Double: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoMSEP in FInitialised) then
    Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPDetail)<=0 then begin
      //                  Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    // parameter error with proxy
    // store the results within FMSEPDetail(1,1)
    PDestDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    ZeroMemory(@FMSEPDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
    if calc_ReduceMemSize in FCalculationFlags then begin
      {$REGION 'calculate rhos each time'}
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                    For ch1 := 0 To FihDependency[i1,FI] do begin
                      h1 := FihDependency[i1,ch1];
                      If h1 >= FI - j Then begin
                        For ch2 := 0 To FihDependency[i2,FI] do begin
                          h2 := FihDependency[i2,ch2];
                          If h2 >= FI - j Then begin
                            For cl1 := 0 To FmlDependency[m1,FM] do begin
                              l1 := FmlDependency[m1,cl1];
                              For cl2 := 0 To FmlDependency[m2,FM] do begin
                                l2 := FmlDependency[m2,cl2];
                                PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                Rho:=Getrho(1,l1,l2,h1,h2,j);
                                TmpValue:=Rho
                                          *FR1[PR11Double]
                                          *FR1[PR12Double]
                                          *FCoorF[PCoorF1Double]
                                          *FCoorF[PCoorF2Double];
                                // diversification only if rho is negative
                                if (calc_VirtualParaErrDiversification in FCalculationFlags)
                                        or (Rho<=0) or (TmpValue>=0) then
                                  FMSEPDetail[PDestDouble]:=FMSEPDetail[PDestDouble]+TmpValue;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end else begin
      {$REGION 'take rhos from memory'}
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                    For ch1 := 0 To FihDependency[i1,FI] do begin
                      h1 := FihDependency[i1,ch1];
                      If h1 >= FI - j Then begin
                        For ch2 := 0 To FihDependency[i2,FI] do begin
                          h2 := FihDependency[i2,ch2];
                          If h2 >= FI - j Then begin
                            For cl1 := 0 To FmlDependency[m1,FM] do begin
                              l1 := FmlDependency[m1,cl1];
                              For cl2 := 0 To FmlDependency[m2,FM] do begin
                                l2 := FmlDependency[m2,cl2];
                                PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                Rho:=Frho[PrhostarDouble];
                                TmpValue:=Rho
                                          *FR1[PR11Double]
                                          *FR1[PR12Double]
                                          *FCoorF[PCoorF1Double]
                                          *FCoorF[PCoorF2Double];
                                // diversification only if rho is negative
                                if (calc_VirtualParaErrDiversification in FCalculationFlags)
                                        or (Rho<=0) or (TmpValue>=0) then
                                  FMSEPDetail[PDestDouble]:=FMSEPDetail[PDestDouble]+TmpValue;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPParaErrProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPParaErrTailProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhostarDouble, PCoorF1Double, PCoorF2Double,
      PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoMSEP in FInitialised) then
    Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPTailDetail)<=0 then begin
      //                      Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_MSEPParaErrTailDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PProxyDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FMSEPTailDetail[PProxyDouble],@FMSEPTailDetail[PDestDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the MSEP tail parameter error proxy'}
      // store the results within FMSEPTailDetail(1,1)
      PDestDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FMSEPTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      if calc_ReduceMemSize in FCalculationFlags then begin
        {$REGION 'calculate rhos each time'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'proc varince proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 +FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'proc varince proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'proc varince proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'proc varince proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        // here we have to take all old accident periods into account, too
                        // therefore, the condition "h1 >= FI - j" has to be ignored
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            // here we have to take all old accident periods into account, too
                            // therefore, the condition "h2 >= FI - j" has to be ignored
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                          end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else begin
        {$REGION 'take rhos from memory'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'proc varince proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'proc varince proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                          ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                          ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'proc varince proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To min(Min(k1, k2),FJ-1) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        If h1 >= FI - j Then begin
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            If h2 >= FI - j Then begin
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                            end;
                          end;
                        end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'proc varince proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      For ch1 := 0 To FihDependency[i1,FI] do begin
                        h1 := FihDependency[i1,ch1];
                        // here we have to take all old accident periods into account, too
                        // therefore, the condition "h1 >= FI - j" has to be ignored
                          For ch2 := 0 To FihDependency[i2,FI] do begin
                            h2 := FihDependency[i2,ch2];
                            // here we have to take all old accident periods into account, too
                            // therefore, the condition "h2 >= FI - j" has to be ignored
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=
                                            ((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=
                                            ((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                          end;
                      end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPParaErrTailProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPProcVar: string;
var StartTime, EndTime: Int64;
    PmmiiDouble, PTotalmDouble, PweightediDouble, PTotalweightedDouble, PMSEPDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVar + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVar + sLineBreak + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_MSEPProcVarDetail in FInitialised) then Result:=InitMSEPProcVarDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEP)<=0 then begin
      //            Proxy Part    m      i
      SetLength(FMSEP,2  *  3  *(FM+2)*(FI+2));
    end;
    if (Init_MSEPProcVarProxy in FInitialised)
        and (calc_VirtualProcVarDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PmmiiDouble:=3*(FM+2)*(FI+2);
      CopyMemory(@FMSEP[0],@FMSEP[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'unweighted estimates'}
      // Proxy=0, part=0
      PMSEPDetailDouble:=0;
      PmmiiDouble:=0;
      ZeroMemory(@FMSEP[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
      PTotalmDouble:=PmmiiDouble;
      Inc(PTotalmDouble,FI+1);
      for m1:=0 to FM do begin
        for i1 := 0 to FI do begin
          // the unweighted estimate of claim property m1 and accident period i1
          for i2 := 0 to i1-1 do begin
            FMSEP[PTotalmDouble]:=FMSEP[PTotalmDouble]+FMSEPDetail[PMSEPDetailDouble];
            Inc(PMSEPDetailDouble,1);
          end;
          FMSEP[PmmiiDouble]:=FMSEPDetail[PMSEPDetailDouble];
          Inc(PmmiiDouble,1);
          FMSEP[PTotalmDouble]:=FMSEP[PTotalmDouble]+FMSEPDetail[PMSEPDetailDouble];
          Inc(PMSEPDetailDouble,1);
          for i2 := i1+1 to FI do begin
            FMSEP[PTotalmDouble]:=FMSEP[PTotalmDouble]+FMSEPDetail[PMSEPDetailDouble];
            Inc(PMSEPDetailDouble,1);
          end;
        end;
        // next m (m1=m2=m)
        Inc(PMSEPDetailDouble,(FM+1)*(FI+1)*(FI+1));
        Inc(PTotalmDouble,FI+2);
        Inc(PmmiiDouble,1);
      end;
      {$ENDREGION}
      {$REGION 'weighted estimates'}
      // Proxy=0, part=0
      PMSEPDetailDouble:=0;
      PTotalweightedDouble:=(FM+1)*(FI+2)+FI+1;
      for m1 := 0 to FM do begin
        for m2 := 0 to FM do begin
          Palpha1Double:=m1*(FI+1);
          for i1 := 0 to FI do begin
            Palpha2Double:=m2*(FI+1);
            PweightediDouble:=(FM+1)*(FI+2)+i1;
            for i2 := 0 to FI do begin
              if i1=i2 then begin
                // weighted estimate for accident period i1=i2
                FMSEP[PweightediDouble]:=FMSEP[PweightediDouble]
                                        +FUsedalpha^[Palpha1Double]
                                        *FUsedalpha^[Palpha2Double]
                                        *FMSEPDetail[PMSEPDetailDouble];
              end;
              // total (over all accident periods) weighted estimate
              FMSEP[PTotalweightedDouble]:=FMSEP[PTotalweightedDouble]
                                          +FUsedalpha^[Palpha1Double]
                                          *FUsedalpha^[Palpha2Double]
                                          *FMSEPDetail[PMSEPDetailDouble];
              Inc(PMSEPDetailDouble,1);
              Inc(Palpha2Double,1);
            end;
            Inc(Palpha1Double,1);
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPProcVar,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVar + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVar + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPProcVarTail: string;
var StartTime, EndTime: Int64;
    PmmiiDouble, PTotalmDouble, PweightediDouble, PTotalweightedDouble, PMSEPDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTail + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_MSEPProcVarTailDetail in FInitialised) then Result:=InitMSEPProcVarTailDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPTail)<=0 then begin
      //                Proxy Part    m      i
      SetLength(FMSEPTail,2  *  3  *(FM+2)*(FI+2));
    end;
    if (Init_MSEPProcVarTailProxy in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PmmiiDouble:=3*(FM+2)*(FI+2);
      CopyMemory(@FMSEPTail[0],@FMSEPTail[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'unweighted estimates'}
      // Proxy=0, part=0
      PMSEPDetailDouble:=0;
      PmmiiDouble:=0;
      ZeroMemory(@FMSEPTail[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
      PTotalmDouble:=PmmiiDouble;
      Inc(PTotalmDouble,FI+1);
      for m1:=0 to FM do begin
        for i1 := 0 to FI do begin
          // the unweighted estimate of claim property m1 and accident period i1
          for i2 := 0 to i1-1 do begin
            FMSEPTail[PTotalmDouble]:=FMSEPTail[PTotalmDouble]+FMSEPTailDetail[PMSEPDetailDouble];
            Inc(PMSEPDetailDouble,1);
          end;
          FMSEPTail[PmmiiDouble]:=FMSEPTailDetail[PMSEPDetailDouble];
          Inc(PmmiiDouble,1);
          FMSEPTail[PTotalmDouble]:=FMSEPTail[PTotalmDouble]+FMSEPTailDetail[PMSEPDetailDouble];
          Inc(PMSEPDetailDouble,1);
          for i2 := i1+1 to FI do begin
            FMSEPTail[PTotalmDouble]:=FMSEPTail[PTotalmDouble]+FMSEPTailDetail[PMSEPDetailDouble];
            Inc(PMSEPDetailDouble,1);
          end;
        end;
        // next m (m1=m2=m)
        Inc(PMSEPDetailDouble,(FM+1)*(FI+1)*(FI+1));
        Inc(PTotalmDouble,FI+2);
        Inc(PmmiiDouble,1);
      end;
      {$ENDREGION}
      {$REGION 'weighted estimates'}
      // Proxy=0, part=0
      PMSEPDetailDouble:=0;
      PTotalweightedDouble:=(FM+1)*(FI+2)+FI+1;
      for m1 := 0 to FM do begin
        for m2 := 0 to FM do begin
          Palpha1Double:=m1*(FI+1);
          for i1 := 0 to FI do begin
            Palpha2Double:=m2*(FI+1);
            PweightediDouble:=(FM+1)*(FI+2)+i1;
            for i2 := 0 to FI do begin
              if i1=i2 then begin
                // weighted estimate for accident period i1=i2
                FMSEPTail[PweightediDouble]:=FMSEPTail[PweightediDouble]
                                            +FUsedalpha^[Palpha1Double]
                                            *FUsedalpha^[Palpha2Double]
                                            *FMSEPTailDetail[PMSEPDetailDouble];
              end;
              // total (over all accident periods) weighted estimate
              FMSEPTail[PTotalweightedDouble]:=FMSEPTail[PTotalweightedDouble]
                                              +FUsedalpha^[Palpha1Double]
                                              *FUsedalpha^[Palpha2Double]
                                              *FMSEPTailDetail[PMSEPDetailDouble];
              Inc(PMSEPDetailDouble,1);
              Inc(Palpha2Double,1);
            end;
            Inc(Palpha1Double,1);
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPProcVarTail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTail + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPProcVarDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PsigmaDouble, PR2Double, PCoorF1Double, PCoorF2Double, PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, h, j, cl1, l1, cl2, l2: Integer;
    sigma, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_sigma, Init_Defaultsigma]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarDetail + sLineBreak
            + rsLSRM_NotInitialisedsigma;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  If not(Init_R2Full in FInitialised) then Result:=InitR2Full();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPDetail)<=0 then begin
      //                  Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_MSEPProcVarProxyDetail in FInitialised)
        and (calc_VirtualProcVarDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=0;
      PProxyDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FMSEPDetail[PDestDouble],@FMSEPDetail[PProxyDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the MSEP process variance without proxy'}
      // store the results within FMSEPDetail(0,0)
      PDestDouble:=0;
      ZeroMemory(@FMSEPDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := 0 To FJ-1 do begin
                    For h := FI-j To Min(Min(i1 + k1, i2 + k2)-j,FI) do begin
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          FMSEPDetail[PDestDouble]:=FMSEPDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPProcVarDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPProcVarProxy: string;
var StartTime, EndTime: Int64;
    PmmiiDouble, PTotalmDouble, PweightediDouble, PTotalweightedDouble, PMSEPDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarProxy + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_MSEPProcVarProxyDetail in FInitialised) then Result:=InitMSEPProcVarProxyDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEP)<=0 then begin
      //            Proxy Part    m      i
      SetLength(FMSEP,2  *  3  *(FM+2)*(FI+2));
    end;
    if (Init_MSEPProcVar in FInitialised)
        and (calc_VirtualProcVarDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PmmiiDouble:=3*(FM+2)*(FI+2);
      CopyMemory(@FMSEP[PmmiiDouble],@FMSEP[0],(FM+2)*(FI+2)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'unweighted estimates'}
      // Proxy=1, part=0
      PMSEPDetailDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PmmiiDouble:=3*(FM+2)*(FI+2);
      ZeroMemory(@FMSEP[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
      PTotalmDouble:=PmmiiDouble;
      Inc(PTotalmDouble,FI+1);
      for m1:=0 to FM do begin
        for i1 := 0 to FI do begin
          // the unweighted estimate of claim property m1 and accident period i1
          for i2 := 0 to i1-1 do begin
            FMSEP[PTotalmDouble]:=FMSEP[PTotalmDouble]+FMSEPDetail[PMSEPDetailDouble];
            Inc(PMSEPDetailDouble,1);
          end;
          FMSEP[PmmiiDouble]:=FMSEPDetail[PMSEPDetailDouble];
          Inc(PmmiiDouble,1);
          FMSEP[PTotalmDouble]:=FMSEP[PTotalmDouble]+FMSEPDetail[PMSEPDetailDouble];
          Inc(PMSEPDetailDouble,1);
          for i2 := i1+1 to FI do begin
            FMSEP[PTotalmDouble]:=FMSEP[PTotalmDouble]+FMSEPDetail[PMSEPDetailDouble];
            Inc(PMSEPDetailDouble,1);
          end;
        end;
        // next m (m1=m2=m)
        Inc(PMSEPDetailDouble,(FM+1)*(FI+1)*(FI+1));
        Inc(PTotalmDouble,FI+2);
        Inc(PmmiiDouble,1);
      end;
      {$ENDREGION}
      {$REGION 'weighted estimates'}
      // Proxy=1, part=0
      PMSEPDetailDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PTotalweightedDouble:=(3*(FM+2)+(FM+1))*(FI+2)+FI+1;
      for m1 := 0 to FM do begin
        for m2 := 0 to FM do begin
          Palpha1Double:=m1*(FI+1);
          for i1 := 0 to FI do begin
            Palpha2Double:=m2*(FI+1);
            PweightediDouble:=(3*(FM+2)+(FM+1))*(FI+2)+i1;
            for i2 := 0 to FI do begin
              if i1=i2 then begin
                // weighted estimate for accident period i1=i2
                FMSEP[PweightediDouble]:=FMSEP[PweightediDouble]
                                        +FUsedalpha^[Palpha1Double]
                                        *FUsedalpha^[Palpha2Double]
                                        *FMSEPDetail[PMSEPDetailDouble];
              end;
              // total (over all accident periods) weighted estimate
              FMSEP[PTotalweightedDouble]:=FMSEP[PTotalweightedDouble]
                                          +FUsedalpha^[Palpha1Double]
                                          *FUsedalpha^[Palpha2Double]
                                          *FMSEPDetail[PMSEPDetailDouble];
              Inc(PMSEPDetailDouble,1);
              Inc(Palpha2Double,1);
            end;
            Inc(Palpha1Double,1);
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPProcVarProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarProxy + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPProcVarProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PsigmaDouble, PR2Double, PCoorF1Double, PCoorF2Double, PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, h, j, cl1, l1, cl2, l2: Integer;
    sigma, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_sigma, Init_Defaultsigma]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedsigma;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  If not(Init_R2Full in FInitialised) then Result:=InitR2Full();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPDetail)<=0 then begin
      //                  Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_MSEPProcVarDetail in FInitialised)
        and (calc_VirtualProcVarDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=0;
      PProxyDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FMSEPDetail[PProxyDouble],@FMSEPDetail[PDestDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the MSEP process variance proxy'}
      // store the results within FMSEPDetail(1,0)
      PDestDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FMSEPDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := 0 To FJ-1 do begin
                    For h := FI-j To Min(Min(i1 + k1, i2 + k2)-j,FI) do begin
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          // diversification only if sigma is negative
                          if (calc_VirtualProcVarDiversification in FCalculationFlags)
                                  or (sigma<=0) or (TmpValue>=0) then
                            FMSEPDetail[PDestDouble]:=FMSEPDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPProcVarProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPProcVarTailDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PsigmaDouble, PR2Double, PCoorF1Double, PCoorF2Double, PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, h, j, cl1, l1, cl2, l2: Integer;
    sigma, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_sigma, Init_Defaultsigma]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailDetail + sLineBreak
            + rsLSRM_NotInitialisedsigma;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  If not(Init_R2Full in FInitialised) then Result:=InitR2Full();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPTailDetail)<=0 then begin
      //                      Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_MSEPProcVarTailProxyDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=0;
      PProxyDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FMSEPTailDetail[PDestDouble],@FMSEPTailDetail[PProxyDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the MSEP tail process variance without proxy'}
      // store the results within FMSEPTailDetail(0,0)
      PDestDouble:=0;
      ZeroMemory(@FMSEPTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              {$REGION 'proc varince non proxy: j Tail, k1 and k2 not'}
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FJ To FJ-1+FP do begin
                    For h := FI-j To Min(Min(i1 + k1, i2 + k2)-j,FI) do begin
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'proc varince non proxy: k1 Tail, k2 and j not'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := 0 To FJ-1 do begin
                    For h := FI-j To Min(Min(i1 + k1, i2 + k2)-j,FI) do begin
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'proc varince non proxy: k2 Tail, k1 and j not'}
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := 0 To FJ-1 do begin
                    For h := FI-j To Min(Min(i1 + k1, i2 + k2)-j,FI) do begin
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'proc varince non proxy: k1, k2 Tail and j not'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := 0 To FJ-1 do begin
                    For h := FI-j To Min(Min(i1 + k1, i2 + k2)-j,FI) do begin
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'proc varince non proxy: k1, k2 and j Tail'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FJ To FJ-1+FP do begin
                    // here we are in the 'total' tail, that means we have to take all old accident periods
                    // into account, too
                    For h := 0 To Min(Min(i1 + k1, i2 + k2)-j,FI) do begin
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'proc varince non proxy: k1 and j Tail, k2 not'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FJ To FJ-1+FP do begin
                    For h := FI-j To Min(Min(i1 + k1, i2 + k2)-j,FI) do begin
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'proc varince non proxy: k2 and j Tail, k1 not'}
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FJ To FJ-1+FP do begin
                    For h := FI-j To Min(Min(i1 + k1, i2 + k2)-j,FI) do begin
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              {$ENDREGION}
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPProcVarTailDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPProcVarTailProxy: string;
var StartTime, EndTime: Int64;
    PmmiiDouble, PTotalmDouble, PweightediDouble, PTotalweightedDouble, PMSEPDetailDouble, Palpha1Double,
      Palpha2Double: IndexDouble;
    m1, m2, i1, i2: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailProxy + sLineBreak
            + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_MSEPProcVarTailProxyDetail in FInitialised) then Result:=InitMSEPProcVarTailProxyDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPTail)<=0 then begin
      //                 Proxy Part    m      i
      SetLength(FMSEPTail,2  *  3  *(FM+2)*(FI+2));
    end;
    if (Init_MSEPProcVarTail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PmmiiDouble:=3*(FM+2)*(FI+2);
      CopyMemory(@FMSEPTail[PmmiiDouble],@FMSEPTail[0],(FM+2)*(FI+2)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'unweighted estimates'}
      // Proxy=1, part=0
      PMSEPDetailDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PmmiiDouble:=3*(FM+2)*(FI+2);
      ZeroMemory(@FMSEPTail[PmmiiDouble],(FM+2)*(FI+2)*SizeOf(Double));
      PTotalmDouble:=PmmiiDouble;
      Inc(PTotalmDouble,FI+1);
      for m1:=0 to FM do begin
        for i1 := 0 to FI do begin
          // the unweighted estimate of claim property m1 and accident period i1
          for i2 := 0 to i1-1 do begin
            FMSEPTail[PTotalmDouble]:=FMSEPTail[PTotalmDouble]+FMSEPTailDetail[PMSEPDetailDouble];
            Inc(PMSEPDetailDouble,1);
          end;
          FMSEPTail[PmmiiDouble]:=FMSEPTailDetail[PMSEPDetailDouble];
          Inc(PmmiiDouble,1);
          FMSEPTail[PTotalmDouble]:=FMSEPTail[PTotalmDouble]+FMSEPTailDetail[PMSEPDetailDouble];
          Inc(PMSEPDetailDouble,1);
          for i2 := i1+1 to FI do begin
            FMSEPTail[PTotalmDouble]:=FMSEPTail[PTotalmDouble]+FMSEPTailDetail[PMSEPDetailDouble];
            Inc(PMSEPDetailDouble,1);
          end;
        end;
        // next m (m1=m2=m)
        Inc(PMSEPDetailDouble,(FM+1)*(FI+1)*(FI+1));
        Inc(PTotalmDouble,FI+2);
        Inc(PmmiiDouble,1);
      end;
      {$ENDREGION}
      {$REGION 'weighted estimates'}
      // Proxy=1, part=0
      PMSEPDetailDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PTotalweightedDouble:=(3*(FM+2)+(FM+1))*(FI+2)+FI+1;
      for m1 := 0 to FM do begin
        for m2 := 0 to FM do begin
          Palpha1Double:=m1*(FI+1);
          for i1 := 0 to FI do begin
            Palpha2Double:=m2*(FI+1);
            PweightediDouble:=(3*(FM+2)+(FM+1))*(FI+2)+i1;
            for i2 := 0 to FI do begin
              if i1=i2 then begin
                // weighted estimate for accident period i1=i2
                FMSEPTail[PweightediDouble]:=FMSEPTail[PweightediDouble]
                                            +FUsedalpha^[Palpha1Double]
                                            *FUsedalpha^[Palpha2Double]
                                            *FMSEPTailDetail[PMSEPDetailDouble];
              end;
              // total (over all accident periods) weighted estimate
              FMSEPTail[PTotalweightedDouble]:=FMSEPTail[PTotalweightedDouble]
                                              +FUsedalpha^[Palpha1Double]
                                              *FUsedalpha^[Palpha2Double]
                                              *FMSEPTailDetail[PMSEPDetailDouble];
              Inc(PMSEPDetailDouble,1);
              Inc(Palpha2Double,1);
            end;
            Inc(Palpha1Double,1);
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPProcVarTailProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailProxy + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPProcVarTailProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PsigmaDouble, PR2Double, PCoorF1Double, PCoorF2Double, PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, h, j, cl1, l1, cl2, l2: Integer;
    sigma, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_sigma, Init_Defaultsigma]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedsigma;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  If not(Init_R2Full in FInitialised) then Result:=InitR2Full();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPTailDetail)<=0 then begin
      //                      Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_MSEPProcVarTailDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=0;
      PProxyDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FMSEPTailDetail[PProxyDouble],@FMSEPTailDetail[PDestDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the MSEP tail process variation proxy'}
      // store the results within FMSEPTailDetail(1,0)
      PDestDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FMSEPTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              {$REGION 'process variance proxy: j Tail, k1 and k2 not'}
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FJ To FJ-1+FP do begin
                    For h := FI-j To Min(Min(i1 + k1, i2 + k2)-j,FI) do begin
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          // diversification only if sigma is negative
                          if (calc_VirtualTailDiversification in FCalculationFlags)
                                  or (sigma<=0) or (TmpValue>=0) then
                            FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'process variance proxy: k1 Tail, k2 and j not'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := 0 To FJ-1 do begin
                    For h := FI-j To Min(Min(i1 + k1, i2 + k2)-j,FI) do begin
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          // diversification only if sigma is negative
                          if (calc_VirtualTailDiversification in FCalculationFlags)
                                  or (sigma<=0) or (TmpValue>=0) then
                            FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'process variance proxy: k2 Tail, k1 and j not'}
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := 0 To FJ-1 do begin
                    For h := FI-j To Min(Min(i1 + k1, i2 + k2)-j,FI) do begin
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          // diversification only if sigma is negative
                          if (calc_VirtualTailDiversification in FCalculationFlags)
                                  or (sigma<=0) or (TmpValue>=0) then
                            FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'process variance proxy: k1 and k2 Tail, j not'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := 0 To FJ-1 do begin
                    For h := FI-j To Min(Min(i1 + k1, i2 + k2)-j,FI) do begin
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          // diversification only if sigma is negative
                          if (calc_VirtualTailDiversification in FCalculationFlags)
                                  or (sigma<=0) or (TmpValue>=0) then
                            FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'process variance proxy: k1, k2 and j Tail'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FJ To FJ-1+FP do begin
                    // here we are in the 'total' tail, that means we have to take all old accident periods
                    // into account, too
                    For h := 0 To Min(Min(i1 + k1, i2 + k2)-j,FI) do begin
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          // diversification only if sigma is negative
                          if (calc_VirtualTailDiversification in FCalculationFlags)
                                  or (sigma<=0) or (TmpValue>=0) then
                            FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'process variance proxy: k1 and j Tail, k2 not'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FJ To FJ-1+FP do begin
                    For h := FI-j To Min(Min(i1 + k1, i2 + k2)-j,FI) do begin
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          // diversification only if sigma is negative
                          if (calc_VirtualTailDiversification in FCalculationFlags)
                                  or (sigma<=0) or (TmpValue>=0) then
                            FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'process variance proxy: k2 and j Tail, k1 not'}
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FJ To FJ-1+FP do begin
                    For h := FI-j To Min(Min(i1 + k1, i2 + k2)-j,FI) do begin
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FI+1)+h)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          // diversification only if sigma is negative
                          if (calc_VirtualTailDiversification in FCalculationFlags)
                                  or (sigma<=0) or (TmpValue>=0) then
                            FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                    end;
                  end;
                end;
              end;
              {$ENDREGION}
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPProcVarTailProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPProxy: string;
var StartTime, EndTime: Int64;
    PProcVarDouble, PParaErrDouble, PTotalDouble: IndexDouble;
    m, i: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPParaErrProxy in FInitialised) then Result:=InitMSEPParaErrProxy();
  if Result<>'' then Exit;
  If not(Init_MSEPProcVarProxy in FInitialised) then Result:=InitMSEPProcVarProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // calculate the sum of the process variance and the parameter error (with proxy)
    // and store the result into FMSEP(1,2)
    PProcVarDouble:=3*(FM+2)*(FI+2);
    PParaErrDouble:=PProcVarDouble;
    Inc(PParaErrDouble,(FM+2)*(FI+2));
    PTotalDouble:=PParaErrDouble;
    Inc(PTotalDouble,(FM+2)*(FI+2));
    for m := 0 to FM+1 do begin
      for i := 0 to FI+1 do begin
        FMSEP[PTotalDouble]:=FMSEP[PProcVarDouble]+FMSEP[PParaErrDouble];
        Inc(PTotalDouble,1);
        Inc(PProcVarDouble,1);
        Inc(PParaErrDouble,1);
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProxy + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPTail: string;
var StartTime, EndTime: Int64;
    PProcVarDouble, PParaErrDouble, PTotalDouble: IndexDouble;
    m, i: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPTail + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPParaErrTail in FInitialised) then Result:=InitMSEPParaErrTail();
  if Result<>'' then Exit;
  If not(Init_MSEPProcVarTail in FInitialised) then Result:=InitMSEPProcVarTail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // calculate the sum of the tail process variance and the tail parameter error (without proxy)
    // and store the result into FMSEP(0,2)
    PProcVarDouble:=0;
    PParaErrDouble:=PProcVarDouble;
    Inc(PParaErrDouble,(FM+2)*(FI+2));
    PTotalDouble:=PParaErrDouble;
    Inc(PTotalDouble,(FM+2)*(FI+2));
    for m := 0 to FM+1 do begin
      for i := 0 to FI+1 do begin
        FMSEPTail[PTotalDouble]:=FMSEPTail[PProcVarDouble]+FMSEPTail[PParaErrDouble];
        Inc(PTotalDouble,1);
        Inc(PProcVarDouble,1);
        Inc(PParaErrDouble,1);
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPTail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPTail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPTail + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitMSEPTailProxy: string;
var StartTime, EndTime: Int64;
    PProcVarDouble, PParaErrDouble, PTotalDouble: IndexDouble;
    m, i: Integer;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPTailProxy + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_MSEPParaErrTailProxy in FInitialised) then Result:=InitMSEPParaErrTailProxy();
  if Result<>'' then Exit;
  If not(Init_MSEPProcVarTailProxy in FInitialised) then Result:=InitMSEPProcVarTailProxy();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // calculate the sum of the tail process variance and the tail parameter error (with proxy)
    // and store the result within FMSEP(1,2)
    PProcVarDouble:=3*(FM+2)*(FI+2);
    PParaErrDouble:=PProcVarDouble;
    Inc(PParaErrDouble,(FM+2)*(FI+2));
    PTotalDouble:=PParaErrDouble;
    Inc(PTotalDouble,(FM+2)*(FI+2));
    for m := 0 to FM+1 do begin
      for i := 0 to FI+1 do begin
        FMSEPTail[PTotalDouble]:=FMSEPTail[PProcVarDouble]+FMSEPTail[PParaErrDouble];
        Inc(PTotalDouble,1);
        Inc(PProcVarDouble,1);
        Inc(PParaErrDouble,1);
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPTailProxy,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPTailProxy + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPTailProxy + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitR1: String;
var m, k, i, l, h, j, cl, ch: Integer;
    PDestDouble, Pgamma1Double, PSDouble: IndexDouble;
    StartTime, EndTime: Int64;
    sign:Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR1 + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_S in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR1 + sLineBreak + rsLSRM_NotInitialisedS;
    Exit;
  end;
  If not(Init_gamma1 in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR1 + sLineBreak + rsLSRM_NotInitialisedgamma1;
    Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array or clear the corresponding memory
    if Length(FR1)<=0 then begin
      //               m     i     k
      SetLength(FR1,(FM+1)*(FI+1)*(FJ+FK));
    end else begin
      ZeroMemory(@FR1[0],(FM+1)*(FI+1)*(FJ+FK)*SizeOf(Double));
    end;
    // calculate
    PDestDouble:=0;
    for m:=0 to FM do begin
      for i:=0 to FI do begin
        sign:=0;
        for k := 0 To Min(FJ - 1 + FP, FI-i) do begin
          for cl:=0 to FmlDependency[m,FM] do begin
            l:=FmlDependency[m,cl];
            for ch:=0 to FihDependency[i,FI] do begin
              h:=FihDependency[i,ch];
              PGamma1Double:=((((m*(FM+1)+l)*(FI+1)+i)*(FJ+FK)+k)*(FI+1)+h)*(FJ+FK);
              PSDouble:=(l*(FI+1)+h)*(FJ+1+FK);
              for j:=0 to Min(i + k - h, k) do begin
                FR1[PDestDouble]:=FR1[PDestDouble]+FGamma1[PGamma1Double]*FS[PSDouble];
                Inc(PGamma1Double,1);
                Inc(PSDouble,1);
              end;
            end;
          end;
          // prevent exposure from changing sign
          if calc_PreventPastExposureFromChangingSign in FCalculationFlags then begin
            if sign=0 then begin
              sign:=FR1[PDestDouble];
            end else begin
              if ((sign>0) and (FR1[PDestDouble]<0)) or ((sign<0) and (FR1[PDestDouble]>0)) then
                FR1[PDestDouble]:=0;
            end;
          end;
          Inc(PDestDouble,1);
        end;
        Inc(PDestDouble,FJ-1+FK-Min(FJ - 1 + FP, FI-i));
     end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_R1,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR1 + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR1 + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitR2: String;
var m1, m2, k, i, l, h, j: Integer;
    PDestDouble, Pgamma2Double, PSDouble: IndexDouble;
    StartTime, EndTime: Int64;
    sign: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2 + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_S in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2 + sLineBreak + rsLSRM_NotInitialisedS;
    Exit;
  end;
  If not(Init_gamma2 in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2 + sLineBreak + rsLSRM_NotInitialisedgamma2;
    Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise the array or clear the corresponding memory
    if Length(FR2)<=0 then begin
      //               m1    m2     i       k
      SetLength(FR2,(FM+1)*(FM+1)*(FI+1)*(FJ+FP));
    end else begin
      ZeroMemory(@FR2[0],(FM+1)*(FM+1)*(FI+1)*(FJ+FP)*SizeOf(Double));
    end;
    // calculate
    PDestDouble:=0;
    for m1:=0 to FM do begin
      for m2:=0 to FM do begin
        for i:=0 to FI do begin
          sign:=0;
          for k := 0 To Min(FJ - 1 + FP, FI-i) do begin
            for l:=0 to FM do begin
              for h:=0 to FI do begin
                PGamma2Double:=(((((m1*(FM+1)+m2)*(FM+1)+l)*(FI+1)+i)*(FJ+FP)+k)*(FI+1)+h)*(FJ+FP);
                PSDouble:=(l*(FI+1)+h)*(FJ+1+FK);
                for j:=0 to Min(i + k - h, k) do begin
                  FR2[PDestDouble]:=FR2[PDestDouble]+FGamma2[PGamma2Double]*FS[PSDouble];
                  Inc(PGamma2Double);
                  Inc(PSDouble);
                end;
              end;
            end;
            // prevent exposure from changing sign
            if calc_PreventPastExposureFromChangingSign in FCalculationFlags then begin
              if sign=0 then begin
                sign:=FR2[PDestDouble];
              end else begin
                if ((sign>0) and (FR2[PDestDouble]<0)) or ((sign<0) and (FR2[PDestDouble]>0)) then
                  FR2[PDestDouble]:=0;
              end;
            end;
            Inc(PDestDouble,1);
          end;
          Inc(PDestDouble,FJ-1+FP-Min(FJ - 1 + FP, FI-i));
        end;
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_R2,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2 + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2 + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitR2Full: String;
var m1, m2, k, i, l, h, j: Integer;
    PDestDouble, Pgamma2Double, PSDouble: IndexDouble;
    StartTime, EndTime: Int64;
    sign: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2Full + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_gamma2 in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2Full + sLineBreak + rsLSRM_NotInitialisedgamma2;
    Exit;
  end;
  If not(Init_R2 in FInitialised) then Result:=InitR2();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // calculate
    PDestDouble:=0;
    for m1:=0 to FM do begin
      for m2:=0 to FM do begin
        for i:=0 to FI do begin
          sign:=0;
          if calc_PreventFutureExposureFromChangingSign in FCalculationFlags then begin
            for k := 0 to Min(FJ - 1 + FP, FI-i) do begin
              if sign=0 then sign:=FR2[PDestDouble];
              Inc(PDestDouble,1);
            end;
          end else begin
            Inc(PDestDouble,Min(FJ - 1 + FP, FI-i)+1);
          end;
          for k := Min(FJ - 1 + FP, FI-i)+1 To FJ-1+FP do begin
            FR2[PDestDouble]:=0;
            for l:=0 to FM do begin
              for h:=0 to FI do begin
                PGamma2Double:=(((((m1*(FM+1)+m2)*(FM+1)+l)*(FI+1)+i)*(FJ+FP)+k)*(FI+1)+h)*(FJ+FP);
                PSDouble:=(l*(FI+1)+h)*(FJ+1+FK);
                for j:=0 to Min(i + k - h, k) do begin
                  FR2[PDestDouble]:=FR2[PDestDouble]+FGamma2[PGamma2Double]*FS[PSDouble];
                  Inc(PGamma2Double,1);
                  Inc(PSDouble,1);
                end;
              end;
            end;
            // prevent exposure from changing sign
            if calc_PreventFutureExposureFromChangingSign in FCalculationFlags then begin
              if sign=0 then begin
                sign:=FR2[PDestDouble];
              end else begin
                if ((sign>0) and (FR2[PDestDouble]<0)) or ((sign<0) and (FR2[PDestDouble]>0)) then
                  FR2[PDestDouble]:=0;
              end;
            end;
            Inc(PDestDouble,1);
          end;
        end;
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_R2Full,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2Full + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2Full + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitrhoCDR: String;
var m1, m2, k, i1, i2: Integer;
    PrhostarDouble, Prhostar1Double, Prhostar2Double, Prhostar12Double, PrhobarDouble,
      Pw1Double, Pw2Double, PsigmaDouble, PR2Double, PR11Double, PR12Double: IndexDouble;
    StartTime, EndTime: Int64;
    TmpDouble: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  if calc_ReduceMemSize in FCalculationFlags then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDR + sLineBreak + rsLSRM_ProxyOnly;
    Exit;
  end;
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitrhoCDR + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_w, Init_Defaultw]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitrhoCDR + sLineBreak + rsLSRM_NotInitialisedw;
    Exit;
  end;
  if ([Init_sigma, Init_Defaultsigma]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitrhoCDR + sLineBreak + rsLSRM_NotInitialisedsigma;
    Exit;
  end;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  If not(Init_R2Full in FInitialised) then Result:=InitR2Full();
  if Result<>'' then Exit;
  If not(Init_rhoMSEP in FInitialised) then Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // calculate
    PrhostarDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP);
    Prhostar1Double:=PrhostarDouble;
    Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
    Prhostar2Double:=Prhostar1Double;
    Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
    Prhostar12Double:=Prhostar2Double;
    Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
    PrhobarDouble:=Prhostar12Double;
    Inc(PrhobarDouble,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
    For m1 := 0 To FM do begin
      For m2 := 0 To FM do begin
        For i1 := 0 To FI do begin
          For i2 := 0 To FI do begin
            For k := 0 To FJ - 1 + FP do begin
              Pw1Double:=((2*(FM+1)+m1)*(FI+1)+FI-k)*(FJ+FP)+k;
              Pw2Double:=((2*(FM+1)+m2)*(FI+1)+FI-k)*(FJ+FP)+k;
              PR11Double:=(m1*(FI+1)+FI-k)*(FJ+FK)+k;
              PR12Double:=(m2*(FI+1)+FI-k)*(FJ+FK)+k;
              PR2Double:=((m1*(FM+1)+m2)*(FI+1)+FI-k)*(FJ+FP)+k;
              PsigmaDouble:=(m1*(FM+1)+m2)*(FJ+FP)+k;
              // rho*1, same as in the article, including the tail k>=FJ
              If ((i1 + k > FI) And (i2 + k >= FI)) or (k>=FJ) Then
                Frho[Prhostar1Double]:=(1-FUsedw^[Pw1Double])*Frho[PrhostarDouble];
              // rho*2, same as in the article, including the tail k>=FJ
              If ((i1 + k >= FI) And (i2 + k > FI)) or (k>=FJ) Then
                Frho[Prhostar2Double]:=(1-FUsedw^[Pw2Double])*Frho[PrhostarDouble];
              // rho*12, same as in the article, including the tail k>=FJ
              If ((i1 + k > FI) And (i2 + k > FI)) or (k>=FJ) Then
                Frho[Prhostar12Double]:=(1-FUsedw^[Pw1Double])*(1-FUsedw^[Pw2Double])*Frho[PrhostarDouble];
              // rho bar, same as in the article, including the tail k>=FJ
              if (FR1[PR11Double] <>0) and (FR1[PR12Double]<>0)  then begin
                TmpDouble:= FUsedsigma^[PsigmaDouble]*FR2[PR2Double]/FR1[PR11Double]/FR1[PR12Double];
                if ((i1 > FI - k) or (k>=FJ)) and ((i2 > FI - k) or (k>=FJ)) then
                  FRho[PrhobarDouble]:=FUsedw^[Pw1Double]*FUsedw^[Pw2Double]*TmpDouble
                else if (i1 = FI - k) and ((i2 > FI - k) or (k>=FJ)) then
                  FRho[PrhobarDouble]:=FUsedw^[Pw2Double]*TmpDouble
                else if ((i1 > FI - k) or (k>=FJ)) and (i2 = FI - k) then
                  FRho[PrhobarDouble]:=FUsedw^[Pw1Double]*TmpDouble
                else if (i1 = FI - k) and (i2 = FI - k) then
                  FRho[PrhobarDouble]:=TmpDouble
                else
                  FRho[PrhobarDouble]:=0;
              end;
              Inc(PRhostarDouble,1);
              Inc(PRhostar1Double,1);
              Inc(PRhostar2Double,1);
              Inc(PRhostar12Double,1);
              Inc(PRhobarDouble,1);
            end;
          end;
        end;
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_rhoCDR,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitrhoCDR + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitrhoCDR + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitrhoMSEP: String;
var m1, m2, i1, i2: Integer;
    PDestDouble, PrhostarFirstDouble, PSourceDouble: IndexDouble;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  if calc_ReduceMemSize in FCalculationFlags then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDR + sLineBreak + rsLSRM_ProxyOnly;
    Exit;
  end;
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitrhoMSEP + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    if Length(Frho)<=0 then
      SetLength(Frho,6*(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
    // calculate
    PrhostarFirstDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP);
    ZeroMemory(@Frho[PrhostarFirstDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP)*SizeOf(Double));
    // note, if i1<>i2 then zero
    //       otherwise a copy of i1=i2=0
    PDestDouble:=PrhostarFirstDouble;
    PSourceDouble:=0;
    For m1 := 0 To FM do begin
      For m2 := 0 To FM do begin
        for i1 := 0 to FI do begin
          for i2 := 0 to FI do begin
            Move(FrhoStar[PSourceDouble],Frho[PDestDouble],(FJ+FP)*Sizeof(Double));
            Inc(PDestDouble,FJ+FP);
          end;
        end;
        Inc(PSourceDouble,FJ+FP);
        Inc(PrhostarFirstDouble,(FI+1)*(FI+1)*(FJ+FP));
        PDestDouble:=PrhostarFirstDouble;
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_rhoMSEP,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitrhoMSEP + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitrhoMSEP + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitrhoStar: String;
var m1, m2, k, i: Integer;
    PDestDouble, Pw1Double, Pw2Double, PsigmaDouble, PR2Double, PR11Double, PR12Double,
      PrhostarPrevDouble, Psig11Double, Psig22Double, Psig11PrevDouble, Psig22PrevDouble,
      PR11PrevDouble, PR12PrevDouble, PR21PrevDouble, PR22PrevDouble,
      PR21Double, PR22Double: IndexDouble;
    StartTime, EndTime: Int64;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitrhoStar + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_sigma, Init_Defaultsigma]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitrhoStar + sLineBreak + rsLSRM_NotInitialisedsigma;
    Exit;
  end;
  If not(Init_R1 in FInitialised) then Result:=InitR1();
  if Result<>'' then Exit;
  If not(Init_R2 in FInitialised) then Result:=InitR2();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array or clear corresponding memory
    if Length(FrhoStar)<=0 then
      SetLength(FrhoStar,(FM+1)*(FM+1)*(FJ+FP))
    else
      ZeroMemory(@FrhoStar[0],(FM+1)*(FM+1)*(FJ+FP)*SizeOf(Double));
    PDestDouble:=0;
    PsigmaDouble:=0;
    For m1 := 0 To FM do begin
      For m2 := 0 To FM do begin
        // i1=i2=0
        For k := 0 To FJ - 1 + FP do begin
          {$REGION 'without tail'}
          Pw1Double:=(((FM+1)+m1)*(FI+1)+0)*(FJ+FP)+k;
          Pw2Double:=(((FM+1)+m2)*(FI+1)+0)*(FJ+FP)+k;
          PR11Double:=(m1*(FI+1)+0)*(FJ+FK)+k;
          PR12Double:=(m2*(FI+1)+0)*(FJ+FK)+k;
          PR2Double:=((m1*(FM+1)+m2)*(FI+1)+0)*(FJ+FP)+k;
          for i := 0 to FI-k-1 do begin
            if (FR1[PR11Double]<>0) and (FR1[PR12Double]<>0) then
              FrhoStar[PDestDouble]:=FrhoStar[PDestDouble]
                                    +FUsedw^[Pw1Double]
                                    *FUsedw^[Pw2Double]
                                    *FR2[PR2Double]
                                    /FR1[PR11Double]
                                    /FR1[PR12Double];
            Inc(Pw1Double,FJ+FP);
            Inc(Pw2Double,FJ+FP);
            Inc(PR11Double,FJ+FK);
            Inc(PR12Double,FJ+FK);
            Inc(PR2Double,FJ+FP);
          end;
          FrhoStar[PDestDouble]:= FrhoStar[PDestDouble] * FUsedsigma^[PsigmaDouble];
          {$ENDREGION}
          {$REGION 'tail'}
          // default tail equals previous rhostar proportional to corresponding
          // SQRT( sigmam1m1_J *sigmam2m2_J * Rm1m1_0,J * Rm2m2_0,J ) / (Rm1_0,J * Rm2_0,J)
          // this may leads to CDR>MSEP, therefore if calc_UseMSEPifTailCDRisLarger is in FCalculationFlags
          // we will use the MSEP for those periods and therefore have to memorise those k
          if (FK>0) and (FJ=FI) and (k=FJ-1+FP) then begin
            Psig11Double:=(m1*(FM+1)+m1)*(FJ+FP)+k;
            Psig22Double:=(m2*(FM+1)+m2)*(FJ+FP)+k;
            Psig11PrevDouble:=Psig11Double;
            Dec(Psig11PrevDouble);
            Psig22PrevDouble:=Psig22Double;
            Dec(Psig22PrevDouble);
            PrhostarPrevDouble:=PDestDouble;
            Dec(PrhostarPrevDouble,1);
            PR11Double:=(m1*(FI+1)+0)*(FJ+FK)+k;
            PR11PrevDouble:=(m1*(FI+1)+0)*(FJ+FK)+k-1;
            PR12Double:=(m2*(FI+1)+0)*(FJ+FK)+k;
            PR12PrevDouble:=(m2*(FI+1)+0)*(FJ+FK)+k-1;
            PR21Double:=((m1*(FM+1)+m1)*(FI+1)+0)*(FJ+FP)+k;
            PR21PrevDouble:=((m1*(FM+1)+m1)*(FI+1)+0)*(FJ+FP)+k-1;
            PR22Double:=((m2*(FM+1)+m2)*(FI+1)+0)*(FJ+FP)+k;
            PR22PrevDouble:=((m2*(FM+1)+m2)*(FI+1)+0)*(FJ+FP)+k-1;
            if (FR1[PR11Double]*FR1[PR12Double]<>0)
                and (FUsedsigma^[Psig11PrevDouble]*FUsedsigma^[Psig22PrevDouble]*FR2[PR21PrevDouble]*FR2[PR22PrevDouble]>0)
                and (FUsedsigma^[Psig11Double]*FUsedsigma^[Psig22Double]*FR2[PR21Double]*FR2[PR22Double]>=0) then
              FrhoStar[PDestDouble]:=FrhoStar[PrhostarPrevDouble]
                                    /SQRT(FUsedsigma^[Psig11PrevDouble]*FUsedsigma^[Psig22PrevDouble]*FR2[PR21PrevDouble]*FR2[PR22PrevDouble])
                                    /FR1[PR11Double]
                                    /FR1[PR12Double]
                                    *SQRT(FUsedsigma^[Psig11Double]*FUsedsigma^[Psig22Double]*FR2[PR21Double]*FR2[PR22Double])
                                    *FR1[PR11PrevDouble]
                                    *FR1[PR12PrevDouble];
          end;
          {$ENDREGION}
          Inc(PsigmaDouble,1);
          Inc(PDestDouble);
        end;
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_rhoStar,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitrhoStar + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitrhoStar + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitS(var S: Double): String;
var HasChanged: boolean;
    StartTime, EndTime: Int64;
    PDestDouble: IndexDouble;
    PSource: PDouble;
    k: Integer;
begin
  Result:='';
  HasChanged:=False;
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitS + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FS)<=0 then begin
      //             m      i      k
      SetLength(FS,(FM+1)*(FI+1)*(FJ+1+FK));
      HasChanged:=True;
    end;
    if (FK=0) then begin
      {$REGION 'we can use fast memory copy and memory compare'}
      if HasChanged or not(Init_S in FInitialised)
          or not(CompareMem(@FS[0],@S,(FM+1)*(FI+1)*(FJ+1)*SizeOf(Double))) then begin
        // copy specified values to array
        Move(S,FS[0],(FM+1)*(FI+1)*(FJ+1)*SizeOf(Double));
        QueryPerformanceCounter(EndTime);
        SetInitFlags(Init_S,EndTime-StartTime);
      end;
      {$ENDREGION}
    end else begin
      {$REGION 'we have to do it value by value'}
      {$REGION 'first check if something has changed'}
      PDestDouble:=0;
      PSource:=@S;
      k:=0;
      while not(HasChanged) and (PDestDouble<=(FM*(FI+1)+FI)*(FJ+1+FK)+FJ) do begin
        HasChanged:=(FS[PDestDouble]<>PSource^);
        if not HasChanged then begin
          Inc(PSource,1);
          if k=FJ then begin
            k:=0;
            Inc(PDestDouble,FK+1);
          end else begin
            Inc(k,1);
            Inc(PDestDouble,1);
          end;
        end;
      end;
      {$ENDREGION}
      {$REGION 'get the values that may have changed and set all dependent objects to not initialised'}
      if HasChanged then begin
        while PDestDouble<=(FM*(FI+1)+FI)*(FJ+1+FK)+FJ do begin
          FS[PDestDouble]:=PSource^;
          Inc(PSource,1);
          if k=FJ then begin
            k:=0;
            Inc(PDestDouble,FK+1);
          end else begin
            Inc(k,1);
            Inc(PDestDouble,1);
          end;
        end;
        QueryPerformanceCounter(EndTime);
        SetInitFlags(Init_S,EndTime-StartTime);
      end;
      {$ENDREGION}
      {$ENDREGION}
    end;
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitS + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitS + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.Initsigma(var sigma: Double): String;
var HasChanged: Boolean;
    StartTime, EndTime: Int64;
begin
  Result:='';
  HasChanged:=False;
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitsigma + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(Fsigma)<=0 then begin
      //                 m1     m2      k
      SetLength(Fsigma,(FM+1)*(FM+1)*(FJ+FP));
      HasChanged:=True;
    end;
    // check if the manual process variances are the same, if not take the new one
    // and set all dependent objects to not initialised
    if HasChanged or not(Init_sigma in FInitialised)
        or not(CompareMem(@Fsigma[0],@sigma,(FM+1)*(FM+1)*(FJ+FP)*SizeOf(Double))) then begin
      Move(sigma,Fsigma[0],(FM+1)*(FM+1)*(FJ+FP)*SizeOf(Double));
      QueryPerformanceCounter(EndTime);
      SetInitFlags(Init_sigma,EndTime-StartTime);
    end;
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitsigma + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitsigma + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitSxS0: string;
var StartTime, EndTime: Int64;
    m1, m2, i1, i2, k1, k2: Integer;
    PDestDouble, PS1Double, PS2Double: IndexDouble;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  if calc_ReduceMemSize in FCalculationFlags then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDR + sLineBreak + rsLSRM_ProxyOnly;
    Exit;
  end;
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitSxS0 + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FSxS)<=0 then begin
      //            Part m1     i1       k1      m2      i2     k2
      Setlength(FSxS,6*(FM+1)*(FI+1)*(FJ+1+FP)*(FM+1)*(FI+1)*(FJ+1+FP));
    end;
    PDestDouble:=0;
    PS1Double:=0;
    For m1 := 0 To FM do begin
      For i1 := 0 To FI do begin
        For k1 := 0 To FJ+FP do begin
          PS2Double:=0;
          For m2 := 0 To FM do begin
            For i2 := 0 To FI do begin
              For k2 := 0 To FJ+FP do begin
                FSxS[PDestDouble]:=FS[PS1Double]*FS[PS2Double];
                Inc(PS2Double,1);
                Inc(PDestDouble,1);
              end;
              Inc(PS2Double,FK-FP);
            end;
          end;
          Inc(PS1Double,1);
        end;
        Inc(PS1Double,FK-FP);
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_SxS0,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitSxS0 + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitSxS0 + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitSxSCDRParaErr: string;
var StartTime, EndTime: Int64;
    PSxSDouble, PrhoDouble: IndexDouble;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  if calc_ReduceMemSize in FCalculationFlags then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDR + sLineBreak + rsLSRM_ProxyOnly;
    Exit;
  end;
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitSxSCDRParaErr + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_Gamma1 in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitSxSCDRParaErr + sLineBreak
            + rsLSRM_NotInitialisedGamma1;
    Exit;
  end;
  If not(Init_SxS0 in FInitialised) then Result:=InitSxS0();
  if Result<>'' then Exit;
  If not(Init_SxSMSEP in FInitialised) then Result:=InitSxSMSEP();
  if Result<>'' then Exit;
  If not(Init_rhoCDR in FInitialised) then Result:=InitrhoCDR();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    {$REGION 'H(rho*1) SxS'}
    PSxSDouble:=2*(FM+1)*(FI+1)*(FJ+1+FP)*(FM+1)*(FI+1)*(FJ+1+FP);
    CopyMemory(@FSxS[PSxSDouble],@FSxS[0],(FM+1)*(FI+1)*(FJ+1+FP)*(FM+1)*(FI+1)*(FJ+1+FP)*SizeOf(Double));
    PrhoDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP);
    Result:=OP_Hn(FI,@Frho[PrhoDouble],@FSxS[PSxSDouble]);
    if Result<>'' then Exit;
    {$ENDREGION}
    {$REGION 'H(rho*2) SxS'}
    Inc(PSxSDouble,(FM+1)*(FI+1)*(FJ+1+FP)*(FM+1)*(FI+1)*(FJ+1+FP));
    CopyMemory(@FSxS[PSxSDouble],@FSxS[0],(FM+1)*(FI+1)*(FJ+1+FP)*(FM+1)*(FI+1)*(FJ+1+FP)*SizeOf(Double));
    Inc(PrhoDouble,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
    Result:=OP_Hn(FI,@Frho[PrhoDouble],@FSxS[PSxSDouble]);
    if Result<>'' then Exit;
    {$ENDREGION}
    {$REGION 'H(rho*12) SxS'}
    Inc(PSxSDouble,(FM+1)*(FI+1)*(FJ+1+FP)*(FM+1)*(FI+1)*(FJ+1+FP));
    CopyMemory(@FSxS[PSxSDouble],@FSxS[0],(FM+1)*(FI+1)*(FJ+1+FP)*(FM+1)*(FI+1)*(FJ+1+FP)*SizeOf(Double));
    Inc(PrhoDouble,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
    Result:=OP_Hn(FI,@Frho[PrhoDouble],@FSxS[PSxSDouble]);
    if Result<>'' then Exit;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_SxSCDRParaErr,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitSxSCDRParaErr + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitSxSCDRParaErr + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitSxSCDRProcVar: string;
var StartTime, EndTime: Int64;
    PSxSDouble, PrhoDouble: IndexDouble;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  if calc_ReduceMemSize in FCalculationFlags then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDR + sLineBreak + rsLSRM_ProxyOnly;
    Exit;
  end;
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitSxSCDRProcVar + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_Gamma1 in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitSxSCDRProcVar + sLineBreak
            + rsLSRM_NotInitialisedGamma1;
    Exit;
  end;
  If not(Init_SxS0 in FInitialised) then Result:=InitSxS0();
  if Result<>'' then Exit;
  If not(Init_SxSMSEP in FInitialised) then Result:=InitSxSMSEP();
  if Result<>'' then Exit;
  If not(Init_rhoCDR in FInitialised) then Result:=InitrhoCDR();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    {$REGION 'H(rhobar) SxS'}
    PSxSDouble:=5*(FM+1)*(FI+1)*(FJ+1+FP)*(FM+1)*(FI+1)*(FJ+1+FP);
    CopyMemory(@FSxS[PSxSDouble],@FSxS[0],(FM+1)*(FI+1)*(FJ+1+FP)*(FM+1)*(FI+1)*(FJ+1+FP)*SizeOf(Double));
    PrhoDouble:=5*(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP);
    Result:=OP_Hn(FI,@Frho[PrhoDouble],@FSxS[PSxSDouble]);
    if Result<>'' then Exit;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_SxSCDRProcVar,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitSxSCDRProcVar + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitSxSCDRProcVar + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitSxSMSEP: string;
var StartTime, EndTime: Int64;
    PSxSDouble, PrhoDouble: IndexDouble;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  if calc_ReduceMemSize in FCalculationFlags then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDR + sLineBreak + rsLSRM_ProxyOnly;
    Exit;
  end;
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitSxSMSEP + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_Gamma1 in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitSxSMSEP + sLineBreak + rsLSRM_NotInitialisedGamma1;
    Exit;
  end;
  If not(Init_SxS0 in FInitialised) then Result:=InitSxS0();
  if Result<>'' then Exit;
  If not(Init_rhoMSEP in FInitialised) then Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    {$REGION 'H(rho*) SxS'}
    PSxSDouble:=(FM+1)*(FI+1)*(FJ+1+FP)*(FM+1)*(FI+1)*(FJ+1+FP);
    CopyMemory(@FSxS[PSxSDouble],@FSxS[0],(FM+1)*(FI+1)*(FJ+1+FP)*(FM+1)*(FI+1)*(FJ+1+FP)*SizeOf(Double));
    PrhoDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP);
    Result:=OP_Hn(FI,@Frho[PrhoDouble],@FSxS[PSxSDouble]);
    if Result<>'' then Exit;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_SxSMSEP,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitSxSMSEP + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitSxSMSEP + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.Initw(var w: Double): String;
var HasChanged: Boolean;
    StartTime, EndTime: Int64;
begin
  Result:='';
  HasChanged:=False;
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitw + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  // exposures are needed for normalization
  If not(Init_R1 in FInitialised) then Result:=InitR1();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(Fw)<=0 then begin
      //           n    m     i       k
      SetLength(Fw,3*(FM+1)*(FI+1)*(FJ+FP));
      HasChanged:=True;
    end;
    // check if manual weight have changed, if so than take the new ones and set all dependent objects to
    // not initialised
    if HasChanged or not(Init_w in FInitialised)
        or not(CompareMem(@Fw[0],@w,(FM+1)*(FI+1)*(FJ+FP)*SizeOf(Double))) then begin
      // copy manual not normalised weight w(0)
      Move(w,Fw[0],(FM+1)*(FI+1)*(FJ+FP)*SizeOf(Double));
      // normalise weights
      NormalizeWeights(Fw);
      QueryPerformanceCounter(EndTime);
      SetInitFlags(Init_w,EndTime-StartTime);
    end;
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitw + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitw + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitUltimate: string;
var m,i: Integer;
    PDestDouble, PUltimateDouble, PalphaDouble: IndexDouble;
    StartTime, EndTime: Int64;
    TmpSum: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimate + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_alpha, Init_Defaultalpha]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimate + sLineBreak + rsLSRM_NotInitialisedalpha;
    Exit;
  end;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // calculate
    {$REGION 'without Tail'}
    PDestDouble:=(FM+1)*(FI+2); // the last coordinate Part=0, m=M+1, i=0
    TmpSum:=0;
    for i:=0 to FI do begin
      PalphaDouble:=i;
      PUltimateDouble:=i;
      FUltimate[PDestDouble]:=0;
      for m:=0 to FM do begin
        FUltimate[PDestDouble]:=FUltimate[PDestDouble]
                                +FUsedalpha^[PalphaDouble]
                                *FUltimate[PUltimateDouble];
        Inc(PUltimateDouble,FI+2); // next m
        Inc(PalphaDouble,FI+1); // next m
      end;
      TmpSum:=TmpSum+FUltimate[pDestDouble];
      Inc(PDestDouble);
    end;
    FUltimate[PDestDouble]:=TmpSum;
    {$ENDREGION}
    {$REGION 'with Tail'}
    PDestDouble:=((FM+2)+(FM+1))*(FI+2); // the last coordinate Part=1, m=M+1, i=0
    TmpSum:=0;
    for i:=0 to FI do begin
      PalphaDouble:=i;
      PUltimateDouble:=(FM+2)*(FI+2)+i;
      FUltimate[PDestDouble]:=0;
      for m:=0 to FM do begin
        FUltimate[PDestDouble]:=FUltimate[PDestDouble]
                                +FUsedalpha^[PalphaDouble]
                                *FUltimate[PUltimateDouble];
        Inc(PUltimateDouble,FI+2); // next m
        Inc(PalphaDouble,FI+1); // next m
      end;
      TmpSum:=TmpSum+FUltimate[pDestDouble];
      Inc(PDestDouble);
    end;
    FUltimate[PDestDouble]:=TmpSum;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_Ultimate,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimate + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimate + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TNormalLSRM.InitUltimateDetail: string;
var m, n, k, i, l, h, j, cl, ch: Integer;
    PR1DestDouble, PSDestDouble, Pgamma1Double, PSDouble, PfDouble, PUltimateDestDouble,
      PUltimateTailDestDouble: IndexDouble;
    StartTime, EndTime: Int64;
    TmpSum, TmpSumTail:Double;
    sign:T1Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimateDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_S in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimateDetail + sLineBreak + rsLSRM_NotInitialisedS;
    Exit;
  end;
  If not(Init_gamma1 in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimateDetail + sLineBreak
            + rsLSRM_NotInitialisedgamma1;
    Exit;
  end;
  if ([Init_f, Init_Defaultf]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimateDetail + sLineBreak + rsLSRM_NotInitialisedf;
    Exit;
  end;
  If not(Init_R1 in FInitialised) then Result:=InitR1();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array or clear corresponding memory
    if Length(FUltimate)<=0 then begin
      //                 Part  m      i
      SetLength(FUltimate,2*(FM+2)*(FI+2));
    end else begin
      ZeroMemory(@FUltimate[0],2*(FM+2)*(FI+2)*SizeOf(Double))
    end;
    {$REGION 'get signs of the first known exposure in case of they should be prevented from changing signs'}
    if calc_PreventFutureExposureFromChangingSign in FCalculationFlags then begin
      setlength(sign,FI+1);
      PR1DestDouble:=0;
      for m:=0 to FM do begin
        for i:=0 to FI do begin
          sign[i]:=0;
          for k := 0 To Min(FJ - 1 + FP, FI-i) do begin
            if sign[i]=0 then
              sign[i]:=FR1[PR1DestDouble];
            Inc(PR1DestDouble,1);
          end;
          Inc(PR1DestDouble,FJ-1+FK-Min(FJ - 1 + FP, FI-i));
        end;
      end;
    end;
    {$ENDREGION}
    {$REGION 'calculate S and R1 diagonal by diagonal: first the tail below the diagonal FI+1'}
    for n:=FJ+1 to FI do begin
      for m:=0 to FM do begin
        for i := max(0,n-FJ-FK) to n-FJ-1 do begin
          // k=n-i
          {$REGION 'R1'}
          PR1DestDouble:=(m*(FI+1)+i)*(FJ+FK)+n-i-1;
          FR1[PR1DestDouble]:=0;
          for cl:=0 to FmlDependency[m,FM] do begin
            l:=FmlDependency[m,cl];
            for ch:=0 to FihDependency[i,FI] do begin
              h:=FihDependency[i,ch];
              PGamma1Double:=((((m*(FM+1)+l)*(FI+1)+i)*(FJ+FK)+n-i-1)*(FI+1)+h)*(FJ+FK);
              PSDouble:=(l*(FI+1)+h)*(FJ+1+FK);
              for j:=0 to Min(n-1 - h, n-i-1) do begin
                FR1[PR1DestDouble]:=FR1[PR1DestDouble]
                                    +FGamma1[PGamma1Double]
                                    *FS[PSDouble];
                Inc(PGamma1Double);
                Inc(PSDouble);
              end;
            end;
          end;
          // prevent exposure from changing sign
          if calc_PreventFutureExposureFromChangingSign in FCalculationFlags then begin
            if sign[i]=0 then begin
              sign[i]:=FR1[PR1DestDouble];
            end else begin
              if ((sign[i]>0) and (FR1[PR1DestDouble]<0)) or ((sign[i]<0) and (FR1[PR1DestDouble]>0)) then
                FR1[PR1DestDouble]:=0;
            end;
          end;
          {$ENDREGION}
          {$REGION 'S'}
          PfDouble:=m*(FJ+FK)+n-i-1;
          PSDestDouble:=(m*(FI+1)+i)*(FJ+1+FK)+n-i;
          FS[PSDestDouble]:=FUsedf^[PfDouble] * FR1[PR1DestDouble];
          {$ENDREGION}
        end;
      end;
    end;
    {$ENDREGION}
    {$REGION 'calculate S and R1 diagonal by diagonal: above the diagonal FI'}
    for n:=FI+1 to FI+FJ+FK do begin
      for m:=0 to FM do begin
        for i := max(0,n-FJ-FK) to FI do begin
          // k=n-i
          {$REGION 'R1'}
          PR1DestDouble:=(m*(FI+1)+i)*(FJ+FK)+n-i-1;
          // first fill up R1(m,i,n-i-1), but only if n>FI+1 or if k>FJ
          if (n>FI+1) or (n-i>FJ) then begin
            FR1[PR1DestDouble]:=0;
            for cl:=0 to FmlDependency[m,FM] do begin
              l:=FmlDependency[m,cl];
              for ch:=0 to FihDependency[i,FI] do begin
                h:=FihDependency[i,ch];
                PGamma1Double:=((((m*(FM+1)+l)*(FI+1)+i)*(FJ+FK)+n-i-1)*(FI+1)+h)*(FJ+FK);
                PSDouble:=(l*(FI+1)+h)*(FJ+1+FK);
                for j:=0 to Min(n-1 - h, n-i-1) do begin
                  FR1[PR1DestDouble]:=FR1[PR1DestDouble]
                                      +FGamma1[PGamma1Double]
                                      *FS[PSDouble];
                  Inc(PGamma1Double);
                  Inc(PSDouble);
                end;
              end;
            end;
            // prevent exposure from changing sign
            if calc_PreventFutureExposureFromChangingSign in FCalculationFlags then begin
              if sign[i]=0 then begin
                sign[i]:=FR1[PR1DestDouble];
              end else begin
                if ((sign[i]>0) and (FR1[PR1DestDouble]<0)) or ((sign[i]<0) and (FR1[PR1DestDouble]>0)) then
                  FR1[PR1DestDouble]:=0;
              end;
            end;
          end;
          {$ENDREGION}
          {$REGION 'S'}
          PfDouble:=m*(FJ+FK)+n-i-1;
          PSDestDouble:=(m*(FI+1)+i)*(FJ+1+FK)+n-i;
          FS[PSDestDouble]:=FUsedf^[PfDouble]*FR1[PR1DestDouble];
          {$ENDREGION}
        end;
      end;
    end;
    {$ENDREGION}
    {$REGION 'fill up the sum over all accident periods'}
    PUltimateDestDouble:=0;
    PUltimateTailDestDouble:=(FM+2)*(FI+2);
    PSDouble:=0;
    for m := 0 to FM do begin
      TmpSum:=0;
      TmpSumTail:=0;
      for i := 0 to FI do begin
        // without Tail
        for k := 0 to FJ do begin
          FUltimate[PUltimateDestDouble]:=FUltimate[PUltimateDestDouble]+FS[PSDouble];
          Inc(PSDouble,1);
        end;
        TmpSum:=TmpSum + FUltimate[PUltimateDestDouble];
        FUltimate[PUltimateTailDestDouble]:=FUltimate[PUltimateDestDouble];
        // Tail
        for k := FJ+1 to FJ+FK do begin
          FUltimate[PUltimateTailDestDouble]:=FUltimate[PUltimateTailDestDouble]+FS[PSDouble];
          Inc(PSDouble,1);
        end;
        TmpSumTail:=TmpSumTail + FUltimate[PUltimateTailDestDouble];
        Inc(PUltimateDestDouble,1);
        Inc(PUltimateTailDestDouble,1);
      end;
      FUltimate[PUltimateDestDouble]:=TmpSum;
      FUltimate[PUltimateTailDestDouble]:=TmpSumTail;
      Inc(PUltimateDestDouble,1);
      Inc(PUltimateTailDestDouble,1);
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_UltimateDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimateDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimateDetail + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

{$ENDREGION}

{$ENDREGION}

{$ENDREGION}

{$REGION 'TSimpleLSRM'}

{$REGION '  TSimpleLSRM protected'}

{$REGION 'property get functions'}

function TSimpleLSRM.GetCoorF(m, i, k, l, h, j: integer): Double;
  var ErrorMsg: string;
begin
  {$REGION 'test indecies'}
  if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (i<0) or (i>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
  if (k<0) or (k>FJ-1+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
  if (l<0) or (l>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (h<0) or (h>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
  if (j<0) or (j>FJ+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
  {$ENDREGION}
  // calculate the coordinates of the operator F, if necessary
  if not(Init_CoorF in FInitialised) then ErrorMsg:=InitCoorF();
  if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
  //return the coordinates (F^m_ik)^l_hj of the operator F
  Result:=FCoorF[((m*(FJ+FP)+k)*(FM+1)+l)*(FJ+1+FP)+j];
end;

class function TSimpleLSRM.GetDelimiters: String;
begin
  result:=TSimpleLSRM.FTokenChars.DecimalChar+TSimpleLSRM.FTokenChars.FormulaSepChar+TSimpleLSRM.FTokenChars.ParaSepChar;
end;

{$ENDREGION}

{$REGION '    operators'}

function TSimpleLSRM.OP_F(n: integer; S: T1Double): string;
var m, i, cl, l, h, j: Integer;
    TmpDestDouble, TmpSDouble, Tmpgamma1Double, TmpfDouble: IndexDouble;
begin
  Result:='';
  try
    For m := 0 To FM do begin
      For i := Max(0, n + 1 - FJ - FP) To Min(FI, n) do begin
        // k = n-i
        TmpDestDouble:=(m*(FI+1)+i)*(FJ+1+FP)+n-i+1;
        S[TmpDestDouble]:=0;
        TmpfDouble:=m*(FJ+FK)+n-i;
        for cl:=0 to FmlDependency[m,FM] do begin
          l:=FmlDependency[m,cl];
            h:=i;
            TmpSDouble:=(l*(FI+1)+h)*(FJ+1+FP);
            Tmpgamma1Double:=((m*(FM+1)+l)*(FJ+FK)+(n-i))*(FJ+FK);
            For j := 0 To Min(n - h, n - i) do begin
              S[TmpDestDouble]:=S[TmpDestDouble] + FGamma1[TmpGamma1Double] * S[TmpSDouble];
              Inc(TmpSDouble,1);
              Inc(Tmpgamma1Double,1);
            end;
        end;
        S[TmpdestDouble]:=S[TmpDestDouble] * FUsedf^[TmpfDouble];
      end;
    end;
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorOP_F + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorOP_F + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.OP_H(n, k: integer; tau, SxS: PDouble): string;
var m1, m2, i1, i2, k1, k2, cl1, cl2, l1, l2, h1, h2, j1, j2: Integer;
    PDestDouble, PSxSDouble, PTauDouble: PDouble;
    PGamma11Double, PGamma12Double, Pf1Double, Pf2Double, PInflation11, PInflation12, PInflation21,
      PInflation22: IndexDouble;
begin
  Result:='';
  try
    Pf1Double:=k;
    For m1 := 0 To FM do begin
      Pf2Double:=k;
      For m2 := 0 To FM do begin
        {$REGION 'k1=k+1 and k2=k+1 and i1+k1>n and i2+k2>n'}
        For i1 := n - k To FI do begin
          For i2 := n - k To FI do begin
            PDestDouble:=SxS;
            Inc(PDestDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k+1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k+1);
            PDestDouble^:=0;
            for cl1:=0 to FmlDependency[m1,FM] do begin
              l1:=FmlDependency[m1,cl1];
              for cl2:=0 to FmlDependency[m2,FM] do begin
                l2:=FmlDependency[m2,cl2];
                  h1:=i1;
                    h2:=i2;
                    For j1 := 0 To k do begin
                      PGamma11Double:=((m1*(FM+1)+l1)*(FJ+FK)+k)*(FJ+FK)+j1;
                      For j2 := 0 To k do begin
                        PGamma12Double:=((m2*(FM+1)+l2)*(FJ+FK)+k)*(FJ+FK)+j2;
                        PSxSDouble:=SxS;
                        Inc(PSxSDouble,((((l1*(FI+1)+h1)*(FJ+1+FP)+j1)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j2);
                        PDestDouble^:=PDestDouble^
                                      +FGamma1[PGamma11Double]
                                      *FGamma1[PGamma12Double]
                                      *PSxSDouble^;
                      end;
                    end;
              end;
            end;
            PtauDouble:=tau;
            Inc(PtauDouble,(((m1*(FM+1)+m2)*(FI+1)+i1)*(FI+1)+i2)*(FJ+FP)+k);
            PDestDouble^:=(FUsedf^[Pf1Double] * FUsedf^[Pf2Double]  + PTauDouble^) * PDestDouble^;
          end;
        end;{$ENDREGION}
        {$REGION 'k1=k+1 and i1+k1>n and k2<k+1'}
        For i1 := n - k To FI do begin
          For i2 := 0 To FI do begin
            For k2 := 0 To k do begin
              PDestDouble:=SxS;
              Inc(PDestDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k+1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k2);
              PDestDouble^:=0;
              for cl1:=0 to FmlDependency[m1,FM] do begin
                l1:=FmlDependency[m1,cl1];
                  h1:=i1;
                  For j1 := 0 To k do begin
                    PGamma11Double:=((m1*(FM+1)+l1)*(FJ+FK)+k)*(FJ+FK)+j1;
                    PSxSDouble:=SxS;
                    Inc(PSxSDouble,((((l1*(FI+1)+h1)*(FJ+1+FP)+j1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k2);
                    PDestDouble^:=PDestDouble^ + FGamma1[PGamma11Double] * PSxSDouble^;
                  end;
              end;
              PDestDouble^:=FUsedf^[Pf1Double] * PDestDouble^;
            end;
          end;
        end; {$ENDREGION}
        {$REGION 'k1=k+1 and i1+k1>n and k2=k+1 and i2+k2<=n'}
        For i1 := n - k To FI do begin
          For i2 := 0 To n - k - 1 do begin
            PDestDouble:=SxS;
            Inc(PDestDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k+1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k+1);
            PDestDouble^:=0;
            for cl1:=0 to FmlDependency[m1,FM] do begin
              l1:=FmlDependency[m1,cl1];
                h1:=i1;
                For j1 := 0 To k do begin
                  PGamma11Double:=((m1*(FM+1)+l1)*(FJ+FK)+k)*(FJ+FK)+j1;
                  PSxSDouble:=SxS;
                  Inc(PSxSDouble,((((l1*(FI+1)+h1)*(FJ+1+FP)+j1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k+1);
                  PDestDouble^:=PDestDouble^ + FGamma1[PGamma11Double] * PSxSDouble^;
                end;
            end;
            PDestDouble^:=FUsedf^[Pf1Double] * PDestDouble^;
          end;
        end; {$ENDREGION}
        {$REGION 'k1<k+1 and k2=k+1 and i2+k2>n'}
        For i1 := 0 To FI do begin
          For i2 := n - k To FI do begin
            For k1 := 0 To k do begin
              PDestDouble:=SxS;
              Inc(PDestDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k+1);
              PDestDouble^:=0;
              for cl2:=0 to FmlDependency[m2,FM] do begin
                l2:=FmlDependency[m2,cl2];
                  h2:=i2;
                  For j2 := 0 To k do begin
                    PGamma12Double:=((m2*(FM+1)+l2)*(FJ+FK)+k)*(FJ+FK)+j2;
                    PSxSDouble:=SxS;
                    Inc(PSxSDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k1)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j2);
                    PDestDouble^:=PDestDouble^ + FGamma1[PGamma12Double] * PSxSDouble^;
                  end;
              end;
              PDestDouble^:=FUsedf^[Pf2Double] * PDestDouble^;
            end;
          end;
        end;{$ENDREGION}
        {$REGION 'k1=k+1 and i1+k1<=n and k2=k+1 and i2+k2>n'}
        For i1 := 0 To n - k - 1 do begin
          For i2 := n - k To FI do begin
            PDestDouble:=SxS;
            Inc(PDestDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k+1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k+1);
            PDestDouble^:=0;
            for cl2:=0 to FmlDependency[m2,FM] do begin
              l2:=FmlDependency[m2,cl2];
                h2:=i2;
                For j2 := 0 To k do begin
                  PGamma12Double:=((m2*(FM+1)+l2)*(FJ+FK)+k)*(FJ+FK)+j2;
                  PSxSDouble:=SxS;
                  Inc(PSxSDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k+1)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j2);
                  PDestDouble^:=PDestDouble^ + FGamma1[PGamma12Double] * PSxSDouble^;
                end;
            end;
            PDestDouble^:=FUsedf^[Pf2Double] * PDestDouble^;
          end;
        end;{$ENDREGION}
        Inc(Pf2Double,FJ+FK);
      end;
      Inc(Pf1Double,FJ+FK);
    end;
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorOP_H + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorOP_H + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

class procedure TSimpleLSRM.SetDelimiters(const Value: String);
  var i: integer;
begin
  if Length(Value)<>3 then
    raise ELSRMInitError.Create('Could not change the delimiter, because the number of specified delimiters is not equal to three.');
  i:=1;
  while (i<=Length(Value)) do begin
    if ((Value[i]>='a') and (Value[i]<='z'))
      or ((Value[i]>='A') and (Value[i]<='Z'))
      or ((Value[i]>='0') and (Value[i]<='9'))
      or CharInSet(Value[i],FTokenChars.WhiteChars)
      or (Value[i]=FTokenChars.PlusChar)
      or (Value[i]=FTokenChars.MinusChar)
      or (Value[i]=FTokenChars.MulChar)
      or (Value[i]=FTokenChars.DivChar)
      or (Value[i]=FTokenChars.PowerChar) then
    raise ELSRMInitError.Create('Could not change the delimiter, because at least one of them is a letter, a digit, an operator or a white character.');
    inc(i);
  end;
  if (Value[1]=Value[2]) or (Value[1]=Value[3]) or (Value[2]=Value[3]) then
    raise ELSRMInitError.Create('Could not change the delimiter, because they are not different from each other.');
  FTokenChars.DecimalChar:=Value[1];
  FTokenChars.FormulaSepChar:=Value[2];
  FTokenChars.ParaSepChar:=Value[3];
end;

{$ENDREGION}

{$ENDREGION}

{$REGION '  TSimpleLSRM public'}

constructor TSimpleLSRM.Create(Name: string);
begin
  inherited;
  FFormulas:=nil;
  try
    FFormulas:=TFormulaParser.Create(FTokenChars);
  except
    FreeAndNil(FFormulas);
  end;
end;

destructor TSimpleLSRM.Destroy;
begin
  FreeAndNil(FFormulas);
  inherited;
end;

{$REGION '    Init functions'}

function TSimpleLSRM.InitCDRParaErrProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhostarDouble, Prhostar1Double, Prhostar2Double, Prhostar12Double,
      PCoorF1Double, PCoorF2Double: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, h1, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoMSEP in FInitialised) then
    Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoCDR in FInitialised) then
    Result:=InitrhoCDR();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRDetail)<=0 then begin
      //                  Proxy Part    m1     m2     i1     i2
      SetLength(FCDRDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    // parameter error with proxy
    // store the results within FCDRDetail(1,1)
    PDestDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    ZeroMemory(@FCDRDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
    if calc_ReduceMemSize in FCalculationFlags then begin
      {$REGION 'calculate rhos each time'}
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                    // skip the summation over h1 and h2
                      h1 := i1;
                          h2 := i2;
                            For cl1 := 0 To FmlDependency[m1,FM] do begin
                              l1 := FmlDependency[m1,cl1];
                              For cl2 := 0 To FmlDependency[m2,FM] do begin
                                l2 := FmlDependency[m2,cl2];
                                PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                Rho:=Getrho(1,l1,l2,h1,h2,j)
                                    -Getrho(2,l1,l2,h1,h2,j)
                                    -Getrho(3,l1,l2,h1,h2,j)
                                    +Getrho(4,l1,l2,h1,h2,j);
                                TmpValue:=Rho
                                          *FR1[PR11Double]
                                          *FR1[PR12Double]
                                          *FCoorF[PCoorF1Double]
                                          *FCoorF[PCoorF2Double];
                                // diversification only if rho is negative
                                if (calc_VirtualParaErrDiversification in FCalculationFlags)
                                        or (Rho<=0) or (TmpValue>=0) then
                                  FCDRDetail[PDestDouble]:=FCDRDetail[PDestDouble]+TmpValue;
                              end;
                            end;
                          end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end else begin
      {$REGION 'take rhos from memory'}
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                    // skip the summation over h1 and h2
                      h1 := i1;
                          h2 := i2;
                            For cl1 := 0 To FmlDependency[m1,FM] do begin
                              l1 := FmlDependency[m1,cl1];
                              For cl2 := 0 To FmlDependency[m2,FM] do begin
                                l2 := FmlDependency[m2,cl2];
                                PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                Prhostar1Double:=PrhostarDouble;
                                Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                Prhostar2Double:=Prhostar1Double;
                                Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                Prhostar12Double:=Prhostar2Double;
                                Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                Rho:=Frho[PrhostarDouble]
                                    -Frho[Prhostar1Double]
                                    -Frho[Prhostar2Double]
                                    +Frho[Prhostar12Double];
                                TmpValue:=Rho
                                          *FR1[PR11Double]
                                          *FR1[PR12Double]
                                          *FCoorF[PCoorF1Double]
                                          *FCoorF[PCoorF2Double];
                                // diversification only if rho is negative
                                if (calc_VirtualParaErrDiversification in FCalculationFlags)
                                        or (Rho<=0) or (TmpValue>=0) then
                                  FCDRDetail[PDestDouble]:=FCDRDetail[PDestDouble]+TmpValue;
                              end;
                            end;
                  end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRParaErrProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.InitCDRParaErrTailDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhostarDouble, Prhostar1Double, Prhostar2Double, Prhostar12Double,
      PCoorF1Double, PCoorF2Double, PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoMSEP in FInitialised) then
    Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoCDR in FInitialised) then
    Result:=InitrhoCDR();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRTailDetail)<=0 then begin
      //                     Proxy Part    m1     m2     i1     i2
      SetLength(FCDRTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_CDRParaErrTailProxyDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PProxyDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FCDRTailDetail[PDestDouble],@FCDRTailDetail[PProxyDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the CDR tail parameter error without proxy'}
      // store the results within FCDRTailDetail(0,1)
      PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FCDRTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      if calc_ReduceMemSize in FCalculationFlags then begin
        {$REGION 'calculate rhos each time'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error non proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 +FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else begin
        {$REGION 'take rhos from memory'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error non proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRParaErrTailDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.InitCDRParaErrTailProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhostarDouble, Prhostar1Double, Prhostar2Double, Prhostar12Double,
      PCoorF1Double, PCoorF2Double, PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoMSEP in FInitialised) then
    Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoCDR in FInitialised) then
    Result:=InitrhoCDR();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRTailDetail)<=0 then begin
      //                     Proxy Part    m1     m2     i1     i2
      SetLength(FCDRTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_CDRParaErrTailDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PProxyDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FCDRTailDetail[PProxyDouble],@FCDRTailDetail[PDestDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the CDR tail parameter error proxy'}
      // store the results within FCDRTailDetail(1,1)
      PDestDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FCDRTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      if calc_ReduceMemSize in FCalculationFlags then begin
        {$REGION 'calculate rhos each time'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 +FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else begin
        {$REGION 'take rhos from memory'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRParaErrTailProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.InitCDRProcVarProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhobarDouble, PCoorF1Double, PCoorF2Double: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, h1, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  If not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoCDR in FInitialised) then
    Result:=InitrhoCDR();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRDetail)<=0 then begin
      //                  Proxy Part    m1     m2     i1     i2
      SetLength(FCDRDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    // parameter error proxy
    // store the results within FCDRDetail(1,0)
    PDestDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    ZeroMemory(@FCDRDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
    if calc_ReduceMemSize in FCalculationFlags then begin
      {$REGION 'calculate rhos each time'}
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1 := i1;
                          h2 := i2;
                            For cl1 := 0 To FmlDependency[m1,FM] do begin
                              l1 := FmlDependency[m1,cl1];
                              For cl2 := 0 To FmlDependency[m2,FM] do begin
                                l2 := FmlDependency[m2,cl2];
                                PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                Rho:=Getrho(5,l1,l2,h1,h2,j);
                                TmpValue:=Rho
                                          *FR1[PR11Double]
                                          *FR1[PR12Double]
                                          *FCoorF[PCoorF1Double]
                                          *FCoorF[PCoorF2Double];
                                // diversification only if rho is negative
                                if (calc_VirtualProcVarDiversification in FCalculationFlags)
                                        or (Rho<=0) or (TmpValue>=0) then
                                  FCDRDetail[PDestDouble]:=FCDRDetail[PDestDouble]+TmpValue;
                              end;
                            end;
                  end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end else begin
      {$REGION 'take rhos from memory'}
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1 := i1;
                          h2 := i2;
                            For cl1 := 0 To FmlDependency[m1,FM] do begin
                              l1 := FmlDependency[m1,cl1];
                              For cl2 := 0 To FmlDependency[m2,FM] do begin
                                l2 := FmlDependency[m2,cl2];
                                PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                Rho:=Frho[PrhobarDouble];
                                TmpValue:=Rho
                                          *FR1[PR11Double]
                                          *FR1[PR12Double]
                                          *FCoorF[PCoorF1Double]
                                          *FCoorF[PCoorF2Double];
                                // diversification only if rho is negative
                                if (calc_VirtualProcVarDiversification in FCalculationFlags)
                                        or (Rho<=0) or (TmpValue>=0) then
                                  FCDRDetail[PDestDouble]:=FCDRDetail[PDestDouble]+TmpValue;
                              end;
                            end;
                  end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRProcVarProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.InitCDRProcVarTailDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhobarDouble, PCoorF1Double, PCoorF2Double,
      PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoCDR in FInitialised) then
    Result:=InitrhoCDR();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  // if calc_UseMAEPifTailCDRisLarger is set the we may need tail MSEP
  if calc_UseMSEPifTailCDRisLarger in FCalculationFlags then begin
    If not(Init_MSEPProcVarTailDetail in FInitialised) then Result:=InitMSEPProcVarTailProxyDetail();
    if Result<>'' then Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRTailDetail)<=0 then begin
      //                     Proxy Part    m1     m2     i1     i2
      SetLength(FCDRTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_CDRProcVarTailProxyDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=0;
      PProxyDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FCDRTailDetail[PDestDouble],@FCDRTailDetail[PProxyDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the CDR tail process variance without proxy'}
      // store the results within FCDRTailDetail(0,0)
      PDestDouble:=0;
      ZeroMemory(@FCDRTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      if calc_ReduceMemSize in FCalculationFlags then begin
        {$REGION 'calculate rhos each time'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'process variance non proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 +FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance non proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance non proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance non proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                if (calc_UseMSEPifTailCDRisLarger in FCalculationFlags)
                    and (FCDRTailDetail[PDestDouble]>FMSEPTailDetail[PDestDouble]) then
                  FCDRTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble];
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else begin
        {$REGION 'take rhos from memory'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'process variance non proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance non proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance non proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance non proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                if (calc_UseMSEPifTailCDRisLarger in FCalculationFlags)
                    and (FCDRTailDetail[PDestDouble]>FMSEPTailDetail[PDestDouble]) then
                  FCDRTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble];
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRProcVarTailDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.InitCDRProcVarTailProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhobarDouble, PCoorF1Double, PCoorF2Double,
      PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoCDR in FInitialised) then
    Result:=InitrhoCDR();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  // if calc_UseMAEPifTailCDRisLarger is set the we may need tail MSEP
  if calc_UseMSEPifTailCDRisLarger in FCalculationFlags then begin
    If not(Init_MSEPProcVarTailDetail in FInitialised) then Result:=InitMSEPProcVarTailProxyDetail();
    if Result<>'' then Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRTailDetail)<=0 then begin
      //                     Proxy Part    m1     m2     i1     i2
      SetLength(FCDRTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_CDRProcVarTailDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=0;
      PProxyDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FCDRTailDetail[PProxyDouble],@FCDRTailDetail[PDestDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the CDR tail process variance proxy'}
      // store the results within FCDRTailDetail(1,0)
      PDestDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FCDRTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      if calc_ReduceMemSize in FCalculationFlags then begin
        {$REGION 'calculate rhos each time'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'process variance proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 +FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                if (calc_UseMSEPifTailCDRisLarger in FCalculationFlags)
                    and (FCDRTailDetail[PDestDouble]>FMSEPTailDetail[PDestDouble]) then
                  FCDRTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble];
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else begin
        {$REGION 'take rhos from memory'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'process variance proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance proxy: k1 k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                if (calc_UseMSEPifTailCDRisLarger in FCalculationFlags)
                    and (FCDRTailDetail[PDestDouble]>FMSEPTailDetail[PDestDouble]) then
                  FCDRTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble];
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRProcVarTailProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.InitCoorF: string;
var m, k, i, cl, l, h, j: Integer;
    PDestDouble, PElhjDouble, PTmp1Double: IndexDouble;
    StartTime, EndTime: Int64;
    Ta:T1Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCoorF + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_Gamma1 in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCoorF + sLineBreak + rsLSRM_NotInitialisedGamma1;
    Exit;
  end;
  if ([Init_f, Init_Defaultf]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitDefaultsigma + sLineBreak + rsLSRM_NotInitialisedf;
    Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCoorF)<=0 then begin
      //                 m       k    l      j
      Setlength(FCoorF,(FM+1)*(FJ+FP)*(FM+1)*(FJ+1+FP));
    end;
    // initialise temporare array
    SetLength(TA,(FM+1)*(FI+1)*(FJ+1+FP));
    For m := 0 To FM do begin
      i:=0;
        For k := 0 To FJ - 1 + FP do begin
          PDestDouble:=((m*(FJ+FP)+k)*(FM+1)+m)*(FJ+1+FP)+k+1;
          // for n=h+j>i+k we have (F^n,m_i,k x)^l_h,j = 1 iff l=m,h=i, j=k+1
          FCoorF[PDestDouble]:=1;
          for cl:=0 to FmlDependency[m,FM] do begin
            l:=FmlDependency[m,cl];
              h:=i;
              PElhjDouble:=(l*(FI+1)+h)*(FJ+1+FP);
              For j := 0 To Min(FJ - 1 + FP, i + k - h) do begin
                // set the (l, h, j) coordinate of the temporare array to one, all other coordinates to zero
                ZeroMemory(@TA[0],(FM+1)*(FI+1)*(FJ+1+FP)*SizeOf(Double));
                Ta[PElhjDouble]:=1;
                // apply F
                Result:=OP_Fn(h+j,i+k,@Ta[0]);
                if Result<>'' then Exit;
                // get the coordinate of F
                PDestDouble:=((m*(FJ+FP)+k)*(FM+1)+l)*(FJ+1+FP)+j;
                PTmp1Double:=(m*(FI+1)+i)*(FJ+1+FP)+k+1;
                FCoorF[PDestDouble]:=TA[PTmp1Double];
                Inc(PElhjDouble,1);
              end;
          end;
        end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CoorF,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCoorF + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCoorF + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.InitGamma1(const Gamma1: OleVariant; Hasgamma1, HasNewInflation: boolean;
                                var NewInflation: Double): String;
var row, m, l, k, j: Integer;
    StartTime, EndTime: Int64;
    Formula: String;
    Dep_ml: TLSRMDependencies; // store which l is needed for which m in double sums
    Formulas: TFormulaParser;
    ParseParas: TParseParas;
    CalcParas: TCalcParas;
    TmpDouble: PDouble;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitGamma1 + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  {$ENDREGION}
  Dep_ml:=nil;
  Formulas:=nil;
  try
    QueryPerformanceCounter(StartTime);
    Dep_ml:=TLSRMDependencies.Create(FM);
    // independend of Gamma1 (operator F is the identity in special cases) we have the dependencies m,m
    For m:=0 to FM do Dep_ml.AddDependsOn(m,m);
    // initialise the array or clear the corresponding memory
    if Length(FGamma1)<=0 then begin
      //                   m     l    k   j
      SetLength(FGamma1,(FM+1)*(FM+1)*(FJ+FK)*(FJ+FK));
    end else begin
      ZeroMemory(@FGamma1[0],(FM+1)*(FM+1)*(FJ+FK)*(FJ+FK)*SizeOf(Double));
    end;
    if (VarType(Gamma1) and VarArray)=0 then begin
      {$REGION 'Gamma1 is not an array, therefore it has to be a formula string'}
      Formula:=Gamma1;
      // parse the formulas
      ParseParas.M:=FM;
      ParseParas.J:=FJ+FK-1;
      ParseParas.OnlyLinear:=true;
      Formulas:=TFormulaParser.Create(FTokenChars);
      Formulas.CaseSensitive:=false;
      try
        Formulas.Parse(Formula,@ParseParas);
      except
        On E:EParseException do begin
          E.Message:=Formulas.LastErrorMessage(E);
          raise;
        end else raise;
      end;
      // check the number of formulas
      if Formulas.Count<FM+1 then begin
        Result:='LSRM ''' + FName + ''':' + rsLSRM_WrongFormulaGamma1_ToofewFormulas;
        FreeAndNil(Dep_ml);
        FreeAndNil(Formulas);
        Exit;
      end;
      if Formulas.Count>FM+1 then begin
        Result:='LSRM ''' + FName + ''':' + rsLSRM_WrongFormulaGamma1_ToomanyFormulas;
        FreeAndNil(Dep_ml);
        FreeAndNil(Formulas);
        Exit;
      end;
      // go through each formula and calculate gamma1
      CalcParas.M:=FM;
      CalcParas.I:=FI;
      CalcParas.J:=FJ+FK-1;
      CalcParas.Mode:=CalcGamma1;
      CalcParas.Dep_ml:=Dep_ml;
      TmpDouble:=@FGamma1[0];
      for m := 0 to FM do begin
        CalcParas.m1:=m;
        Formulas.Formula[m].Calc(TmpDouble,@CalcParas);
        inc(TmpDouble,(FM+1)*(FJ+FK)*(FJ+FK));
      end;
      {$ENDREGION}
    end else begin
      {$REGION 'Gamma1 contains an array of values'}
      for row:=VarArrayLowBound(Gamma1,1) to VarArrayHighBound(Gamma1,1) do begin
        // set value
        m:=Gamma1[row,0];
        if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
        l:=Gamma1[row,1];
        if (l<0) or (l>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
        k:=Gamma1[row,2];
        if (k<0) or (k>FJ-1+FK) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
        j:=Gamma1[row,3];
        if (j<0) or (j>FJ-1+FK) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
        FGamma1[((m*(FM+1)+l)*(FJ+FK)+k)*(FJ+FK)+j]:=Gamma1[row,4];
        // add dependencies
        Dep_ml.AddDependsOn(Gamma1[row,0],Gamma1[row,1]);
      end;
      {$ENDREGION}
    end;
    // fill the corresponding dependency arrays
    Dep_ml.FillDependenceArray(FmlDependency);
    // clean up
    FreeAndNil(Dep_ml);
    FreeAndNil(Formulas);
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_Gamma1,EndTime-StartTime);
  except
    On E:Exception do begin
      FreeAndNil(Dep_ml);
      FreeAndNil(Formulas);
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitGamma1 + sLineBreak + E.Message;
    end else begin
      FreeAndNil(Dep_ml);
      FreeAndNil(Formulas);
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitGamma1 + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.InitGamma2(const Gamma2: OleVariant; Hasgamma2, HasNewInflation: boolean;
                                var NewInflation: Double): String;
var row, m1, m2, l, k, j: Integer;
    StartTime, EndTime: Int64;
    Formula: String;
    ParseParas: TParseParas;
begin
  // we do not try to reduce the complexity in sums over m1,m2,l, because
  // such sums are only appear in the calculation of R^m1,m2_i,k
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitGamma2 + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    if (VarType(Gamma2) and VarArray)=0 then begin
      {$REGION 'Gamma2 is not an array, therefore it has to be a formula string'}
      Formula:=Gamma2;
      // parse the formulas
      ParseParas.M:=FM;
      ParseParas.J:=FJ+FP-1;
      ParseParas.OnlyLinear:=false;
      FFormulas.CaseSensitive:=false;
      try
        FFormulas.Parse(Formula,@ParseParas);
      except
        On E:EParseException do begin
          E.Message:=FFormulas.LastErrorMessage(E);
          raise;
        end else raise;
      end;
      // check the number of formulas
      if FFormulas.Count<(FM+1)*(FM+1) then begin
        Result:='LSRM ''' + FName + ''':' + rsLSRM_WrongFormulaGamma2_ToofewFormulas;
        Exit;
      end;
      if FFormulas.Count>(FM+1)*(FM+1) then begin
        Result:='LSRM ''' + FName + ''':' + rsLSRM_WrongFormulaGamma2_ToomanyFormulas;
        Exit;
      end;
      {$ENDREGION}
    end else begin
      {$REGION 'Gamma2 contains an array of values'}
      // initialise the array or clear the corresponding memory
      if Length(FGamma2)<=0 then begin
        //                   m1     m2     l   k  j
        SetLength(FGamma2,(FM+1)*(FM+1)*(FM+1)*(FJ+FP)*(FJ+FP));
      end else begin
        ZeroMemory(@FGamma2[0],(FM+1)*(FM+1)*(FM+1)*(FJ+FP)*(FJ+FP)*SizeOf(Double));
      end;
      for row:=VarArrayLowBound(Gamma2,1) to VarArrayHighBound(Gamma2,1) do begin
        // set value
        m1:=Gamma2[row,0];
        if (m1<0) or (m1>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
        m2:=Gamma2[row,1];
        if (m2<0) or (m2>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
        l:=Gamma2[row,2];
        if (l<0) or (l>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
        k:=Gamma2[row,3];
        if (k<0) or (k>FJ-1+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
        j:=Gamma2[row,4];
        if (j<0) or (j>FJ-1+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
        FGamma2[(((m1*(FM+1)+m2)*(FM+1)+l)*(FJ+FP)+k)*(FJ+FP)+j]:=Gamma2[row,5];
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_Gamma2,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitGamma2 + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitGamma2 + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.InitMSEPParaErrProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhostarDouble, PCoorF1Double, PCoorF2Double: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, h1, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  If not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoMSEP in FInitialised) then
    Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPDetail)<=0 then begin
      //                  Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    // parameter error proxy
    // store the results within FMSEPDetail(1,1)
    PDestDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    ZeroMemory(@FMSEPDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
    if calc_ReduceMemSize in FCalculationFlags then begin
      {$REGION 'calculate rhos each time'}
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1 := i1;
                          h2 := i2;
                            For cl1 := 0 To FmlDependency[m1,FM] do begin
                              l1 := FmlDependency[m1,cl1];
                              For cl2 := 0 To FmlDependency[m2,FM] do begin
                                l2 := FmlDependency[m2,cl2];
                                PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                Rho:=Getrho(1,l1,l2,h1,h2,j);
                                TmpValue:=Rho
                                          *FR1[PR11Double]
                                          *FR1[PR12Double]
                                          *FCoorF[PCoorF1Double]
                                          *FCoorF[PCoorF2Double];
                                // diversification only if rho is negative
                                if (calc_VirtualParaErrDiversification in FCalculationFlags)
                                        or (Rho<=0) or (TmpValue>=0) then
                                  FMSEPDetail[PDestDouble]:=FMSEPDetail[PDestDouble]+TmpValue;
                              end;
                            end;
                  end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end else begin
      {$REGION 'take rhos from memory'}
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1 := i1;
                          h2 := i2;
                            For cl1 := 0 To FmlDependency[m1,FM] do begin
                              l1 := FmlDependency[m1,cl1];
                              For cl2 := 0 To FmlDependency[m2,FM] do begin
                                l2 := FmlDependency[m2,cl2];
                                PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                Rho:=Frho[PrhostarDouble];
                                TmpValue:=Rho
                                          *FR1[PR11Double]
                                          *FR1[PR12Double]
                                          *FCoorF[PCoorF1Double]
                                          *FCoorF[PCoorF2Double];
                                // diversification only if rho is negative
                                if (calc_VirtualParaErrDiversification in FCalculationFlags)
                                        or (Rho<=0) or (TmpValue>=0) then
                                  FMSEPDetail[PDestDouble]:=FMSEPDetail[PDestDouble]+TmpValue;
                              end;
                            end;
                  end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPParaErrProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.InitMSEPParaErrTailDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhostarDouble, PCoorF1Double, PCoorF2Double,
      PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoMSEP in FInitialised) then
    Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPTailDetail)<=0 then begin
      //                      Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_MSEPParaErrTailProxyDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PProxyDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FMSEPTailDetail[PDestDouble],@FMSEPTailDetail[PProxyDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the MSEP tail parameter error without proxy'}
      // store the results within FMSEPTailDetail(0,1)
      PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FMSEPTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      if calc_ReduceMemSize in FCalculationFlags then begin
        {$REGION 'calculate rhos each time'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error non proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 +FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else begin
        {$REGION 'take rhos from memory'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error non proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPParaErrTailDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.InitMSEPParaErrTailProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhostarDouble, PCoorF1Double, PCoorF2Double,
      PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoMSEP in FInitialised) then
    Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPTailDetail)<=0 then begin
      //                      Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_MSEPParaErrTailDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PProxyDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FMSEPTailDetail[PProxyDouble],@FMSEPTailDetail[PDestDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the MSEP tail parameter error proxy'}
      // store the results within FMSEPTailDetail(1,1)
      PDestDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FMSEPTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      if calc_ReduceMemSize in FCalculationFlags then begin
        {$REGION 'calculate rhos each time'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 +FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1^, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else begin
        {$REGION 'take rhos from memory'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPParaErrTailProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.InitMSEPProcVarDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PsigmaDouble, PR2Double, PCoorF1Double, PCoorF2Double, PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, h, j, cl1, l1, cl2, l2: Integer;
    sigma, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_sigma, Init_Defaultsigma]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarDetail + sLineBreak
            + rsLSRM_NotInitialisedsigma;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  If not(Init_R2Full in FInitialised) then Result:=InitR2Full();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPDetail)<=0 then begin
      //                  Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_MSEPProcVarProxyDetail in FInitialised)
        and (calc_VirtualProcVarDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=0;
      PProxyDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FMSEPDetail[PDestDouble],@FMSEPDetail[PProxyDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the MSEP process variance without proxy'}
      // store the results within FMSEPDetail(0,0)
      PDestDouble:=0;
      ZeroMemory(@FMSEPDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            // only for i1=i2=h both FCoorF factors are possible not zero
            i2:=i1;
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FI-i1 To Min(k1,k2) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          FMSEPDetail[PDestDouble]:=FMSEPDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              // next i1=i2
              Inc(PDestDouble,FI+2);
          end;
          // next i1=i1 is at 0=0 and not at FI+1=FI+1
          Dec(PDestDouble,FI+1);
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPProcVarDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.InitMSEPProcVarProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PsigmaDouble, PR2Double, PCoorF1Double, PCoorF2Double, PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, h, j, cl1, l1, cl2, l2: Integer;
    sigma, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_sigma, Init_Defaultsigma]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedsigma;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  If not(Init_R2Full in FInitialised) then Result:=InitR2Full();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPDetail)<=0 then begin
      //                  Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_MSEPProcVarDetail in FInitialised)
        and (calc_VirtualProcVarDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=0;
      PProxyDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FMSEPDetail[PProxyDouble],@FMSEPDetail[PDestDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the MSEP process variance proxy'}
      // store the results within FMSEPDetail(1,0)
      PDestDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FMSEPDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            // only for i1=i2=h both FCoorF factors are possible not zero
            i2:=i1;
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := Min(FI-i1,FJ-1) To Min(k1,k2) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          // diversification only if sigma is negative
                          if (calc_VirtualProcVarDiversification in FCalculationFlags)
                                  or (sigma<=0) or (TmpValue>=0) then
                            FMSEPDetail[PDestDouble]:=FMSEPDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              // next i1=i2
              Inc(PDestDouble,FI+2);
          end;
          // next i1=i1 is at 0=0 and not at FI+1=FI+1
          Dec(PDestDouble,FI+1);
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPProcVarProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.InitMSEPProcVarTailDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PsigmaDouble, PR2Double, PCoorF1Double, PCoorF2Double, PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, h, j, cl1, l1, cl2, l2: Integer;
    sigma, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_sigma, Init_Defaultsigma]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailDetail + sLineBreak
            + rsLSRM_NotInitialisedsigma;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  If not(Init_R2Full in FInitialised) then Result:=InitR2Full();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPTailDetail)<=0 then begin
      //                      Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_MSEPProcVarTailProxyDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=0;
      PProxyDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FMSEPTailDetail[PDestDouble],@FMSEPTailDetail[PProxyDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the MSEP tail process variance without proxy'}
      // store the results within FMSEPTailDetail(0,0)
      PDestDouble:=0;
      ZeroMemory(@FMSEPTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            // only for i1=i2=h both FCoorF factors are possible not zero
            i2:=i1;
              {$REGION 'proc varince non proxy: j Tail, k1 and k2 not'}
                // not possible since j<=Min(k1,k2)
              {$ENDREGION}
              {$REGION 'proc varince non proxy: k1 Tail, k2 and j not'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FI-i1 To Min(k1,k2) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'proc varince non proxy: k2 Tail, k1 and j not'}
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FI-i1 To Min(k1,k2) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'proc varince non proxy: k1 and k2 Tail, j not'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FI-i1 To Min(Min(k1,k2),FJ-1) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'proc varince non proxy: k1, k2 and j Tail'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FJ To Min(k1,k2) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'proc varince non proxy: k1 and j Tail, k2 not'}
                // not possible since j<=Min(k1,k2)
              {$ENDREGION}
              {$REGION 'proc varince non proxy: k2 and j Tail, k1 not'}
                // not possible since j<=Min(k1,k2)
              {$ENDREGION}
              // next i1=i2
              Inc(PDestDouble,FI+2);
          end;
          // next i1=i1 is at 0=0 and not at FI+1=FI+1
          Dec(PDestDouble,FI+1);
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPProcVarTailDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.InitMSEPProcVarTailProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PsigmaDouble, PR2Double, PCoorF1Double, PCoorF2Double, PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, h, j, cl1, l1, cl2, l2: Integer;
    sigma, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_sigma, Init_Defaultsigma]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedsigma;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  If not(Init_R2Full in FInitialised) then Result:=InitR2Full();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPTailDetail)<=0 then begin
      //                      Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_MSEPProcVarTailDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=0;
      PProxyDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FMSEPTailDetail[PProxyDouble],@FMSEPTailDetail[PDestDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the MSEP tail process variance proxy'}
      // store the results within FMSEPTailDetail(1,0)
      PDestDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FMSEPTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            // only for i1=i2=h both FCoorF factors are possible not zero
            i2:=i1;
              {$REGION 'proc varince proxy: j Tail, k1 and k2 not'}
                // not possible since j<=Min(k1,k2)
              {$ENDREGION}
              {$REGION 'proc varince proxy: k1 Tail, k2 and j not'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FI-i1 To Min(k1,k2) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          // diversification only if sigma is negative
                          if (calc_VirtualTailDiversification in FCalculationFlags)
                                  or (sigma<=0) or (TmpValue>=0) then
                            FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'proc varince proxy: k2 Tail, k1 and j not'}
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FI-i1 To Min(k1,k2) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          // diversification only if sigma is negative
                          if (calc_VirtualTailDiversification in FCalculationFlags)
                                  or (sigma<=0) or (TmpValue>=0) then
                            FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'proc varince proxy: k1 and k2 Tail, j not'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FI-i1 To Min(Min(k1,k2),FJ-1) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          // diversification only if sigma is negative
                          if (calc_VirtualTailDiversification in FCalculationFlags)
                                  or (sigma<=0) or (TmpValue>=0) then
                            FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'proc varince proxy: k1, k2 and j Tail'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FJ To Min(k1,k2) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=((m1*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=((m2*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          // diversification only if sigma is negative
                          if (calc_VirtualTailDiversification in FCalculationFlags)
                                  or (sigma<=0) or (TmpValue>=0) then
                            FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'proc varince proxy: k1 and j Tail, k2 not'}
                // not possible since j<=Min(k1,k2)
              {$ENDREGION}
              {$REGION 'proc varince proxy: k2 and j Tail, k1 not'}
                // not possible since j<=Min(k1,k2)
              {$ENDREGION}
              // next i1=i2
              Inc(PDestDouble,FI+2);
          end;
          // next i1=i1 is at 0=0 and not at FI+1=FI+1
          Dec(PDestDouble,FI+1);
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPProcVarTailProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.InitR1: String;
var m, k, i, l, h, j, cl: Integer;
    PDestDouble, Pgamma1Double, PSDouble: IndexDouble;
    StartTime, EndTime: Int64;
    sign: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR1 + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_S in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR1 + sLineBreak + rsLSRM_NotInitialisedS;
    Exit;
  end;
  If not(Init_gamma1 in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR1 + sLineBreak + rsLSRM_NotInitialisedgamma1;
    Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array or clear corresponding memory
    if Length(FR1)<=0 then begin
      //               m     i       k
      SetLength(FR1,(FM+1)*(FI+1)*(FJ+FK));
    end else begin
      ZeroMemory(@FR1[0],(FM+1)*(FI+1)*(FJ+FK)*SizeOf(Double));
    end;
    // calculate
    PDestDouble:=0;
    for m:=0 to FM do begin
      for i:=0 to FI do begin
        sign:=0;
        for k := 0 To Min(FJ - 1+FP, FI-i) do begin
          for cl:=0 to FmlDependency[m,FM] do begin
            l:=FmlDependency[m,cl];
              h:=i;
              PGamma1Double:=((m*(FM+1)+l)*(FJ+FK)+k)*(FJ+FK);
              PSDouble:=(l*(FI+1)+h)*(FJ+1+FK);
              for j:=0 to k do begin
                FR1[PDestDouble]:=FR1[PDestDouble]
                                  +FGamma1[PGamma1Double]
                                  *FS[PSDouble];
                Inc(PGamma1Double);
                Inc(PSDouble);
              end;
          end;
          // prevent exposure from changing sign
          if calc_PreventPastExposureFromChangingSign in FCalculationFlags then begin
            if sign=0 then begin
              sign:=FR1[PDestDouble];
            end else begin
              if ((sign>0) and (FR1[PDestDouble]<0)) or ((sign<0) and (FR1[PDestDouble]>0)) then
                FR1[PDestDouble]:=0;
            end;
          end;
          Inc(PDestDouble,1);
        end;
        Inc(PDestDouble,FJ-1+FK-Min(FJ - 1 + FP, FI-i));
     end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_R1,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR1 + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR1 + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.InitR2: String;
var m1, m2, k, i, l, h, j: Integer;
    PDestDouble, PGamma2Double, PSDouble: IndexDouble;
    StartTime, EndTime: Int64;
    sign: Double;
    CalcParas: TCalcParas;
    TmpDouble: PDouble;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2 + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_S in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2 + sLineBreak + rsLSRM_NotInitialisedS;
    Exit;
  end;
  If not(Init_gamma2 in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2 + sLineBreak + rsLSRM_NotInitialisedgamma2;
    Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array or clear corresponding memory
    if Length(FR2)<=0 then begin
      //               m1    m2     i     k
      SetLength(FR2,(FM+1)*(FM+1)*(FI+1)*(FJ+FP));
    end else begin
      ZeroMemory(@FR2[0],(FM+1)*(FM+1)*(FI+1)*(FJ+FP)*SizeOf(Double));
    end;
    // calculate
    if length(FGamma2)>0 then begin
      {$REGION 'R2 has been specified by an gamma2 array'}
      PDestDouble:=0;
      for m1:=0 to FM do begin
        for m2:=0 to FM do begin
          for i:=0 to FI do begin
            sign:=0;
            for k := 0 To Min(FJ - 1 + FP, FI-i) do begin
              for l:=0 to FM do begin
                h:=i;
                  PGamma2Double:=(((m1*(FM+1)+m2)*(FM+1)+l)*(FJ+FP)+k)*(FJ+FP);
                  PSDouble:=(l*(FI+1)+h)*(FJ+1+FK);
                  for j:=0 to k do begin
                    FR2[PDestDouble]:=FR2[PDestDouble]
                                      +FGamma2[PGamma2Double]
                                      *FS[PSDouble];
                    Inc(PGamma2Double,1);
                    Inc(PSDouble,1);
                  end;
              end;
              // prevent exposure from changing sign
              if calc_PreventPastExposureFromChangingSign in FCalculationFlags then begin
                if sign=0 then begin
                  sign:=FR2[PDestDouble];
                end else begin
                  if ((sign>0) and (FR2[PDestDouble]<0)) or ((sign<0) and (FR2[PDestDouble]>0)) then
                    FR2[PDestDouble]:=0;
                end;
              end;
              Inc(PDestDouble,1);
            end;
            Inc(PDestDouble,FJ-1+FP-Min(FJ - 1 + FP, FI-i));
          end;
        end;
      end;
      {$ENDREGION}
    end else begin
      {$REGION 'R2 has been specified by a formula'}
      // go through each formula and calculate R2
      CalcParas.M:=FM;
      CalcParas.I:=FI;
      CalcParas.J:=FJ+FP-1;
      CalcParas.Mode:=CalcGamma2;
      CalcParas.LastDiag:=FI;
      CalcParas.LastColumn:=FJ+FP-1;
      CalcParas.ClaimProperty.S:=@FS[0];
      CalcParas.ClaimProperty.V:=nil; // no inflation
      CalcParas.ClaimProperty.J:=FJ+FK;
      TmpDouble:=@FR2[0];
      for m1 := 0 to FM do begin
        for m2 := 0 to FM do begin
          FFormulas.Formula[m1*(FM+1)+m2].Calc(TmpDouble,@CalcParas);
          inc(TmpDouble,(FI+1)*(FJ+FP));
        end;
      end;
      // prevent exposure from changing sign
      PDestDouble:=0;
      if calc_PreventPastExposureFromChangingSign in FCalculationFlags then begin
        for m1:=0 to FM do begin
          for m2:=0 to FM do begin
            for i:=0 to FI do begin
              sign:=0;
              for k := 0 To Min(FJ - 1 + FP, FI-i) do begin
                if sign=0 then begin
                  sign:=FR2[PDestDouble];
                end else begin
                  if ((sign>0) and (FR2[PDestDouble]<0)) or ((sign<0) and (FR2[PDestDouble]>0)) then
                    FR2[PDestDouble]:=0;
                end;
                Inc(PDestDouble,1);
              end;
              Inc(PDestDouble,FJ-1+FP-Min(FJ - 1 + FP, FI-i));
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_R2,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2 + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2 + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.InitR2Full: String;
var m1, m2, k, i, l, h, j: Integer;
    PDestDouble, PGamma2Double, PSDouble: IndexDouble;
    StartTime, EndTime: Int64;
    sign: Double;
    CalcParas: TCalcParas;
    TmpDouble: PDouble;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2Full + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_gamma2 in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2Full + sLineBreak + rsLSRM_NotInitialisedgamma2;
    Exit;
  end;
  if not(Init_R2 in FInitialised) then Result:=InitR2();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // calculate
    if length(FGamma2)>0 then begin
      {$REGION 'R2 has been specified by a gamma2 array'}
      PDestDouble:=0;
      for m1:=0 to FM do begin
        for m2:=0 to FM do begin
          for i:=0 to FI do begin
            sign:=0;
            if calc_PreventFutureExposureFromChangingSign in FCalculationFlags then begin
              for k := 0 to Min(FJ - 1 + FP, FI-i) do begin
                if sign=0 then sign:=FR2[PDestDouble];
                Inc(PDestDouble,1);
              end;
            end else begin
              Inc(PDestDouble,Min(FJ - 1 + FP, FI-i)+1);
            end;
            for k := Min(FJ - 1 + FP, FI-i)+1 To FJ-1+FP do begin
              FR2[PDestDouble]:=0;
              for l:=0 to FM do begin
                h:=i;
                  PGamma2Double:=(((m1*(FM+1)+m2)*(FM+1)+l)*(FJ+FP)+k)*(FJ+FP);
                  PSDouble:=(l*(FI+1)+h)*(FJ+1+FK);
                  for j:=0 to Min(i + k - h, k) do begin
                    FR2[PDestDouble]:=FR2[PDestDouble]
                                      +FGamma2[PGamma2Double]
                                      *FS[PSDouble];
                    Inc(PGamma2Double,1);
                    Inc(PSDouble,1);
                  end;
              end;
              // prevent exposure from changing sign
              if calc_PreventFutureExposureFromChangingSign in FCalculationFlags then begin
                if sign=0 then begin
                  sign:=FR2[PDestDouble];
                end else begin
                  if ((sign>0) and (FR2[PDestDouble]<0)) or ((sign<0) and (FR2[PDestDouble]>0)) then
                    FR2[PDestDouble]:=0;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end else begin
      {$REGION 'R2 has been specified by a formula'}
      ZeroMemory(@FR2[0],(FM+1)*(FM+1)*(FI+1)*(FJ+FP)*SizeOf(Double));
      // go through each formula and calculate R2
      CalcParas.M:=FM;
      CalcParas.I:=FI;
      CalcParas.J:=FJ+FP-1;
      CalcParas.Mode:=CalcGamma2;
      CalcParas.LastDiag:=FI+FJ+FP-1;
      CalcParas.LastColumn:=FJ+FP-1;
      CalcParas.ClaimProperty.S:=@FS[0];
      CalcParas.ClaimProperty.V:=nil; // no inflation
      CalcParas.ClaimProperty.J:=FJ+FK;
      TmpDouble:=@FR2[0];
      for m1 := 0 to FM do begin
        for m2 := 0 to FM do begin
          FFormulas.Formula[m1*(FM+1)+m2].Calc(TmpDouble,@CalcParas);
          inc(TmpDouble,(FI+1)*(FJ+FP));
        end;
      end;
      // prevent exposure from changing sign
      PDestDouble:=0;
      if [calc_PreventPastExposureFromChangingSign,calc_PreventFutureExposureFromChangingSign]*FCalculationFlags<>[] then begin
        for m1:=0 to FM do begin
          for m2:=0 to FM do begin
            for i:=0 to FI do begin
              sign:=0;
              for k := 0 To Min(FJ - 1 + FP, FI-i) do begin
                if [calc_PreventPastExposureFromChangingSign,calc_PreventFutureExposureFromChangingSign] * FCalculationFlags<>[] then begin
                  if sign=0 then begin
                    sign:=FR2[PDestDouble];
                  end else if calc_PreventPastExposureFromChangingSign in FCalculationFlags then begin
                    if ((sign>0) and (FR2[PDestDouble]<0)) or ((sign<0) and (FR2[PDestDouble]>0)) then
                      FR2[PDestDouble]:=0;
                  end;
                end;
                Inc(PDestDouble,1);
              end;
              for k := Min(FJ - 1 + FP, FI-i)+1 To FJ - 1 + FP do begin
                if calc_PreventFutureExposureFromChangingSign in FCalculationFlags then begin
                  if sign=0 then begin
                    sign:=FR2[PDestDouble];
                  end else begin
                    if ((sign>0) and (FR2[PDestDouble]<0)) or ((sign<0) and (FR2[PDestDouble]>0)) then
                      FR2[PDestDouble]:=0;
                  end;
                end;
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_R2Full,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2Full + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2Full + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleLSRM.InitUltimateDetail: string;
var m, n, k, i, l, h, j, cl: Integer;
    PR1DestDouble, PSDestDouble, Pgamma1Double, PSDouble, PfDouble, PUltimateDestDouble,
      PUltimateTailDestDouble: IndexDouble;
    StartTime, EndTime: Int64;
    TmpSum,TmpSumTail:Double;
    sign:T1Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimateDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_S in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimateDetail + sLineBreak + rsLSRM_NotInitialisedS;
    Exit;
  end;
  If not(Init_gamma1 in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimateDetail + sLineBreak
            + rsLSRM_NotInitialisedgamma1;
    Exit;
  end;
  if ([Init_f, Init_Defaultf]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimateDetail + sLineBreak + rsLSRM_NotInitialisedf;
    Exit;
  end;
  If not(Init_R1 in FInitialised) then Result:=InitR1();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array or clear corresponding memory
    if Length(FUltimate)<=0 then begin
      //                    m   i
      SetLength(FUltimate,2*(FM+2)*(FI+2));
    end else begin
      ZeroMemory(@FUltimate[0],2*(FM+2)*(FI+2)*SizeOf(Double))
    end;
    {$REGION 'get signs of the first known exposure in case of they should be prevented from changing signs'}
    if calc_PreventFutureExposureFromChangingSign in FCalculationFlags then begin
      setlength(sign,FI+1);
      PR1DestDouble:=0;
      for m:=0 to FM do begin
        for i:=0 to FI do begin
          sign[i]:=0;
          for k := 0 To Min(FJ - 1 + FP, FI-i) do begin
            if sign[i]=0 then
              sign[i]:=FR1[PR1DestDouble];
            Inc(PR1DestDouble,1);
          end;
          Inc(PR1DestDouble,FJ-1+FK-Min(FJ - 1 + FP, FI-i));
        end;
      end;
    end;
    {$ENDREGION}
    {$REGION 'calculate S and R1 diagonal by diagonal: first the tail below the diagonal FI+1'}
    for n:=FJ+1 to FI do begin
      for m:=0 to FM do begin
        for i := max(0,n-FJ-FK) to n-FJ-1 do begin
          // k=n-i
          {$REGION 'R1'}
          PR1DestDouble:=(m*(FI+1)+i)*(FJ+FK)+n-i-1;
          FR1[PR1DestDouble]:=0;
          for cl:=0 to FmlDependency[m,FM] do begin
            l:=FmlDependency[m,cl];
              h:=i;
              PGamma1Double:=((m*(FM+1)+l)*(FJ+FK)+n-i-1)*(FJ+FK);
              PSDouble:=(l*(FI+1)+h)*(FJ+1+FK);
              for j:=0 to Min(n-1 - h, n-i-1) do begin
                FR1[PR1DestDouble]:=FR1[PR1DestDouble]
                                    +FGamma1[PGamma1Double]
                                    *FS[PSDouble];
                Inc(PGamma1Double);
                Inc(PSDouble);
              end;
          end;
          // prevent exposure from changing sign
          if calc_PreventFutureExposureFromChangingSign in FCalculationFlags then begin
            if sign[i]=0 then begin
              sign[i]:=FR1[PR1DestDouble];
            end else begin
              if ((sign[i]>0) and (FR1[PR1DestDouble]<0)) or ((sign[i]<0) and (FR1[PR1DestDouble]>0)) then
                FR1[PR1DestDouble]:=0;
            end;
          end;
          {$ENDREGION}
          {$REGION 'S'}
          PfDouble:=m*(FJ+FK)+n-i-1;
          PSDestDouble:=(m*(FI+1)+i)*(FJ+1+FK)+n-i;
          FS[PSDestDouble]:=FUsedf^[PfDouble] * FR1[PR1DestDouble];
          {$ENDREGION}
        end;
      end;
    end;
    {$ENDREGION}
    {$REGION 'calculate S and R1 diagonal by diagonal: above the diagonal FI'}
    for n:=FI+1 to FI+FJ+FK do begin
      for m:=0 to FM do begin
        for i := max(0,n-FJ-FK) to FI do begin
          // k=n-i
          {$REGION 'R1'}
          PR1DestDouble:=(m*(FI+1)+i)*(FJ+FK)+n-i-1;
          // first fill up R1(m,i,n-i-1), but only if n>FI+1 or if k>FJ
          if (n>FI+1) or (n-i>FJ) then begin
            FR1[PR1DestDouble]:=0;
            for cl:=0 to FmlDependency[m,FM] do begin
              l:=FmlDependency[m,cl];
                h:=i;
                PGamma1Double:=((m*(FM+1)+l)*(FJ+FK)+n-i-1)*(FJ+FK);
                PSDouble:=(l*(FI+1)+h)*(FJ+1+FK);
                for j:=0 to Min(n-1 - h, n-i-1) do begin
                  FR1[PR1DestDouble]:=FR1[PR1DestDouble]
                                      +FGamma1[PGamma1Double]
                                      *FS[PSDouble];
                  Inc(PGamma1Double,1);
                  Inc(PSDouble,1);
                end;
            end;
            // prevent exposure from changing sign
            if calc_PreventFutureExposureFromChangingSign in FCalculationFlags then begin
              if sign[i]=0 then begin
                sign[i]:=FR1[PR1DestDouble];
              end else begin
                if ((sign[i]>0) and (FR1[PR1DestDouble]<0)) or ((sign[i]<0) and (FR1[PR1DestDouble]>0)) then
                  FR1[PR1DestDouble]:=0;
              end;
            end;
          end;
          {$ENDREGION}
          {$REGION 'S'}
          PfDouble:=m*(FJ+FK)+n-i-1;
          PSDestDouble:=(m*(FI+1)+i)*(FJ+1+FK)+n-i;
          FS[PSDestDouble]:=FUsedf^[PfDouble]*FR1[PR1DestDouble];
          {$ENDREGION}
        end;
      end;
    end;
    {$ENDREGION}
    {$REGION 'fill up the sum over all accident periods'}
    PUltimateDestDouble:=0;
    PUltimateTailDestDouble:=(FM+2)*(FI+2);
    PSDouble:=0;
    for m := 0 to FM do begin
      TmpSum:=0;
      TmpSumTail:=0;
      for i := 0 to FI do begin
        {$REGION 'without Tail'}
        for k := 0 to FJ do begin
          FUltimate[PUltimateDestDouble]:=FUltimate[PUltimateDestDouble]+FS[PSDouble];
          Inc(PSDouble,1);
        end;
        TmpSum:=TmpSum + FUltimate[PUltimateDestDouble];
        FUltimate[PUltimateTailDestDouble]:=FUltimate[PUltimateDestDouble];
        {$ENDREGION}
        {$REGION 'Tail'}
        for k := FJ+1 to FJ+FK do begin
          FUltimate[PUltimateTailDestDouble]:=FUltimate[PUltimateTailDestDouble]+FS[PSDouble];
          Inc(PSDouble,1);
        end;
        TmpSumTail:=TmpSumTail + FUltimate[PUltimateTailDestDouble];
        {$ENDREGION}
        Inc(PUltimateDestDouble,1);
        Inc(PUltimateTailDestDouble,1);
      end;
      FUltimate[PUltimateDestDouble]:=TmpSum;
      FUltimate[PUltimateTailDestDouble]:=TmpSumTail;
      Inc(PUltimateDestDouble,1);
      Inc(PUltimateTailDestDouble,1);
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_UltimateDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimateDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimateDetail + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

{$ENDREGION}

{$ENDREGION}

{$ENDREGION}

{$REGION 'TSimpleInflationLSRM'}

{$REGION '  TSimpleInflationLSRM private'}

procedure TSimpleInflationLSRM.SetInitFlags(InitFlag: TLSRMInitialisedElement; UsedTime: Int64;
  UpdateTime: Boolean);
begin
  case InitFlag of
    Init_InflationRate: begin {$REGION}
      include(FInitialised,InitFlag);
      // since gammas and inflation are handled separately gammas do not depend in inflation
      FInitialised:=FInitialised-(cInfluencedByInflationRate-[Init_Gamma1,Init_Gamma2]);
      if UpdateTime then UpdateCalcInfo(InitFlag, cInfluencedByInflationRate, UsedTime);
    end; {$ENDREGION}
    else inherited;
  end;
end;

{$ENDREGION}

{$REGION '  TSimpleInflationLSRM protected'}

{$REGION 'property get functions'}

function TSimpleInflationLSRM.GetCoorF(m, i, k, l, h, j: integer): Double;
  var ErrorMsg: string;
begin
  {$REGION 'test indecies'}
  if (m<0) or (m>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (i<0) or (i>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
  if (k<0) or (k>FJ-1+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
  if (l<0) or (l>FM) then raise ELSRMRangeError.Create(rsLSRM_NoSuchClaimProperty);
  if (h<0) or (h>FI) then raise ELSRMRangeError.Create(rsLSRM_NoSuchAccidentPeriod);
  if (j<0) or (j>FJ+FP) then raise ELSRMRangeError.Create(rsLSRM_NoSuchDevelopmentPeriod);
  {$ENDREGION}
  // calculate the coordinates of the operator F, if necessary
  if not(Init_CoorF in FInitialised) then ErrorMsg:=InitCoorF();
  if ErrorMsg<>'' then raise ELSRMInitError.Create(ErrorMsg);
  //return the coordinates (F^m_ik)^l_hj of the operator F
  Result:=FCoorF[(((m*(FI+1)+i)*(FJ+FP)+k)*(FM+1)+l)*(FJ+1+FP)+j]
end;

{$ENDREGION}

{$REGION '    operators'}

function TSimpleInflationLSRM.OP_F(n: integer; S: T1Double): string;
var m, i, cl, l, h, j: Integer;
    TmpDestDouble, TmpSDouble, Tmpgamma1Double, TmpfDouble, PInflation1, PInflation2: IndexDouble;
begin
  Result:='';
  try
    For m := 0 To FM do begin
      For i := Max(0, n + 1 - FJ - FP) To Min(FI, n) do begin
        // k = n-i
        TmpDestDouble:=(m*(FI+1)+i)*(FJ+1+FP)+n-i+1;
        PInflation1:=(m*(FI+1)+i)*(FJ+1+FK)+n-i+1;
        S[TmpDestDouble]:=0;
        TmpfDouble:=m*(FJ+FK)+n-i;
        for cl:=0 to FmlDependency[m,FM] do begin
          l:=FmlDependency[m,cl];
            h:=i;
            TmpSDouble:=(l*(FI+1)+h)*(FJ+1+FP);
            Tmpgamma1Double:=((m*(FM+1)+l)*(FJ+FK)+(n-i))*(FJ+FK);
            PInflation2:=(l*(FI+1)+h)*(FJ+1+FK);
            For j := 0 To Min(n - h, n - i) do begin
              S[TmpDestDouble]:=S[TmpDestDouble]
                                + FGamma1[TmpGamma1Double]
                                *FInflation[PInflation1]
                                /FInflation[PInflation2]
                                *S[TmpSDouble];
              Inc(TmpSDouble,1);
              Inc(Tmpgamma1Double,1);
              Inc(PInflation2,1);
            end;
        end;
        S[TmpdestDouble]:=S[TmpDestDouble] * FUsedf^[TmpfDouble];
      end;
    end;
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorOP_F + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorOP_F + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleInflationLSRM.OP_H(n, k: integer; tau, SxS: PDouble): string;
var m1, m2, i1, i2, k1, k2, cl1, cl2, l1, l2, h1, h2, j1, j2: Integer;
    PDestDouble, PSxSDouble, PTauDouble: PDouble;
    PGamma11Double, PGamma12Double, Pf1Double, Pf2Double, PInflation11, PInflation12, PInflation21,
      PInflation22: IndexDouble;
begin
  Result:='';
  try
    Pf1Double:=k;
    For m1 := 0 To FM do begin
      Pf2Double:=k;
      For m2 := 0 To FM do begin
        {$REGION 'k1=k+1 and k2=k+1 and i1+k1>n and i2+k2>n'}
        For i1 := n - k To FI do begin
          PInflation11:=(m1*(FI+1)+i1)*(FJ+1+FK)+k+1;
          For i2 := n - k To FI do begin
            PInflation21:=(m2*(FI+1)+i2)*(FJ+1+FK)+k+1;
            PDestDouble:=SxS;
            Inc(PDestDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k+1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k+1);
            PDestDouble^:=0;
            for cl1:=0 to FmlDependency[m1,FM] do begin
              l1:=FmlDependency[m1,cl1];
              for cl2:=0 to FmlDependency[m2,FM] do begin
                l2:=FmlDependency[m2,cl2];
                  h1:=i1;
                    h2:=i2;
                    PGamma11Double:=((m1*(FM+1)+l1)*(FJ+FK)+k)*(FJ+FK);
                    PInflation12:=(l1*(FI+1)+h1)*(FJ+1+FK);
                    For j1 := 0 To k do begin
                      PGamma12Double:=((m2*(FM+1)+l2)*(FJ+FK)+k)*(FJ+FK);
                      PInflation22:=(l2*(FI+1)+h2)*(FJ+1+FK);
                      For j2 := 0 To k do begin
                        PSxSDouble:=SxS;
                        Inc(PSxSDouble,((((l1*(FI+1)+h1)*(FJ+1+FP)+j1)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j2);
                        PDestDouble^:=PDestDouble^
                                      +FGamma1[PGamma11Double]
                                      *FInflation[PInflation11]
                                      /FInflation[PInflation12]
                                      *FGamma1[PGamma12Double]
                                      *FInflation[PInflation21]
                                      /FInflation[PInflation22]
                                      *PSxSDouble^;
                        Inc(PGamma12Double,1);
                        Inc(PInflation22,1);
                      end;
                      Inc(PGamma11Double,1);
                      Inc(PInflation12,1);
                    end;
              end;
            end;
            PtauDouble:=tau;
            Inc(PtauDouble,(((m1*(FM+1)+m2)*(FI+1)+i1)*(FI+1)+i2)*(FJ+FP)+k);
            PDestDouble^:=(FUsedf^[Pf1Double] * FUsedf^[Pf2Double]  + PTauDouble^) * PDestDouble^;
          end;
        end;{$ENDREGION}
        {$REGION 'k1=k+1 and i1+k1>n and k2<k+1'}
        For i1 := n - k To FI do begin
          PInflation11:=(m1*(FI+1)+i1)*(FJ+1+FK)+k+1;
          For i2 := 0 To FI do begin
            For k2 := 0 To k do begin
              PDestDouble:=SxS;
              Inc(PDestDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k+1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k2);
              PDestDouble^:=0;
              for cl1:=0 to FmlDependency[m1,FM] do begin
                l1:=FmlDependency[m1,cl1];
                  h1:=i1;
                  PInflation12:=(l1*(FI+1)+h1)*(FJ+1+FK);
                  PGamma11Double:=((m1*(FM+1)+l1)*(FJ+FK)+k)*(FJ+FK);
                  For j1 := 0 To k do begin
                    PSxSDouble:=SxS;
                    Inc(PSxSDouble,((((l1*(FI+1)+h1)*(FJ+1+FP)+j1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k2);
                    PDestDouble^:=PDestDouble^
                                  +FGamma1[PGamma11Double]
                                  *FInflation[PInflation11]
                                  /FInflation[PInflation12]
                                  *PSxSDouble^;
                    Inc(PGamma11Double,1);
                    Inc(PInflation12,1);
                  end;
              end;
              PDestDouble^:=FUsedf^[Pf1Double] * PDestDouble^;
            end;
          end;
        end; {$ENDREGION}
        {$REGION 'k1=k+1 and i1+k1>n and k2=k+1 and i2+k2<=n'}
        For i1 := n - k To FI do begin
          PInflation11:=(m1*(FI+1)+i1)*(FJ+1+FK)+k+1;
          For i2 := 0 To n - k - 1 do begin
            PDestDouble:=SxS;
            Inc(PDestDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k+1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k+1);
            PDestDouble^:=0;
            for cl1:=0 to FmlDependency[m1,FM] do begin
              l1:=FmlDependency[m1,cl1];
                h1:=i1;
                PInflation12:=(l1*(FI+1)+h1)*(FJ+1+FK);
                PGamma11Double:=((m1*(FM+1)+l1)*(FJ+FK)+k)*(FJ+FK);
                For j1 := 0 To k do begin
                  PSxSDouble:=SxS;
                  Inc(PSxSDouble,((((l1*(FI+1)+h1)*(FJ+1+FP)+j1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k+1);
                  PDestDouble^:=PDestDouble^
                                +FGamma1[PGamma11Double]
                                *FInflation[PInflation11]
                                /FInflation[PInflation12]
                                *PSxSDouble^;
                  Inc(PGamma11Double,1);
                  Inc(PInflation12,1);
                end;
            end;
            PDestDouble^:=FUsedf^[Pf1Double] * PDestDouble^;
          end;
        end; {$ENDREGION}
        {$REGION 'k1<k+1 and k2=k+1 and i2+k2>n'}
        For i1 := 0 To FI do begin
          For i2 := n - k To FI do begin
            PInflation21:=(m2*(FI+1)+i2)*(FJ+1+FK)+k+1;
            For k1 := 0 To k do begin
              PDestDouble:=SxS;
              Inc(PDestDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k+1);
              PDestDouble^:=0;
              for cl2:=0 to FmlDependency[m2,FM] do begin
                l2:=FmlDependency[m2,cl2];
                  h2:=i2;
                  PInflation22:=(l2*(FI+1)+h2)*(FJ+1+FK);
                  PGamma12Double:=((m2*(FM+1)+l2)*(FJ+FK)+k)*(FJ+FK);
                  For j2 := 0 To k do begin
                    PSxSDouble:=SxS;
                    Inc(PSxSDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k1)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j2);
                    PDestDouble^:=PDestDouble^
                                  +FGamma1[PGamma12Double]
                                  *FInflation[PInflation21]
                                  /FInflation[PInflation22]
                                  *PSxSDouble^;
                    Inc(PGamma12Double,1);
                    Inc(PInflation22,1);
                  end;
              end;
              PDestDouble^:=FUsedf^[Pf2Double] * PDestDouble^;
            end;
          end;
        end;{$ENDREGION}
        {$REGION 'k1=k+1 and i1+k1<=n and k2=k+1 and i2+k2>n'}
        For i1 := 0 To n - k - 1 do begin
          For i2 := n - k To FI do begin
            PInflation21:=(m2*(FI+1)+i2)*(FJ+1+FK)+k+1;
            PDestDouble:=SxS;
            Inc(PDestDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k+1)*(FM+1)+m2)*(FI+1)+i2)*(FJ+1+FP)+k+1);
            PDestDouble^:=0;
            for cl2:=0 to FmlDependency[m2,FM] do begin
              l2:=FmlDependency[m2,cl2];
                h2:=i2;
                PInflation22:=(l2*(FI+1)+h2)*(FJ+1+FK);
                PGamma12Double:=((m2*(FM+1)+l2)*(FJ+FK)+k)*(FJ+FK);
                For j2 := 0 To k do begin
                  PSxSDouble:=SxS;
                  Inc(PSxSDouble,((((m1*(FI+1)+i1)*(FJ+1+FP)+k+1)*(FM+1)+l2)*(FI+1)+h2)*(FJ+1+FP)+j2);
                  PDestDouble^:=PDestDouble^
                                +FGamma1[PGamma12Double]
                                *FInflation[PInflation21]
                                /FInflation[PInflation22]
                                *PSxSDouble^;
                  Inc(PGamma12Double,1);
                  Inc(PInflation22,1);
                end;
            end;
            PDestDouble^:=FUsedf^[Pf2Double] * PDestDouble^;
          end;
        end;{$ENDREGION}
        Inc(Pf2Double,FJ+FK);
      end;
      Inc(Pf1Double,FJ+FK);
    end;
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorOP_H + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorOP_H + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

{$ENDREGION}

{$ENDREGION}

{$REGION '  TSimpleInflationLSRM public'}

{$REGION '    Init functions'}

function TSimpleInflationLSRM.InitCDRParaErrProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhostarDouble, Prhostar1Double, Prhostar2Double, Prhostar12Double,
      PCoorF1Double, PCoorF2Double: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, h1, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoMSEP in FInitialised) then
    Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoCDR in FInitialised) then
    Result:=InitrhoCDR();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRDetail)<=0 then begin
      //                  Proxy Part    m1     m2     i1     i2
      SetLength(FCDRDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    // parameter error  proxy
    // store the results within FCDRDetail(1,1)
    PDestDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    ZeroMemory(@FCDRDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
    if calc_ReduceMemSize in FCalculationFlags then begin
      {$REGION 'calculate rhos each time'}
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1 := i1;
                          h2 := i2;
                            For cl1 := 0 To FmlDependency[m1,FM] do begin
                              l1 := FmlDependency[m1,cl1];
                              For cl2 := 0 To FmlDependency[m2,FM] do begin
                                l2 := FmlDependency[m2,cl2];
                                PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                Rho:=Getrho(1,l1,l2,h1,h2,j)
                                    -Getrho(2,l1,l2,h1,h2,j)
                                    -Getrho(3,l1,l2,h1,h2,j)
                                    +Getrho(4,l1,l2,h1,h2,j);
                                TmpValue:=Rho
                                          *FR1[PR11Double]
                                          *FR1[PR12Double]
                                          *FCoorF[PCoorF1Double]
                                          *FCoorF[PCoorF2Double];
                                // diversification only if rho is negative
                                if (calc_VirtualParaErrDiversification in FCalculationFlags)
                                        or (Rho<=0) or (TmpValue>=0) then
                                  FCDRDetail[PDestDouble]:=FCDRDetail[PDestDouble]+TmpValue;
                              end;
                            end;
                          end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end else begin
      {$REGION 'take rhos from memory'}
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1 := i1;
                          h2 := i2;
                            For cl1 := 0 To FmlDependency[m1,FM] do begin
                              l1 := FmlDependency[m1,cl1];
                              For cl2 := 0 To FmlDependency[m2,FM] do begin
                                l2 := FmlDependency[m2,cl2];
                                PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                Prhostar1Double:=PrhostarDouble;
                                Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                Prhostar2Double:=Prhostar1Double;
                                Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                Prhostar12Double:=Prhostar2Double;
                                Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                Rho:=Frho[PrhostarDouble]
                                    -Frho[Prhostar1Double]
                                    -Frho[Prhostar2Double]
                                    +Frho[Prhostar12Double];
                                TmpValue:=Rho
                                          *FR1[PR11Double]
                                          *FR1[PR12Double]
                                          *FCoorF[PCoorF1Double]
                                          *FCoorF[PCoorF2Double];
                                // diversification only if rho is negative
                                if (calc_VirtualParaErrDiversification in FCalculationFlags)
                                        or (Rho<=0) or (TmpValue>=0) then
                                  FCDRDetail[PDestDouble]:=FCDRDetail[PDestDouble]+TmpValue;
                              end;
                            end;
                  end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRParaErrProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleInflationLSRM.InitCDRParaErrTailDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhostarDouble, Prhostar1Double, Prhostar2Double, Prhostar12Double,
      PCoorF1Double, PCoorF2Double, PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoMSEP in FInitialised) then
    Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoCDR in FInitialised) then
    Result:=InitrhoCDR();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRTailDetail)<=0 then begin
      //                     Proxy Part    m1     m2     i1     i2
      SetLength(FCDRTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_CDRParaErrTailProxyDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PProxyDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FCDRTailDetail[PDestDouble],@FCDRTailDetail[PProxyDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the CDR tail parameter error without proxy'}
      // store the results within FCDRTailDetail(0,1)
      PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FCDRTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      if calc_ReduceMemSize in FCalculationFlags then begin
        {$REGION 'calculate rhos each time'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error non proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 +FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else begin
        {$REGION 'take rhos from memory'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error non proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRParaErrTailDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleInflationLSRM.InitCDRParaErrTailProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhostarDouble, Prhostar1Double, Prhostar2Double, Prhostar12Double,
      PCoorF1Double, PCoorF2Double, PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoMSEP in FInitialised) then
    Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoCDR in FInitialised) then
    Result:=InitrhoCDR();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRTailDetail)<=0 then begin
      //                     Proxy Part    m1     m2     i1     i2
      SetLength(FCDRTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_CDRParaErrTailDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PProxyDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FCDRTailDetail[PProxyDouble],@FCDRTailDetail[PDestDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the CDR tail parameter error proxy'}
      // store the results within FCDRTailDetail(1,1)
      PDestDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FCDRTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      if calc_ReduceMemSize in FCalculationFlags then begin
        {$REGION 'calculate rhos each time'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 +FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j)
                                      -Getrho(2,l1,l2,h1,h2,j)
                                      -Getrho(3,l1,l2,h1,h2,j)
                                      +Getrho(4,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else begin
        {$REGION 'take rhos from memory'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  Prhostar1Double:=PrhostarDouble;
                                  Inc(Prhostar1Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar2Double:=Prhostar1Double;
                                  Inc(Prhostar2Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  Prhostar12Double:=Prhostar2Double;
                                  Inc(Prhostar12Double,(FM+1)*(FM+1)*(FI+1)*(FI+1)*(FJ+FP));
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble]
                                      -Frho[Prhostar1Double]
                                      -Frho[Prhostar2Double]
                                      +Frho[Prhostar12Double];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRParaErrTailProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRParaErrTailProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleInflationLSRM.InitCDRProcVarProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhobarDouble, PCoorF1Double, PCoorF2Double: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, h1, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  If not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoCDR in FInitialised) then
    Result:=InitrhoCDR();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRDetail)<=0 then begin
      //                  Proxy Part    m1     m2     i1     i2
      SetLength(FCDRDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    // process variance proxy
    // store the results within FCDRDetail(1,0)
    PDestDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    ZeroMemory(@FCDRDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
    if calc_ReduceMemSize in FCalculationFlags then begin
      {$REGION 'calculate rhos each time'}
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1 := i1;
                          h2 := i2;
                            For cl1 := 0 To FmlDependency[m1,FM] do begin
                              l1 := FmlDependency[m1,cl1];
                              For cl2 := 0 To FmlDependency[m2,FM] do begin
                                l2 := FmlDependency[m2,cl2];
                                PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                Rho:=Getrho(5,l1,l2,h1,h2,j);
                                TmpValue:=Rho
                                          *FR1[PR11Double]
                                          *FR1[PR12Double]
                                          *FCoorF[PCoorF1Double]
                                          *FCoorF[PCoorF2Double];
                                // diversification only if rho is negative
                                if (calc_VirtualProcVarDiversification in FCalculationFlags)
                                        or (Rho<=0) or (TmpValue>=0) then
                                  FCDRDetail[PDestDouble]:=FCDRDetail[PDestDouble]+TmpValue;
                              end;
                            end;
                  end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end else begin
      {$REGION 'take rhos from memory'}
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1 := i1;
                          h2 := i2;
                            For cl1 := 0 To FmlDependency[m1,FM] do begin
                              l1 := FmlDependency[m1,cl1];
                              For cl2 := 0 To FmlDependency[m2,FM] do begin
                                l2 := FmlDependency[m2,cl2];
                                PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                Rho:=Frho[PrhobarDouble];
                                TmpValue:=Rho
                                          *FR1[PR11Double]
                                          *FR1[PR12Double]
                                          *FCoorF[PCoorF1Double]
                                          *FCoorF[PCoorF2Double];
                                // diversification only if rho is negative
                                if (calc_VirtualProcVarDiversification in FCalculationFlags)
                                        or (Rho<=0) or (TmpValue>=0) then
                                  FCDRDetail[PDestDouble]:=FCDRDetail[PDestDouble]+TmpValue;
                              end;
                            end;
                  end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRProcVarProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleInflationLSRM.InitCDRProcVarTailDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhobarDouble, PCoorF1Double, PCoorF2Double,
      PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoCDR in FInitialised) then
    Result:=InitrhoCDR();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  // if calc_UseMAEPifTailCDRisLarger is set the we may need tail MSEP
  if calc_UseMSEPifTailCDRisLarger in FCalculationFlags then begin
    If not(Init_MSEPProcVarTailDetail in FInitialised) then Result:=InitMSEPProcVarTailProxyDetail();
    if Result<>'' then Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRTailDetail)<=0 then begin
      //                     Proxy Part    m1     m2     i1     i2
      SetLength(FCDRTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_CDRProcVarTailProxyDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=0;
      PProxyDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FCDRTailDetail[PDestDouble],@FCDRTailDetail[PProxyDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the CDR tail process variance without proxy'}
      // store the results within FCDRTailDetail(0,0)
      PDestDouble:=0;
      ZeroMemory(@FCDRTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      if calc_ReduceMemSize in FCalculationFlags then begin
        {$REGION 'calculate rhos each time'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'process variance non proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 +FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance non proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance non proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance non proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                if (calc_UseMSEPifTailCDRisLarger in FCalculationFlags)
                    and (FCDRTailDetail[PDestDouble]>FMSEPTailDetail[PDestDouble]) then
                  FCDRTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble];
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else begin
        {$REGION 'take rhos from memory'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'process variance non proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance non proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance non proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To MIn(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance non proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                if (calc_UseMSEPifTailCDRisLarger in FCalculationFlags)
                    and (FCDRTailDetail[PDestDouble]>FMSEPTailDetail[PDestDouble]) then
                  FCDRTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble];
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRProcVarTailDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleInflationLSRM.InitCDRProcVarTailProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhobarDouble, PCoorF1Double, PCoorF2Double,
      PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoCDR in FInitialised) then Result:=InitrhoCDR();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  // if calc_UseMAEPifTailCDRisLarger is set the we may need tail MSEP
  if calc_UseMSEPifTailCDRisLarger in FCalculationFlags then begin
    If not(Init_MSEPProcVarTailDetail in FInitialised) then Result:=InitMSEPProcVarTailProxyDetail();
    if Result<>'' then Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCDRTailDetail)<=0 then begin
      //                     Proxy Part    m1     m2     i1     i2
      SetLength(FCDRTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_CDRProcVarTailDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=0;
      PProxyDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FCDRTailDetail[PProxyDouble],@FCDRTailDetail[PDestDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the CDR tail process variance proxy'}
      // store the results within FCDRTailDetail(1,0)
      PDestDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FCDRTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      if calc_ReduceMemSize in FCalculationFlags then begin
        {$REGION 'calculate rhos each time'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'process variance proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 +FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(5,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                if (calc_UseMSEPifTailCDRisLarger in FCalculationFlags)
                    and (FCDRTailDetail[PDestDouble]>FMSEPTailDetail[PDestDouble]) then
                  FCDRTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble];
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else begin
        {$REGION 'take rhos from memory'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'process variance proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'process variance proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhobarDouble:=((((5*(FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhobarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FCDRTailDetail[PDestDouble]:=FCDRTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                if (calc_UseMSEPifTailCDRisLarger in FCalculationFlags)
                    and (FCDRTailDetail[PDestDouble]>FMSEPTailDetail[PDestDouble]) then
                  FCDRTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble];
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CDRProcVarTailProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCDRProcVarTailProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleInflationLSRM.InitCoorF: string;
var m, k, i, cl, l, h, j: Integer;
    PDestDouble, PElhjDouble, PTmp1Double: IndexDouble;
    StartTime, EndTime: Int64;
    Ta:T1Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCoorF + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_Gamma1 in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCoorF + sLineBreak + rsLSRM_NotInitialisedGamma1;
    Exit;
  end;
  if ([Init_f, Init_Defaultf]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitDefaultsigma + sLineBreak + rsLSRM_NotInitialisedf;
    Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FCoorF)<=0 then begin
      //                 m      i       k       l      j
      Setlength(FCoorF,(FM+1)*(FI+1)*(FJ+FP)*(FM+1)*(FJ+1+FP));
    end;
    // initialise temporare array
    SetLength(TA,(FM+1)*(FI+1)*(FJ+1+FP));
    For m := 0 To FM do begin
      for i := 0 to FI do begin
        For k := 0 To FJ - 1 + FP do begin
          PDestDouble:=(((m*(FI+1)+i)*(FJ+FP)+k)*(FM+1)+m)*(FJ+1+FP)+k+1;
          // for n=h+j>i+k we have (F^n,m_i,k x)^l_h,j = 1 iff l=m,h=i, j=k+1
          FCoorF[PDestDouble]:=1;
          for cl:=0 to FmlDependency[m,FM] do begin
            l:=FmlDependency[m,cl];
              h:=i;
              PElhjDouble:=(l*(FI+1)+h)*(FJ+1+FP);
              For j := 0 To Min(FJ - 1 + FP, i + k - h) do begin
                // set the (l,h,j)-coordinate of the temporare array to one and all other coordinates to zero
                ZeroMemory(@TA[0],(FM+1)*(FI+1)*(FJ+1+FP)*SizeOf(Double));
                Ta[PElhjDouble]:=1;
                // apply F
                Result:=OP_Fn(h+j,i+k,@Ta[0]);
                if Result<>'' then Exit;
                // get the coordinate of F
                PDestDouble:=(((m*(FI+1)+i)*(FJ+FP)+k)*(FM+1)+l)*(FJ+1+FP)+j;
                PTmp1Double:=(m*(FI+1)+i)*(FJ+1+FP)+k+1;
                FCoorF[PDestDouble]:=TA[PTmp1Double];
                Inc(PElhjDouble,1);
              end;
          end;
        end;
      end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_CoorF,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCoorF + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitCoorF + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleInflationLSRM.InitMSEPParaErrProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhostarDouble, PCoorF1Double, PCoorF2Double: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, h1, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  If not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoMSEP in FInitialised) then
    Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPDetail)<=0 then begin
      //                  Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    // parameter error proxy
    // store the results within FMSEPDetail(1,1)
    PDestDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
    ZeroMemory(@FMSEPDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
    if calc_ReduceMemSize in FCalculationFlags then begin
      {$REGION 'calculate rhos each time'}
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1 := i1;
                          h2 := i2;
                            For cl1 := 0 To FmlDependency[m1,FM] do begin
                              l1 := FmlDependency[m1,cl1];
                              For cl2 := 0 To FmlDependency[m2,FM] do begin
                                l2 := FmlDependency[m2,cl2];
                                PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                Rho:=Getrho(1,l1,l2,h1,h2,j);
                                TmpValue:=Rho
                                          *FR1[PR11Double]
                                          *FR1[PR12Double]
                                          *FCoorF[PCoorF1Double]
                                          *FCoorF[PCoorF2Double];
                                // diversification only if rho is negative
                                if (calc_VirtualParaErrDiversification in FCalculationFlags)
                                        or (Rho<=0) or (TmpValue>=0) then
                                  FMSEPDetail[PDestDouble]:=FMSEPDetail[PDestDouble]+TmpValue;
                              end;
                            end;
                  end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end else begin
      {$REGION 'take rhos from memory'}
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            For i2 := 0 To FI do begin
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1 := i1;
                          h2 := i2;
                            For cl1 := 0 To FmlDependency[m1,FM] do begin
                              l1 := FmlDependency[m1,cl1];
                              For cl2 := 0 To FmlDependency[m2,FM] do begin
                                l2 := FmlDependency[m2,cl2];
                                PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                Rho:=Frho[PrhostarDouble];
                                TmpValue:=Rho
                                          *FR1[PR11Double]
                                          *FR1[PR12Double]
                                          *FCoorF[PCoorF1Double]
                                          *FCoorF[PCoorF2Double];
                                // diversification only if rho is negative
                                if (calc_VirtualParaErrDiversification in FCalculationFlags)
                                        or (Rho<=0) or (TmpValue>=0) then
                                  FMSEPDetail[PDestDouble]:=FMSEPDetail[PDestDouble]+TmpValue;
                              end;
                            end;
                  end;
                end;
              end;
              Inc(PDestDouble,1);
            end;
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPParaErrProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleInflationLSRM.InitMSEPParaErrTailDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhostarDouble, PCoorF1Double, PCoorF2Double,
      PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoMSEP in FInitialised) then
    Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPTailDetail)<=0 then begin
      //                      Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_MSEPParaErrTailProxyDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PProxyDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FMSEPTailDetail[PDestDouble],@FMSEPTailDetail[PProxyDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the MSEP tail parameter error without proxy'}
      // store the results within FMSEPTailDetail(0,1)
      PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FMSEPTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      if calc_ReduceMemSize in FCalculationFlags then begin
        {$REGION 'calculate rhos each time'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error non proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 +FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else begin
        {$REGION 'take rhos from memory'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error non proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error non proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPParaErrTailDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleInflationLSRM.InitMSEPParaErrTailProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PR11Double, PR12Double, PrhostarDouble, PCoorF1Double, PCoorF2Double,
      PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, j, cl1, l1, cl2, l2, ch1, h1, ch2, h2: Integer;
    Rho, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  if not(Init_rhoStar in FInitialised) then Result:=InitrhoStar();
  if Result<>'' then Exit;
  If not(calc_ReduceMemSize in FCalculationFlags) and not(Init_rhoMSEP in FInitialised) then
    Result:=InitrhoMSEP();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // uinitialise array, if necessary
    if Length(FMSEPTailDetail)<=0 then begin
      //                      Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_MSEPParaErrTailDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=(FM+1)*(FM+1)*(FI+1)*(FI+1);
      PProxyDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FMSEPTailDetail[PProxyDouble],@FMSEPTailDetail[PDestDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the MSEP tail parameter error proxy'}
      // store the results within FMSEPTailDetail(1,1)
      PDestDouble:=3*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FMSEPTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      if calc_ReduceMemSize in FCalculationFlags then begin
        {$REGION 'calculate rhos each time'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 +FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),FJ-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Getrho(1,l1,l2,h1,h2,j);
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end else begin
        {$REGION 'take rhos from memory'}
        For m1 := 0 To FM do begin
          For m2 := 0 To FM do begin
            For i1 := 0 To FI do begin
              For i2 := 0 To FI do begin
                {$REGION 'parameter error proxy: k1 Tail, k2 not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FI - i2 To FJ - 1 do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k2 Tail, k1 not'}
                For k1 := FI - i1 To FJ - 1 do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1 and k2 Tail, j not'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FI - Min(i1, i2) To Min(Min(k1, k2),Fj-1) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                {$REGION 'parameter error proxy: k1, k2 and j Tail'}
                For k1 := FJ To FJ - 1 + FP do begin
                  For k2 := FJ To FJ - 1 + FP do begin
                    For j := FJ To Min(k1, k2) do begin
                      h1:=i1;
                        h2:=i2;
                              For cl1 := 0 To FmlDependency[m1,FM] do begin
                                l1 := FmlDependency[m1,cl1];
                                For cl2 := 0 To FmlDependency[m2,FM] do begin
                                  l2 := FmlDependency[m2,cl2];
                                  PrhostarDouble:=(((((FM+1)+l1)*(FM+1)+l2)*(FI+1)+h1)*(FI+1)+h2)*(FJ+FP)+j;
                                  PR11Double:=(l1*(FI+1)+h1)*(FJ+FK)+j;
                                  PR12Double:=(l2*(FI+1)+h2)*(FJ+FK)+j;
                                  PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                                  PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                                  Rho:=Frho[PrhostarDouble];
                                  TmpValue:=Rho
                                            *FR1[PR11Double]
                                            *FR1[PR12Double]
                                            *FCoorF[PCoorF1Double]
                                            *FCoorF[PCoorF2Double];
                                  // diversification only if rho is negative
                                  if (calc_VirtualTailDiversification in FCalculationFlags)
                                          or (Rho<=0) or (TmpValue>=0) then
                                    FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                                end;
                              end;
                    end;
                  end;
                end;
                {$ENDREGION}
                Inc(PDestDouble,1);
              end;
            end;
          end;
        end;
        {$ENDREGION}
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPParaErrTailProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPParaErrTailProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleInflationLSRM.InitMSEPProcVarDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PsigmaDouble, PR2Double, PCoorF1Double, PCoorF2Double, PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, h, j, cl1, l1, cl2, l2: Integer;
    sigma, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_sigma, Init_Defaultsigma]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarDetail + sLineBreak
            + rsLSRM_NotInitialisedsigma;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  If not(Init_R2Full in FInitialised) then Result:=InitR2Full();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPDetail)<=0 then begin
      //                  Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_MSEPProcVarProxyDetail in FInitialised)
        and (calc_VirtualProcVarDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=0;
      PProxyDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FMSEPDetail[PDestDouble],@FMSEPDetail[PProxyDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the MSEP process variance withou proxy'}
      // store the results within FMSEPDetail(0,0)
      PDestDouble:=0;
      ZeroMemory(@FMSEPDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            // only for i1=i2=h both FCoorF factors are possible not zero
            i2:=i1;
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FI-i1 To Min(k1,k2) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          FMSEPDetail[PDestDouble]:=FMSEPDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              // next i1=i2
              Inc(PDestDouble,FI+2);
          end;
          // next i1=i1 is at 0=0 and not at FI+1=FI+1
          Dec(PDestDouble,FI+1);
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPProcVarDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleInflationLSRM.InitMSEPProcVarProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PsigmaDouble, PR2Double, PCoorF1Double, PCoorF2Double, PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, h, j, cl1, l1, cl2, l2: Integer;
    sigma, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_sigma, Init_Defaultsigma]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedsigma;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  If not(Init_R2Full in FInitialised) then Result:=InitR2Full();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPDetail)<=0 then begin
      //                  Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_MSEPProcVarDetail in FInitialised)
        and (calc_VirtualProcVarDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=0;
      PProxyDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FMSEPDetail[PProxyDouble],@FMSEPDetail[PDestDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the MSEP process variance proxy'}
      // store the results within FMSEPDetail(1,0)
      PDestDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FMSEPDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            // only for i1=i2=h both FCoorF factors are possible not zero
            i2:=i1;
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := Min(FI-i1,FJ-1) To Min(k1,k2) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          // diversification only if sigma is negative, only for the proxy
                          if (calc_VirtualProcVarDiversification in FCalculationFlags)
                                  or (sigma<=0) or (TmpValue>=0) then
                            FMSEPDetail[PDestDouble]:=FMSEPDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              // next i1=i2
              Inc(PDestDouble,FI+2);
          end;
          // next i1=i1 is at 0=0 and not at FI+1=FI+1
          Dec(PDestDouble,FI+1);
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPProcVarProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarProxyDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleInflationLSRM.InitMSEPProcVarTailDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PsigmaDouble, PR2Double, PCoorF1Double, PCoorF2Double, PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, h, j, cl1, l1, cl2, l2: Integer;
    sigma, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_sigma, Init_Defaultsigma]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailDetail + sLineBreak
            + rsLSRM_NotInitialisedsigma;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  If not(Init_R2Full in FInitialised) then Result:=InitR2Full();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPTailDetail)<=0 then begin
      //                      Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_MSEPProcVarTailProxyDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=0;
      PProxyDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FMSEPTailDetail[PDestDouble],@FMSEPTailDetail[PProxyDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the MSEP tail process variance without proxy'}
      // store the results within FMSEPTailDetail(0,0)
      PDestDouble:=0;
      ZeroMemory(@FMSEPTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            // only for i1=i2=h both FCoorF factors are possible not zero
            i2:=i1;
              {$REGION 'process variance non proxy: j Tail, k1 and k2 not'}
                // not possible since j<=Min(k1,k2)
              {$ENDREGION}
              {$REGION 'process variance non proxy: k1 Tail, k2 and j not'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FI-i1 To Min(k1,k2) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'process variance non proxy: k2 Tail, k1 and j not'}
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FI-i1 To Min(k1,k2) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'process variance non proxy: k1 and k2 Tail, j not'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FI-i1 To Min(Min(k1,k2), FJ-1) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'process variance non proxy: k1, k2 and j Tail'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FJ To Min(k1,k2) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'process variance non proxy: k1 and j Tail, k2 not'}
                // not possible since j<=Min(k1,k2)
              {$ENDREGION}
              {$REGION 'process variance non proxy: k2 and j Tail, k1 not'}
                // not possible since j<=Min(k1,k2)
              {$ENDREGION}
              // next i1=i2
              Inc(PDestDouble,FI+2);
          end;
          // next i1=i1 is at 0=0 and not at FI+1=FI+1
          Dec(PDestDouble,FI+1);
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPProcVarTailDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailDetail + sLineBreak
              + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleInflationLSRM.InitMSEPProcVarTailProxyDetail: string;
var StartTime, EndTime: Int64;
    PDestDouble, PsigmaDouble, PR2Double, PCoorF1Double, PCoorF2Double, PProxyDouble: IndexDouble;
    m1, m2, i1, i2, k1, k2, h, j, cl1, l1, cl2, l2: Integer;
    sigma, TmpValue: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  if ([Init_sigma, Init_Defaultsigma]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailProxyDetail + sLineBreak
            + rsLSRM_NotInitialisedsigma;
    Exit;
  end;
  If not(Init_CoorF in FInitialised) then Result:=InitCoorF();
  if Result<>'' then Exit;
  If not(Init_R2Full in FInitialised) then Result:=InitR2Full();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array, if necessary
    if Length(FMSEPTailDetail)<=0 then begin
      //                      Proxy Part    m1     m2     i1     i2
      SetLength(FMSEPTailDetail,2  *  2  *(FM+1)*(FM+1)*(FI+1)*(FI+1));
    end;
    if (Init_MSEPProcVarTailDetail in FInitialised)
        and (calc_VirtualTailDiversification in FCalculationFlags) then begin
      {$REGION 'there is no special approximation, therefore make a copy'}
      PDestDouble:=0;
      PProxyDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      CopyMemory(@FMSEPTailDetail[PProxyDouble],@FMSEPTailDetail[PDestDouble],
                  (FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      {$ENDREGION}
    end else begin
      {$REGION 'calculate the details of the MSEP tail process variance proxy'}
      // store the results within FMSEPTailDetail(1,0)
      PDestDouble:=2*(FM+1)*(FM+1)*(FI+1)*(FI+1);
      ZeroMemory(@FMSEPTailDetail[PDestDouble],(FM+1)*(FM+1)*(FI+1)*(FI+1)*SizeOf(Double));
      For m1 := 0 To FM do begin
        For m2 := 0 To FM do begin
          For i1 := 0 To FI do begin
            // only for i1=i2=h both FCoorF factors are possible not zero
            i2:=i1;
              {$REGION 'process variance proxy: j Tail, k1 and k2 not'}
                // not possible since j<=Min(k1,k2)
              {$ENDREGION}
              {$REGION 'process variance proxy: k1 Tail, k2 and j not'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FI - i2 To FJ - 1 do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FI-i1 To Min(k1,k2) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          // diversification only if sigma is negative
                          if (calc_VirtualTailDiversification in FCalculationFlags)
                                  or (sigma<=0) or (TmpValue>=0) then
                            FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'process variance proxy: k2 Tail, k1 and j not'}
              For k1 := FI - i1 To FJ - 1 do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FI-i1 To Min(k1,k2) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          // diversification only if sigma is negative
                          if (calc_VirtualTailDiversification in FCalculationFlags)
                                  or (sigma<=0) or (TmpValue>=0) then
                            FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'process variance proxy: k1 and k2 Tail, j not'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FI-i1 To Min(Min(k1,k2),FJ-1) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          // diversification only if sigma is negative
                          if (calc_VirtualTailDiversification in FCalculationFlags)
                                  or (sigma<=0) or (TmpValue>=0) then
                            FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'process variance proxy: k1, k2 and j Tail'}
              For k1 := FJ To FJ - 1 + FP do begin
                For k2 := FJ  To FJ - 1 + FP do begin
                  // change the order of summation from n and j to j and h:=n-1-j
                  // this will make it easier to speed up the calculation for simple models
                  For j := FJ To Min(k1,k2) do begin
                    h:=i1;
                      For cl1 := 0 To FmlDependency[m1,FM] do begin
                        l1 := FmlDependency[m1,cl1];
                        For cl2 := 0 To FmlDependency[m2,FM] do begin
                          l2 := FmlDependency[m2,cl2];
                          PsigmaDouble:=(l1*(FM+1)+l2)*(FJ+FP)+j;
                          PR2Double:=((l1*(FM+1)+l2)*(FI+1)+h)*(FJ+FP)+j;
                          PCoorF1Double:=(((m1*(FI+1)+i1)*(FJ+FP)+k1)*(FM+1)+l1)*(FJ+1+FP)+j+1;
                          PCoorF2Double:=(((m2*(FI+1)+i2)*(FJ+FP)+k2)*(FM+1)+l2)*(FJ+1+FP)+j+1;
                          sigma:=FUsedsigma^[PsigmaDouble];
                          TmpValue:=sigma
                                    *FR2[PR2Double]
                                    *FCoorF[PCoorF1Double]
                                    *FCoorF[PCoorF2Double];
                          // diversification only if sigma is negative
                          if (calc_VirtualTailDiversification in FCalculationFlags)
                                  or (sigma<=0) or (TmpValue>=0) then
                            FMSEPTailDetail[PDestDouble]:=FMSEPTailDetail[PDestDouble]+TmpValue;
                        end;
                      end;
                  end;
                end;
              end;
              {$ENDREGION}
              {$REGION 'process variance proxy: k1 and j Tail, k2 not'}
                // not possible since j<=Min(k1,k2)
              {$ENDREGION}
              {$REGION 'process variance proxy: k2 and j Tail, k1 not'}
                // not possible since j<=Min(k1,k2)
              {$ENDREGION}
              // next i1=i2
              Inc(PDestDouble,FI+2);
          end;
          // next i1=i1 is at 0=0 and not at FI+1=FI+1
          Dec(PDestDouble,FI+1);
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_MSEPProcVarTailProxyDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailProxyDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitMSEPProcVarTailProxyDetail
              + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleInflationLSRM.InitR1: String;
var m, k, i, l, h, j, cl: Integer;
    PDestDouble, Pgamma1Double, PSDouble, PInflation1, PInflation2: IndexDouble;
    StartTime, EndTime: Int64;
    sign: Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR1 + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_S in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR1 + sLineBreak + rsLSRM_NotInitialisedS;
    Exit;
  end;
  If not(Init_gamma1 in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR1 + sLineBreak + rsLSRM_NotInitialisedgamma1;
    Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array or clear the corresponding memory
    if Length(FR1)<=0 then begin
      //               m     i     k
      SetLength(FR1,(FM+1)*(FI+1)*(FJ+FK));
    end else begin
      ZeroMemory(@FR1[0],(FM+1)*(FI+1)*(FJ+FK)*SizeOf(Double));
    end;
    // calculate
    PDestDouble:=0;
    PInflation1:=1;
    for m:=0 to FM do begin
      for i:=0 to FI do begin
        sign:=0;
        for k := 0 To Min(FJ - 1 + FP, FI-i) do begin
          for cl:=0 to FmlDependency[m,FM] do begin
            l:=FmlDependency[m,cl];
              h:=i;
              PGamma1Double:=((m*(FM+1)+l)*(FJ+FK)+k)*(FJ+FK);
              PSDouble:=(l*(FI+1)+h)*(FJ+1+FK);
              PInflation2:=(l*(FI+1)+h)*(FJ+1+FK);
              for j:=0 to k do begin
                FR1[PDestDouble]:=FR1[PDestDouble]
                                  +FGamma1[PGamma1Double]
                                  *FInflation[PInflation1]
                                  /FInflation[PInflation2]
                                  *FS[PSDouble];
                Inc(PGamma1Double,1);
                Inc(PSDouble,1);
                Inc(PInflation2,1);
              end;
          end;
          // prevent exposure from changing sign
          if calc_PreventPastExposureFromChangingSign in FCalculationFlags then begin
            if sign=0 then begin
              sign:=FR1[PDestDouble];
            end else begin
              if ((sign>0) and (FR1[PDestDouble]<0)) or ((sign<0) and (FR1[PDestDouble]>0)) then
                FR1[PDestDouble]:=0;
            end;
          end;
          Inc(PDestDouble,1);
          Inc(PInflation1,1);
        end;
        Inc(PDestDouble,FJ-1+FK-Min(FJ - 1 + FP, FI-i));
        Inc(PInflation1,FJ+FK-Min(FJ - 1+ FP, FI-i));
     end;
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_R1,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR1 + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR1 + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleInflationLSRM.InitR2: String;
var m1, m2, k, i, l, h, j: Integer;
    PDestDouble, PGamma2Double, PSDouble, PInflation1, PInflation2, PInflation3: IndexDouble;
    StartTime, EndTime: Int64;
    sign: Double;
    CalcParas: TCalcParas;
    TmpDouble: PDouble;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2 + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_S in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2 + sLineBreak + rsLSRM_NotInitialisedS;
    Exit;
  end;
  If not(Init_gamma2 in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2 + sLineBreak + rsLSRM_NotInitialisedgamma2;
    Exit;
  end;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array or clear the corresponding memory
    if Length(FR2)<=0 then begin
      //               m1    m2     i     k
      SetLength(FR2,(FM+1)*(FM+1)*(FI+1)*(FJ+FP));
    end else begin
      ZeroMemory(@FR2[0],(FM+1)*(FM+1)*(FI+1)*(FJ+FP)*SizeOf(Double));
    end;
    // calculate
    if length(FGamma2)>0 then begin
      {$REGION 'R2 has been specified by a gamma2 array'}
      PDestDouble:=0;
      for m1:=0 to FM do begin
        PInflation2:=1;
        for m2:=0 to FM do begin
          PInflation1:=m1*(FI+1)*(FJ+1+FK)+1;
          for i:=0 to FI do begin
            sign:=0;
            for k := 0 To Min(FJ - 1 + FP, FI-i) do begin
              for l:=0 to FM do begin
                h:=i;
                  PGamma2Double:=(((m1*(FM+1)+m2)*(FM+1)+l)*(FJ+FP)+k)*(FJ+FP);
                  PSDouble:=(l*(FI+1)+h)*(FJ+1+FK);
                  PInflation3:=(l*(FI+1)+h)*(FJ+1+FK);
                  for j:=0 to k do begin
                    FR2[PDestDouble]:=FR2[PDestDouble]
                                      +FGamma2[PGamma2Double]
                                      *FInflation[PInflation1]
                                      *FInflation[PInflation2]
                                      /FInflation[PInflation3]
                                      *FS[PSDouble];
                    Inc(PGamma2Double,1);
                    Inc(PSDouble,1);
                    Inc(PInflation3,1);
                  end;
              end;
              // prevent exposure from changing sign
              if calc_PreventPastExposureFromChangingSign in FCalculationFlags then begin
                if sign=0 then begin
                  sign:=FR2[PDestDouble];
                end else begin
                  if ((sign>0) and (FR2[PDestDouble]<0)) or ((sign<0) and (FR2[PDestDouble]>0)) then
                    FR2[PDestDouble]:=0;
                end;
              end;
              Inc(PDestDouble,1);
              Inc(PInflation1,1);
              Inc(PInflation2,1);
            end;
            Inc(PDestDouble,FJ-1+FP-Min(FJ - 1 + FP, FI-i));
            Inc(PInflation1,FJ+FK-Min(FJ - 1 + FP, FI-i));
            Inc(PInflation2,FJ+FK-Min(FJ - 1 + FP, FI-i));
          end;
        end;
      end;
      {$ENDREGION}
    end else begin
      {$REGION 'R2 has benn specified by a formula'}
      // go through each formula and calculate R2
      CalcParas.M:=FM;
      CalcParas.I:=FI;
      CalcParas.J:=FJ+FP-1;
      CalcParas.Mode:=CalcGamma2;
      CalcParas.LastDiag:=FI;
      CalcParas.LastColumn:=FJ+FP-1;
      CalcParas.ClaimProperty.S:=@FS[0];
      CalcParas.ClaimProperty.V:=@FInflation[0];
      CalcParas.ClaimProperty.J:=FJ+FK;
      TmpDouble:=@FR2[0];
      for m1 := 0 to FM do begin
        for m2 := 0 to FM do begin
          FFormulas.Formula[m1*(FM+1)+m2].Calc(TmpDouble,@CalcParas);
          inc(TmpDouble,(FI+1)*(FJ+FP));
        end;
      end;
      // prevent exposure from changing sign and inflate R2
      PDestDouble:=0;
      for m1:=0 to FM do begin
        PInflation2:=1;
        for m2:=0 to FM do begin
          PInflation1:=m1*(FI+1)*(FJ+1+FK)+1;
          for i:=0 to FI do begin
            sign:=0;
            for k := 0 To Min(FJ - 1 + FP, FI-i) do begin
              FR2[PDestDouble]:=FR2[PDestDouble]
                                *FInflation[PInflation1]
                                *FInflation[PInflation2];
              if calc_PreventPastExposureFromChangingSign in FCalculationFlags then begin
                if sign=0 then begin
                  sign:=FR2[PDestDouble];
                end else begin
                  if ((sign>0) and (FR2[PDestDouble]<0)) or ((sign<0) and (FR2[PDestDouble]>0)) then
                    FR2[PDestDouble]:=0;
                end;
              end;
              Inc(PDestDouble,1);
              Inc(PInflation1,1);
              Inc(PInflation2,1);
            end;
            Inc(PDestDouble,FJ-1+FP-Min(FJ - 1 + FP, FI-i));
            Inc(PInflation1,FJ+FK-Min(FJ - 1 + FP, FI-i));
            Inc(PInflation2,FJ+FK-Min(FJ - 1 + FP, FI-i));
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_R2,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2 + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2 + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleInflationLSRM.InitR2Full: String;
var m1, m2, k, i, l, h, j: Integer;
    PDestDouble, PGamma2Double, PSDouble, PInflation1, PInflation2, PInflation3: IndexDouble;
    StartTime, EndTime: Int64;
    sign: Double;
    CalcParas: TCalcParas;
    TmpDouble: PDouble;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2Full + sLineBreak + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_gamma2 in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2Full + sLineBreak + rsLSRM_NotInitialisedgamma2;
    Exit;
  end;
  if not(Init_R2 in FInitialised) then Result:=InitR2();
  if Result<>'' then Exit;
  If not(Init_UltimateDetail in FInitialised) then Result:=InitUltimateDetail();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // calculate
    if length(FGamma2)>0 then begin
      {$REGION 'R2 has been specified by a gamma2 array'}
      PDestDouble:=0;
      for m1:=0 to FM do begin
        PInflation2:=1;
        for m2:=0 to FM do begin
          PInflation1:=m1*(FI+1)*(FJ+1+FK)+1;
          for i:=0 to FI do begin
            sign:=0;
            if calc_PreventFutureExposureFromChangingSign in FCalculationFlags then begin
              for k := 0 to Min(FJ - 1 + FP, FI-i) do begin
                if sign=0 then sign:=FR2[PDestDouble];
                Inc(PDestDouble,1);
              end;
            end else begin
              Inc(PDestDouble,Min(FJ - 1 + FP, FI-i)+1);
            end;
            Inc(PInflation1,Min(FJ - 1 + FP, FI-i)+1);
            Inc(PInflation2,Min(FJ - 1 + FP, FI-i)+1);
            for k := Min(FJ - 1 + FP, FI-i)+1 To FJ - 1 + FP do begin
              FR2[PDestDouble]:=0;
              for l:=0 to FM do begin
                h:=i;
                  PGamma2Double:=(((m1*(FM+1)+m2)*(FM+1)+l)*(FJ+FP)+k)*(FJ+FP);
                  PSDouble:=(l*(FI+1)+h)*(FJ+1+FK);
                  PInflation3:=(l*(FI+1)+h)*(FJ+1+FK);
                  for j:=0 to k do begin
                    FR2[PDestDouble]:=FR2[PDestDouble]
                                      +FGamma2[PGamma2Double]
                                      *FInflation[PInflation1]
                                      *FInflation[PInflation2]
                                      /FInflation[PInflation3]
                                      *FS[PSDouble];
                    Inc(PGamma2Double,1);
                    Inc(PSDouble,1);
                    Inc(PInflation3,1);
                  end;
              end;
              // prevent exposure from changing sign
              if calc_PreventFutureExposureFromChangingSign in FCalculationFlags then begin
                if sign=0 then begin
                  sign:=FR2[PDestDouble];
                end else begin
                  if ((sign>0) and (FR2[PDestDouble]<0)) or ((sign<0) and (FR2[PDestDouble]>0)) then
                    FR2[PDestDouble]:=0;
                end;
              end;
              Inc(PDestDouble,1);
              Inc(PInflation1,1);
              Inc(PInflation2,1);
            end;
            Inc(PInflation1,FK-FP+1);
            Inc(PInflation2,FK-FP+1);
          end;
        end;
      end;
      {$ENDREGION}
    end else begin
      {$REGION 'R2 has been specified by a formula'}
      ZeroMemory(@FR2[0],(FM+1)*(FM+1)*(FI+1)*(FJ+FP)*SizeOf(Double));
      // go through each formula and calculate R2
      CalcParas.M:=FM;
      CalcParas.I:=FI;
      CalcParas.J:=FJ+FP-1;
      CalcParas.Mode:=CalcGamma2;
      CalcParas.LastDiag:=FI+FJ+FP-1;
      CalcParas.LastColumn:=FJ+FP-1;
      CalcParas.ClaimProperty.S:=@FS[0];
      CalcParas.ClaimProperty.V:=@FInflation[0];
      CalcParas.ClaimProperty.J:=FJ+FK;
      TmpDouble:=@FR2[0];
      for m1 := 0 to FM do begin
        for m2 := 0 to FM do begin
          FFormulas.Formula[m1*(FM+1)+m2].Calc(TmpDouble,@CalcParas);
          inc(TmpDouble,(FI+1)*(FJ+FP));
        end;
      end;
      // prevent exposure from changing sign and inflate R2
      PDestDouble:=0;
      for m1:=0 to FM do begin
        PInflation2:=1;
        for m2:=0 to FM do begin
          PInflation1:=m1*(FI+1)*(FJ+1+FK)+1;
          for i:=0 to FI do begin
            sign:=0;
            for k := 0 To Min(FJ - 1 + FP, FI-i) do begin
              FR2[PDestDouble]:=FR2[PDestDouble]
                                *FInflation[PInflation1]
                                *FInflation[PInflation2];
              if [calc_PreventPastExposureFromChangingSign,calc_PreventFutureExposureFromChangingSign]*FCalculationFlags<>[] then begin
                if sign=0 then begin
                  sign:=FR2[PDestDouble];
                end else if calc_PreventPastExposureFromChangingSign in FCalculationFlags then begin
                  if ((sign>0) and (FR2[PDestDouble]<0)) or ((sign<0) and (FR2[PDestDouble]>0)) then
                    FR2[PDestDouble]:=0;
                end;
              end;
              Inc(PDestDouble,1);
              Inc(PInflation1,1);
              Inc(PInflation2,1);
            end;
            for k := Min(FJ - 1 + FP, FI-i)+1 To FJ - 1 + FP do begin
              FR2[PDestDouble]:=FR2[PDestDouble]
                                *FInflation[PInflation1]
                                *FInflation[PInflation2];
              if calc_PreventFutureExposureFromChangingSign in FCalculationFlags then begin
                if sign=0 then begin
                  sign:=FR2[PDestDouble];
                end else begin
                  if ((sign>0) and (FR2[PDestDouble]<0)) or ((sign<0) and (FR2[PDestDouble]>0)) then
                    FR2[PDestDouble]:=0;
                end;
              end;
              Inc(PDestDouble,1);
              Inc(PInflation1,1);
              Inc(PInflation2,1);
            end;
            Inc(PInflation1,FK - FP + 1);
            Inc(PInflation2,FK - FP + 1);
          end;
        end;
      end;
      {$ENDREGION}
    end;
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_R2Full,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2Full + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitR2Full + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

function TSimpleInflationLSRM.InitUltimateDetail: string;
var m, n, k, i, l, h, j, cl: Integer;
    PR1DestDouble, PSDestDouble, Pgamma1Double, PSDouble, PfDouble, PUltimateDestDouble,
      PUltimateTailDestDouble, PInflation1, PInflation2: IndexDouble;
    StartTime, EndTime: Int64;
    TmpSum, TmpSumTail:Double;
    sign: T1Double;
begin
  Result:='';
  {$REGION 'test if all necessary elements have been calculated'}
  If not(Init_Geometry in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimateDetail + sLineBreak
            + rsLSRM_NotInitialisedGeometry;
    Exit;
  end;
  If not(Init_S in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimateDetail + sLineBreak + rsLSRM_NotInitialisedS;
    Exit;
  end;
  If not(Init_gamma1 in FInitialised) then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimateDetail + sLineBreak
            + rsLSRM_NotInitialisedgamma1;
    Exit;
  end;
  if ([Init_f, Init_Defaultf]*FInitialised)=[] then begin
    Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimateDetail + sLineBreak + rsLSRM_NotInitialisedf;
    Exit;
  end;
  If not(Init_R1 in FInitialised) then Result:=InitR1();
  if Result<>'' then Exit;
  {$ENDREGION}
  try
    QueryPerformanceCounter(StartTime);
    // initialise array or clear the corresponding memory
    if Length(FUltimate)<=0 then begin
      //                       m      i
      SetLength(FUltimate,2*(FM+2)*(FI+2));
    end else begin
      ZeroMemory(@FUltimate[0],2*(FM+2)*(FI+2)*SizeOf(Double))
    end;
    {$REGION 'get signs of the first known exposure in case of they should be prevented from changing signs'}
    if calc_PreventFutureExposureFromChangingSign in FCalculationFlags then begin
      setlength(sign,FI+1);
      PR1DestDouble:=0;
      for m:=0 to FM do begin
        for i:=0 to FI do begin
          sign[i]:=0;
          for k := 0 To Min(FJ - 1 + FP, FI-i) do begin
            if sign[i]=0 then
              sign[i]:=FR1[PR1DestDouble];
            Inc(PR1DestDouble,1);
          end;
          Inc(PR1DestDouble,FJ-1+FK-Min(FJ - 1 + FP, FI-i));
        end;
      end;
    end;
    {$ENDREGION}
    {$REGION 'calculate S and R1 diagonal by diagonal: first the tail below the diagonal FI+1'}
    for n:=FJ+1 to FI do begin
      for m:=0 to FM do begin
        for i := max(0,n-FJ-FK) to n-FJ-1 do begin
          // k=n-i
          {$REGION 'R1'}
          PR1DestDouble:=(m*(FI+1)+i)*(FJ+FK)+n-i-1;
          PInflation1:=(m*(FI+1)+i)*(FJ+1+FK)+n-i;
          FR1[PR1DestDouble]:=0;
          for cl:=0 to FmlDependency[m,FM] do begin
            l:=FmlDependency[m,cl];
              h:=i;
              PGamma1Double:=((m*(FM+1)+l)*(FJ+FK)+n-i-1)*(FJ+FK);
              PSDouble:=(l*(FI+1)+h)*(FJ+1+FK);
              PInflation2:=(l*(FI+1)+h)*(FJ+1+FK);
              for j:=0 to Min(n-1 - h, n-i-1) do begin
                FR1[PR1DestDouble]:=FR1[PR1DestDouble]
                                    +FGamma1[PGamma1Double]
                                    *FInflation[PInflation1]
                                    /FInflation[PInflation2]
                                    *FS[PSDouble];
                Inc(PGamma1Double,1);
                Inc(PSDouble,1);
                Inc(PInflation2,1);
              end;
          end;
          // prevent exposure from changing sign
          if calc_PreventFutureExposureFromChangingSign in FCalculationFlags then begin
            if sign[i]=0 then begin
              sign[i]:=FR1[PR1DestDouble];
            end else begin
              if ((sign[i]>0) and (FR1[PR1DestDouble]<0)) or ((sign[i]<0) and (FR1[PR1DestDouble]>0)) then
                FR1[PR1DestDouble]:=0;
            end;
          end;
          {$ENDREGION}
          {$REGION 'S'}
          PfDouble:=m*(FJ+FK)+n-i-1;
          PSDestDouble:=(m*(FI+1)+i)*(FJ+1+FK)+n-i;
          FS[PSDestDouble]:=FUsedf^[PfDouble]*FR1[PR1DestDouble];
          {$ENDREGION}
        end;
      end;
    end;
    {$ENDREGION}
    {$REGION 'calculate S and R1 diagonal by diagonal: above the diagonal FI'}
    for n:=FI+1 to FI+FJ+FK do begin
      for m:=0 to FM do begin
        for i := max(0,n-FJ-FK) to FI do begin
          // k=n-i
          {$REGION 'R1'}
          PR1DestDouble:=(m*(FI+1)+i)*(FJ+FK)+n-i-1;
          PInflation1:=(m*(FI+1)+i)*(FJ+1+FK)+n-i;
          // first fill up R1(m,i,n-i-1), but only if n>FI+1 or if k>FJ
          if (n>FI+1) or (n-i>FJ) then begin
            FR1[PR1DestDouble]:=0;
            for cl:=0 to FmlDependency[m,FM] do begin
              l:=FmlDependency[m,cl];
                h:=i;
                PGamma1Double:=((m*(FM+1)+l)*(FJ+FK)+n-i-1)*(FJ+FK);
                PSDouble:=(l*(FI+1)+h)*(FJ+1+FK);
                PInflation2:=(l*(FI+1)+h)*(FJ+1+FK);
                for j:=0 to Min(n-1 - h, n-i-1) do begin
                  FR1[PR1DestDouble]:=FR1[PR1DestDouble]
                                      +FGamma1[PGamma1Double]
                                      *FInflation[PInflation1]
                                      /FInflation[PInflation2]
                                      *FS[PSDouble];
                  Inc(PGamma1Double,1);
                  Inc(PSDouble,1);
                  Inc(PInflation2,1);
                end;
            end;
            // prevent exposure from changing sign
            if calc_PreventFutureExposureFromChangingSign in FCalculationFlags then begin
              if sign[i]=0 then begin
                sign[i]:=FR1[PR1DestDouble];
              end else begin
                if ((sign[i]>0) and (FR1[PR1DestDouble]<0)) or ((sign[i]<0) and (FR1[PR1DestDouble]>0)) then
                  FR1[PR1DestDouble]:=0;
              end;
            end;
          end;
          {$ENDREGION}
          {$REGION 'S'}
          PfDouble:=m*(FJ+FK)+n-i-1;
          PSDestDouble:=(m*(FI+1)+i)*(FJ+1+FK)+n-i;
          FS[PSDestDouble]:=FUsedf^[PfDouble]*FR1[PR1DestDouble];
          {$ENDREGION}
        end;
      end;
    end;
    {$ENDREGION}
    {$REGION 'fill up the sum over all accident periods'}
    PUltimateDestDouble:=0;
    PUltimateTailDestDouble:=(FM+2)*(FI+2);
    PSDouble:=0;
    for m := 0 to FM do begin
      TmpSum:=0;
      TmpSumTail:=0;
      for i := 0 to FI do begin
        {$REGION 'without Tail'}
        for k := 0 to FJ do begin
          FUltimate[PUltimateDestDouble]:=FUltimate[PUltimateDestDouble]+FS[PSDouble];
          Inc(PSDouble,1);
        end;
        TmpSum:=TmpSum + FUltimate[PUltimateDestDouble];
        FUltimate[PUltimateTailDestDouble]:=FUltimate[PUltimateDestDouble];
        {$ENDREGION}
        {$REGION 'Tail'}
        for k := FJ+1 to FJ+FK do begin
          FUltimate[PUltimateTailDestDouble]:=FUltimate[PUltimateTailDestDouble]+FS[PSDouble];
          Inc(PSDouble,1);
        end;
        TmpSumTail:=TmpSumTail + FUltimate[PUltimateTailDestDouble];
        {$ENDREGION}
        Inc(PUltimateDestDouble,1);
        Inc(PUltimateTailDestDouble,1);
      end;
      FUltimate[PUltimateDestDouble]:=TmpSum;
      FUltimate[PUltimateTailDestDouble]:=TmpSumTail;
      Inc(PUltimateDestDouble,1);
      Inc(PUltimateTailDestDouble,1);
    end;
    {$ENDREGION}
    QueryPerformanceCounter(EndTime);
    SetInitFlags(Init_UltimateDetail,EndTime-StartTime);
  except
    On E:Exception do begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimateDetail + sLineBreak + E.Message;
    end else begin
      Result:='LSRM ''' + FName + ''':' + rsLSRM_ErrorInitUltimateDetail + sLineBreak + rsLSRM_UnknownError;
    end;
  end;
end;

{$ENDREGION}

{$ENDREGION}

{$ENDREGION}

{$REGION 'TLSRMDependencies'}

{$REGION '  TLSRMDependencies private'}

procedure TLSRMDependencies.AddDependsOn(Dependant, i:Integer);
var n: Integer;
begin
  // detect if Dependant already depends on i
  If DependsOn[Dependant, i] Then Exit;
  // add the dependence on i to Dependant and to all objects that depend on Dependant
  FDependsOn[Dependant].Bits[i] := True;
  For n := 0 To High(FDependsOn) do begin
    If DoesInfluence[Dependant, n] Then AddDependsOn(n,i);
    // add the dependence on each of the object i that depends onto Dependant and to all objects that
    // depend on Dependant
    If DependsOn[i, n] Then AddDependsOn(Dependant,n);
  end;
  // add the influence on Dependant to i
  AddDoesInfluence(i,Dependant);
end;

procedure TLSRMDependencies.AddDoesInfluence(Influencer, i:Integer);
var n: Integer;
begin
  // detect if Influencer is already influences i
  If DoesInfluence[Influencer, i] Then Exit;
  // add the influence of Influencer on i to Influencer and to all objects that have influence on Influencer
  FDoesInfluence[Influencer].Bits[i] := True;
  For n := 0 To High(FDependsOn) do begin
    If DependsOn[Influencer, n] Then AddDoesInfluence(n,i);
    // add the influence on each of the object i that does influence Influencer and to all objects that
    // influence Influencer
    If DoesInfluence[i, n] Then AddDoesInfluence(Influencer,n);
  end;
  // add the dependence on Influencer to i
  AddDependsOn(i,Influencer);
end;

function TLSRMDependencies.GetDependsOn(Dependant, i:Integer): Boolean;
begin
  Result:=FDependsOn[Dependant].Bits[i];
end;

function TLSRMDependencies.GetDoesInfluence(Influencer, i:Integer): Boolean;
begin
  Result:=FDoesInfluence[Influencer].Bits[i];
end;

{$ENDREGION}

{$REGION '  TLSRMDependencies public'}

constructor TLSRMDependencies.Create(MaxInstance: Integer);
var i: Integer;
begin
  inherited Create;
  SetLength(FDependsOn,MaxInstance+1);
  SetLength(FDoesInfluence,MaxInstance+1);
  for i:=0 to MaxInstance do begin
    FDependsOn[i]:=TBits.Create;
    FDependsOn[i].Size:=MaxInstance+1;
    FDoesInfluence[i]:=TBits.Create;
    FDoesInfluence[i].Size:=MaxInstance+1;
  end;
end;

destructor TLSRMDependencies.Destroy;
var i: Integer;
begin
  for i:=Low(FDependsOn) to High(FDependsOn) do begin
    FreeAndNil(FDependsOn[i]);
    FreeAndNil(FDoesInfluence[i]);
  end;
  inherited;
end;

procedure TLSRMDependencies.FillDependenceArray(var Dependencies: T2IntegerArray);
var n, m, Index: Integer;
begin
  SetLength(Dependencies,Length(FDependsOn),Length(FDependsOn));
  For n := 0 To High(FDependsOn) do begin
    Index := 0;
    For m := 0 To High(FDependsOn) do begin
      If DependsOn[n, m] Then begin
        Dependencies[n, Index] := m;
        Inc(Index);
      End;
    End;
    Dependencies[n, High(FDependsOn)] := Index - 1;
  end;
end;

{$ENDREGION}

{$ENDREGION}

initialization
  // initialise the frequency that is used to convert time measured by the frequency of the high-resolution
  // performance counter into miliseconds
  QueryPerformanceFrequency(FFrequence);

  {$REGION 'initialise the short descriptions (or titles) of the initialisation steps of LSRMs'}
  FCalcInfoNames[Init_Geometry] := rsLSRM_InitName_Geometry;
  FCalcInfoNames[Init_Gamma1] := rsLSRM_InitName_Gamma1;
  FCalcInfoNames[Init_S] := rsLSRM_InitName_S;
  FCalcInfoNames[Init_InflationRate] := rsLSRM_InitName_InflationRate;
  FCalcInfoNames[Init_R1] := rsLSRM_InitName_R1;
  FCalcInfoNames[Init_w] := rsLSRM_InitName_w;
  FCalcInfoNames[Init_Defaultw] := rsLSRM_InitName_Defaultw;
  FCalcInfoNames[Init_f] := rsLSRM_InitName_f;
  FCalcInfoNames[Init_Defaultf] := rsLSRM_InitName_Defaultf;
  FCalcInfoNames[Init_alpha] := rsLSRM_InitName_alpha;
  FCalcInfoNames[Init_Defaultalpha] := rsLSRM_InitName_Defaultalpha;
  FCalcInfoNames[Init_UltimateDetail] := rsLSRM_InitName_UltimateDetail;
  FCalcInfoNames[Init_Ultimate] := rsLSRM_InitName_Ultimate;
  FCalcInfoNames[Init_Gamma2] := rsLSRM_InitName_Gamma2;
  FCalcInfoNames[Init_R2] := rsLSRM_InitName_R2;
  FCalcInfoNames[Init_R2Full] := rsLSRM_InitName_R2Full;
  FCalcInfoNames[Init_sigma] := rsLSRM_InitName_sigma;
  FCalcInfoNames[Init_Defaultsigma] := rsLSRM_InitName_Defaultsigma;
  FCalcInfoNames[Init_Eigenvalues] := rsLSRM_InitName_Eigenvalues;
  FCalcInfoNames[Init_CoorF] := rsLSRM_InitName_CoorF;
  FCalcInfoNames[Init_rhoStar] := rsLSRM_InitName_rhoStar;
  FCalcInfoNames[Init_rhoMSEP] := rsLSRM_InitName_rhoMSEP;
  FCalcInfoNames[Init_rhoCDR] := rsLSRM_InitName_rhoCDR;
  FCalcInfoNames[Init_MSEPParaErrProxyDetail] := rsLSRM_InitName_MSEPParaErrProxyDetail;
  FCalcInfoNames[Init_MSEPParaErrProxy] := rsLSRM_InitName_MSEPParaErrProxy;
  FCalcInfoNames[Init_MSEPProcVarProxyDetail] := rsLSRM_InitName_MSEPProcVarProxyDetail;
  FCalcInfoNames[Init_MSEPProcVarProxy] := rsLSRM_InitName_MSEPProcVarProxy;
  FCalcInfoNames[Init_MSEPProxy] := rsLSRM_InitName_MSEPProxy;
  FCalcInfoNames[Init_MSEPParaErrTailProxyDetail] := rsLSRM_InitName_MSEPParaErrTailProxyDetail;
  FCalcInfoNames[Init_MSEPParaErrTailProxy] := rsLSRM_InitName_MSEPParaErrTailProxy;
  FCalcInfoNames[Init_MSEPProcVarTailProxyDetail] := rsLSRM_InitName_MSEPProcVarTailProxyDetail;
  FCalcInfoNames[Init_MSEPProcVarTailProxy] := rsLSRM_InitName_MSEPProcVarTailProxy;
  FCalcInfoNames[Init_MSEPTailProxy] := rsLSRM_InitName_MSEPTailProxy;
  FCalcInfoNames[Init_CDRParaErrProxyDetail] := rsLSRM_InitName_CDRParaErrProxyDetail;
  FCalcInfoNames[Init_CDRParaErrProxy] := rsLSRM_InitName_CDRParaErrProxy;
  FCalcInfoNames[Init_CDRProcVarProxyDetail] := rsLSRM_InitName_CDRProcVarProxyDetail;
  FCalcInfoNames[Init_CDRProcVarProxy] := rsLSRM_InitName_CDRProcVarProxy;
  FCalcInfoNames[Init_CDRProxy] := rsLSRM_InitName_CDRProxy;
  FCalcInfoNames[Init_CDRParaErrTailProxyDetail] := rsLSRM_InitName_CDRParaErrTailProxyDetail;
  FCalcInfoNames[Init_CDRParaErrTailProxy] := rsLSRM_InitName_CDRParaErrTailProxy;
  FCalcInfoNames[Init_CDRProcVarTailProxyDetail] := rsLSRM_InitName_CDRProcVarTailProxyDetail;
  FCalcInfoNames[Init_CDRProcVarTailProxy] := rsLSRM_InitName_CDRProcVarTailProxy;
  FCalcInfoNames[Init_CDRTailProxy] := rsLSRM_InitName_CDRTailProxy;
  FCalcInfoNames[Init_SxS0] := rsLSRM_InitName_SxS0;
  FCalcInfoNames[Init_SxSMSEP] := rsLSRM_InitName_SxSMSEP;
  FCalcInfoNames[Init_SxSCDRParaErr] := rsLSRM_InitName_SxSCDRParaErr;
  FCalcInfoNames[Init_SxSCDRProcVar] := rsLSRM_InitName_SxSCDRProcVar;
  FCalcInfoNames[Init_MSEPParaErrDetail] := rsLSRM_InitName_MSEPParaErrDetail;
  FCalcInfoNames[Init_MSEPParaErr] := rsLSRM_InitName_MSEPParaErr;
  FCalcInfoNames[Init_MSEPProcVarDetail] := rsLSRM_InitName_MSEPProcVarDetail;
  FCalcInfoNames[Init_MSEPProcVar] := rsLSRM_InitName_MSEPProcVar;
  FCalcInfoNames[Init_MSEP] := rsLSRM_InitName_MSEP;
  FCalcInfoNames[Init_MSEPParaErrTailDetail] := rsLSRM_InitName_MSEPParaErrTailDetail;
  FCalcInfoNames[Init_MSEPParaErrTail] := rsLSRM_InitName_MSEPParaErrTail;
  FCalcInfoNames[Init_MSEPProcVarTailDetail] := rsLSRM_InitName_MSEPProcVarTailDetail;
  FCalcInfoNames[Init_MSEPProcVarTail] := rsLSRM_InitName_MSEPProcVarTail;
  FCalcInfoNames[Init_MSEPTail] := rsLSRM_InitName_MSEPTail;
  FCalcInfoNames[Init_CDRParaErrDetail] := rsLSRM_InitName_CDRParaErrDetail;
  FCalcInfoNames[Init_CDRParaErr] := rsLSRM_InitName_CDRParaErr;
  FCalcInfoNames[Init_CDRProcVarDetail] := rsLSRM_InitName_CDRProcVarDetail;
  FCalcInfoNames[Init_CDRProcVar] := rsLSRM_InitName_CDRProcVar;
  FCalcInfoNames[Init_CDR] := rsLSRM_InitName_CDR;
  FCalcInfoNames[Init_CDRParaErrTailDetail] := rsLSRM_InitName_CDRParaErrTailDetail;
  FCalcInfoNames[Init_CDRParaErrTail] := rsLSRM_InitName_CDRParaErrTail;
  FCalcInfoNames[Init_CDRProcVarTailDetail] := rsLSRM_InitName_CDRProcVarTailDetail;
  FCalcInfoNames[Init_CDRProcVarTail] := rsLSRM_InitName_CDRProcVarTail;
  FCalcInfoNames[Init_CDRTail] := rsLSRM_InitName_CDRTail;
  FCalcInfoNames[Init_MSEPCovParaErrProxy] := rsLSRM_InitName_MSEPCovParaErrProxy;
  FCalcInfoNames[Init_MSEPCovProcVarProxy] := rsLSRM_InitName_MSEPCovProcVarProxy;
  FCalcInfoNames[Init_MSEPCovProxy] := rsLSRM_InitName_MSEPCovProxy;
  FCalcInfoNames[Init_MSEPCovParaErrTailProxy] := rsLSRM_InitName_MSEPCovParaErrTailProxy;
  FCalcInfoNames[Init_MSEPCovProcVarTailProxy] := rsLSRM_InitName_MSEPCovProcVarTailProxy;
  FCalcInfoNames[Init_MSEPCovTailProxy] := rsLSRM_InitName_MSEPCovTailProxy;
  FCalcInfoNames[Init_CDRCovParaErrProxy] := rsLSRM_InitName_CDRCovParaErrProxy;
  FCalcInfoNames[Init_CDRCovProcVarProxy] := rsLSRM_InitName_CDRCovProcVarProxy;
  FCalcInfoNames[Init_CDRCovProxy] := rsLSRM_InitName_CDRCovProxy;
  FCalcInfoNames[Init_CDRCovParaErrTailProxy] := rsLSRM_InitName_CDRCovParaErrTailProxy;
  FCalcInfoNames[Init_CDRCovProcVarTailProxy] := rsLSRM_InitName_CDRCovProcVarTailProxy;
  FCalcInfoNames[Init_CDRCovTailProxy] := rsLSRM_InitName_CDRCovTailProxy;
  FCalcInfoNames[Init_MSEPCovParaErr] := rsLSRM_InitName_MSEPCovParaErr;
  FCalcInfoNames[Init_MSEPCovProcVar] := rsLSRM_InitName_MSEPCovProcVar;
  FCalcInfoNames[Init_MSEPCov] := rsLSRM_InitName_MSEPCov;
  FCalcInfoNames[Init_MSEPCovParaErrTail] := rsLSRM_InitName_MSEPCovParaErrTail;
  FCalcInfoNames[Init_MSEPCovProcVarTail] := rsLSRM_InitName_MSEPCovProcVarTail;
  FCalcInfoNames[Init_MSEPCovTail] := rsLSRM_InitName_MSEPCovTail;
  FCalcInfoNames[Init_CDRCovParaErr] := rsLSRM_InitName_CDRCovParaErr;
  FCalcInfoNames[Init_CDRCovProcVar] := rsLSRM_InitName_CDRCovProcVar;
  FCalcInfoNames[Init_CDRCov] := rsLSRM_InitName_CDRCov;
  FCalcInfoNames[Init_CDRCovParaErrTail] := rsLSRM_InitName_CDRCovParaErrTail;
  FCalcInfoNames[Init_CDRCovProcVarTail] := rsLSRM_InitName_CDRCovProcVarTail;
  FCalcInfoNames[Init_CDRCovTail] := rsLSRM_InitName_CDRCovTail;
  FCalcInfoNames[Init_MSEPCovEigenvalueParaErrProxy] := rsLSRM_InitName_MSEPCovEigenvalueParaErrProxy;
  FCalcInfoNames[Init_MSEPCovEigenvalueProcVarProxy] := rsLSRM_InitName_MSEPCovEigenvalueProcVarProxy;
  FCalcInfoNames[Init_MSEPCovEigenvalueProxy] := rsLSRM_InitName_MSEPCovEigenvalueProxy;
  FCalcInfoNames[Init_MSEPCovEigenvalueParaErrTailProxy] := rsLSRM_InitName_MSEPCovEigenvalueParaErrTailProxy;
  FCalcInfoNames[Init_MSEPCovEigenvalueProcVarTailProxy] := rsLSRM_InitName_MSEPCovEigenvalueProcVarTailProxy;
  FCalcInfoNames[Init_MSEPCovEigenvalueTailProxy] := rsLSRM_InitName_MSEPCovEigenvalueTailProxy;
  FCalcInfoNames[Init_CDRCovEigenvalueParaErrProxy] := rsLSRM_InitName_CDRCovEigenvalueParaErrProxy;
  FCalcInfoNames[Init_CDRCovEigenvalueProcVarProxy] := rsLSRM_InitName_CDRCovEigenvalueProcVarProxy;
  FCalcInfoNames[Init_CDRCovEigenvalueProxy] := rsLSRM_InitName_CDRCovEigenvalueProxy;
  FCalcInfoNames[Init_CDRCovEigenvalueParaErrTailProxy] := rsLSRM_InitName_CDRCovEigenvalueParaErrTailProxy;
  FCalcInfoNames[Init_CDRCovEigenvalueProcVarTailProxy] := rsLSRM_InitName_CDRCovEigenvalueProcVarTailProxy;
  FCalcInfoNames[Init_CDRCovEigenvalueTailProxy] := rsLSRM_InitName_CDRCovEigenvalueTailProxy;
  FCalcInfoNames[Init_MSEPCovEigenvalueParaErr] := rsLSRM_InitName_MSEPCovEigenvalueParaErr;
  FCalcInfoNames[Init_MSEPCovEigenvalueProcVar] := rsLSRM_InitName_MSEPCovEigenvalueProcVar;
  FCalcInfoNames[Init_MSEPCovEigenvalue] := rsLSRM_InitName_MSEPCovEigenvalue;
  FCalcInfoNames[Init_MSEPCovEigenvalueParaErrTail] := rsLSRM_InitName_MSEPCovEigenvalueParaErrTail;
  FCalcInfoNames[Init_MSEPCovEigenvalueProcVarTail] := rsLSRM_InitName_MSEPCovEigenvalueProcVarTail;
  FCalcInfoNames[Init_MSEPCovEigenvalueTail] := rsLSRM_InitName_MSEPCovEigenvalueTail;
  FCalcInfoNames[Init_CDRCovEigenvalueParaErr] := rsLSRM_InitName_CDRCovEigenvalueParaErr;
  FCalcInfoNames[Init_CDRCovEigenvalueProcVar] := rsLSRM_InitName_CDRCovEigenvalueProcVar;
  FCalcInfoNames[Init_CDRCovEigenvalue] := rsLSRM_InitName_CDRCovEigenvalue;
  FCalcInfoNames[Init_CDRCovEigenvalueParaErrTail] := rsLSRM_InitName_CDRCovEigenvalueParaErrTail;
  FCalcInfoNames[Init_CDRCovEigenvalueProcVarTail] := rsLSRM_InitName_CDRCovEigenvalueProcVarTail;
  FCalcInfoNames[Init_CDRCovEigenvalueTail] := rsLSRM_InitName_CDRCovEigenvalueTail;
  {$ENDREGION}
  TNormalLSRM.FMaxVarCoef1:=100000000;
  TNormalLSRM.FMaxVarCoef2:=100000000;
  with TSimpleLSRM.FTokenChars do begin
    WhiteChars:=[AnsiChar(VK_TAB),
                 AnsiChar(VK_Space),
                 AnsiChar(VK_Return),
                 AnsiChar(#10)];
    DecimalChar:=FormatSettings.DecimalSeparator;
    PlusChar:='+';
    MinusChar:='-';
    MulChar:='*';
    DivChar:='/';
    PowerChar:='^';
    OpeningBraceChar:='(';
    ClosingBraceChar:=')';
    BeginVecChar:='[';
    EndVecChar:=']';
    FormulaSepChar:=';';
    ParaSepChar:=',';
    FuncCumString:='Cum';
    FuncIncString:='Inc';
    FuncMaxString:='Max';
    FuncMinString:='Min';
    FuncAbsString:='Abs';
    FuncSqrtString:='Sqrt';
    ClaimPropString:='S';
    FuncColumnString:='Column';
  end;
end.
